#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <string.h>
#include "names.h"

//Status list 0: Power-ups
//Status list 1: Monsters
//Status list 2: Bullets
//Status list 3: Explosions

#define XDIM 320
#define YDIM 200
#define MAXTILES 4096
#define NUMOPTIONS 4
#define MAXSTATUS 1024
#define MAXSPRITES 4096
#define MAXPLAYERS 16
#define MAXCONNECTIONS 100

	//Player variables
static long posx[MAXPLAYERS], posy[MAXPLAYERS], zoom[MAXPLAYERS];
static short ang[MAXPLAYERS], vel[MAXPLAYERS], svel[MAXPLAYERS];
static short angvel[MAXPLAYERS], zoomvel[MAXPLAYERS];
static long lastshoottime[MAXPLAYERS], health[MAXPLAYERS];
static long deadspinx[MAXPLAYERS], deadspiny[MAXPLAYERS];

static long oposx[MAXPLAYERS], oposy[MAXPLAYERS], ozoom[MAXPLAYERS];
static short oang[MAXPLAYERS];

static char palette[768], boardfilename[13];
static short boardnum = 0;

	//Important extern variables
extern char option[NUMOPTIONS];
extern short nummaps, tilesizx[MAXTILES], tilesizy[MAXTILES];
extern short board2d[4096], sintable[2048];
extern long picanm[MAXTILES], waloff[MAXTILES];
extern short lmost[YDIM], rmost[YDIM], umost[XDIM], dmost[XDIM];
extern unsigned char keystatus[256];
extern long totalclock, lockclock, randomseed, pivotx, pivoty;
extern void __interrupt __far ksmhandler(void);

	//Sprite variables
static struct spritetype
{
	long vel, birthtime;
	short angvel, owner, health;
};
struct spritetype sprite[MAXSPRITES];
extern short headsprite[MAXSTATUS+1];
extern short prevsprite[MAXSPRITES];
extern short nextsprite[MAXSPRITES];
extern short spritestat[MAXSPRITES];   //DO NOT MODIFY DIRECTLY!
extern long spritex[MAXSPRITES];
extern long spritey[MAXSPRITES];
extern long spritezoom[MAXSPRITES];
extern short spritepicnum[MAXSPRITES];
extern short spriteang[MAXSPRITES];

	//Multiplayer variables
extern long *lastpacket2clock, *totalclockinreal;
static char tempbuf[546];
static short masterconnectnum, screenpeek = 0, myconnectindex = 0;
extern short numplayers, myconnectnum;
extern short connectindex[MAXCONNECTIONS];
extern short connecthead, connectnum[MAXPLAYERS], connectpoint2[MAXPLAYERS];

	//Send/Receiving syncing variables
static char synctics, lastsynctics;
static signed char syncmousx[MAXPLAYERS+1], osyncmousx[MAXPLAYERS+1];
static signed char syncmousy[MAXPLAYERS+1], osyncmousy[MAXPLAYERS+1];
static short syncbits[MAXPLAYERS+1], osyncbits[MAXPLAYERS+1];

	//Record/Playback variables
static long reccnt, recstat = 1;
static char recsynctics[16384];
static signed char recsyncmousx[16384][2];
static signed char recsyncmousy[16384][2];
static short recsyncbits[16384][2];

#pragma aux getshort =\
	"mov ax, word ptr [eax]",\
	parm [eax]\

#pragma aux putshort =\
	"mov word ptr [ebx], ax",\
	parm [ebx][eax]\

main(short int argc,char **argv)
{
	short waitplayers;
	long x, y, i, j, k, l, templong;

	if (argc >= 2)
	{
		strcpy(&boardfilename,argv[1]);
		if(strchr(boardfilename,'.') == 0)
			strcat(boardfilename,".map");
	}
	else
		strcpy(&boardfilename,"boards.map");

	init2draw();                      //Initialize 2draw engine
	initsb(165);                      //Initialize Sound Blaster at 11000hz
	loadpics("tiles000.art");

	if (option[2] == 1)
	{
		loadmusic("2drawsng.ksm");
		musicon(); kenchaintimer(ksmhandler,120);
	}

	loadpalette("palette.dat");
	setviewboundaries();
	permanentwritesprite(0,0,STATUSBAR);

	if (option[4] > 0)
	{
		for(i=0;i<MAXPLAYERS;i++)
		{
			syncmousx[i] = 0; osyncmousx[i] = 0;
			syncmousy[i] = 0; osyncmousy[i] = 0;
			syncbits[i] = 0; osyncbits[i] = 0;
		}

		printext256(68L,92L,31,0,"Waiting for other player...");
		tempbuf[0] = 0;
		sendpacket(-1,tempbuf,1);

		waitplayers = 2;

		if ((option[4] >= 1) && (option[4] <= 4))
			waitplayers = 2;

		while (numplayers < waitplayers)
		{
			getpackets();
			if (keystatus[1] > 0)
			{
				if (numplayers > 1)     //Signing off
				{
					tempbuf[0] = 255;
					sendpacket(-1,tempbuf,1);
				}
				if (option[2] == 1)
				{
					kenchaintimer(0,0); musicoff();
				}
				uninitsb();
				uninit2draw();
				exit(0);
			}
		}
		screenpeek = myconnectindex;
	}

	for(i=0;i<65536;i++)
		i += 0;

	reccnt = 0;
	prepareboard(boardfilename,boardnum);

	while (keystatus[1] == 0)       //Main loop starts here
	{
		if (numplayers <= 0)
		{
			drawscreen(0,256L);
			getinput(0);
			movethings();
		}
		else
		{
			sync();
			if ((myconnectnum != masterconnectnum) && (lastsynctics != 0))
				drawscreen(screenpeek,((*totalclockinreal-*lastpacket2clock)<<8)/lastsynctics);
			else
				drawscreen(screenpeek,256L);
		}
	}

	for(i=0;i<65536;i++)
		i += 0;

	if (numplayers > 1)            //Signing off
	{
		tempbuf[0] = 255;
		sendpacket(-1,tempbuf,1);
	}

	if (option[2] == 1)
	{
		kenchaintimer(0,0); musicoff();
	}

	uninitsb();
	templong = uninit2draw();            //Prepare engine for exit to DOS
	printf("%d.%1d%1d frames per second\n",(int)(templong/100),(int)((templong/10)%10),(int)(templong%10));
}

drawscreen(short snum, long smoothratio)
{
	long i, xvect, yvect;
	long daposx[MAXPLAYERS], daposy[MAXPLAYERS], dazoom[MAXPLAYERS];
	short daang[MAXPLAYERS];

		//Smoothratio is for frame interpolation
		//interpolates positions between last and 2nd to last frames
	if (smoothratio < 0) smoothratio = 0;
	if (smoothratio > 256) smoothratio = 256;

	i = connecthead;
	while (i != -1)
	{
		daposx[i] = oposx[i] + (((posx[i]-oposx[i])*smoothratio)>>8);
		daposy[i] = oposy[i] + (((posy[i]-oposy[i])*smoothratio)>>8);
		daang[i] = oang[i] + (((((ang[i]+1024-oang[i])&2047)-1024)*smoothratio)>>8);
		dazoom[i] = ozoom[i] + (((zoom[i]-ozoom[i])*smoothratio)>>8);
		i = connectpoint2[i];
	}

			//Draw the full 2-D board
	draw2board(daposx[snum],daposy[snum],dazoom[snum],daang[snum]);

		//Draw power-ups
	i = headsprite[0];
	while (i != -1)
	{
		drawsprite(spritex[i],spritey[i],spritezoom[i],spriteang[i],spritepicnum[i]);
		i = nextsprite[i];
	}

		//Draw bullets
	i = headsprite[2];
	while (i != -1)
	{
		drawsprite(spritex[i],spritey[i],spritezoom[i],spriteang[i],spritepicnum[i]);
		i = nextsprite[i];
	}

		//Draw monsters
	i = headsprite[1];
	while (i != -1)
	{
		drawsprite(spritex[i],spritey[i],spritezoom[i],spriteang[i],spritepicnum[i]);
		i = nextsprite[i];
	}

		//Draw players
	i = connecthead;
	while (i != -1)
	{
		if (health[i] >= 0)
			drawsprite(daposx[i],daposy[i],65536,daang[i],AIRPLANE);
		else
			drawsprite(deadspinx[i],deadspiny[i],65536+(health[i]<<8),ang[i]+(health[i]<<3),BURNINGAIRPLANE);
		i = connectpoint2[i];
	}

	i = headsprite[3];                   //Draw explosions over it
	while (i != -1)
	{
		drawsprite(spritex[i],spritey[i],spritezoom[i],spriteang[i],spritepicnum[i]);
		i = nextsprite[i];
	}

		//Tell who's master or slave
	if (lockclock < 120)
	{
		if (numplayers < 2)
			printext256(152L,92L,31,-1,"Alone");
		else
		{
			if (myconnectnum == masterconnectnum)
				printext256(152L,92L,31,-1,"Master");
			else
				printext256(152L,92L,31,-1,"Slave");
		}
	}
	nextpage();
}

getinput(short snum)
{
	long i;
	short mousx, mousy, bstatus;

	if (keystatus[0x1c] > 0)  //Left ENTER
	{
		keystatus[0x1c] = 0;
		screenpeek = connectpoint2[screenpeek];
		if (screenpeek == -1) screenpeek = connecthead;
	}

	if ((keystatus[0x2a]&keystatus[0x36]&keystatus[0x13]) > 0)   //Sh.Sh.R (replay)
	{
		keystatus[0x13] = 0;
		playback();
		keystatus[1] = 1;
	}

	if (option[3] != 0)
		getmousevalues(&mousx,&mousy,&bstatus);
	else
	{
		mousx = 0;
		mousy = 0;
		bstatus = 0;
	}

	if (mousx > 127) syncmousx[snum] = 127;
	else if (mousx < -128) syncmousx[snum] = -128;
	else syncmousx[snum] = (signed char)mousx;

	if (mousy > 127) syncmousy[snum] = 127;
	else if (mousy < -128) syncmousy[snum] = -128;
	else syncmousy[snum] = (signed char)mousy;

	syncbits[snum] = ((bstatus&1)<<15) + (((bstatus&6)>0)<<14);
	syncbits[snum] += (keystatus[0xd1]<<11) + (keystatus[0xc9]<<10);
	syncbits[snum] += (keystatus[0x1d]<<9) + (keystatus[0x9d]<<8);
	syncbits[snum] += (keystatus[0x2a]<<7) + (keystatus[0x2c]<<6);
	syncbits[snum] += (keystatus[0x1e]<<5) + (keystatus[0xcd]<<4);
	syncbits[snum] += (keystatus[0xcb]<<3) + (keystatus[0xd0]<<2);
	syncbits[snum] += (keystatus[0xc8]<<1) + keystatus[0x1];
	if ((syncbits[snum]&1024) > 0) keystatus[0xc9] = 0; //Only press PGUP once
	if ((syncbits[snum]&2048) > 0) keystatus[0xd1] = 0; //Only press PGDN once

	i = totalclock-lockclock;
	if (i > 255) i = 255;
	synctics = (char)i;
}

movethings()
{
	short daang;
	long x, y, i, j, k, l, nexti, nextj;
	long dist1, dist2, mindist, xvect, yvect;

	//if (option[4] > 0)
	//{
	//   printf("%x",synctics);
	//   printf("%x",randomseed);
	//   for(i=0;i<numplayers;i++)
	//   {
	//      printf("%x ",syncmousx[i]);
	//      printf("%x ",syncmousy[i]);
	//      printf("%x",syncbits[i]);
	//   }
	//   printf("\n");
	//}

	lastsynctics = synctics;
	k = connecthead;
	while (k != -1)
	{
		oposx[k] = posx[k];
		oposy[k] = posy[k];
		ozoom[k] = zoom[k];
		oang[k] = ang[k];
		k = connectpoint2[k];
	}

	if ((numplayers <= 2) && (recstat == 1))
	{
		i = connecthead; j = 0;
		while (i != -1)
		{
			recsyncmousx[reccnt][j] = syncmousx[i];
			recsyncmousy[reccnt][j] = syncmousy[i];
			recsyncbits[reccnt][j] = syncbits[i];

			j++;
			i = connectpoint2[i];
		}
		recsynctics[reccnt] = synctics;
		reccnt++; if (reccnt > 16383) reccnt = 16383;
	}

	lockclock += ((long)synctics);

		//Go through power-up sprites
	i = headsprite[0];
	while (i != -1)
	{
		nexti = nextsprite[i];

		spriteang[i] = ((spriteang[i]+sprite[i].angvel+2048)&2047);

		k = connecthead;
		while (k != -1)
		{
			if (labs(spritex[i]-posx[k])+labs(spritey[i]-posy[k]) < 8388608)
			{
				if (spritepicnum[i] == COIN)
				{
					wsay("getstuff.wav",255);
					health[k] += 1024;
				}
				if (spritepicnum[i] == SUPERSPIRAL)
				{
					wsay("spiral.wav",255);
					explode((short)i,8,SUPERSPIRALGET);
					health[k] += 4096;
				}
				deletesprite(i);
				goto nextsprite0;
			}
			k = connectpoint2[k];
		}

nextsprite0:
		i = nexti;
	}

		//Go through monster sprites
	i = headsprite[1];
	while (i != -1)
	{
		nexti = nextsprite[i];            //Back up pointer to next sprite

			//Move monster synctics steps
		xvect = ((sprite[i].vel*((long)synctics)*sintable[(2048+spriteang[i])&2047])>>2);
		yvect = ((-sprite[i].vel*((long)synctics)*sintable[(2560+spriteang[i])&2047])>>2);
		clipmove(&spritex[i],&spritey[i],xvect,yvect,2097152,1);
		spriteang[i] = ((spriteang[i]+sprite[i].angvel+2048)&2047);

		if ((spritepicnum[i] == AIRPLANE) || (spritepicnum[i] == CAR) || (spritepicnum[i] == CAR2))
		{
				//k = which player to attack
			mindist = 0x7fffffff;
			j = connecthead; k = j;
			while (j != -1)
			{
				dist1 = (posx[j]>>16)-(spritex[i]>>16); dist1 *= dist1;
				dist2 = (posy[j]>>16)-(spritey[i]>>16); dist2 *= dist2;
				if ((dist1+dist2) < mindist)
				{
					mindist = (dist1+dist2);
					k = j;
				}
				j = connectpoint2[j];
			}

			if (spritepicnum[i] == AIRPLANE)
				if (mindist < (8<<18))
				{
					daang = ((getangle((posx[k]>>8)-(spritex[i]>>8),(posy[k]>>8)-(spritey[i]>>8))+512)&2047);
					if (((spriteang[i]+2048-daang)&2047) < 1024)
						sprite[i].angvel = -(((spriteang[i]+2048-daang)&2047)>>4);
					else
						sprite[i].angvel = (((daang+2048-spriteang[i])&2047)>>4);
						//Enemy airplane randomly shoots a bullet
					if ((((spriteang[i]+2048-daang)&2047) < 128) || (((daang+2048-spriteang[i])&2047) < 128))
						if ((krand()&7) == 0)
							if (canmovethrough(posx[k],posy[k],spritex[i],spritey[i],2) == 1)
								shootbullet((short)i);
				}
			if (spritepicnum[i] == CAR)
				if (mindist < (32<<18))
				{
					daang = ((getangle((posx[k]>>8)-(spritex[i]>>8),(posy[k]>>8)-(spritey[i]>>8))+512)&2047);
					if (((spriteang[i]+2048-daang)&2047) < 1024)
						sprite[i].angvel = -(((spriteang[i]+2048-daang)&2047)>>6);
					else
						sprite[i].angvel = (((daang+2048-spriteang[i])&2047)>>6);
					//Enemy car randomly shoots a bullet
					if ((((spriteang[i]+2048-daang)&2047) < 128) || (((daang+2048-spriteang[i])&2047) < 128))
						if ((krand()&31) == 0)   //Enemy car randomly shoots a bullet
							if (canmovethrough(posx[k],posy[k],spritex[i],spritey[i],2) == 1)
								shootbullet((short)i);
				}
			if (spritepicnum[i] == CAR2)
				if (mindist < (32<<18))
				{
					daang = ((getangle((posx[k]>>8)-(spritex[i]>>8),(posy[k]>>8)-(spritey[i]>>8))+512)&2047);
					if (((spriteang[i]+2048-daang)&2047) < 1024)
						sprite[i].angvel = -(((spriteang[i]+2048-daang)&2047)>>7);
					else
						sprite[i].angvel = (((daang+2048-spriteang[i])&2047)>>7);
					//Enemy car randomly shoots a bullet
					if ((((spriteang[i]+2048-daang)&2047) < 256) || (((daang+2048-spriteang[i])&2047) < 256))
						if ((krand()&63) == 0)   //Enemy car randomly shoots a bullet
							if (canmovethrough(posx[k],posy[k],spritex[i],spritey[i],2) == 1)
								shootbullet((short)i);
				}
		}

		if ((spritepicnum[i] == BROKENCAR) || (spritepicnum[i] == BROKENCAR2))
		{                    //Broken car slows down then shrinks
			if (sprite[i].vel > 0) {sprite[i].vel -= ((((long)synctics)>>1)+1); if (sprite[i].vel < 0) sprite[i].vel = 0;}
			if (sprite[i].vel < 0) {sprite[i].vel += ((((long)synctics)>>1)+1); if (sprite[i].vel > 0) sprite[i].vel = 0;}
			if (sprite[i].vel == 0) spritezoom[i] -= (((long)synctics)<<9);
			if (spritezoom[i] <= 0)
			{                //When car shrinks to nothing, delete it
				deletesprite(i);
				goto nextsprite1;
			}
		}
nextsprite1:
		i = nexti;
	}

		//Go through bullet sprites...
	i = headsprite[2];
	while (i != -1)
	{
		nexti = nextsprite[i];            //Back up pointer to next sprite

			//Bullet gets deleted after 4 seconds of life (1sec. = 120ticks)
		if (lockclock > sprite[i].birthtime+960)
		{
			deletesprite(i);
			goto nextsprite2;
		}

			//Move bullet synctics steps
		xvect = ((sprite[i].vel*((long)synctics)*sintable[(2048+spriteang[i])&2047])>>2);
		yvect = ((-sprite[i].vel*((long)synctics)*sintable[(2560+spriteang[i])&2047])>>2);
		clipmove(&spritex[i],&spritey[i],xvect,yvect,2097152,2);
		spriteang[i] = ((spriteang[i]+sprite[i].angvel+2048)&2047);

			 //Check if the bullet hit any players
		j = connecthead;
		while (j != -1)
		{
			if (labs(spritex[i]-posx[j]) < 6291456)
				if (labs(spritey[i]-posy[j]) < 6291456)
					if (sprite[i].owner != j+4096)
						if (health[j] >= 0)
						{
							wsay("ouch.wav",255);
							health[j] -= 1024;
							if (health[j] < 0)
							{
								explode((short)j+4096,16,EXPLOSION);
								deadspinx[j] = posx[j];
								deadspiny[j] = posy[j];
								wsay("death.wav",255);
								health[j] = -1;
							}
							deletesprite(i);
							goto nextsprite2;
						}

			j = connectpoint2[j];
		}

			//Check if bullet hit any monsters
		j = headsprite[1];
		while (j != -1)
		{
			nextj = nextsprite[j];

			if (labs(spritex[i]-spritex[j]) < 6291456)
				if (labs(spritey[i]-spritey[j]) < 6291456)
					if (sprite[i].owner != j)
						switch(spritepicnum[j])
						{
							case AIRPLANE:
								sprite[j].health -= 1024;
								if (sprite[j].health < 0)
								{
									spritepicnum[j] = BURNINGAIRPLANE;
									spritezoom[j] = 49152;
									sprite[j].vel = 0;
									sprite[j].angvel = 0;
									explode((short)j,32,EXPLOSION);
									if (sprite[i].owner >= 4096)
										wsay("blowup.wav",255);
								}
								else if (sprite[i].owner >= 4096)
									wsay("bullseye.wav",255);

								deletesprite(i);
								goto nextsprite2;
							case CAR:
								sprite[j].health -= 2048;
								if (sprite[j].health < 0)
								{
									sprite[j].angvel = 0;
									explode((short)j,8,EXPLOSION);
									spritepicnum[j] = BROKENCAR;
									if (sprite[i].owner >= 4096)
										wsay("blowup.wav",255);
								}
								else if (sprite[i].owner >= 4096)
									wsay("bullseye.wav",255);

								deletesprite(i);
								goto nextsprite2;
							case CAR2:
								if (sprite[i].owner < 4096)    //Don't want car2 to disappear too quickly!
									sprite[j].health -= 1024;
								else
									sprite[j].health -= 16384;
								if (sprite[j].health < 0)
								{
									sprite[j].angvel = 0;
									explode((short)j,8,EXPLOSION);
									spritepicnum[j] = BROKENCAR2;
									if (sprite[i].owner >= 4096)
										wsay("blowup.wav",255);
								}
								deletesprite(i);
								goto nextsprite2;
							case SKATEBOARDDUDE:
								if (sprite[i].owner >= 4096)    //Only a player can kill him!
								{
									sprite[j].angvel = 0;
									explode((short)j,16,BODYPART);
									wsay("explode.wav",255);
									deletesprite(j);
									goto nextsprite2;
								}
						}

			j = nextj;
		}

nextsprite2:
		i = nexti;
	}

	i = headsprite[3];
	while (i != -1)                      //Go through explosion sprites
	{
		nexti = nextsprite[i];

		if ((sprite[i].owner >= 0) && (sprite[i].owner < 4096))
			if (spritestat[sprite[i].owner] < MAXSTATUS)
			{
					//Keep momentum while dying
				spritex[i] += ((sprite[sprite[i].owner].vel*((long)synctics)*sintable[(2048+spriteang[sprite[i].owner])&2047])>>3);
				spritey[i] += ((-sprite[sprite[i].owner].vel*((long)synctics)*sintable[(2560+spriteang[sprite[i].owner])&2047])>>3);
			}
		spritex[i] += ((sprite[i].vel*((long)synctics)*sintable[(2048+spriteang[i])&2047])>>2);
		spritey[i] += ((-sprite[i].vel*((long)synctics)*sintable[(2560+spriteang[i])&2047])>>2);
		spriteang[i] = ((spriteang[i]+sprite[i].angvel+2048)&2047);
		spritezoom[i] -= (((long)synctics)<<9);

		if ((lockclock > sprite[i].birthtime+240) || (spritezoom[i] <= 16))
		{
			deletesprite(i);
			goto nextsprite3;
		}

nextsprite3:
		i = nexti;
	}

		//Go through players
	k = connecthead;
	while (k != -1)
	{
		input2draw((short)k);
		k = connectpoint2[k];
	}
}

input2draw(short snum)
{
	long xvect, yvect;
	short i, j, movez, mousx, mousy, mstrafe;

	if (health[snum] < 0)
	{
		health[snum] -= ((long)synctics);
		if (health[snum] <= -256)
		{
			warptorandomspot(&posx[snum],&posy[snum],&ang[snum]);
			health[snum] = 16383;
		}
		else
		{
			xvect = (((long)sintable[(2048-ang[snum])&2047])*((long)vel[snum]));
			yvect = (((long)sintable[(2560-ang[snum])&2047])*((long)vel[snum]));
			xvect += (((long)sintable[(2560-ang[snum])&2047])*((long)svel[snum]));
			yvect += (((long)sintable[(1024-ang[snum])&2047])*((long)svel[snum]));
			clipmove(&deadspinx[snum],&deadspiny[snum],xvect,yvect,2097152,1);
		}
		return;
	}

	mousx = 0;
	mousy = 0;
	mstrafe = 0;
	if (option[3] != 0)
	{
		mousx = syncmousx[snum];
		mousy = (((short)syncmousy[snum])<<3);
	}

	if (((syncbits[snum]&256) > 0) || ((syncbits[snum]&16384) != 0))   //Rt. Ctrl or Rt. mouse button
	{
		mstrafe = (mousx<<3);
		mousx = 0;
		if ((syncbits[snum]&8) > 0) mstrafe -= (((short)synctics)<<(4+((syncbits[snum]&128)>0)));
		if ((syncbits[snum]&16) > 0) mstrafe += (((short)synctics)<<(4+((syncbits[snum]&128)>0)));
	}
	else                       //Else turn
	{
		if ((syncbits[snum]&8) > 0) mousx -= (((short)synctics)<<(2+((syncbits[snum]&128)>0)));
		if ((syncbits[snum]&16) > 0) mousx += (((short)synctics)<<(2+((syncbits[snum]&128)>0)));
	}                          //Move forward / backward
	if ((syncbits[snum]&2) > 0) mousy -= (((short)synctics)<<(4+((syncbits[snum]&128)>0)));
	if ((syncbits[snum]&4) > 0) mousy += (((short)synctics)<<(4+((syncbits[snum]&128)>0)));

	i = angvel[snum];
	angvel[snum] = (((angvel[snum]<<2)+(angvel[snum]<<1)+angvel[snum]+mousx)>>3);
	if (i == angvel[snum])
	{
		if (mousx < angvel[snum]) angvel[snum]--;
		if (mousx > angvel[snum]) angvel[snum]++;
	}
	i = vel[snum];
	vel[snum] = (((vel[snum]<<2)+(vel[snum]<<1)+vel[snum]+mousy)>>3);
	if (i == vel[snum])
	{
		if (mousy < vel[snum]) vel[snum]--;
		if (mousy > vel[snum]) vel[snum]++;
	}
	i = svel[snum];
	svel[snum] = (((svel[snum]<<2)+(svel[snum]<<1)+svel[snum]+mstrafe)>>3);
	if (i == svel[snum])
	{
		if (mstrafe < svel[snum]) svel[snum]--;
		if (mstrafe > svel[snum]) svel[snum]++;
	}

	ang[snum] = (ang[snum]+2048+angvel[snum])&2047;
	xvect = (((long)sintable[(2048-ang[snum])&2047])*((long)vel[snum]));
	yvect = (((long)sintable[(2560-ang[snum])&2047])*((long)vel[snum]));
	xvect += (((long)sintable[(2560-ang[snum])&2047])*((long)svel[snum]));
	yvect += (((long)sintable[(1024-ang[snum])&2047])*((long)svel[snum]));
	clipmove(&posx[snum],&posy[snum],xvect,yvect,2097152,1);   //Actually move player!

	movez = 0;
	if ((syncbits[snum]&32) > 0) movez -= (((long)synctics)<<(7+(((syncbits[snum]&128)>0)<<1)));
	if ((syncbits[snum]&64) > 0) movez += (((long)synctics)<<(7+(((syncbits[snum]&128)>0)<<1)));

	i = zoomvel[snum];
	zoomvel[snum] = (((zoomvel[snum]<<2)+(zoomvel[snum]<<1)+zoomvel[snum]+movez)>>3);
	if (i == zoomvel[snum])
	{
		if (movez < zoomvel[snum]) zoomvel[snum]--;
		if (movez > zoomvel[snum]) zoomvel[snum]++;
	}
	zoom[snum] += zoomvel[snum];
	if (zoom[snum] < 16384) zoom[snum] = 16384;
	if (zoom[snum] > 1048576) zoom[snum] = 1048576;

	if (((syncbits[snum]&512) > 0) || ((syncbits[snum]&32768) > 0))  //Left CTRL or Left mouse button
		if (lockclock > lastshoottime[snum]+20)
		{
			lastshoottime[snum] = lockclock;
			shootbullet(snum+4096);                 //You shoot a bullet!
		}

	if ((syncbits[snum]&1024) > 0)   //PGUP - Jump to last level
	{
		boardnum--; if (boardnum < 0) boardnum = nummaps-1;
		prepareboard(boardfilename,boardnum);
	}
	if ((syncbits[snum]&2048) > 0)   //PGDN - Jump to next level
	{
		boardnum++; if (boardnum >= nummaps) boardnum = 0;
		prepareboard(boardfilename,boardnum);
	}
}

shootbullet(short owner)
{
	long i, j;

	j = insertsprite(2);
	sprite[j].birthtime = lockclock;
	sprite[j].owner = owner;

	if ((owner >= 0) && (owner < 4096))
	{
		spritex[j] = spritex[owner];
		spritey[j] = spritey[owner];
		spriteang[j] = spriteang[owner];

		if (spritepicnum[owner] == AIRPLANE)
		{
			spritezoom[j] = 65536+32768;
			spritepicnum[j] = BULLET;
			sprite[j].vel = 192;
			sprite[j].angvel = 0;
		}
		if (spritepicnum[owner] == CAR)
		{
			spritezoom[j] = 65536;
			spritepicnum[j] = BULLET;
			sprite[j].vel = 128;
			sprite[j].angvel = 0;
		}
		if (spritepicnum[owner] == CAR2)
		{
			spritezoom[j] = 32768;
			spritepicnum[j] = BULLET;
			sprite[j].vel = 64;
			sprite[j].angvel = 0;
		}
	}
	else if (owner >= 4096)
	{
		wsay("zipguns.wav",255);

		spritex[j] = posx[owner-4096];
		spritey[j] = posy[owner-4096];
		spriteang[j] = ang[owner-4096];
		spritezoom[j] = 65536;
		spritepicnum[j] = SILVERBULLET;
		sprite[j].vel = 128;
		sprite[j].angvel = 0;
	}
}

explode(short owner, short numfragments, short dapicnum)
{
	long i, j;

	for(i=0;i<numfragments;i++)
	{
		j = insertsprite(3);
		if (owner >= 4096)
		{
			spritex[j] = posx[owner-4096];
			spritey[j] = posy[owner-4096];
		}
		else if (owner >= 0)
		{
			spritex[j] = spritex[owner];
			spritey[j] = spritey[owner];
		}
		spritezoom[j] = 65536;
		spriteang[j] = (krand()&2047);
		spritepicnum[j] = dapicnum;
		sprite[j].vel = (krand()&63);
		sprite[j].angvel = (krand()&63)-32;
		sprite[j].birthtime = lockclock;
		sprite[j].owner = owner;
	}
}

warptorandomspot(long *x, long *y, short *ang)
{
	long i;

	do
	{
		i = (krand()&4095);
	}
	while (((picanm[board2d[i]]>>24)&255) != 0);
	*x = ((i>>6)<<23)+4194304;
	*y = ((i&63)<<23)+4194304;
	*ang = (krand()&0x600);
}

loadpalette(char filename[20])
{
	long i, fil;

	if ((fil = open(filename,O_BINARY|O_RDWR,S_IREAD)) != -1)
	{
		read(fil,&palette[0],768);
		close(fil);
	}
	outp(0x3c8,0);
	for(i=0;i<768;i++)
		outp(0x3c9,palette[i]);
}

setviewboundaries()
{
	long i;

	for(i=0;i<YDIM;i++)   //Set up engine boundaries for each horizontal line.
	{
		lmost[i] = 32;
		rmost[i] = (XDIM-1);
	}
	for(i=0;i<XDIM;i++)
	{
		if (i < 32)
		{
			umost[i] = 1;
			dmost[i] = 0;
		}
		else
		{
			umost[i] = 0;
			dmost[i] = (YDIM-1);
		}
	}
	pivotx = 176;     //Screen coordinate pivot point that whole screen
	pivoty = 176;     //rotates around (used to be 160, 100)
}

prepareboard(char *boardfilename, short daboardnum)
{
	long i, j, nextj;

	randomseed = 17L;

	load2board(boardfilename,daboardnum,&posx[0],&posy[0],&zoom[0],&ang[0]);
	for(i=1;i<MAXPLAYERS;i++)
	{
		posx[i] = posx[0];
		posy[i] = posy[0];
		zoom[i] = zoom[0];
		ang[i] = ang[0];
	}

	j = headsprite[1];                      //Set up sprite variables
	while (j != -1)
	{
		nextj = nextsprite[j];

		sprite[j].health = 16383;
		sprite[j].birthtime = -1;
		sprite[j].angvel = 0;
		sprite[j].vel = 0;
		sprite[j].owner = -1;

		if (spritepicnum[j] == AIRPLANE)
		{
			sprite[j].vel = 100;
			sprite[j].angvel = 0;
			changespritestat(j,1);
		}
		else if ((spritepicnum[j] == CAR) || (spritepicnum[j] == CAR2) || (spritepicnum[j] == SKATEBOARDDUDE))
		{
			sprite[j].vel = 50;
			sprite[j].angvel = (krand()&15)-8;
			changespritestat(j,1);
		}
		else if ((spritepicnum[j] == COIN) || (spritepicnum[j] == SUPERSPIRAL))
		{
			sprite[j].angvel = ((((spritex[j]>>23)&1)^((spritey[j]>>23)&1))<<1)-1;
			if (spritepicnum[j] == COIN) sprite[j].angvel <<= 4;
			if (spritepicnum[j] == SUPERSPIRAL) sprite[j].angvel <<= 5;
			changespritestat(j,0);
		}

		j = nextj;
	}

	for(i=0;i<MAXPLAYERS;i++)
	{
		vel[i] = 0;
		svel[i] = 0;
		zoomvel[i] = 0;
		angvel[i] = 0;
		health[i] = 16383;                       //Initialize my variables
		lastshoottime[i] = 0;
	}

	totalclock = 0;
	lockclock = 0;
}

sync()
{
	char otherconnectnum;
	long i, j;

	if (myconnectnum == masterconnectnum)   //I am the MASTER!
	{
		getinput(connectindex[myconnectnum]);
		getpackets();

		tempbuf[0] = 2;
		tempbuf[1] = synctics;
		tempbuf[2] = 0;       //bits of which bytes to send field

		i = connecthead; j = 3;
		while (i != -1)
		{
			if ((syncmousx[i] != osyncmousx[i]) || (syncmousy[i] != osyncmousy[i]))
			{
				tempbuf[2] |= (1<<i);
				tempbuf[j] = syncmousx[i]; osyncmousx[i] = syncmousx[i];
				tempbuf[j+1] = syncmousy[i]; osyncmousy[i] = syncmousy[i];
				j += 2;
			}
			if (syncbits[i] != osyncbits[i])
			{
				tempbuf[2] |= (16<<i);
				putshort((long)(&tempbuf[j]),syncbits[i]);
				osyncbits[i] = syncbits[i];
				j += 2;
			}
			i = connectpoint2[i];
		}

		if (numplayers == 2)
		{
			if (myconnectnum == connectnum[connecthead])
				otherconnectnum = connectnum[connectpoint2[connecthead]];
			else
				otherconnectnum = connectnum[connecthead];
			sendpacket(otherconnectnum,tempbuf,j);
		}
		else if (numplayers > 2)
			sendpacket(-1,tempbuf,j);

		movethings();  //Move EVERYTHING (you too!)
	}
	else    //I am a SLAVE
	{
		getinput(MAXPLAYERS);

		tempbuf[0] = 3;
		tempbuf[1] = 0;       //bits of which bytes to send field

		j = 2;
		if ((syncmousx[MAXPLAYERS] != osyncmousx[MAXPLAYERS]) || (syncmousy[MAXPLAYERS] != osyncmousy[MAXPLAYERS]))
		{
			tempbuf[1] |= 1;
			tempbuf[j++] = syncmousx[MAXPLAYERS];
			tempbuf[j++] = syncmousy[MAXPLAYERS];
			osyncmousx[MAXPLAYERS] = syncmousx[MAXPLAYERS];
			osyncmousy[MAXPLAYERS] = syncmousy[MAXPLAYERS];
		}
		if (syncbits[MAXPLAYERS] != osyncbits[MAXPLAYERS])
		{
			tempbuf[1] |= 2;
			putshort((long)(&tempbuf[j]),syncbits[MAXPLAYERS]);
			osyncbits[MAXPLAYERS] = syncbits[MAXPLAYERS];
			j += 2;
		}
		sendpacket(masterconnectnum,tempbuf,j);
		getpackets();
	}
}

getpackets()
{
	short otherconnectnum, otherconnectindex, tempbufleng;
	long i, j;

	while ((tempbufleng = getpacket(&otherconnectnum,tempbuf)) > 0)
	{
		switch(tempbuf[0])
		{
			case 0:  //[0] (login, if myconnectnum's lowest, then respond with packet type 1)
				if (option[4] < 5)
				{
					tempbuf[0] = 1;
					sendpacket(-1,tempbuf,1);
					myconnectnum = 0;
					myconnectindex = 0;
					masterconnectnum = 0;
					numplayers = 2;
				}
				else if (option[4] == 5)
				{
					netinitconnection(otherconnectnum);

						//Set masterconnectnum to first connection number
					masterconnectnum = myconnectnum;
					for(i=0;i<numplayers;i++)
						if (connectnum[i] != otherconnectnum)
							if (connectnum[i] < masterconnectnum)
								masterconnectnum = connectnum[i];

					if (masterconnectnum == myconnectnum)
					{
						tempbuf[0] = 1;
						j = 1;
						i = connecthead;
						while (i != -1)
						{
							if (connectnum[i] != otherconnectnum)
								tempbuf[j++] = connectnum[i];
							i = connectpoint2[i];
						}
						tempbuf[j++] = connectindex[otherconnectnum];
						sendpacket(otherconnectnum,tempbuf,j);
					}
				}
				break;
			case 1:  //[1][connectnum][connectnum]...(Packet type 0 response)
				if (option[4] < 5)
				{
					myconnectnum = 1;
					myconnectindex = 1;
					masterconnectnum = 0;
					numplayers = 2;
				}
				else if (option[4] == 5)
				{
					for(i=1;i<tempbufleng-1;i++)
						netinitconnection(tempbuf[i]);
					myconnectindex = tempbuf[tempbufleng-1];
					masterconnectnum = otherconnectnum;
				}
				break;
			case 2:  //[2] (receive master sync buffer)
				synctics = tempbuf[1];

				i = connecthead; j = 3;
				while (i != -1)
				{
					if ((tempbuf[2]&(1<<i)) > 0)
					{
						syncmousx[i] = tempbuf[j];
						syncmousy[i] = tempbuf[j+1];
						j += 2;
					}
					if ((tempbuf[2]&(16<<i)) > 0)
					{
						syncbits[i] = getshort((long)(&tempbuf[j]));
						j += 2;
					}
					i = connectpoint2[i];
				}

				movethings();        //Move all players and sprites

				break;
			case 3:  //[3] (receive slave sync buffer)
				otherconnectindex = connectindex[otherconnectnum];
				j = 2;
				if ((tempbuf[1]&1) > 0)
				{
					syncmousx[otherconnectindex] = tempbuf[j];
					syncmousy[otherconnectindex] = tempbuf[j+1];
					j += 2;
				}
				if ((tempbuf[1]&2) > 0)
					syncbits[otherconnectindex] = getshort((long)(&tempbuf[j]));
				break;
			case 255:  //[255] (logout)
				keystatus[1] = 1;
				netuninitconnection(otherconnectnum);
				break;
		}
	}
}

playback()
{
	long i, j, k;

	recstat = 0;

	while (keystatus[1] == 0)
	{
		boardnum = 0;
		prepareboard(boardfilename,boardnum);

		i = 0;
		while (i < reccnt)
		{
			if (keystatus[1] > 0) break;

			if (keystatus[0x1c] > 0)
			{
				keystatus[0x1c] = 0;
				screenpeek = connectpoint2[screenpeek];
				if (screenpeek == -1) screenpeek = connecthead;
			}

			while ((totalclock > lockclock) && (i < reccnt))
			{
				j = connecthead; k = 0;
				while (j != -1)
				{
					syncmousx[j] = recsyncmousx[i][k];
					syncmousy[j] = recsyncmousy[i][k];
					syncbits[j] = recsyncbits[i][k];
					k++;
					j = connectpoint2[j];
				}
				synctics = recsynctics[i];

				movethings();
				i++;
			}

			while (totalclock < lockclock)
			{
				if (lastsynctics != 0)
					drawscreen(screenpeek,256L+((totalclock-lockclock)<<8)/lastsynctics);
				else
					drawscreen(screenpeek,256L);
			}
		}
	}
}
