// "Build Engine & Tools" Copyright (c) 1993-1997 Ken Silverman
// Ken Silverman's official web site: "http://www.advsys.net/ken"
// See the included license file "BUILDLIC.TXT" for license info.

#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>

#define MAXSECTORS 1024
#define MAXWALLS 4096
#define MAXSPRITES 4096

static struct sectortype
{
	short unsigned int wallptr, wallnum;
	char ceilingstat, ceilingosc, ceilingshade;
	long ceilingz;
	short ceilingpicnum, ceilingheinum;
	char floorstat, floorosc, floorshade;
	long floorz;
	short floorpicnum, floorheinum;
	long tag;
};

static struct walltype
{
	long x, y;
	short point2;
	char cstat, shade;
	unsigned char xrepeat, yrepeat;
	short picnum;
	short nextsector1, nextwall1;
	short nextsector2, nextwall2;
	long tag;
};

static struct spritetype
{
	long x, y, z;
	char cstat, shade;
	unsigned char xrepeat, yrepeat;
	short picnum, ang, vel, zvel, owner;
	short sectnum, statnum;
	long tag;
};

//**************************************************************************

static struct sectortype sector[MAXSECTORS];
static struct walltype wall[MAXWALLS];
static struct spritetype sprite[MAXSPRITES];

static long posx, posy, posz, mapversion;
static short int ang, cursectnum, numsectors, numwalls, numsprites;

main(int argc, char **argv)
{
	char filename[80];
	long i, j;

	if (argc != 2)
	{
		printf("CONVMAP [filename]\n");
		printf("This CONVMAP.EXE converts map version 0 to map version 1\n");
		exit(0);
	}
	strcpy(filename,argv[1]);
	if(strchr(filename,'.') == 0)
		strcat(filename,".map");

	printf("Converts from map version 1 to map version 2\n");
	printf("Converting bit 1 of sprite cstat to all non - sector[].floorz dependent.\n");

	if ((i = loadoldboard(filename)) < 0)
	{
		if (i == -1) printf("%s not found.\n",filename);
		if (i == -2) printf("Cannot convert %s.\n",filename);
		exit(0);
	}

	for(i=0;i<numsprites;i++)
	{
		if ((sprite[i].cstat&2) == 0)
			sprite[i].z += sector[sprite[i].sectnum].floorz;
		sprite[i].cstat &= (255-2);
	}

	mapversion = 2L;
	if (savenewboard(filename) == 0)
		printf("%s successfully converted.\n",filename);
	else
		printf("%s NOT converted for some stupid reason.\n",filename);
}

loadoldboard(char filename[20])
{
	short int fil, i;

	if ((fil = open(filename,O_BINARY|O_RDWR,S_IREAD)) == -1)
		return(-1);

	read(fil,&mapversion,4);
	if (mapversion != 1L) return(-2);
	read(fil,&posx,4);
	read(fil,&posy,4);
	read(fil,&posz,4);
	read(fil,&ang,2);
	read(fil,&cursectnum,2);
	read(fil,&numsectors,2);
	read(fil,&numwalls,2);
	read(fil,&numsprites,2);
	for(i=0;i<numsectors;i++) read(fil,&sector[i],sizeof(struct sectortype));
	for(i=0;i<numwalls;i++) read(fil,&wall[i],sizeof(struct walltype));
	for(i=0;i<numsprites;i++) read(fil,&sprite[i],sizeof(struct spritetype));
	close(fil);
	return(0);
}

savenewboard(char filename[20])
{
	short int fil, i;

	if ((fil = open(filename,O_BINARY|O_TRUNC|O_CREAT|O_WRONLY,S_IWRITE)) == -1)
		return(-1);

	write(fil,&mapversion,4);
	write(fil,&posx,4);
	write(fil,&posy,4);
	write(fil,&posz,4);
	write(fil,&ang,2);
	write(fil,&cursectnum,2);
	write(fil,&numsectors,2);
	write(fil,&numwalls,2);
	write(fil,&numsprites,2);
	for(i=0;i<numsectors;i++) write(fil,&sector[i],sizeof(struct sectortype));
	for(i=0;i<numwalls;i++) write(fil,&wall[i],sizeof(struct walltype));
	for(i=0;i<numsprites;i++) write(fil,&sprite[i],sizeof(struct spritetype));
	close(fil);
	return(0);
}
