#include <dos.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>

#define MAXTILES 4096
#define XDIM 320
#define YDIM 200

static char palette[768];
static long sqrtable[2048];

static char parallaxbuf[1024*256];
static long lookup[64000];

char picsiz[MAXTILES];
long artversion;
static char *pic;
short tilesizx[MAXTILES], tilesizy[MAXTILES];
long numtiles, picanm[MAXTILES], waloff[MAXTILES];

#pragma aux setvmode =\
	"int 0x10",\
	parm [eax]\

#pragma aux drawpixel =\
	"mov byte ptr [edi], al",\
	parm [edi][eax]\

#pragma aux drawpixelses =\
	"mov dword ptr [edi], eax",\
	parm [edi][eax]\

#pragma aux ksqrtasm =\
	"bsr ecx, eax",\
	"cmp ecx, 11",\
	"jge over2048",\
	"mov eax, dword ptr sqrtable[eax*4]",\
	"shr eax, 10",\
	"jmp endksqrtasm",\
	"over2048:",\
	"sub cl, 9",\
	"and cl, 0xfe",\
	"shr eax, cl",\
	"mov eax, dword ptr sqrtable[eax*4]",\
	"shr cl, 1",\
	"shl eax, cl",\
	"shr eax, 10",\
	"endksqrtasm:",\
	parm [eax]\
	modify [ebx ecx edx]\

main()
{
	char *ptr;
	short ang;
	long x, y, nx, ny, vx, vy, dist, dat, p, picnume, fil, i;

	setvmode(0x13);

	for(i=0;i<2048;i++)
		sqrtable[i] = ksqrt(i<<20);

	loadpics("tiles.art");
	if ((fil = open("palette.dat",O_BINARY|O_RDWR,S_IREAD)) != -1)
	{
		read(fil,&palette[0],768);
		outp(0x3c8,0);
		for(i=0;i<768;i++)
			outp(0x3c9,palette[i]);
		close(fil);
	}

	picnume = 122;
	for(x=0;x<XDIM;x++)   //Generate True-Sphere PSKY lookup table
		for(y=0;y<200;y++)
		{
			vx = x-160;
			vy = y-100;
			dist = ksqrtasm((vx*vx + vy*vy + 160*160)<<16);
			nx = ((vx*160)<<8) / dist;
			ny = ((vy*160)<<8) / dist;

			lookup[y*320+x] = (nx&1023) + (ny+100)*1024;
		}

	for(x=0;x<1024;x++)
		for(y=0;y<200;y++)
		{
			ptr = (char *)(waloff[picnume]+x*200+y);
			parallaxbuf[y*1024+x] = *ptr;
		}

	ang = 0;
	while (kbhit() == 0)
	{
		p = 0xa0000;
		for(p=0;p<64000;p+=4)
		{
			dat = ((long)parallaxbuf[(lookup[p+0]&0x3fc00)+((lookup[p+0]+ang)&1023)]);
			dat += ((long)parallaxbuf[(lookup[p+1]&0x3fc00)+((lookup[p+1]+ang)&1023)]<<8);
			dat += ((long)parallaxbuf[(lookup[p+2]&0x3fc00)+((lookup[p+2]+ang)&1023)]<<16);
			dat += ((long)parallaxbuf[(lookup[p+3]&0x3fc00)+((lookup[p+3]+ang)&1023)]<<24);
			drawpixelses(0xa0000+p,dat);
		}
		ang++;
	}

	free(pic);
	setvmode(0x3);
}

loadpics(char filename[20])
{
	long totpicsiz, offscount, siz;
	short int fil, i, j;

	if ((fil = open(filename,O_BINARY|O_RDONLY,S_IREAD)) == -1)
		return(-1);
	read(fil,&artversion,4);
	if (artversion != 0L) return(-1);
	read(fil,&numtiles,4);
	read(fil,&tilesizx[0],numtiles<<1);
	read(fil,&tilesizy[0],numtiles<<1);
	read(fil,&picanm[0],numtiles<<2);

	for(i=numtiles;i<MAXTILES;i++)
	{
		tilesizx[i] = 0;
		tilesizy[i] = 0;
		picanm[i] = 0L;
	}

	for(i=0;i<MAXTILES;i++)
	{
		j = 15;
		while ((j > 0) && ((1<<j) > tilesizx[i]))
			j--;
		picsiz[i] = ((char)j);
		j = 15;
		while ((j > 0) && ((1<<j) > tilesizy[i]))
			j--;
		picsiz[i] += ((char)(j<<4));
	}

	totpicsiz = 0L;
	for(i=0;i<numtiles;i++)
		totpicsiz += (long)(tilesizx[i]*tilesizy[i]);

	pic = (char *)malloc(totpicsiz);
	offscount = 0L;
	for(i=0;i<MAXTILES;i++)
	{
		waloff[i] = offscount+FP_OFF(pic);
		siz = tilesizx[i]*tilesizy[i];
		if (siz > 0)
		{
			read(fil,&pic[offscount],siz);
			offscount += siz;
		}
	}
	close(fil);
	return(0);
}

ksqrt(long num)
{
	long root, temp;

	root = 128;
	do
	{
		temp = root;
		root = ((root+(num/root))>>1);
	}
	while (labs(temp-root) > 1);
	return(root);
}
