#include <dos.h>
#include <malloc.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>

#define XDIM 320
#define YDIM 200
#define MAXXDIM 320
#define MAXYDIM 200
#define BSIZ 8
#define MAXPOINTS 4096
#define MAXPOLYS 1024
#define MAXSPLITCNT 1024
#define MAXPALOOKUPS 64
#define MAXTILES 4096
#define NUMOPTIONS 8
#define NUMKEYS 19

#define AVERAGEFRAMES 16
static long frameval[AVERAGEFRAMES], framecnt = 0;

typedef struct
{
	long x, y;
} point2d;

typedef struct
{
	long x, y, z;
} point3d;

	//Board layout
static char board[BSIZ*BSIZ*BSIZ];

	//6 cube planes in clockwise order
static long of[6*4*3] =
{
	1,-1,-1,1,-1,1,1,1,1,1,1,-1,     //posx > x
	-1,1,-1,1,1,-1,1,1,1,-1,1,1,     //posy > y
	-1,-1,1,-1,1,1,1,1,1,1,-1,1,     //posz > z
	-1,-1,-1,-1,1,-1,-1,1,1,-1,-1,1, //posx < x
	-1,-1,-1,-1,-1,1,1,-1,1,1,-1,-1, //posy < y
	-1,-1,-1,1,-1,-1,1,1,-1,-1,1,-1, //posz < z
};

	//positions/angles and their velocities
static long posx, posy, posz, a1, a2, a3;
static long fvel = 0, svel = 0, hvel = 0, a1vel = 0, a2vel = 0, a3vel = 0;
  
	//New polygon variables
static point3d r1[MAXPOINTS], r2[MAXPOINTS];
static point2d sc[MAXPOINTS];
static short point2[MAXPOINTS], numpoints, poly[MAXPOLYS], numpolys;
static short plist[MAXPOINTS];

	//2D screen bsp variables
static point2d bp[MAXPOINTS], bn[MAXPOINTS];
static short front[MAXPOINTS], back[MAXPOINTS], pnum;
static short pt[MAXPOINTS], polytag[MAXPOLYS], splittag[MAXPOINTS];
static short split[MAXSPLITCNT];

static long arealeft;

	//Diagonal filling variables
static long hplc[MAXXDIM+1], hinc[MAXXDIM+1];
static char hind[MAXXDIM+1];
static short h1[(MAXXDIM<<1)+4], n1[MAXXDIM<<5], p1[MAXXDIM<<5];
static short h2[(MAXXDIM<<1)+4], n2[MAXXDIM<<5], p2[MAXXDIM<<5];

static long gxx, gxy, gxz, gyx, gyy, gyz, gzx, gzy, gzz;       //rotation
static long gm, gd, gdx, gdy, gu, gux, guy, gv, gvx, gvy, gt;  //texturizing
static long gxoffs, gyoffs;

static char *gbuf;
static long gshade;

static char *screen;
static long frameplace, ylookup[MAXYDIM], zoom, xreszoom, ydimscale;

	//Palette variables
static char paletteloaded = 0, translucloaded = 0;
static short numpalookups;
static char palette[768], whitecol, blackcol;
static char palookup[MAXPALOOKUPS][256], *transluc;

	//Art variables
static char buf[20480], palette[768];
static short tilesizx[MAXTILES], tilesizy[MAXTILES];
static long picanm[MAXTILES];

static short sintable[2048];
static long sqrtable[2048];
static char britable[16][64];
static char textfont[1024], smalltextfont[1024];
char option[NUMOPTIONS], keys[NUMKEYS];
long randomseed;

long totalclock, lockclock, lockspeed, numframes;
static void (__interrupt __far *oldtimerhandler)();
static void __interrupt __far timerhandler(void);

volatile char keystatus[256], readch, oldreadch, extended;
static void (__interrupt __far *oldkeyhandler)();
static void __interrupt __far keyhandler(void);

#pragma aux setvmode =\
	"int 0x10",\
	parm [eax]\

#pragma aux drawpixel =\
	"mov byte ptr [edi], al",\
	parm [edi][eax]\

#pragma aux klabs =\
	"test eax, eax",\
	"jns skipnegate",\
	"neg eax",\
	"skipnegate:",\
	parm [eax]\

#pragma aux koutp =\
	"out dx, al",\
	parm [edx][eax]\

#pragma aux koutpw =\
	"out dx, ax",\
	parm [edx][eax]\

#pragma aux kinp =\
	"in al, dx",\
	parm [edx]\

#pragma aux readtimer =\
	"mov al, 0xd2",\
	"out 0x43, al",\
	"in al, 0x40",\
	"mov ah, al",\
	"in al, 0x40",\
	"xchg al, ah",\
	"and eax, 0x000ffff",\
	modify [eax]\

#pragma aux clearbuf =\
	"rep stosd",\
	parm [edi][ecx][eax]\

#pragma aux clearbufbyte =\
	"shr ecx, 1",\
	"jnc skip1",\
	"stosb",\
	"skip1: shr ecx, 1",\
	"jnc skip2",\
	"stosw",\
	"skip2: test ecx, ecx",\
	"jz skip3",\
	"rep stosd",\
	"skip3:",\
	parm [edi][ecx][eax]\

#pragma aux copybuf =\
	"rep movsd",\
	parm [esi][edi][ecx]\

#pragma aux copybufbyte =\
	"shr ecx, 1",\
	"jnc skip1",\
	"movsb",\
	"skip1: shr ecx, 1",\
	"jnc skip2",\
	"movsw",\
	"skip2: test ecx, ecx",\
	"jz skip3",\
	"rep movsd",\
	"skip3:",\
	parm [esi][edi][ecx]\

#pragma aux scale =\
	"imul ebx",\
	"idiv ecx",\
	parm [eax][ebx][ecx]\
	modify [eax edx]\

#pragma aux mulscale =\
	"imul ebx",\
	"shrd eax, edx, cl",\
	parm [eax][ebx][ecx]\
	modify [edx]\

#pragma aux mulscale8 =\
	"imul ebx",\
	"shrd eax, edx, 8",\
	parm [eax][ebx]\
	modify [edx]\

#pragma aux mulscale12 =\
	"imul ebx",\
	"shrd eax, edx, 12",\
	parm [eax][ebx]\
	modify [edx]\

#pragma aux mulscale14 =\
	"imul ebx",\
	"shrd eax, edx, 14",\
	parm [eax][ebx]\
	modify [edx]\

#pragma aux mulscale16 =\
	"imul ebx",\
	"shrd eax, edx, 16",\
	parm [eax][ebx]\
	modify [edx]\

#pragma aux mulscale28 =\
	"imul ebx",\
	"shrd eax, edx, 28",\
	parm [eax][ebx]\
	modify [edx]\

#pragma aux mulscale32 =\
	"imul ebx",\
	"mov eax, edx",\
	parm [eax][ebx]\
	modify [edx]\

#pragma aux dmulscale =\
	"imul ebx",\
	"xchg esi, eax",\
	"mov ebx, edx",\
	"imul edi",\
	"add eax, esi",\
	"adc edx, ebx",\
	"shrd eax, edx, cl",\
	parm [eax][ebx][esi][edi][ecx]\
	modify [edx]\

#pragma aux dmulscale8 =\
	"imul ebx",\
	"xchg esi, eax",\
	"mov ebx, edx",\
	"imul edi",\
	"add eax, esi",\
	"adc edx, ebx",\
	"shrd eax, edx, 8",\
	parm [eax][ebx][esi][edi]\
	modify [edx]\

#pragma aux dmulscale18 =\
	"imul ebx",\
	"xchg esi, eax",\
	"mov ebx, edx",\
	"imul edi",\
	"add eax, esi",\
	"adc edx, ebx",\
	"shrd eax, edx, 18",\
	parm [eax][ebx][esi][edi]\
	modify [edx]\

#pragma aux dmulscale24 =\
	"imul ebx",\
	"xchg esi, eax",\
	"mov ebx, edx",\
	"imul edi",\
	"add eax, esi",\
	"adc edx, ebx",\
	"shrd eax, edx, 24",\
	parm [eax][ebx][esi][edi]\
	modify [edx]\

#pragma aux divscale =\
	"cdq",\
	"shld edx, eax, cl",\
	"sal eax, cl",\
	"idiv ebx",\
	parm [eax][ebx][ecx]\
	modify [edx]\

#pragma aux divscale12 =\
	"mov edx, eax",\
	"sar edx, 31",\
	"shld edx, eax, 12",\
	"sal eax, 12",\
	"idiv ebx",\
	parm [eax][ebx]\
	modify [edx]\

#pragma aux divscale16 =\
	"mov edx, eax",\
	"sar edx, 31",\
	"shld edx, eax, 16",\
	"sal eax, 16",\
	"idiv ebx",\
	parm [eax][ebx]\
	modify [edx]\

#pragma aux divscale28 =\
	"mov edx, eax",\
	"sar edx, 31",\
	"shld edx, eax, 28",\
	"sal eax, 28",\
	"idiv ebx",\
	parm [eax][ebx]\
	modify [edx]\

#pragma aux divscale32 =\
	"mov edx, eax",\
	"xor eax, eax",\
	"idiv ebx",\
	parm [eax][ebx]\
	modify [edx]\

#pragma aux tmulscale16 =\
	"imul edx",\
	"xchg eax, ebx",\
	"xchg edx, ecx",\
	"imul edx",\
	"add ebx, eax",\
	"adc ecx, edx",\
	"mov eax, esi",\
	"imul edi",\
	"add eax, ebx",\
	"adc edx, ecx",\
	"shrd eax, edx, 16",\
	parm [eax][edx][ebx][ecx][esi][edi]\

#pragma aux tmulscale28 =\
	"imul edx",\
	"xchg eax, ebx",\
	"xchg edx, ecx",\
	"imul edx",\
	"add ebx, eax",\
	"adc ecx, edx",\
	"mov eax, esi",\
	"imul edi",\
	"add eax, ebx",\
	"adc edx, ecx",\
	"shrd eax, edx, 28",\
	parm [eax][edx][ebx][ecx][esi][edi]\

// 11,12 is 2,1-10
// 13,14 is 4,2-10
// 15,16 is 6,3-10
#pragma aux ksqrtasm =\
	"or eax, eax",\
	"jz endksqrtasm",\
	"jns skipnegate",\
	"neg eax",\
	"skipnegate: bsr ecx, eax",\
	"cmp ecx, 11",\
	"jge over2048",\
	"mov eax, dword ptr sqrtable[eax*4]",\
	"shr eax, 10",\
	"jmp endksqrtasm",\
	"over2048:",\
	"sub cl, 9",\
	"and cl, 0xfe",\
	"shr eax, cl",\
	"mov eax, dword ptr sqrtable[eax*4]",\
	"shr cl, 1",\
	"shl eax, cl",\
	"shr eax, 10",\
	"endksqrtasm:",\
	parm [eax]\
	modify [eax ebx ecx edx]\

#pragma aux readkey =\
	"in al, 0x60",\
	"mov readch, al",\
	"in al, 0x61",\
	"or al, 0x80",\
	"out 0x61, al",\
	"and al, 0x7f",\
	"out 0x61, al",\
	modify [eax]\

#pragma aux qinterpolatedown16 =\
	"mov ebx, ecx",\
	"shr ecx, 1",\
	"jz skipbegcalc",\
	"begqcalc: lea edi, [edx+esi]",\
	"sar edx, 16",\
	"mov dword ptr [eax], edx",\
	"lea edx, [edi+esi]",\
	"sar edi, 16",\
	"mov dword ptr [eax+4], edi",\
	"add eax, 8",\
	"dec ecx",\
	"jnz begqcalc",\
	"test ebx, 1",\
	"jz skipbegqcalc2",\
	"skipbegcalc: sar edx, 16",\
	"mov dword ptr [eax], edx",\
	"skipbegqcalc2:",\
	parm [eax][ecx][edx][esi]\
	modify [ebx edi]\

void __interrupt __far timerhandler()
{
	totalclock++;
	keytimerstuff();

	_chain_intr(oldtimerhandler);
	koutp(0x20,0x20);
}

keytimerstuff()
{
		//Acceleration
	if (keystatus[keys[0]]) fvel = min(fvel+8,127);
	if (keystatus[keys[1]]) fvel = max(fvel-8,-128);
	if (keystatus[0x2c]) hvel = min(hvel+8,127);
	if (keystatus[0x1e]) hvel = max(hvel-8,-128);
	if (keystatus[keys[5]])
	{
		if (keystatus[keys[3]]) svel = min(svel+8,127);
		if (keystatus[keys[2]]) svel = max(svel-8,-128);
	}
	else
	{
		if (keystatus[keys[3]]) a3vel = min(a3vel+16,127);
		if (keystatus[keys[2]]) a3vel = max(a3vel-16,-128);
	}
	if (keystatus[0x33]) a1vel = min(a1vel+16,127);
	if (keystatus[0x34]) a1vel = max(a1vel-16,-128);
	if (keystatus[keys[11]]) a2vel = min(a2vel+16,127);
	if (keystatus[keys[10]]) a2vel = max(a2vel-16,-128);

		//Friction
	if (fvel < 0) fvel = min(fvel+2,0);
	if (fvel > 0) fvel = max(fvel-2,0);
	if (svel < 0) svel = min(svel+2,0);
	if (svel > 0) svel = max(svel-2,0);
	if (hvel < 0) hvel = min(hvel+2,0);
	if (hvel > 0) hvel = max(hvel-2,0);
	if (a1vel < 0) a1vel = min(a1vel+12,0);
	if (a1vel > 0) a1vel = max(a1vel-12,0);
	if (a2vel < 0) a2vel = min(a2vel+12,0);
	if (a2vel > 0) a2vel = max(a2vel-12,0);
	if (a3vel < 0) a3vel = min(a3vel+12,0);
	if (a3vel > 0) a3vel = max(a3vel-12,0);
}

void __interrupt __far keyhandler()
{
	oldreadch = readch;
	readkey();
	if ((readch|1) == 0xe1)
		extended = 128;
	else
	{
		if (oldreadch != readch)
			keystatus[(readch&127)+extended] = ((readch>>7)^1);
		extended = 0;
	}
	koutp(0x20,0x20);
	if (keystatus[0x46])
	{
		koutp(0x43,54); koutp(0x40,255); koutp(0x40,255);
		_dos_setvect(0x9, oldkeyhandler);
		_dos_setvect(0x8, oldtimerhandler);
		free(screen);
		setvmode(0x3);
		exit(0);
	}
}

main()
{
	long i, j, x, y, z, xx, yy, zz, xp, yp, zp, xvect, yvect, zvect, templong;
	long xlist[BSIZ], ylist[BSIZ], zlist[BSIZ];
	char shiftstate, tempbuf[256];

	setvmode(0x13);

	if ((screen = (char *)malloc(XDIM*(YDIM+256))) == NULL)
	{
		setvmode(0x3);
		printf("Not enough memory for screen buffer\n");
		exit(0);
	}
	frameplace = FP_OFF(screen)+(XDIM*128);

	loadtables();
	loadpalette();

	outp(0x3c8,0);
	for(i=0;i<768;i++)
		outp(0x3c9,palette[i]);

	loadnewall(0L,20L);
	for(j=4096;j<20480;j+=4096)
	{
		do
		{
			i = (krand()&255);
		} while ((tilesizx[i] != 64) || (tilesizy[i] != 64));
		loadnewall(j,i);
	}

	j = 0;
	for(i=0;i<MAXYDIM;i++) ylookup[i] = j, j += XDIM;

	srand(readtimer());

	initboard();

	oldtimerhandler = _dos_getvect(0x8);
	_dos_setvect(0x8, timerhandler);
	koutp(0x43,54); koutp(0x40,9942&255); koutp(0x40,9942>>8);
	oldkeyhandler = _dos_getvect(0x9);
	_disable(); _dos_setvect(0x9, keyhandler); _enable();
	totalclock = 0L;
	lockclock = 0L;
	numframes = 0;
	lockspeed = 0;

	zoom = 65536;
	xreszoom = mulscale(zoom,XDIM,1);
	ydimscale = divscale32(YDIM,XDIM*zoom);

	while (keystatus[1] == 0)
	{
		if (keystatus[0x39])
		{
			keystatus[0x39] = 0;
			for(j=0;j<20480;j+=4096)
			{
				do
				{
					i = (krand()&255);
				} while ((tilesizx[i] != 64) || (tilesizy[i] != 64));
				loadnewall(j,i);
			}
		}

		setuprotate3d();
  
		if (keystatus[0x29] != 0)
			{ keystatus[0x29] = 2; frameplace = 0xa0000; }
		else
			frameplace = FP_OFF(screen)+(XDIM*128);

		clearbuf(frameplace,(XDIM*YDIM)>>2,0L);

		xp = max(min(posx>>10,BSIZ-1),0);
		yp = max(min(posy>>10,BSIZ-1),0);
		zp = max(min(posz>>10,BSIZ-1),0);

		i = 0;
		for(z=0;z<xp;z++) xlist[i++] = z;
		for(z=BSIZ-1;z>xp;z--) xlist[i++] = z;
		xlist[i] = xp;

		i = 0;
		for(z=0;z<yp;z++) ylist[i++] = z;
		for(z=BSIZ-1;z>yp;z--) ylist[i++] = z;
		ylist[i] = yp;

		i = 0;
		for(z=0;z<zp;z++) zlist[i++] = z;
		for(z=BSIZ-1;z>zp;z--) zlist[i++] = z;
		zlist[i] = zp;

		arealeft = mulscale(XDIM*YDIM,1023,5); pnum = 0;

		for(xx=BSIZ-1;xx>=0;xx--)
		{
			x = xlist[xx];
			for(yy=BSIZ-1;yy>=0;yy--)
			{
				y = ylist[yy];
				for(zz=BSIZ-1;zz>=0;zz--)
				{
					z = zlist[zz];
					if (board[x*BSIZ*BSIZ+y*BSIZ+z] != 0) drawcube(x,y,z);
					if (arealeft <= 0) goto endraw;
				}
			}
		}
endraw:
		i = totalclock;
		if (i != frameval[framecnt])
		{
			sprintf(tempbuf,"%ld",(120*AVERAGEFRAMES)/(i-frameval[framecnt]));
			printext256(0L,0L,31,-1,tempbuf,1);
			frameval[framecnt] = i;
		}
		framecnt = ((framecnt+1)&(AVERAGEFRAMES-1));

		copybuf(frameplace,0xa0000,16000);

		shiftstate = (keystatus[0x2a]|keystatus[0x36]);
		if ((keystatus[0x34] > 0) && (a1vel > (-128<<shiftstate))) a1vel -= (lockspeed<<(3+shiftstate));
		if ((keystatus[0x33] > 0) && (a1vel < (128<<shiftstate))) a1vel += (lockspeed<<(3+shiftstate));
		if ((keystatus[0x2c] > 0) && (hvel < (192<<shiftstate))) hvel += (lockspeed<<(2+shiftstate));
		if ((keystatus[0x1e] > 0) && (hvel > (-192<<shiftstate))) hvel -= (lockspeed<<(2+shiftstate));

		xvect = (svel<<(shiftstate+1));
		yvect = hvel;
		zvect = (fvel<<(shiftstate+1));
		if ((xvect|yvect|zvect) != 0)
		{
			xvect *= (lockspeed<<7);
			yvect *= (lockspeed<<7);
			zvect *= (lockspeed<<7);
			backrotate3d(&xvect,&yvect,&zvect);
			clipmove(&posx,&posy,&posz,xvect>>11,yvect>>11,zvect>>11);

			posx = min(max(posx,1024+128),(BSIZ<<10)-(1024+128));
			posy = min(max(posy,1024+128),(BSIZ<<10)-(1024+128));
			posz = min(max(posz,1024+128),(BSIZ<<10)-(1024+128));
		}

		if (keystatus[0x52] > 0)    //0 on keypad (return to normal angle)
		{
			if ((keystatus[0x1d]|keystatus[0x9d]) > 0)
			{
				a2 = ((a2+1536)&2047);
				if (a2 < 1024) { a2 -= (lockspeed<<1); if (a2 < 0) a2 = 0; }
				else           { a2 += (lockspeed<<1); if (a2 >= 2048) a2 = 0; }
				a2 = ((a2+512)&2047);
			}
			else
			{
				if (a2 < 1024) { a2 -= (lockspeed<<1); if (a2 < 0) a2 = 0; }
				else           { a2 += (lockspeed<<1); if (a2 >= 2048) a2 = 0; }
			}
			if (a1 < 1024) { a1 -= (lockspeed<<1); if (a1 < 0) a1 = 0; }
			else           { a1 += (lockspeed<<1); if (a1 >= 2048) a1 = 0; }
		}

		a1 = ((a1+((a1vel*lockspeed)>>6)+2048)&2047);
		a2 = ((a2+((a2vel*lockspeed)>>(5-shiftstate))+2048)&2047);
		a3 = ((a3+((a3vel*lockspeed)>>(5-shiftstate))+2048)&2047);
			//Auto-Tilting
		if (a3vel == 0)
		{
			if (a1 < 1024)
			{
				a1 -= (lockspeed<<1);
				if (a1 < 0) a1 = 0;
			}
			else
			{
				a1 += (lockspeed<<1);
				if (a1 >= 2048) a1 = 0;
			}
		}
		else
		{
			a1 = (((((((a1+1024)&2047)*7) + (1024-a3vel))>>3)+1024)&2047);
		}

		if (hvel > 0) {hvel -= (lockspeed<<1); if (hvel < 0) hvel = 0;}
		if (hvel < 0) {hvel += (lockspeed<<1); if (hvel > 0) hvel = 0;}
		if (a1vel > 0) {a1vel -= (lockspeed<<2); if (a1vel < 0) a1vel = 0;}
		if (a1vel < 0) {a1vel += (lockspeed<<2); if (a1vel > 0) a1vel = 0;}

		lockspeed = totalclock-lockclock;
		lockclock += lockspeed;
		numframes++;
	}
	_dos_setvect(0x9, oldkeyhandler);
	_dos_setvect(0x8, oldtimerhandler);
	koutp(0x43,54); koutp(0x40,255); koutp(0x40,255);
	setvmode(0x3);
	if (totalclock != 0)
	{
		templong = (numframes*12000L)/totalclock;
		printf("%d.%1d%1d frames per second\n",(short)(templong/100),(short)((templong/10)%10),(short)(templong%10));
	}
	free(screen);
}

	//This routine is for convex polygons with a 90 viewing angle
	//It also clips to a z-plane first so no need to check for overflow
clippoly(long npoints)
{
	long i, z, zz, s1, s2, ss1, ss2, t1, t2, npoints2, start2;

		//Clip to close z
	npoints2 = 0; ss1 = r1[0].z-4096;
	for(z=0;z<npoints;z++)
	{
		zz = z+1; if (zz == npoints) zz = 0;
		s1 = ss1; ss1 = r1[zz].z-4096;
		if (s1 >= 0) copybuf(&r1[z],&r2[npoints2++],3);
		if ((s1^ss1) < 0)
		{
			t1 = divscale28(s1,s1-ss1);
			r2[npoints2].x = r1[z].x + mulscale28(t1,r1[zz].x-r1[z].x);
			r2[npoints2].y = r1[z].y + mulscale28(t1,r1[zz].y-r1[z].y);
			r2[npoints2].z = 4096;
			npoints2++;
		}
	}
	if (npoints2 < 3) return(0);

		//Clip left & right
	start2 = npoints2;
	ss1 = r2[0].z+r2[0].x; ss2 = r2[0].z-r2[0].x;
	for(z=0;z<start2;z++)
	{
		zz = z+1; if (zz == start2) zz = 0;
		s1 = ss1; s2 = ss2;
		ss1 = r2[zz].z+r2[zz].x; ss2 = r2[zz].z-r2[zz].x;
		if ((s1|s2) >= 0) copybuf(&r2[z],&r2[npoints2++],3);
		if ((s1^ss1) < 0)
		{
			if ((s2^ss2) < 0)
			{
				if (s1 < 0)
					{ t1 = divscale28(s1,s1-ss1); t2 = divscale28(s2,s2-ss2); }
				else
					{ t1 = divscale28(s2,s2-ss2); t2 = divscale28(s1,s1-ss1); }
				r2[npoints2].x = r2[z].x + mulscale28(t1,r2[zz].x-r2[z].x);
				r2[npoints2+1].x = r2[z].x + mulscale28(t2,r2[zz].x-r2[z].x);
				r2[npoints2].y = r2[z].y + mulscale28(t1,r2[zz].y-r2[z].y);
				r2[npoints2+1].y = r2[z].y + mulscale28(t2,r2[zz].y-r2[z].y);
				r2[npoints2].z = r2[z].z + mulscale28(t1,r2[zz].z-r2[z].z);
				r2[npoints2+1].z = r2[z].z + mulscale28(t2,r2[zz].z-r2[z].z);
				npoints2 += 2;
			}
			else
			{
				t1 = divscale28(s1,s1-ss1);
				r2[npoints2].x = r2[z].x + mulscale28(t1,r2[zz].x-r2[z].x);
				r2[npoints2].y = r2[z].y + mulscale28(t1,r2[zz].y-r2[z].y);
				r2[npoints2].z = -r2[npoints2].x;   //90 optimization
				npoints2++;
			}
		}
		else if ((s2^ss2) < 0)
		{
			t1 = divscale28(s2,s2-ss2);
			r2[npoints2].x = r2[z].x + mulscale28(t1,r2[zz].x-r2[z].x);
			r2[npoints2].y = r2[z].y + mulscale28(t1,r2[zz].y-r2[z].y);
			r2[npoints2].z = r2[npoints2].x;       //90 optimization
			npoints2++;
		}
	}
	if (npoints2-start2 < 3) return(0);

		//Clip top & bottom
	npoints = 0;
	i = mulscale16(r2[start2].z,ydimscale);
	ss2 = i-r2[start2].y; ss1 = i+r2[start2].y;
	for(z=start2;z<npoints2;z++)
	{
		zz = z+1; if (zz == npoints2) zz = start2;
		s1 = ss1; s2 = ss2;
		i = mulscale16(r2[zz].z,ydimscale);
		ss2 = i-r2[zz].y; ss1 = i+r2[zz].y;
		if ((s1|s2) >= 0) copybuf(&r2[z],&r1[npoints++],3);
		if ((s1^ss1) < 0)
		{
			if ((s2^ss2) < 0)
			{
				if (s1 < 0)
					{ t1 = divscale28(s1,s1-ss1); t2 = divscale28(s2,s2-ss2); }
				else
					{ t1 = divscale28(s2,s2-ss2); t2 = divscale28(s1,s1-ss1); }
				r1[npoints].x = r2[z].x + mulscale28(t1,r2[zz].x-r2[z].x);
				r1[npoints+1].x = r2[z].x + mulscale28(t2,r2[zz].x-r2[z].x);
				r1[npoints].y = r2[z].y + mulscale28(t1,r2[zz].y-r2[z].y);
				r1[npoints+1].y = r2[z].y + mulscale28(t2,r2[zz].y-r2[z].y);
				r1[npoints].z = r2[z].z + mulscale28(t1,r2[zz].z-r2[z].z);
				r1[npoints+1].z = r2[z].z + mulscale28(t2,r2[zz].z-r2[z].z);
				npoints += 2;
			}
			else
			{
				t1 = divscale28(s1,s1-ss1);
				r1[npoints].x = r2[z].x + mulscale28(t1,r2[zz].x-r2[z].x);
				r1[npoints].y = r2[z].y + mulscale28(t1,r2[zz].y-r2[z].y);
				r1[npoints].z = r2[z].z + mulscale28(t1,r2[zz].z-r2[z].z);
				npoints++;
			}
		}
		else if ((s2^ss2) < 0)
		{
			t1 = divscale28(s2,s2-ss2);
			r1[npoints].x = r2[z].x + mulscale28(t1,r2[zz].x-r2[z].x);
			r1[npoints].y = r2[z].y + mulscale28(t1,r2[zz].y-r2[z].y);
			r1[npoints].z = r2[z].z + mulscale28(t1,r2[zz].z-r2[z].z);
			npoints++;
		}
	}
	return(npoints);
}

drawcube (long x, long y, long z)
{
	long i, j, t, p, cx, cy, cz, newarea, firstime, *optr;

	cx = (((x<<10)+512-posx)<<8);
	cy = (((y<<10)+512-posy)<<8);
	cz = (((z<<10)+512-posz)<<8);
	gbuf = (char *)(FP_OFF(buf)+((board[x*BSIZ*BSIZ+y*BSIZ+z]-1)<<12));

	for(j=0;j<=2;j++)
	{
		switch(j)
		{
			case 0:
				if (cx < (-512<<8))
				{
					if (x < BSIZ-1) if (board[(x+1)*BSIZ*BSIZ+y*BSIZ+z] != 0) continue;
					optr = (long *)(of+0*12);
				}
				else
				{
					if (x > 0) if (board[(x-1)*BSIZ*BSIZ+y*BSIZ+z] != 0) continue;
					if (cx <= (512<<8)) continue;
					optr = (long *)(of+3*12);
				}
				break;
			case 1:
				if (cy < (-512<<8))
				{
					if (y < BSIZ-1) if (board[x*BSIZ*BSIZ+(y+1)*BSIZ+z] != 0) continue;
					optr = (long *)(of+1*12);
				}
				else
				{
					if (cy <= (512<<8)) continue;
					if (y > 0) if (board[x*BSIZ*BSIZ+(y-1)*BSIZ+z] != 0) continue;
					optr = (long *)(of+4*12);
				}
				break;
			case 2:
				if (cz < (-512<<8))
				{
					if (z < BSIZ-1) if (board[x*BSIZ*BSIZ+y*BSIZ+(z+1)] != 0) continue;
					optr = (long *)(of+2*12);
				}
				else
				{
					if (cz <= (512<<8)) continue;
					if (z > 0) if (board[x*BSIZ*BSIZ+y*BSIZ+(z-1)] != 0) continue;
					optr = (long *)(of+5*12);
				}
				break;
		}
                
		for(i=0;i<4;i++)
		{
			r1[i].x = optr[0]+cx;
			r1[i].y = optr[1]+cy;
			r1[i].z = optr[2]+cz;
			optr += 3;
			rotate3d(&r1[i].x,&r1[i].y,&r1[i].z);
		}

		setupgetreadpos();

		numpoints = clippoly(4);

		if (numpoints >= 3)
		{
			for(i=0;i<numpoints;i++)
			{
				sc[i].x = scale(xreszoom,r1[i].x,r1[i].z) + (XDIM<<15)+32768;
				sc[i].y = scale(xreszoom,r1[i].y,r1[i].z) + (YDIM<<15)+32768;

				if (sc[i].x < 0) sc[i].x = 0;
				if (sc[i].x >= (XDIM<<16)) sc[i].x = (XDIM<<16)-1;
				if (sc[i].y < 0) sc[i].y = 0;
				if (sc[i].y >= (YDIM<<16)) sc[i].y = (YDIM<<16)-1;

				point2[i] = i+1;
			}
			point2[numpoints-1] = 0;
			poly[0] = 0; poly[1] = numpoints; numpolys = 1;

			gshade = ((j-1)<<2)+4;

			addpoly();
			firstime = 0;
			for(p=numpolys-1;p>=0;p--)
				if ((polytag[p] == -2) && (poly[p+1] >= poly[p]+3))
				{
					newarea = 0;
					for(i=poly[p];i<poly[p+1];i++)
						newarea += mulscale28(sc[i].x+sc[point2[i]].x,sc[point2[i]].y-sc[i].y);
					if (newarea > 0)
					{
						if (firstime == 0) { initfasttextureplane(); firstime = 1; }
						drawfasttextureplane((long)poly[p],(long)poly[p+1]-1);
						arealeft -= newarea; if (arealeft < 0) return;
					}
				}
		}
	}
}

initboard()
{
	long i, j, k;

	for(i=BSIZ*BSIZ*BSIZ-1;i>=0;i--)
	{
		if ((rand()&3) == 0)
			board[i] = (krand()&3)+2;
		else
			board[i] = 0;
	}
	for(i=0;i<BSIZ;i++)
		for(j=0;j<BSIZ;j++)
		{
			board[i*BSIZ*BSIZ+j*BSIZ+0] = 1;
			board[i*BSIZ*BSIZ+j*BSIZ+(BSIZ-1)] = 1;
			board[i*BSIZ*BSIZ+0*BSIZ+j] = 1;
			board[i*BSIZ*BSIZ+(BSIZ-1)*BSIZ+j] = 1;
			board[0*BSIZ*BSIZ+i*BSIZ+j] = 1;
			board[(BSIZ-1)*BSIZ*BSIZ+i*BSIZ+j] = 1;
		}
        
	do
	{
		posx = ((rand()%(BSIZ-2))<<10) + 1536;
		posy = ((rand()%(BSIZ-2))<<10) + 1536;
		posz = ((rand()%(BSIZ-2))<<10) + 1536;
	} while (board[(posx>>10)*BSIZ*BSIZ+(posy>>10)*BSIZ+(posz>>10)] != 0);
	a1 = 0; a2 = 0; a3 = (rand()&2047);

	for(i=6*4*3-1;i>=0;i--) of[i] <<= 17;
}

rotate3d (long *nx, long *ny, long *nz)
{
	long x, y, z;

	x = *nx; y = *ny; z = *nz;
	*nx = tmulscale28(x,gxx,y,gyx,z,gzx);
	*ny = tmulscale28(x,gxy,y,gyy,z,gzy);
	*nz = tmulscale28(x,gxz,y,gyz,z,gzz);
}

backrotate3d (long *nx, long *ny, long *nz)
{
	long x, y, z;

	x = *nx; y = *ny; z = *nz;
	*nx = tmulscale28(x,gxx,y,gxy,z,gxz);
	*ny = tmulscale28(x,gyx,y,gyy,z,gyz);
	*nz = tmulscale28(x,gzx,y,gzy,z,gzz);
}

setupgetreadpos()
{
	long x0, y0, z0, x1, y1, z1, x2, y2, z2;

	x0 = r1[0].x; x1 = r1[1].x-r1[0].x; x2 = r1[3].x-r1[0].x;
	y0 = r1[0].y; y1 = r1[1].y-r1[0].y; y2 = r1[3].y-r1[0].y;
	z0 = r1[0].z; z1 = r1[1].z-r1[0].z; z2 = r1[3].z-r1[0].z;
	gdx = dmulscale24(y1,z2,-y2,z1); gdy = dmulscale24(x2,z1,-x1,z2);
	gux = dmulscale24(y2,z0,-y0,z2); guy = dmulscale24(x0,z2,-x2,z0);
	gvx = dmulscale24(y0,z1,-y1,z0); gvy = dmulscale24(x1,z0,-x0,z1);
	gd = mulscale16(dmulscale24(x1,y2,-x2,y1),xreszoom)-mulscale16(gdx,(XDIM<<15)+32768)-mulscale16(gdy,(YDIM<<15)+32768);
	gu = mulscale16(dmulscale24(x2,y0,-x0,y2),xreszoom)-mulscale16(gux,(XDIM<<15)+32768)-mulscale16(guy,(YDIM<<15)+32768);
	gv = mulscale16(dmulscale24(x0,y1,-x1,y0),xreszoom)-mulscale16(gvx,(XDIM<<15)+32768)-mulscale16(gvy,(YDIM<<15)+32768);
	gt = mulscale28(tmulscale16(x0,gdx,x1,gux,x2,gvx),xreszoom);
}

setuprotate3d()
{
	long c1, s1, c2, s2, c3, s3, c1c3, c1s3, s1c3, s1s3;

	c1 = sintable[(a1+512)&2047]; s1 = sintable[a1&2047];
	c2 = sintable[(a2+512)&2047]; s2 = sintable[a2&2047];
	c3 = sintable[(a3+512)&2047]; s3 = sintable[a3&2047];

	c1c3 = c1*c3; c1s3 = c1*s3; s1c3 = s1*c3; s1s3 = s1*s3;

	gxx = mulscale14(s1s3,s2) + c1c3;
	gxy = -mulscale14(c1s3,s2) + s1c3;
	gxz = c2*s3;
	gyx = -s1*c2;
	gyy = c1*c2;
	gyz = (s2<<14);
	gzx = mulscale14(s1c3,s2) - c1s3;
	gzy = -mulscale14(c1c3,s2) - s1s3;
	gzz = c2*c3;
}

loadtables()
{
	long i, fil;

	for(i=0;i<2048;i++)
		sqrtable[i] = calcksqrt(i<<20);

	if ((fil = open("tables.dat",O_BINARY|O_RDWR,S_IREAD)) != -1)
	{
		read(fil,&sintable[0],4096);
		lseek(fil,4096+4096+640,SEEK_SET);
		read(fil,&textfont[0],1024);
		read(fil,&smalltextfont[0],1024);
		read(fil,&britable[0],1024);
		close(fil);
	}
	if ((fil = open("setup.dat",O_BINARY|O_RDWR,S_IREAD)) != -1)
	{
		read(fil,&option[0],NUMOPTIONS);
		read(fil,&keys[0],NUMKEYS);
		close(fil);
	}
}

loadpalette()
{
	long i, j, fil;

	if (paletteloaded == 0)
	{
		if ((fil = open("palette.dat",O_BINARY|O_RDWR,S_IREAD)) != -1)
		{
			numpalookups = ((filelength(fil)-768)>>7);
			if ((numpalookups&1) > 0)
			{   //Transluscent palette found in palette.dat
				numpalookups = ((numpalookups-255)>>1);

				if (translucloaded == 0)
					if ((transluc = (char *)malloc(65536L)) != 0)
					{
						//fixtransluscence(FP_OFF(transluc));
						translucloaded = 1;
					}
			}
			else
			{    //No transluscent palette in palette.dat
				numpalookups >>= 1;
			}
			read(fil,&palette[0],768);
			read(fil,&palookup[0][0],(numpalookups<<8));
			if (translucloaded == 1)
			{
				for(i=0;i<255;i++)
				{
					read(fil,&transluc[(i<<8)+i+1],255-i);
					for(j=i+1;j<256;j++)
						transluc[(j<<8)+i] = transluc[(i<<8)+j];
				}
				for(i=0;i<256;i++)
					transluc[(i<<8)+i] = i;
			}

			blackcol = 0; j = 0x7fffffff;
			for(i=0;i<256;i++)
				if (palette[i*3]+palette[i*3+1]+palette[i*3+2] < j)
				{
					j = palette[i*3]+palette[i*3+1]+palette[i*3+2];
					blackcol = i;
				}

			whitecol = 0; j = 0x80000000;
			for(i=0;i<256;i++)
				if (palette[i*3]+palette[i*3+1]+palette[i*3+2] > j)
				{
					j = palette[i*3]+palette[i*3+1]+palette[i*3+2];
					whitecol = i;
				}

			close(fil);
		}
		paletteloaded = 1;
	}
}

ksqrt(long num)
{
	return(ksqrtasm(num));
}

calcksqrt(long num)
{
	long root, temp;

	if (num < 0) return(-1);
	root = 128;
	do
	{
		temp = root;
		root = ((root+(num/root))>>1);
	}
	while (klabs(temp-root) > 1);
	return(root);
}

clipmove (long *x, long *y, long *z, long xv, long yv, long zv)
{
	*x = (*x) + xv;
	*y = (*y) + yv;
	*z = (*z) + zv;
}

void initfasttextureplane()
{
	//dapicnum = gtext+animateoffs(gtext);
	//if (waloff[dapicnum] == 0) loadtile(dapicnum);
	//if (waloff[dapicnum] == 0) return;

	if (klabs(gdx) <= klabs(gdy))
	{
		if (gdx == 0)
			gm = 0;
		else
		{
			gm = divscale16(gdx,gdy);   //d& = y& + mulscale&(x&, m&, 16)
			if (klabs(gm) < 32) gm = 0;
			if (gm > 65536-64) gm = 65536;
			if (gm < -65536+64) gm = -65536;
		}
		if (gm > 0) hplc[0] = 0L; else hplc[0] = XDIM;
		gxoffs = gux-mulscale16(guy,gm);
		gyoffs = gvx-mulscale16(gvy,gm);
	}
	else
	{
		if (gdy == 0)
			gm = 0;
		else
		{
			gm = divscale16(gdy,gdx);
			if (klabs(gm) < 32) gm = 0;
			if (gm > 65536-64) gm = 65536;
			if (gm < -65536+64) gm = -65536;
		}
		if (gm > 0) hplc[0] = 0L; else hplc[0] = XDIM;
		gxoffs = guy-mulscale16(gux,gm);
		gyoffs = gvy-mulscale16(gvx,gm);
	}
}

void drawfasttextureplane(long pstart, long pend)
{
	long i, j, z, zz, x, y, x1, y1, x2, y2, d1, d2, dapicnum;
	long plc, inc, minx, maxx, miny, maxy, c1, c2;

	if (keystatus[0x38])
	{
		for(z=pstart;z<=pend;z++)
		{
			zz = point2[z];
			drawline256(sc[z].x,sc[z].y,sc[zz].x,sc[zz].y,15);
		}
		return;
	}

	if (klabs(gdx) <= klabs(gdy))
	{
		for(z=pstart;z<=pend;z++) sc[z].x += 65535;

		minx = sc[pstart].x; maxx = sc[pstart].x;
		for(z=pstart+1;z<=pend;z++)
		{
			x = sc[z].x;
			if (x < minx) minx = x;
			if (x > maxx) maxx = x;
		}
		minx >>= 16; maxx >>= 16;

		qinterpolatedown16(&hplc[minx],maxx-minx+1,(((long)hplc[0])<<16)+gm*minx+32768,gm);

		if (gm != 0)
		{
			if (gm < 0) i = XDIM; else i = -XDIM;
			for(x=minx;x<=maxx;x++)
			{
				hinc[x] = 1; hind[x] = 0;
				if (hplc[x] != hplc[x+1]) hinc[x] += i, hind[x] = 1;
			}
			hinc[maxx] = 0;
		}
		else
		{
			for(x=minx;x<=maxx;x++)
				hinc[x] = 1, hind[x] = 0;
		}

		miny = 0x7fffffff; maxy = 0x80000000;
		for(z=pstart;z<=pend;z++)
		{
			y = (sc[z].y>>16) + hplc[sc[z].x>>16];
			if (y < miny) miny = y;
			if (y > maxy) maxy = y;
		}
		miny--; maxy++;

		c1 = 1; clearbuf(&h1[miny],((maxy+1-miny)+1)>>1,0L);
		c2 = 1; clearbuf(&h2[miny],((maxy+1-miny)+1)>>1,0L);

		for(z=pstart;z<=pend;z++)
		{
			zz = point2[z];
			x1 = sc[z].x; y1 = sc[z].y; x2 = sc[zz].x; y2 = sc[zz].y;

			if (((x1^x2)&0xffff0000) == 0)
			{
				x = (x1>>16);
				y = (y1>>16) + hplc[x];
				j = (y2>>16) + hplc[x];
				while (j < y) { p1[c1] = x; n1[c1] = h1[y]; h1[y] = c1; c1++; y--; }
				while (j > y) { y++; p2[c2] = x; n2[c2] = h2[y]; h2[y] = c2; c2++; }
			}
			else
			{
				if ((klabs(y2-y1)>>15) < klabs(x2-x1))
					inc = divscale16(y2-y1,x2-x1);
				else
					inc = 0;

				d1 = (x1>>16); d2 = (x2>>16);
				if (x1 < x2)
				{
					plc = y1 + mulscale16((-x1)&65535,inc);
					y = (y1>>16) + hplc[d1];
					for(x=d1;x<d2;x++)
					{
						j = (plc>>16) + hplc[x]; plc += inc;
						while (j < y) { p1[c1] = x; n1[c1] = h1[y]; h1[y] = c1; c1++; y--; }
						while (j > y) { y++; p2[c2] = x; n2[c2] = h2[y]; h2[y] = c2; c2++; }
					}
					j = (y2>>16) + hplc[d2];
					while (j < y) { p1[c1] = x; n1[c1] = h1[y]; h1[y] = c1; c1++; y--; }
					while (j > y) { y++; p2[c2] = x; n2[c2] = h2[y]; h2[y] = c2; c2++; }
				}
				else
				{
					plc = y2 + mulscale16((-x2)&65535,inc);
					y = (y2>>16) + hplc[d2];
					for(x=d2;x<d1;x++)
					{
						j = (plc>>16) + hplc[x]; plc += inc;
						while (j < y) { p2[c2] = x; n2[c2] = h2[y]; h2[y] = c2; c2++; y--; }
						while (j > y) { y++; p1[c1] = x; n1[c1] = h1[y]; h1[y] = c1; c1++; }
					}
					j = (y1>>16) + hplc[d1];
					while (j < y) { p2[c2] = x; n2[c2] = h2[y]; h2[y] = c2; c2++; y--; }
					while (j > y) { y++; p1[c1] = x; n1[c1] = h1[y]; h1[y] = c1; c1++; }
				}
			}
		}

		for(y=miny;y<=maxy;y++)
			for(y1=h1[y],y2=h2[y];y1!=0;y1=n1[y1],y2=n2[y2])
			{
				d1 = y1; d2 = y2;
				for(i=n1[y1],j=n2[y2];i!=0;i=n1[i],j=n2[j])
				{
					if (p1[i] < p1[d1]) d1 = i;
					if (p2[j] < p2[d2]) d2 = j;
				}
				if (p1[d1] < p2[d2]) diaghline((long)p1[d1],(long)p2[d2]-1,y);
				p1[d1] = p1[y1]; p2[d2] = p2[y2];
			}
	}
	else
	{
		for(z=pstart;z<=pend;z++) sc[z].y += 65535;

		miny = sc[pstart].y; maxy = sc[pstart].y;
		for(z=pstart+1;z<=pend;z++)
		{
			y = sc[z].y;
			if (y < miny) miny = y;
			if (y > maxy) maxy = y;
		}
		miny >>= 16; maxy >>= 16;

		qinterpolatedown16(&hplc[miny],maxy-miny+1,(((long)hplc[0])<<16)+gm*miny+32768,gm);

		if (gm != 0)
		{
			if (gm < 0) i = 1; else i = -1;
			for(y=miny;y<maxy;y++)
			{
				hinc[y] = XDIM, hind[y] = 0;
				if (hplc[y] != hplc[y+1]) hinc[y] += i, hind[y] = 1;
			}
			hinc[maxy] = 0;
		}
		else
		{
			for(y=miny;y<=maxy;y++)
				hinc[y] = XDIM, hind[y] = 0;
		}

		minx = 0x7fffffff; maxx = 0x80000000;
		for(z=pstart;z<=pend;z++)
		{
			x = (sc[z].x>>16) + hplc[sc[z].y>>16];
			if (x < minx) minx = x;
			if (x > maxx) maxx = x;
		}
		minx--; maxx++;

		c1 = 1; clearbuf(&h1[minx],((maxx+1-minx)+1)>>1,0L);
		c2 = 1; clearbuf(&h2[minx],((maxx+1-minx)+1)>>1,0L);

		for(z=pstart;z<=pend;z++)
		{
			zz = point2[z];
			x1 = sc[z].x; y1 = sc[z].y; x2 = sc[zz].x; y2 = sc[zz].y;

			if (((y1^y2)&0xffff0000) == 0)
			{
				y = (y1>>16);
				x = (x1>>16) + hplc[y];
				j = (x2>>16) + hplc[y];
				while (j < x) { p1[c1] = y; n1[c1] = h1[x]; h1[x] = c1; c1++; x--; }
				while (j > x) { x++; p2[c2] = y; n2[c2] = h2[x]; h2[x] = c2; c2++; }
			}
			else
			{
				if ((klabs(x2-x1)>>15) < klabs(y2-y1))
					inc = divscale16(x2-x1,y2-y1);
				else
					inc = 0;

				d1 = (y1>>16); d2 = (y2>>16);
				if (y1 < y2)
				{
					plc = x1 + mulscale16((-y1)&65535,inc);
					x = (x1>>16) + hplc[d1];
					for(y=d1;y<d2;y++)
					{
						j = (plc>>16) + hplc[y]; plc += inc;
						while (j < x) { p1[c1] = y; n1[c1] = h1[x]; h1[x] = c1; c1++; x--; }
						while (j > x) { x++; p2[c2] = y; n2[c2] = h2[x]; h2[x] = c2; c2++; }
					}
					j = (x2>>16) + hplc[d2];
					while (j < x) { p1[c1] = y; n1[c1] = h1[x]; h1[x] = c1; c1++; x--; }
					while (j > x) { x++; p2[c2] = y; n2[c2] = h2[x]; h2[x] = c2; c2++; }
				}
				else
				{
					plc = x2 + mulscale16((-y2)&65535,inc);
					x = (x2>>16) + hplc[d2];
					for(y=d2;y<d1;y++)
					{
						j = (plc>>16) + hplc[y]; plc += inc;
						while (j < x) { p2[c2] = y; n2[c2] = h2[x]; h2[x] = c2; c2++; x--; }
						while (j > x) { x++; p1[c1] = y; n1[c1] = h1[x]; h1[x] = c1; c1++; }
					}
					j = (x1>>16) + hplc[d1];
					while (j < x) { p2[c2] = y; n2[c2] = h2[x]; h2[x] = c2; c2++; x--; }
					while (j > x) { x++; p1[c1] = y; n1[c1] = h1[x]; h1[x] = c1; c1++; }
				}
			}
		}

		for(x=minx;x<=maxx;x++)
			for(x1=h1[x],x2=h2[x];x1!=0;x1=n1[x1],x2=n2[x2])
			{
				d1 = x1; d2 = x2;
				for(i=n1[x1],j=n2[x2];i!=0;i=n1[i],j=n2[j])
				{
					if (p1[i] < p1[d1]) d1 = i;
					if (p2[j] < p2[d2]) d2 = j;
				}
				if (p2[d2] < p1[d1]) diagvline((long)p2[d2],(long)p1[d1]-1,x);
				p1[d1] = p1[x1]; p2[d2] = p2[x2];
			}
	}
}

diaghline (long sx, long end, long sy)
{
	long i, p, d, t, ui[2], vi[2], pinc[2];
	unsigned long u, v;
	char *dapal;

	d = gd + (sy-hplc[0])*gdy; if (d <= 512) return; d = divscale(65536,d,24);

	dapal = (char *)(palookup[0]+(max(min(gshade+mulscale(gt,d,29)+((keystatus[0x20]<<2)&sy),numpalookups-1),0)<<8));
	sy -= hplc[sx]; p = ylookup[sy]+sx+frameplace;
	if (sx < end)
	{
		if (((gm == 0) || (mulscale(gt,d,29) == 0) || (keystatus[0xb8] != 0)) && (keystatus[0x36] == 0))
		{
			sy = ((sy+hplc[sx]-hplc[0])<<16)-sx*gm;
			u = mulscale8(gu+sx*gux+mulscale16(sy,guy),d);
			v = mulscale8(gv+sx*gvx+mulscale16(sy,gvy),d);
			ui[0] = mulscale8(gxoffs,d);
			vi[0] = mulscale8(gyoffs,d);
			do
			{
				drawpixel(p,dapal[gbuf[((u>>20)&0xfc0)+(v>>26)]]);
				u += ui[0]; v += vi[0]; p += hinc[sx];
			} while (++sx < end);
		}
		else
		{
			u = mulscale8(gu+sx*gux+sy*guy,d);
			v = mulscale8(gv+sx*gvx+sy*gvy,d);
			ui[0] = mulscale8(gux,d); vi[0] = mulscale8(gvx,d);
			ui[1] = mulscale8(guy,d); vi[1] = mulscale8(gvy,d);
			pinc[0] = 1;
			if (gm < 0) ui[1] = ui[0]+ui[1], vi[1] = vi[0]+vi[1], pinc[1] = 321;
					 else ui[1] = ui[0]-ui[1], vi[1] = vi[0]-vi[1], pinc[1] = -319;
			do
			{
				drawpixel(p,dapal[gbuf[((u>>20)&0xfc0)+(v>>26)]]);
				i = hind[sx]; u += ui[i]; v += vi[i]; p += pinc[i];
			} while (++sx < end);
		}
	}
	else
	{
		u = mulscale8(gu+sx*gux+sy*guy,d);
		v = mulscale8(gv+sx*gvx+sy*gvy,d);
	}
	drawpixel(p,dapal[gbuf[((u>>20)&0xfc0)+(v>>26)]]);
}

diagvline (long sy, long end, long sx)
{
	long i, p, d, t, ui[2], vi[2], pinc[2];
	unsigned long u, v;
	char *dapal;

	d = gd + (sx-hplc[0])*gdx; if (d <= 512) return; d = divscale(65536,d,24);

	dapal = (char *)(palookup[0]+(max(min(gshade+mulscale(gt,d,29)+((keystatus[0x20]<<2)&sx),numpalookups-1),0)<<8));
	sx -= hplc[sy]; p = ylookup[sy]+sx+frameplace;
	if (sy < end)
	{
		if (((gm == 0) || (mulscale(gt,d,29) == 0) || (keystatus[0xb8] != 0)) && (keystatus[0x36] == 0))
		{
			sx = ((sx+hplc[sy]-hplc[0])<<16)-sy*gm;
			u = mulscale8(gu+mulscale16(sx,gux)+sy*guy,d);
			v = mulscale8(gv+mulscale16(sx,gvx)+sy*gvy,d);
			ui[0] = mulscale8(gxoffs,d); vi[0] = mulscale8(gyoffs,d);
			do
			{
				drawpixel(p,dapal[gbuf[((u>>20)&0xfc0)+(v>>26)]]);
				u += ui[0]; v += vi[0]; p += hinc[sy];
			} while (++sy < end);
		}
		else
		{
			u = mulscale8(gu+sx*gux+sy*guy,d);
			v = mulscale8(gv+sx*gvx+sy*gvy,d);
			ui[0] = mulscale8(guy,d); vi[0] = mulscale8(gvy,d);
			ui[1] = mulscale8(gux,d); vi[1] = mulscale8(gvx,d);
			pinc[0] = XDIM;
			if (gm < 0) ui[1] = ui[0]+ui[1], vi[1] = vi[0]+vi[1], pinc[1] = XDIM+1;
					 else ui[1] = ui[0]-ui[1], vi[1] = vi[0]-vi[1], pinc[1] = XDIM-1;
			do
			{
				drawpixel(p,dapal[gbuf[((u>>20)&0xfc0)+(v>>26)]]);
				i = hind[sy]; u += ui[i]; v += vi[i]; p += pinc[i];
			} while (++sy < end);
		}
	}
	else
	{
		u = mulscale8(gu+sx*gux+sy*guy,d);
		v = mulscale8(gv+sx*gvx+sy*gvy,d);
	}
	drawpixel(p,dapal[gbuf[((u>>20)&0xfc0)+(v>>26)]]);
}

addpoly()
{
	long z, p, pcnt;

	for(z=0;z<numpoints;z++) splittag[z] = -1;

	if (pnum == 0)
	{
		polytag[0] = -2;
		back[0] = 0;
		add2bsp(0L);
		return;
	}
	polytag[0] = 0; pcnt = 0;
	while (pcnt < numpolys)
	{
		p = polytag[pcnt];
		if ((p >= 0) && (poly[pcnt+1] >= poly[pcnt]+3))
		{
			polytag[numpolys+0] = front[p];
			polytag[numpolys+1] = back[p];
			splitpoly2(p,pcnt,bp[p].x,bp[p].y,bn[p].x,bn[p].y);
			if ((back[p] == -2) && (poly[numpolys] >= poly[numpolys-1]+3))
				back[p] = add2bsp(numpolys-1L);
		}
		pcnt++;
	}
}

add2bsp (long polynum)
{
	long z, zz, dap, i, pcnt, plistnum, daparent;
	long bx1, by1, bx2, by2, t1, t2, fz, bz;

	plistnum = 0;
	for(z=poly[polynum];z<poly[polynum+1];z++)
		if (splittag[z] < 0) plist[plistnum++] = z;
	if (plistnum == 0) return(-1);

	daparent = pnum; pcnt = pnum+1;

	for(dap=plistnum-1;dap>=0;dap--)
	{
		i = 0; //mulscale16(krand(),dap+1);
		z = plist[i]; plist[i] = plist[dap];
		zz = point2[z];
		bx1 = sc[z].x; bx2 = sc[zz].x; bp[pnum].x = bx1; bn[pnum].x = bx2-bx1;
		by1 = sc[z].y; by2 = sc[zz].y; bp[pnum].y = by1; bn[pnum].y = by2-by1;
		front[pnum] = -1; back[pnum] = -2; pt[pnum] = daparent;
		pnum++;

		while (pcnt < pnum)
		{
			z = pt[pcnt];
			do
			{
				t1 = dmulscale18(bn[z].x,by1-bp[z].y,-bn[z].y,bx1-bp[z].x);
				t2 = dmulscale18(bn[z].x,by2-bp[z].y,-bn[z].y,bx2-bp[z].x);

				if ((t1 >= 0) && (t2 >= 0))
				{
					if (front[z] >= 0) z = front[z];
					else { front[z] = pcnt; break; }
				}
				else if ((t1 <= 0) && (t2 <= 0))
				{
					if (back[z] >= 0) z = back[z];
					else { back[z] = pcnt; break; }
				}
				else
				{
					copybuf(&bp[pcnt],&bp[pnum],2);
					copybuf(&bn[pcnt],&bn[pnum],2);
					front[pnum] = -1; back[pnum] = -2;

					fz = front[z]; bz = back[z];
					if (t1 > 0)
					{
						if (bz >= 0) pt[pnum] = bz; else { back[z] = pnum; pt[pnum] = -1; }
						pnum++;
						if (fz >= 0) z = fz; else { front[z] = pcnt; break; }
					}
					else
					{
						if (fz >= 0) pt[pnum] = fz; else { front[z] = pnum; pt[pnum] = -1; }
						pnum++;
						if (bz >= 0) z = bz; else { back[z] = pcnt; break; }
					}
				}
			} while (z >= 0);
			pcnt++;
		}
	}
	return(daparent);
}

splitpoly2 (long p, long polynum, long x1, long y1, long dx, long dy)
{
	long pstart, pend, numpoints2, start2, start22, s1, s2, s3, s4;
	long i, z, zsgn, zz, zzsgn, splitcnt, splitcnt2, intx, inty;

	pstart = poly[polynum]; pend = poly[polynum+1]-1;

	i = dx; dx = dy; dy = -i;

	numpoints2 = MAXPOINTS-1;
	start2 = numpoints; start22 = numpoints2;
	z = pstart; splitcnt = 0; splitcnt2 = MAXSPLITCNT-1;
	do
	{
		zzsgn = dmulscale18(sc[z].x-x1,dx,sc[z].y-y1,dy);
		do
		{
			zz = point2[z]; point2[z] |= 0x8000;

			zsgn = zzsgn;
			zzsgn = dmulscale18(sc[zz].x-x1,dx,sc[zz].y-y1,dy);

			if (zsgn < 0)
			{
				sc[numpoints].x = sc[z].x;
				sc[numpoints].y = sc[z].y;
				splittag[numpoints] = splittag[z];
				point2[numpoints] = numpoints+1;
				numpoints++;
			}
			else if (zsgn > 0)
			{
				sc[numpoints2].x = sc[z].x;
				sc[numpoints2].y = sc[z].y;
				splittag[numpoints2] = splittag[z];
				point2[numpoints2] = numpoints2-1;
				numpoints2--;
			}
			if (((zsgn^zzsgn) < 0) || (((-zsgn)^(-zzsgn)) < 0))
			{
				intx = sc[z].x + scale(sc[zz].x-sc[z].x,zsgn,zsgn-zzsgn);
				inty = sc[z].y + scale(sc[zz].y-sc[z].y,zsgn,zsgn-zzsgn);
				if ((zsgn^zzsgn) < 0)
				{
					sc[numpoints].x = intx;
					sc[numpoints].y = inty;
					point2[numpoints] = numpoints+1;
					if (zsgn < 0)
						{ split[splitcnt++] = numpoints; splittag[numpoints] = p; }
					else
						splittag[numpoints] = splittag[z];
					numpoints++;
				}
				if (((-zsgn)^(-zzsgn)) < 0)
				{
					sc[numpoints2].x = intx;
					sc[numpoints2].y = inty;
					point2[numpoints2] = numpoints2-1;
					if (zsgn > 0)
						{ split[splitcnt2--] = numpoints2; splittag[numpoints2] = p; }
					else
						splittag[numpoints2] = splittag[z];
					numpoints2--;
				}
			}
			z = zz;
		} while (point2[z] >= 0);

		if (numpoints >= start2+3)
			{ point2[numpoints-1] = start2; start2 = numpoints; }
		else
			numpoints = start2;

		if (numpoints2 <= start22-3)
			{ point2[numpoints2+1] = start22; start22 = numpoints2; }
		else
			numpoints2 = start22;

		z = pend;
		while ((z > pstart) && (point2[z] < 0)) z--;
	} while (z > pstart);

	if (klabs(dx) >= klabs(dy))
	{
		for(z=1;z<splitcnt;z++)
			for(zz=0;zz<z;zz++)
			{
				s1 = split[z]; s2 = point2[s1]; s3 = split[zz]; s4 = point2[s3];
				if (klabs(sc[s1].x-sc[s4].x)+klabs(sc[s3].x-sc[s2].x) < klabs(sc[s1].x-sc[s2].x) + klabs(sc[s3].x-sc[s4].x))
					{ i = point2[s1]; point2[s1] = point2[s3]; point2[s3] = i; }
			}
		for(z=1;z<MAXSPLITCNT-splitcnt2-1;z++)
			for(zz=0;zz<z;zz++)
			{
				s1 = split[MAXSPLITCNT-1-z]; s2 = point2[s1]; s3 = split[MAXSPLITCNT-1-zz]; s4 = point2[s2];
				if (klabs(sc[s1].x-sc[s4].x)+klabs(sc[s3].x-sc[s2].x) < klabs(sc[s1].x-sc[s2].x) + klabs(sc[s3].x-sc[s4].x))
					{ i = point2[s1]; point2[s1] = point2[s3]; point2[s3] = i; }
			}
	}
	else
	{
		for(z=1;z<splitcnt;z++)
			for(zz=0;zz<z;zz++)
			{
				s1 = split[z]; s2 = point2[s1]; s3 = split[zz]; s4 = point2[s3];
				if (klabs(sc[s1].y-sc[s4].y)+klabs(sc[s3].y-sc[s2].y) < klabs(sc[s1].y-sc[s2].y) + klabs(sc[s3].y-sc[s4].y))
					{ i = point2[s1]; point2[s1] = point2[s3]; point2[s3] = i; }
			}
		for(z=1;z<MAXSPLITCNT-splitcnt2-1;z++)
			for(zz=0;zz<z;zz++)
			{
				s1 = split[MAXSPLITCNT-1-z]; s2 = point2[s1]; s3 = split[MAXSPLITCNT-1-zz]; s4 = point2[s2];
				if (klabs(sc[s1].y-sc[s4].y)+klabs(sc[s3].y-sc[s2].y) < klabs(sc[s1].y-sc[s2].y) + klabs(sc[s3].y-sc[s4].y))
					{ i = point2[s1]; point2[s1] = point2[s3]; point2[s3] = i; }
			}
	}

	for(z=pstart;z<=pend;z++) point2[z] &= 0x7fff;

	numpolys++; poly[numpolys] = numpoints;

	zz = (MAXPOINTS-1) + numpoints;
	for(z=MAXPOINTS-1;z>numpoints2;z--)
	{
		sc[numpoints].x = sc[z].x;
		sc[numpoints].y = sc[z].y;
		splittag[numpoints] = splittag[z];
		point2[numpoints] = zz-point2[z];
		numpoints++;
	}
	numpolys++; poly[numpolys] = numpoints;
}

krand()
{
	randomseed = ((randomseed*21+1)&65535);
	return(randomseed);
}

loadnewall (long bufplc, long wallnum)
{
	long fil, version, numtiles, localtilestart, localtileend, totsiz, z;

	if ((fil = open("tiles000.art",O_BINARY|O_RDONLY,S_IREAD)) == -1) return;

	read(fil,&version,4);
	read(fil,&numtiles,4);
	read(fil,&localtilestart,4);
	read(fil,&localtileend,4);
	read(fil,&tilesizx[0],(localtileend-localtilestart+1)<<1);
	read(fil,&tilesizy[0],(localtileend-localtilestart+1)<<1);
	read(fil,&picanm[0],(localtileend-localtilestart+1)<<2);

	totsiz = 0;
	for(z=0;z<wallnum;z++) totsiz += tilesizx[z]*tilesizy[z];
	lseek(fil,totsiz+16+((localtileend-localtilestart+1)<<3),SEEK_SET);
	read(fil,&buf[bufplc],tilesizx[wallnum]*tilesizy[wallnum]);

	close(fil);
}

static long wx1 = (0<<12);
static long wy1 = (0<<12);
static long wx2 = (XDIM<<12);
static long wy2 = (YDIM<<12);

drawline256 (long x1, long y1, long x2, long y2, char col)
{
	long dx, dy, i, j, p, inc, plc, daend;

	//col = palookup[0][col];


	x1 >>= 4; y1 >>= 4; x2 >>= 4; y2 >>= 4;


	dx = x2-x1; dy = y2-y1;
	if (dx >= 0)
	{
		if ((x1 >= wx2) || (x2 < wx1)) return;
		if (x1 < wx1) y1 += scale(wx1-x1,dy,dx), x1 = wx1;
		if (x2 > wx2) y2 += scale(wx2-x2,dy,dx), x2 = wx2;
	}
	else
	{
		if ((x2 >= wx2) || (x1 < wx1)) return;
		if (x2 < wx1) y2 += scale(wx1-x2,dy,dx), x2 = wx1;
		if (x1 > wx2) y1 += scale(wx2-x1,dy,dx), x1 = wx2;
	}
	if (dy >= 0)
	{
		if ((y1 >= wy2) || (y2 < wy1)) return;
		if (y1 < wy1) x1 += scale(wy1-y1,dx,dy), y1 = wy1;
		if (y2 > wy2) x2 += scale(wy2-y2,dx,dy), y2 = wy2;
	}
	else
	{
		if ((y2 >= wy2) || (y1 < wy1)) return;
		if (y2 < wy1) x2 += scale(wy1-y2,dx,dy), y2 = wy1;
		if (y1 > wy2) x1 += scale(wy2-y1,dx,dy), y1 = wy2;
	}

	if (klabs(dx) >= klabs(dy))
	{
		if (dx == 0) return;
		if (dx < 0)
		{
			i = x1; x1 = x2; x2 = i;
			i = y1; y1 = y2; y2 = i;
		}

		inc = divscale12(dy,dx);
		plc = y1+mulscale12((2047-x1)&4095,inc);
		i = ((x1+2048)>>12); daend = ((x2+2048)>>12);
		for(;i<daend;i++)
		{
			drawpixel(ylookup[plc>>12]+i+frameplace,col);
			plc += inc;
		}
	}
	else
	{
		if (dy < 0)
		{
			i = x1; x1 = x2; x2 = i;
			i = y1; y1 = y2; y2 = i;
		}

		inc = divscale12(dx,dy);
		plc = x1+mulscale12((2047-y1)&4095,inc);
		i = ((y1+2048)>>12); daend = ((y2+2048)>>12);
		p = ylookup[i]+frameplace;
		for(;i<daend;i++)
		{
			drawpixel((plc>>12)+p,col);
			plc += inc; p += XDIM;
		}
	}
}

printext256(long xpos, long ypos, short col, short backcol, char name[82], char fontsize)
{
	long stx, zx, stupidbitoffset, x, y, charxsiz;
	short dat, dacol;
	char *fontptr;

	stx = xpos;

	if (fontsize == 1)
	{
		fontptr = smalltextfont;
		charxsiz = 4;
		stupidbitoffset = 1;
	}
	else
	{
		fontptr = textfont;
		charxsiz = 8;
		stupidbitoffset = 0;
	}

	for(zx=0;zx<strlen(name);zx++)
	{
		dat = name[zx];

		for(y=0;y<8;y++)
			for(x=0;x<charxsiz;x++)
			{
				if ((fontptr[(dat<<3)+y]&(1<<(7-stupidbitoffset-x))) > 0)
					dacol = (short)col;
				else if (backcol >= 0)
					dacol = (short)backcol;
				else
					dacol = -1;

				if (dacol >= 0)
					drawpixel(ylookup[y+ypos]+(x+stx-stupidbitoffset)+frameplace,(long)dacol);
			}

		stx += charxsiz;
	}
}
