// "Build Engine & Tools" Copyright (c) 1993-1997 Ken Silverman
// Ken Silverman's official web site: "http://www.advsys.net/ken"
// See the included license file "BUILDLIC.TXT" for license info.

#include <string.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <conio.h>

#define maxsector 1024
#define maxwalls 4096
#define maxtiles 4096
#define numpalookups 32
#define numtimecounts 4
#define xdim 320
#define ydim 200
#define texttilenum 85

static char timenames[numtimecounts][20] =
{
	"Calculations       ",
	"Vline drawing      ",
	"Hline drawing      ",
	"Screen copying     ",
};

static struct sectortype
{
	short unsigned int wallptr, wallnum;
	char cstat1, zosc1, shade1;             //Parallax, Groudraw, Clip XOR
	short z1, picnum1, heinum1;
	char cstat2, zosc2, shade2;             //Parallax, Groudraw, Clip XOR
	short z2, picnum2, heinum2;
};

static struct walltype
{
	long xa1, ya1;
	short point2;
	char cstat, shade;
	unsigned char xrepeat, yrepeat;
	short picnum;
	short nextsector1, nextwall1;
	short nextsector2, nextwall2;
};

struct sectortype sector[maxsector];
struct walltype wall[maxwalls];

static unsigned char tempbuf[maxwalls];

static char picsiz[maxtiles], picanm[maxtiles];
static long waloff[maxtiles];
static char *pic;

static short sintable[2048];
static long tantable[1024];
static short radarang[320];

static long xb1[maxwalls], yb1[maxwalls], xb2[maxwalls], yb2[maxwalls];
static short thesector[maxwalls], thewall[maxwalls];
static short llinum[maxwalls], rlinum[maxwalls];

static short umost[xdim], dmost[xdim], startumost[xdim], startdmost[xdim];
static long h[xdim], uplc[xdim], dplc[xdim], linum[xdim];

static char got[maxsector], screen[64000], visibility, tablesloaded = 0;
static long lastx[ydim+2];
static char palette[768], palookup[numpalookups][256];

static unsigned char clipbuf[128][128];

static short numsectors, numwalls, numscans, cursectnum, numhits;
static long posx, posy, posz;
static short ang, posfil, capturecount;
static long poscapture = 0L;
static long vel, svel, hvel, angvel, nowx, nowy;
static short int mousx, mousy, moustat, bstatus, obstatus;

static short searchit, editstatus, asksave;
static long searchx, searchy;                          //search input
static short searchsector, searchwall, searchstat;     //search output
static short temppicnum;
static char tempshade, somethingintab;
static unsigned short grid, anglelock;
static short linehighlight;
static long zoom;

static long numframes, frameplace, clockspeed, totalclock;
static short page, videocard;

volatile long timetakercount[numtimecounts], takernum;
volatile unsigned char keystatus[256], readch, oldreadch, extended;

static void (__interrupt __far *oldtimerhandler)();
static void __interrupt __far timerhandler(void);
static void (__interrupt __far *oldkeyhandler)();
static void __interrupt __far keyhandler(void);

#pragma aux int5 =\
	"int 0x5",\

#pragma aux setvmode =\
	"int 0x10",\
	parm [eax]\

#pragma aux setupmouse =\
	"mov ax, 0",\
	"int 33h",\
	"mov moustat,1",\
	modify [eax]\

#pragma aux readmouse =\
	"mov ax, 11d",\
	"int 33h",\
	"sar cx, 1",\
	"sar dx, 1",\
	"mov mousx, cx",\
	"mov mousy, dx",\
	"mov ax, 5d",\
	"int 33h",\
	"mov bx, bstatus",\
	"mov obstatus, bx",\
	"mov bstatus, ax",\
	modify [eax ecx edx]\

#pragma aux getpixel =\
	"xor eax, eax",\
	"mov al, [edi]",\
	parm [edi]\

#pragma aux drawpixel =\
	"mov byte ptr [edi], al",\
	parm [edi][eax]\

#pragma aux drawpixels =\
	"mov word ptr [edi], ax",\
	parm [edi][eax]\

#pragma aux drawpixeles =\
	"mov dword ptr [edi], eax",\
	parm [edi][eax]\

#pragma aux drawpixel640400 =\
	"mov ah, al",\
	"mov cx, di",\
	"mov dx, 0x3c5",\
	"mov al, 0x11",\
	"rol al, cl",\
	"out dx, al",\
	"shr edi, 2",\
	"mov [edi+0xa0000], ah",\
	parm [edi][eax]\
	modify [eax ecx edx edi]\

#pragma aux limitrate =\
	"mov dx, 0x3da",\
	"wait1: in al, dx",\
	"test al, 8",\
	"jnz wait1",\
	"wait2: in al, dx",\
	"test al, 8",\
	"jz wait2",\

#pragma aux fillscreen =\
	"mov ah, al",\
	"mov cx, ax",\
	"shl ecx, 16",\
	"or eax, ecx",\
	"mov edi, 0xa0000",\
	"mov ecx, 16384",\
	"rep stosd",\
	parm [eax]\
	modify [eax ecx edi]\

#pragma aux clearseg =\
	"mov edi, 0xa0000",\
	"mov ecx, 16384",\
	"xor eax, eax",\
	"rep stosd",\
	modify [eax edi ecx]\

#pragma aux checkvesa =\
	"mov ax, 0x4f00",\
	"int 0x10",\
	"cmp ax, 0x004f",\
	"sete al",\
	"and eax, 0x000000ff",\
	modify [eax]\

#pragma aux vesasetmode =\
	"mov ax, 0x4f02",\
	"int 0x10",\
	parm [ebx]\
	modify [eax ebx]\

#pragma aux vesasetpage =\
	"mov ax, 0x4f05",\
	"xor bx, bx",\
	"int 0x10",\
	parm [edx]\
	modify [eax ebx]\

#pragma aux copytovideo =\
	"cld",\
	"mov esi, frameplace",\
	"mov edi, 0xa0000",\
	"mov ecx, 16000",\
	"rep movsd",\
	modify [esi edi ecx]\

#pragma aux videomove =\
	"cld",\
	"mov esi, 0xa0000",\
	"mov edi, 0xa0000",\
	"mov ecx, 16000",\
	"rep movsd",\
	modify [esi edi ecx]\

#pragma aux interruptend =\
	"mov al, 20h",\
	"out 20h, al",\
	modify [eax]\

#pragma aux readkey =\
	"in al, 0x60",\
	"mov readch, al",\
	"in al, 0x61",\
	"or al, 0x80",\
	"out 0x61, al",\
	"and al, 0x7f",\
	"out 0x61, al",\
	modify [eax]\

#pragma aux clearbuf =\
	"rep stosd",\
	parm [edi][ecx][eax]\

#pragma aux memcopybyte =\
	"mov al, byte ptr [ecx]",\
	"add al, dl",\
	"mov byte ptr [ebx], al",\
	parm [ebx][ecx][edx]\

#pragma aux readrow =\
	"mov ecx, 80d",\
	"add edi, offset tempbuf",\
	"rep movsd",\
	parm [esi][edi]\
	modify [ecx edi]\

#pragma aux drawrow =\
	"cld",\
	"mov ecx, 80d",\
	"mov esi, offset tempbuf",\
	"rep movsd",\
	parm [edi]\
	modify [edi esi ecx]\

#pragma aux gethline64 =\
	"xor ebx, ebx",\
	"shld ebx, eax, 6",\
	"shld ebx, edx, 6",\
	"mov bl, byte ptr [esi+ebx]",\
	"mov bh, cl",\
	"xor eax, eax",\
	"mov al, byte ptr palookup[ebx]",\
	parm [eax][ecx][edx][esi]\
	modify [ebx]\

#pragma aux plotvline =\
	"shr eax, 24",\
	"and eax, ebx",\
	"xor ebx, ebx",\
	"mov bl, byte ptr [esi+eax]",\
	"mov bh, cl",\
	"mov al, byte ptr palookup[ebx]",\
	"mov byte ptr [edi], al",\
	parm [eax][ebx][ecx][esi][edi]\

void __interrupt __far timerhandler()
{
	clockspeed++;
	timetakercount[takernum]++;
	interruptend();
}

void __interrupt __far keyhandler()
{
	oldreadch = readch;
	readkey();
	if ((readch|1) == 0xe1)
		extended = 128;
	else
	{
		if (oldreadch != readch)
			keystatus[(readch&127)+extended] = ((readch>>7)^1);
		extended = 0;
	}
	interruptend();
}

main(short int argc,char **argv)
{
	char boardfilename[80], ch;
	long i;

	if (argc >= 2)
	{
		strcpy(&boardfilename,argv[1]);
		if(strchr(boardfilename,'.') == 0)
			strcat(boardfilename,".dat");
	}
	else
		strcpy(&boardfilename,"secboard.dat");

	loadpics("pics.dat");
	if (loadboard(boardfilename) == -1)
	{
		printf("Board not found\n");
		//Ask if want to create new board here! - don't quit!
		return(-1);
	}
	initengine();
	setgamemode();

	//loadpcx("captur00.pcx",frameplace);
	for (i=0;i<xdim;i++)                      //define 3d picture boundaries
	{
		startumost[i] = 0;
		startdmost[i] = ydim;
	}

	while (keystatus[1] == 0)
	{
		draw3dscreen();
		//printext(8L, 0L, "BUILD! - programmed by Ken Silverman.");

		if (editstatus > 0)
			editinput();
		engineinput();

		nextpage();
	}
	uninitengine();
	setvmode(0x3);

	if (asksave == 0)
		showengineinfo();
	else
	{
		printf("Save changes?\n");
		do
		{
			ch = getch();
		}
		while ((ch != 'y') && (ch != 'Y') && (ch != 'n') && (ch != 'N'));
		if ((ch == 'y') || (ch == 'Y'))
			saveboard(boardfilename);
	}
	return(0);
}

draw3dscreen()
{
	short i, j, k, tempshort, closest;
	long templong;

	for(i=0;i<numsectors;i++)
		got[i] = 0;

	for(i=0;i<xdim;i++)
	{
		umost[i] = startumost[i];
		dmost[i] = startdmost[i];
	}

	getsector();

	numhits = xdim;
	numscans = 0;
	scansector(cursectnum);
	while (numscans > 0)
	{
		tempbuf[0] = 1;
		for(i=1;i<numscans;i++)
			tempbuf[i] = 0;

		closest = 0;
		for(i=0;i<numscans;i++)
			if (tempbuf[i] == 0)
				if ((xb1[i] < xb2[closest]) && (xb1[closest] < xb2[i]))
					if (front(i,closest) == i)
					{
						tempbuf[i] = 1;
						closest = i;
						i = 0;
					}

		drawall(closest);
		if (numhits < 0)
			return(0);
	}
	return(0);
}

scansector (short sectnum)
{
	short z, startwall, endwall;

	got[sectnum] = 1;
	startwall = sector[sectnum].wallptr;
	endwall = startwall + sector[sectnum].wallnum - 1;
	for(z=startwall;z<=endwall;z++)
		addboxlinum(sectnum, z);
	return(0);
}

front (short l1, short l2)
{
	long x1l1, y1l1, x2l1, y2l1, x1l2, y1l2, x2l2, y2l2;
	long templong0, templong1, templong2, templong3, templong4;

	x1l1 = wall[thewall[l1]].xa1;
	y1l1 = wall[thewall[l1]].ya1;
	x2l1 = wall[wall[thewall[l1]].point2].xa1;
	y2l1 = wall[wall[thewall[l1]].point2].ya1;
	x1l2 = wall[thewall[l2]].xa1;
	y1l2 = wall[thewall[l2]].ya1;
	x2l2 = wall[wall[thewall[l2]].point2].xa1;
	y2l2 = wall[wall[thewall[l2]].point2].ya1;

	templong1 = (x1l2-x1l1)*(y2l1-y1l1) - (x2l1-x1l1)*(y1l2-y1l1); //p1(l2) vs. l1
	templong2 = (x2l2-x1l1)*(y2l1-y1l1) - (x2l1-x1l1)*(y2l2-y1l1); //p2(l2) vs. l1
	if ((templong1|templong2) == 0) return(-1);
	if (templong1 == 0) templong1 = templong2;
	if (templong2 == 0) templong2 = templong1;
	if ((templong1 < 0) == (templong2 < 0))
	{
		templong0 = (posx-x1l1)*(y2l1-y1l1) - (x2l1-x1l1)*(posy-y1l1);  //pos vs. l1
		if ((templong1<0) == (templong0<0))
			return(l2);
		else
			return(l1);
	}

	templong1 = (x1l1-x1l2)*(y2l2-y1l2) - (x2l2-x1l2)*(y1l1-y1l2); //p1(l1) vs. l2
	templong2 = (x2l1-x1l2)*(y2l2-y1l2) - (x2l2-x1l2)*(y2l1-y1l2); //p2(l1) vs. l2
	if ((templong1|templong2) == 0) return(-1);
	if (templong1 == 0) templong1 = templong2;
	if (templong2 == 0) templong2 = templong1;
	if ((templong1 < 0) == (templong2 < 0))
	{
		templong0 = (posx-x1l2)*(y2l2-y1l2) - (x2l2-x1l2)*(posy-y1l2);  //pos vs. l2
		if ((templong1<0) == (templong0<0))
			return(l1);
		else
			return(l2);
	}
	return(-2);
}

drawall (short z)
{
	short wallnum, sectnum, nextsectnum, tempshort;
	long i, j, top, topinc, bot, botinc, x, y, templong;

	wallnum = thewall[z];
	sectnum = thesector[z];
	nextsectnum = wall[wallnum].nextsector1;

	top = 0; topinc = (rlinum[z]-llinum[z])*yb1[z];
	bot = (xb2[z]-xb1[z])*yb2[z]; botinc = yb1[z]-yb2[z];

	yb1[z] = (2097152*64)/yb1[z];
	yb2[z] = (2097152*64)/yb2[z];
	if (yb1[z] > 4194304) yb1[z] = 4194304;
	if (yb2[z] > 4194304) yb2[z] = 4194304;

	for(x=xb1[z];x<=xb2[z];x++)
	{
		if (xb1[z] != xb2[z])
		{
			h[x] = (yb1[z]*(xb2[z]-x) + yb2[z]*(x-xb1[z]))/(xb2[z]-xb1[z]);
		}
		else
		{
			h[x] = yb1[z];
		}
		uplc[x] = (ydim>>1) + ((h[x]*(sector[sectnum].z1-(posz>>8)))>>16);
		dplc[x] = (ydim>>1) + ((h[x]*(sector[sectnum].z2-(posz>>8)))>>16);
		if (uplc[x] < umost[x]) uplc[x] = umost[x];
		if (uplc[x] > dmost[x]) uplc[x] = dmost[x];
		if (dplc[x] < umost[x]) dplc[x] = umost[x];
		if (dplc[x] > dmost[x]) dplc[x] = dmost[x];

		if (bot != 0) linum[x] = llinum[z] + (top/bot);
		linum[x] = ((linum[x]*wall[wallnum].xrepeat)>>3);
		top += topinc;
		bot += botinc;
		if ((top&0xc0000000) != 0)
		{
			top >>= 1;
			bot >>= 1;
			topinc >>= 1;
			botinc >>= 1;
		}
	}

	if (searchit > 0)
	{
		if ((searchx >= xb1[z]) && (searchx <= xb2[z]))
		{
			if (searchy <= uplc[searchx])
			{
				searchsector = sectnum;
				searchwall = wallnum;
				searchstat = 1;  //ceiling
				searchit = 0;
			}
			else if (searchy >= dplc[searchx])
			{
				searchsector = sectnum;
				searchwall = wallnum;
				searchstat = 2;  //floor
				searchit = 0;
			}
			else if (nextsectnum == -1)
			{
				searchsector = sectnum;
				searchwall = wallnum;
				searchstat = 0;  //wall
				searchit = 0;
			}
			else
			{
				if (sector[nextsectnum].z1 > sector[sectnum].z1)
				{
					y = (ydim>>1) + ((h[searchx]*(sector[nextsectnum].z1-(posz>>8)))>>16);
					if (searchy <= y)
					{
						searchsector = sectnum;
						searchwall = wallnum;
						searchstat = 0;  //wall
						searchit = 0;
					}
				}
				if (sector[nextsectnum].z2 < sector[sectnum].z2)
				{
					y = (ydim>>1) + ((h[searchx]*(sector[nextsectnum].z2-(posz>>8)))>>16);
					if (searchy >= y)
					{
						searchsector = sectnum;
						searchwall = wallnum;
						searchstat = 0;  //wall
						searchit = 0;
					}
				}
			}
		}
	}

	if ((sector[sectnum].cstat1&1) > 0)
	{
		for(j=xb1[z];j<=xb2[z];j++)
		{
			x = (long)((sintable[(radarang[j]+2048)&2047]/92) + ((ang&2047)>>1));
			y = (long)((sintable[(radarang[j]+2560)&2047]/108));
			templong = ((((long)(((20480L/y)*(xdim))))/320L)<<16);
			vline(j,(100L-((templong>>8)/320L))<<16,(uplc[j]<<16),(templong<<2),8L,(long)(x<<7),sector[sectnum].picnum1,sector[sectnum].shade1);
		}
	}
	else
		ceilscan(xb1[z],xb2[z],sectnum);

	/*if ((sector[sectnum].cstat2&1) > 0)
	{
		for(j=xb1[z];j<=xb2[z];j++)
		{
			x = (long)((sintable[(radarang[j]+2048)&2047]/92) + ((ang&2047)>>1));
			y = (long)((sintable[(radarang[j]+2560)&2047]/108));
			templong = ((((long)(((20480L/y)*(xdim))))/320L)<<16);
			vline(j,(dplc[j]<<16),(100L+((templong>>8)/320L))<<16,(templong<<2),8L,(long)(x<<7),sector[sectnum].picnum2,sector[sectnum].shade2);
		}
	}
	else*/
		florscan(xb1[z],xb2[z],sectnum);

	if (nextsectnum == -1)
	{
		for(x=xb1[z];x<=xb2[z];x++)
			if (umost[x] <= dmost[x])
			{
				uplc[x] = (ydim<<15)+(h[x]*(sector[sectnum].z1-(posz>>8)));
				dplc[x] = (ydim<<15)+(h[x]*(sector[sectnum].z2-(posz>>8)));
				vline(x,uplc[x],dplc[x],h[x]<<9,(long)wall[wallnum].yrepeat,(long)linum[x],wall[wallnum].picnum,wall[wallnum].shade);
				umost[x] = ydim;
				dmost[x] = 0;
				numhits--;
			}
	}
	else
	{
		if (sector[nextsectnum].z1 <= sector[sectnum].z1)
		{
			for(x=xb1[z];x<=xb2[z];x++)
				if (umost[x] <= dmost[x])
				{
					if (uplc[x] >= ydim)
						uplc[x] = ydim;
					if (uplc[x] > umost[x])
					{
						umost[x] = uplc[x];
						if (umost[x] > dmost[x])
							numhits--;
					}
				}
		}
		else
		{
			for(x=xb1[z];x<=xb2[z];x++)
				if (umost[x] <= dmost[x])
				{
					uplc[x] = (ydim<<15)+(h[x]*(sector[sectnum].z1-(posz>>8)));
					if (sector[sectnum].z2 < sector[nextsectnum].z1)
						y = (ydim<<15)+(h[x]*(sector[sectnum].z2-(posz>>8)));
					else
						y = (ydim<<15)+(h[x]*(sector[nextsectnum].z1-(posz>>8)));

					vline(x,uplc[x],y,h[x]<<9,(long)wall[wallnum].yrepeat,(long)linum[x],wall[wallnum].picnum,wall[wallnum].shade);

					y >>= 16;
					if (y > ydim)
						y = ydim;
					if (y > umost[x])
					{
						umost[x] = y;
						if (umost[x] > dmost[x])
							numhits--;
					}
				}
		}
		if (sector[nextsectnum].z2 >= sector[sectnum].z2)
		{
			for(x=xb1[z];x<=xb2[z];x++)
				if (umost[x] <= dmost[x])
				{
					if (dplc[x] < 0)
						dplc[x] = 0;
					if (dplc[x] < dmost[x])
					{
						dmost[x] = dplc[x];
						if (umost[x] > dmost[x])
							numhits--;
					}
				}
		}
		else
		{
			for(x=xb1[z];x<=xb2[z];x++)
				if (umost[x] <= dmost[x])
				{
					dplc[x] = (ydim<<15)+(h[x]*(sector[sectnum].z2-(posz>>8)));
					if (sector[sectnum].z1 > sector[nextsectnum].z2)
						y = (ydim<<15)+(h[x]*(sector[sectnum].z1-(posz>>8)));
					else
						y = (ydim<<15)+(h[x]*(sector[nextsectnum].z2-(posz>>8)));

					vline(x,y,dplc[x],h[x]<<9,(long)wall[wallnum].yrepeat,(long)linum[x],wall[wallnum].picnum,wall[wallnum].shade);

					y >>= 16;
					if (y < 0)
						y = 0;
					if (y < dmost[x])
					{
						dmost[x] = y;
						if (umost[x] > dmost[x])
							numhits--;
					}
				}
		}
	}
	if (numhits < 0)
		return(0);

	i = xb1[z];
	j = xb2[z];

	numscans--;
	if (z != numscans)
	{
		xb1[z] = xb1[numscans];
		yb1[z] = yb1[numscans];
		xb2[z] = xb2[numscans];
		yb2[z] = yb2[numscans];
		llinum[z] = llinum[numscans];
		rlinum[z] = rlinum[numscans];
		thesector[z] = thesector[numscans];
		thewall[z] = thewall[numscans];
	}

	if ((nextsectnum != -1) && (got[nextsectnum] == 0))
		for(x=i;x<=j;x++)
			if (umost[x] < dmost[x])
			{
				scansector(nextsectnum);
				return(0);
			}

	return(0);
}

addboxlinum(short sectore, short walle)
{
	short good;
	long x1, y1, x2, y2, xp1, yp1, xp2, yp2, xc, yc, templong, templong1;

	x1 = wall[walle].xa1 - posx;
	y1 = wall[walle].ya1 - posy;
	x2 = wall[wall[walle].point2].xa1 - posx;
	y2 = wall[wall[walle].point2].ya1 - posy;

	xp1 = ((x1*sintable[(ang+0)&2047] - y1*sintable[(ang+512)&2047])>>14);
	yp1 = ((x1*sintable[(ang+512)&2047] + y1*sintable[(ang+0)&2047])>>14);
	xp2 = ((x2*sintable[(ang+0)&2047] - y2*sintable[(ang+512)&2047])>>14);
	yp2 = ((x2*sintable[(ang+512)&2047] + y2*sintable[(ang+0)&2047])>>14);
 
	good = 7;
	if ((labs(xp1) <= yp1) && (yp1 > 0)) good--;
	if ((labs(xp2) <= yp2) && (yp2 > 0)) good -= 2;
	if ((good == 7) && ((yp1 > 0) || (yp2 > 0)) && ((xp1 < 0) == (0 < xp2)))
		good = 3;
	yb1[numscans] = -1;
	yb2[numscans] = -1;

	if (good < 7)
	{
		if ((good&3) > 0)
			templong = xp2*yp1 - xp1*yp2;
		if ((good&1) == 0)
		{
			xb1[numscans] = 160 - ((xp1*160) / yp1);
			yb1[numscans] = yp1;
			llinum[numscans] = 0;
		}
		else
		{
			if (xp1 < 0)
			{
				if ((xp1 < 0) != (0 < xp2)) return(0);
				if ((yp1 < 0) != (0 < yp2)) return(0);
				if (labs(xp1*yp2) > labs(xp2*yp1)) return(0);
			}
			xb1[numscans] = 0;
			templong1 = yp1-yp2+xp2-xp1;
			if (templong1 == 0)
				yb1[numscans] = 1;
			else
				yb1[numscans] = templong/templong1;
        
			xc = ((160-xb1[numscans])*yb1[numscans]) / 160;
			yc = yb1[numscans];
			llinum[numscans] = (4095*(labs(xc-xp1)+labs(yc-yp1))) / (labs(xp2-xp1) + labs(yp2-yp1));
		}
		if ((good&2) == 0)
		{
			xb2[numscans] = 160 - ((xp2*160) / yp2);
			yb2[numscans] = yp2;
			rlinum[numscans] = 4095;
		}
		else
		{
			if (xp2 > 0)
			{
				if ((xp1 < 0) != (0 < xp2)) return(0);
				if ((yp1 < 0) != (0 < yp2)) return(0);
				if (labs(xp1*yp2) < labs(xp2*yp1)) return(0);
			}
			xb2[numscans] = xdim-1;
			templong1 = yp2-yp1+xp2-xp1;
			if (templong1 == 0)
				yb2[numscans] = 1;
			else
				yb2[numscans] = templong/templong1;

			xc = ((160-xb2[numscans])*yb2[numscans]) / 160;
			yc = yb2[numscans];
			rlinum[numscans] = (4095*(labs(xc-xp1)+labs(yc-yp1))) / (labs(xp2-xp1) + labs(yp2-yp1));
		}
		if ((yb1[numscans] > 0) && (yb2[numscans] > 0))
		{
			if (xb1[numscans] < xb2[numscans])
			{
				thesector[numscans] = sectore;
				thewall[numscans] = walle;
				numscans++;
			}
		}
	}
	return(0);
}

ceilscan (long x1, long x2, short sectnum)
{
	long j, m, um, up, theplc;

	clearbuf((long)(&lastx[0]),(ydim+2)>>1,0xffffffff);
	um = uplc[x1];
	up = um-1;
	if (dmost[x1]-1 <= up)
	{
		up = dmost[x1]-1;
		um = up+1;
	}
	if (um < -1) um = -1;
	if (up < -1) up = -1;
	if (um > ydim) um = ydim;
	if (up > ydim) up = ydim;

	for(j=x1;j<=x2;j++)
	{
		theplc = uplc[j]-1;
		if (theplc < -1) theplc = -1;
		if (theplc > ydim) theplc = ydim;
		if (theplc >= dmost[j]) theplc = dmost[j]-1;

		while (umost[j] > um)
		{
			if (lastx[um+1] >= 0)
			{
				hline(lastx[um+1],j-1,um,sectnum);
				lastx[um+1] = -1;
			}
			um++;
		}
		while (theplc < up)
		{
			if (lastx[up+1] >= 0)
			{
				hline(lastx[up+1],j-1,up,sectnum);
				lastx[up+1] = -1;
			}
			up--;
		}
		if (umost[j] < um)
		{
			if (um > up+1) um = up+1;
			while (umost[j] < um)
			{
				um--;
				lastx[um+1] = j;
			}
			um = umost[j];
		}
		if (theplc > up)
		{
			if (um > up+1) up = um-1;
			while (theplc > up)
			{
				up++;
				lastx[up+1] = j;
			}
			up = theplc;
		}
	}
	for(m=um;m<=up;m++)
		if (lastx[m+1] >= 0)
		{
			hline(lastx[m+1],x2,m,sectnum);
			lastx[m+1] = -1;
		}
	return(0);
}

florscan (long x1, long x2, short sectnum)
{
	long j, m, dm, dp, theplc;

	clearbuf((long)(&lastx[0]),(ydim+2)>>1,0xffffffff);
	dm = dplc[x1]-1;
	dp = dm+1;
	if (umost[x1] >= dp)
	{
		dp = umost[x1];
		dm = dp-1;
	}
	if (dm < -1) dm = -1;
	if (dp < -1) dp = -1;
	if (dm > ydim) dm = ydim;
	if (dp > ydim) dp = ydim;

	for(j=x1;j<=x2;j++)
	{
		theplc = dplc[j];
		if (theplc < -1) theplc = -1;
		if (theplc > ydim) theplc = ydim;
		if (theplc < umost[j]) theplc = umost[j];
		while (dmost[j]-1 < dm)
		{
			if (lastx[dm+1] >= 0)
			{
				hline(lastx[dm+1],j-1,dm,sectnum);
				lastx[dm+1] = -1;
			}
			dm--;
		}
		while (theplc > dp)
		{
			if (lastx[dp+1] >= 0)
			{
				hline(lastx[dp+1],j-1,dp,sectnum);
				lastx[dp+1] = -1;
			}
			dp++;
		}
		while (dmost[j]-1 > dm)
		{
			dm++;
			if (dp <= dm)
				lastx[dm+1] = j;
		}
		while (theplc < dp)
		{
			dp--;
			if (dp <= dm)
				lastx[dp+1] = j;
		}
	}
	for(m=dp;m<=dm;m++)
		if (lastx[m+1] >= 0)
		{
			hline(lastx[m+1],x2,m,sectnum);
			lastx[m+1] = -1;
		}
	return(0);
}

intersect (long x1l1, long y1l1, long x2l1, long y2l1, short l)
{
	short value;
	long x1l2, y1l2, x2l2, y2l2, templong1, templong2;

	x1l2 = wall[l].xa1;
	y1l2 = wall[l].ya1;
	x2l2 = wall[wall[l].point2].xa1;
	y2l2 = wall[wall[l].point2].ya1;

	value = 0;

	if ((x1l1 < x2l1) == (x2l1 < x2l1))
		if ((x1l1 < x2l2) == (x2l1 < x2l2))
			if ((x1l1 < x2l1) == (x1l1 < x2l2))
				if ((y1l1 < y2l1) == (y2l1 < y2l1))
					if ((y1l1 < y2l2) == (y2l1 < y2l2))
						if ((y1l1 < y2l1) == (y1l1 < y2l2))
							value = -1;

	if (value == 0)
	{
		templong1 = (x1l2-x1l1)*(y2l1-y1l1) - (x2l1-x1l1)*(y1l2-y1l1);  //p1(l2) vs. l1
		templong2 = (x2l2-x1l1)*(y2l1-y1l1) - (x2l1-x1l1)*(y2l2-y1l1);  //p2(l2) vs. l1
		if ((templong1 < 0) == (templong2 < 0)) value = -1;
		if ((templong1 == 0) || (templong2 == 0)) value = 0;
	}

	if (value == 0)
	{
		templong1 = (x1l1-x1l2)*(y2l2-y1l2) - (x2l2-x1l2)*(y1l1-y1l2);  //p1(l1) vs. l2
		templong2 = (x2l1-x1l2)*(y2l2-y1l2) - (x2l2-x1l2)*(y2l1-y1l2);  //p2(l1) vs. l2
		if ((templong1 < 0) == (templong2 < 0)) value = -1;
		if ((templong1 == 0) || (templong2 == 0)) value = 0;
	}

	if (value >= 0)
	{
		templong2 = (x2l1-x1l2)*(y2l2-y1l2) - (x2l2-x1l2) * (y2l1-y1l2);  //p2(l1) vs. l2
		if (templong2 >= 0)
		{
			value = 1;
		}
		else
		{
			value = 2;
		}
	}

	return(value);
}

loadboard(filename)
char filename[20];
{
	short int fil, i;

	if ((fil = open(filename,O_BINARY|O_RDWR,S_IREAD)) == -1)
		return(-1);

	read(fil,&posx,4);
	read(fil,&posy,4);
	read(fil,&posz,4);
	read(fil,&ang,2);
	read(fil,&numsectors,2);
	read(fil,&numwalls,2);
	read(fil,&cursectnum,2);

	for(i=0;i<numsectors;i++)
		read(fil,&sector[i],sizeof(struct sectortype));

	for(i=0;i<numwalls;i++)
		read(fil,&wall[i],sizeof(struct walltype));

	nowx = 0x80000000;
	nowy = 0x80000000;

	close(fil);
	return(0);
}

saveboard(filename)
char filename[20];
{
	short int fil, i;

	if ((fil = open(filename,O_BINARY|O_TRUNC|O_CREAT|O_WRONLY,S_IWRITE)) == -1)
		return(-1);

	write(fil,&posx,4);
	write(fil,&posy,4);
	write(fil,&posz,4);
	write(fil,&ang,2);
	write(fil,&numsectors,2);
	write(fil,&numwalls,2);
	write(fil,&cursectnum,2);

	for(i=0;i<numsectors;i++)
		write(fil,&sector[i],sizeof(struct sectortype));

	for(i=0;i<numwalls;i++)
		write(fil,&wall[i],sizeof(struct walltype));

	close(fil);
	return(0);
}

loadtables()
{
	short fil;

	if (tablesloaded == 0)
	{
		if ((fil = open("tables.dat",O_BINARY|O_RDWR,S_IREAD)) != -1)
		{
			read(fil,&sintable[0],4096);
			read(fil,&tantable[0],4096);
			read(fil,&radarang[0],640);
			close(fil);
			tablesloaded = 1;
		}
	}
	return(0);
}

setgamemode()
{
	short i, fil;

	setvmode(0x13);
	if ((fil = open("palette.dat",O_BINARY|O_RDWR,S_IREAD)) != -1)
	{
		read(fil,&palette[0],768);
		outp(0x3c8,0);
		for(i=0;i<768;i++)
			outp(0x3c9,palette[i]);
		read(fil,&palookup[0][0],(numpalookups<<8));
		close(fil);
	}
	return(0);
}

hline (long xl, long xr, long yp, short sectnum)
{
	char *ptr;
	signed char shadeoffs;
	unsigned long dat;
	short shade, siz;
	long zx, x, y, xc, yc, hd, zd, p, bufplc;

	if ((xl < 0) || (xr >= xdim) || (yp < 0) || (yp >= ydim) || (yp == (ydim>>1)))
		return(0);
	if (yp < (ydim>>1))
	{
		zd = ((sector[sectnum].z1-(posz>>8))<<8);
		bufplc = waloff[sector[sectnum].picnum1];
		shadeoffs = sector[sectnum].shade1;
	}
	else
	{
		zd = (((posz>>8)-sector[sectnum].z2)<<8);
		bufplc = waloff[sector[sectnum].picnum2];
		shadeoffs = sector[sectnum].shade2;
	}

	if (xl > xr)
		zx = xl, xl = xr, xr = zx;
	siz = xr-xl+1;

	hd = (zd<<14) / (xdim*labs((ydim>>1)-yp));
	xc = ((hd*sintable[(2048+ang)&2047])>>2);
	yc = ((hd*sintable[(2560+ang)&2047])>>2);
	x = xc*xl - (xc+yc)*(xdim>>1) + (posx<<22);
	y = yc*xl + (xc-yc)*(xdim>>1) - (posy<<22);
	if (yp > (ydim>>1))
	{
		y ^= 0xffffffff;
		yc ^= 0xffffffff;
	}

	shade = (labs(hd)>>visibility)+shadeoffs;
	if (shade < 0) shade = 0;
	if (shade >= numpalookups) shade = numpalookups-1;

	p = yp*320+xl+frameplace;

	takernum = 2;
	for(zx=siz;zx>=4;zx-=4)
	{
		dat = gethline64((long)x,(long)shade,(long)y,(long)bufplc);
		x += xc, y += yc;
		dat += (gethline64((long)x,(long)shade,(long)y,(long)bufplc)<<8);
		x += xc, y += yc;
		dat += (gethline64((long)x,(long)shade,(long)y,(long)bufplc)<<16);
		x += xc, y += yc;
		dat += (gethline64((long)x,(long)shade,(long)y,(long)bufplc)<<24);
		x += xc, y += yc;
		drawpixeles(p,dat);
		p += 4;
	}
	if ((siz&2) > 0)
	{
		dat = gethline64((long)x,(long)shade,(long)y,(long)bufplc);
		x += xc, y += yc;
		dat += (gethline64((long)x,(long)shade,(long)y,(long)bufplc)<<8);
		x += xc, y += yc;
		drawpixels(p,dat);
		p += 2;
	}
	if ((siz&1) > 0)
	{
		dat = gethline64((long)x,(long)shade,(long)y,(long)bufplc);
		drawpixel(p,dat);
	}
	takernum = 0;
	return(0);
}

vline (long x, long y1, long y2, long siz, long yrepeat, long picplc, short tilenum, signed char shadeoffs)
{
	char *ptr;
	short y1v, y2v, sizv, thesiz, zx, shade;
	long vplc, vinc, p, bufplc, mask;

	if (((y1 < 0) && (y2 < 0)) || ((y1 > (ydim<<16)) && (y2 > (ydim<<16))) || (siz < 2048))
		return(0);

	y1v = (y1>>16);
	y2v = (y2>>16);
	sizv = (siz>>16);
	thesiz = y2v-y1v;
 
	if ((y1v > dmost[x]) || (y2v < umost[x]) || (thesiz <= 0) || (sizv <= 0) || (x < 0) || (x >= xdim))
		return(0);

	vinc = (33554432*32) / (siz>>11);

	shade = (vinc>>(visibility+2))+shadeoffs;
	if (shade < 0) shade = 0;
	if (shade >= numpalookups) shade = numpalookups-1;

	vinc = ((vinc*yrepeat)<<5);
	vplc = ((((y1&65535^65535) * vinc)>>8)&65535);
	if (y1v < umost[x])
	{
		vplc += (umost[x]-y1v)*vinc;
		y1v = umost[x];
	}
	if (y2v > dmost[x]) y2v = dmost[x];
	thesiz = y2v-y1v;
	if ((thesiz < 1) || (thesiz > dmost[x])) return(0);

	picplc &= ((0xffff>>(16-(picsiz[tilenum]&15)))<<(picsiz[tilenum]>>4));
	bufplc = waloff[tilenum]+picplc;

	mask = (1<<(picsiz[tilenum]>>4))-1;

	p = y1v*320 + x+frameplace;
	takernum = 1;
	for(zx=0;zx<thesiz;zx++)
	{
		plotvline((long)vplc,(long)mask,(long)shade,(long)bufplc,(long)p);
		vplc += vinc;
		p += 320;
	}
	takernum = 0;
	return(0);
}

initengine()
{
	long i;

	searchx = 160;
	searchy = 100;
	searchit = 0;
	searchstat = -1;
	asksave = 0;
	somethingintab = 0;
	zoom = 768, grid = 1, anglelock = 1;
	loadtables();
	for(i=0;i<xdim;i++)
	{
		startumost[i] = 0;
		startdmost[i] = ydim;
	}
	for(i=0;i<=ydim+1;i++)
		lastx[i] = -1;
	oldkeyhandler = _dos_getvect(0x9);
	_disable(); _dos_setvect(0x9, keyhandler); _enable();
	resettiming();
	visibility = 13;
	capturecount = 0;
	clearbuf((long)(&keystatus[0]),64L,0L);
	takernum = 0;
	editstatus = 0;
	mousx = 0;
	mousy = 0;
	page = 0;
	setupmouse();
	videocard = getcard();
	if (videocard == 1)
	{
		page = 0; outp(0x3cd,page);
		frameplace = 0xa0000;
	}
	else
		frameplace = FP_OFF(screen);
	oldtimerhandler = _dos_getvect(0x8);
	_dos_setvect(0x8, timerhandler);
	outp(0x43,54); outp(0x40,4972&255); outp(0x40,4972>>8);
	return(0);
}

uninitengine()
{
	_dos_setvect(0x8, oldtimerhandler);
	_dos_setvect(0x9, oldkeyhandler);
	outp(0x43,54); outp(0x40,255); outp(0x40,255);
	free(pic);
}

nextpage()
{
	if (videocard == 1)
	{
		outp(0x3d4,0xc); outp(0x3d5,page*64);
		page++;
		if (page == 3) page = 0;
		outp(0x3cd,page);
	}
	else
	{
		takernum = 3;
		copytovideo();
		takernum = 0;
	}

	numframes++;
	totalclock += ((long)clockspeed), clockspeed = 0;
}

loadpics(filename)
char filename[20];
{
	long totpicsiz, offscount, siz;
	short fil, i;

	if ((fil = open(filename,O_BINARY|O_RDONLY,S_IREAD)) == -1)
		return(-1);
	read(fil,&picsiz[0],maxtiles);
	read(fil,&picanm[0],maxtiles);
	totpicsiz = 0L;
	for(i=0;i<maxtiles;i++)
		if (picsiz[i] > 0)
			totpicsiz += (1<<((long)((picsiz[i]&15)+(picsiz[i]>>4))));
	pic = (char *)malloc(totpicsiz);
	offscount = 0L;
	for(i=0;i<maxtiles;i++)
	{
		waloff[i] = offscount+FP_OFF(pic);
		if (picsiz[i] > 0)
		{
			siz = (1<<((long)((picsiz[i]&15)+(picsiz[i]>>4))));
			read(fil,&pic[offscount],siz);
			offscount += siz;
		}
	}
	close(fil);
	return(0);
}

showengineinfo()
{
	short i, j, k;
	long templong, templong1;

	templong = 0;
	for(i=0;i<numtimecounts;i++)
		templong += timetakercount[i];
	if (templong != 0)
		for(i=0;i<numtimecounts;i++)
		{
			templong1 = (10000L*timetakercount[i])/templong;
			printf("%d - %2d.%1d%1d%% - %s",i,(short)(templong1/100),(short)((templong1/10)%10),(short)(templong1%10),timenames[i]);
			j = (short)((templong1+50)/100);
			for(k=0;k<(j>>1);k++)
				printf("");
			if ((j&1) > 0)
				printf("");
			printf("\n");
		}
	if (totalclock != 0)
	{
		templong1 = (numframes*24000L)/totalclock;
		printf("%d.%1d%1d frames per second\n",(short)(templong1/100),(short)((templong1/10)%10),(short)(templong1%10));
	}
}

resettiming()
{
	long i;

	for(i=0;i<numtimecounts;i++)
		timetakercount[i] = 0L;
	numframes = 0L;
	clockspeed = 0L;
	totalclock = 0L;
}

clipmove (long vel)
{
	char poszchar;
	long i, x, y, zx;
	long incr1, incr2, xdir, ydir, xdiredge, ydiredge, d, bad, bad1, bad2;

	poszchar = 255;
	incr1 = vel * sintable[(ang+2560)&2047];
	incr2 = vel * sintable[(ang+2048)&2047];
	if (incr1 < 0)
	{
		xdir = -1;
		incr1 = -incr1;
	}
	else if (incr1 > 0)
	{
		xdir = 1;
	}
	else
	{
		xdir = 0;
	}
	if (incr2 < 0)
	{
		ydir = -1;
		incr2 = -incr2;
	}
	else if (incr2 > 0)
	{
		ydir = 1;
	}
	else
	{
		ydir = 0;
	}
	xdiredge = (xdir<<3); ydiredge = (ydir<<3);
	d = incr2 - incr1;
	bad1 = 0; bad2 = 0;
	x = ((posx-nowx)>>4); y = ((posy-nowy)>>4);
	if ((x <= 8) || (x >= 119) || (y <= 8) || (y >= 119))
	{
		nowx = ((posx-256)&0xfc00);
		nowy = ((posy-256)&0xfc00);
		x = ((posx-nowx)>>4); y = ((posy-nowy)>>4);
		cliploadlocal();
	}
	for(i=0;i<=((incr1+incr2)>>19);i++)
	{
		if (d < 0)
		{
			d += incr2;
			bad1 = 0;
			if (clipbuf[x+xdiredge][y-7] < poszchar) bad1 = 2;
			if (clipbuf[x+xdiredge][y+7] < poszchar) bad1 = 2;
			for(zx=-6;zx<=6;zx++)
				if (clipbuf[x+xdiredge][y+zx] < poszchar) bad1 = 1;
			if (bad1 == 0) x += xdir;
		}
		else
		{
			d -= incr1;
			bad2 = 0;
			if (clipbuf[x-7][y+ydiredge] < poszchar) bad2 = 2;
			if (clipbuf[x+7][y+ydiredge] < poszchar) bad2 = 2;
			for(zx=-6;zx<=6;zx++)
				if (clipbuf[x+zx][y+ydiredge] < poszchar) bad2 = 1;
			if (bad2 == 0) y += ydir;
		}
		if (((bad1 == 2) || (bad2 == 2)) && ((bad1 > 0) && (bad2 > 0)))
		{
			if (bad1 == 2)
			{
				bad = 0;
				for(zx=-7;zx<=7;zx++)
					bad |= (255-clipbuf[x+zx][y-ydiredge]);
				if (bad == 0)
				{
					y -= ydir;
					d = 0;
				}
			}
			if (bad2 == 2)
			{
				bad = 0;
				for(zx=-7;zx<=7;zx++)
					bad |= (255-clipbuf[x-xdiredge][y+zx]);
				if (bad == 0)
				{
					x -= xdir;
					d = -1;
				}
			}
		}
		bad = 0;
		if (x <= 8)
		{
			nowx -= 1024;
			x += 64;
			bad = 1;
		}
		if (x >= 119)
		{
			nowx += 1024;
			x -= 64;
			bad = 1;
		}
		if (y <= 8)
		{
			nowy -= 1024;
			y += 64;
			bad = 1;
		}
		if (y >= 119)
		{
			nowy += 1024;
			y -= 64;
			bad = 1;
		}
		if (bad == 1) cliploadlocal();
	}
	posx = nowx + (x<<4);
	posy = nowy + (y<<4);
}

cliploadlocal()
{
	char clipheight;
	short z, startwall, endwall;
	long i, j, xz, yz, p, linti, xc1, yc1, xc2, yc2;
	long incr1, incr2, xdir, ydir, xc, x, y;

	clearbuf((long)(&clipbuf),4096L,0xffffffff);

	startwall = sector[cursectnum].wallptr;
	endwall = startwall + sector[cursectnum].wallnum - 1;
	for(z=startwall;z<=endwall;z++)
		if (wall[z].nextsector1 < 0)
		{
			clipheight = 0;

			xc1 = ((wall[z].xa1-nowx)>>4);
			yc1 = ((wall[z].ya1-nowy)>>4);
			xc2 = ((wall[wall[z].point2].xa1-nowx)>>4);
			yc2 = ((wall[wall[z].point2].ya1-nowy)>>4);

			incr1 = xc2 - xc1;
			incr2 = yc2 - yc1;
			if (incr1 < 0)
			{
				xdir = -1;
				incr1 = -incr1;
			}
			else if (incr1 > 0)
			{
				xdir = 1;
			}
			else
			{
				xdir = 0;
			}
			if (incr2 < 0)
			{
				ydir = -1;
				incr2 = -incr2;
			}
			else if (incr2 > 0)
			{
				ydir = 1;
			}
			else
			{
				ydir = 0;
			}
			xc = 0;
			x = xc1; y = yc1;
			do
			{
				if (((x|y)&0xffffff80) == 0) clipbuf[x][y] = clipheight;
				if (xc < 0)
				{
					xc += incr2;
					x += xdir;
				}
				else
				{
					xc -= incr1;
					y += ydir;
				}
			}
			while ((x != xc2) || (y != yc2));
		}
}

engineinput()
{
	int i, doubvel;
	long oldposx, oldposy, newposz;

	if (keystatus[0x12] > 0)
	{
		keystatus[0x12] = 0;
		editstatus = 1 - editstatus;
		if (editstatus == 0)
			searchstat = -1;
	}
	if (keystatus[0x2f] > 0)
	{
		keystatus[0x2f] = 0;
		visibility--;
		if (visibility == 7)
			visibility = 15;
	}
	if (vel < 0)
	{
		vel+=((clockspeed>>3)+1);
		if (vel > 0) vel = 0;
	}
	if (vel > 0)
	{
		vel-=((clockspeed>>3)+1);
		if (vel < 0) vel = 0;
	}
	if (svel < 0)
	{
		svel+=((clockspeed>>2)+1);
		if (svel > 0) svel = 0;
	}
	if (svel > 0)
	{
		svel-=((clockspeed>>2)+1);
		if (svel < 0) svel = 0;
	}
	if (hvel < 0)
	{
		hvel += 4;
		if (hvel > 0) hvel = 0;
	}
	if (hvel > 0)
	{
		hvel -= 4;
		if (hvel < 0) hvel = 0;
	}
	if (angvel < 0)
	{
		angvel+=((clockspeed>>1)+1);
		if (angvel > 0) angvel = 0;
	}
	if (angvel > 0)
	{
		angvel-=((clockspeed>>1)+1);
		if (angvel < 0) angvel = 0;
	}
	if (editstatus == 0)
	{
		if (moustat == 1)
			readmouse();
		angvel += mousx;
		vel -= mousy;
	}
	if (keystatus[0x2a] > 0)
		doubvel = 2;
	else
		doubvel = 1;

	if (keystatus[0xb7] > 0)
	{
		keystatus[0xb7] = 0;
		int5();
	}
	if (keystatus[0x9d] == 0)
	{
		if (keystatus[0xcb] > 0) angvel -= 10;
		if (keystatus[0xcd] > 0) angvel += 10;
	}
	else
	{
		if (keystatus[0xcb] > 0) svel += 6*doubvel;
		if (keystatus[0xcd] > 0) svel -= 6*doubvel;
	}
	if (keystatus[0x1e] > 0) hvel -= 6*doubvel;
	if (keystatus[0x2c] > 0) hvel += 6*doubvel;
	if (keystatus[0xc8] > 0) vel += 6*doubvel;
	if (keystatus[0xd0] > 0) vel -= 6*doubvel;

	if (vel < -16*doubvel) vel = -16*doubvel;
	if (vel > 16*doubvel) vel = 16*doubvel;
	if (svel < -16*doubvel) svel = -16*doubvel;
	if (svel > 16*doubvel) svel = 16*doubvel;
	if (hvel < -16*doubvel) hvel = -16*doubvel;
	if (hvel > 16*doubvel) hvel = 16*doubvel;
	if (angvel < -24) angvel = -24;
	if (angvel > 24) angvel = 24;

	if (keystatus[0x13] > 0)
		resettiming();
	if (keystatus[0x19] > 0)
	{
		keystatus[0x19] = 0;
		while ((keystatus[0x19] == 0) && (keystatus[1] == 0));
		keystatus[0x19] = 0;
	}
	if (keystatus[88] > 0)
	{
		screencapture();
		keystatus[88] = 0;
	}
	if ((keystatus[0x2e] > 0) && (poscapture == 0))
	{
		poscapture = 1L;
		posfil = open("poscapt.dat",O_BINARY|O_TRUNC|O_CREAT|O_WRONLY,S_IWRITE);
		write(posfil,&poscapture,4);
	}
	if (angvel != 0)
	{
		ang += ((angvel*clockspeed)>>3);
		ang = (ang+2048)&2047;
	}

	oldposx = posx; oldposy = posy;
	if (vel != 0)
		clipmove(vel*clockspeed);
	if (svel != 0)
	{
		ang = ((ang+1536)&2047);
		clipmove(svel*clockspeed);
		ang = ((ang+512)&2047);
	}

	newposz = sector[cursectnum].z2-32;
	if (newposz < sector[cursectnum].z1+32)
		newposz = ((sector[cursectnum].z1+sector[cursectnum].z2)>>1);
	if (hvel != 0)
		newposz += hvel;
	posz = (posz+posz+posz+(newposz<<8))>>2;

	if (poscapture > 0)
	{
		write(posfil,&posx,4);
		write(posfil,&posy,4);
		write(posfil,&posz,4);
		write(posfil,&ang,2);
		poscapture++;
	}
}

screencapture()
{
	char filename[15];
	short fil;
	long i, j, k, bufplc;

	strcpy(filename,"capturxx.bmp");
	filename[6] = (capturecount/10)+48;
	filename[7] = (capturecount%10)+48;
	capturecount++;
	if ((fil=open(filename,O_BINARY|O_CREAT|O_WRONLY,S_IWRITE))==-1)
	{
		capturecount--;
		return(-1);
	}
	tempbuf[0] = 66, tempbuf[1] = 77, tempbuf[2] = 90, tempbuf[3] = 34;
	tempbuf[4] = 0, tempbuf[5] = 0, tempbuf[6] = 0, tempbuf[7] = 0;
	tempbuf[8] = 0, tempbuf[9] = 0, tempbuf[10] = 26, tempbuf[11] = 3;
	tempbuf[12] = 0, tempbuf[13] = 0, tempbuf[14] = 12, tempbuf[15] = 0;
	tempbuf[16] = 0, tempbuf[17] = 0, tempbuf[18] = 64, tempbuf[19] = 1;
	tempbuf[20] = 200, tempbuf[21] = 0, tempbuf[22] = 1, tempbuf[23] = 0;
	tempbuf[24] = 8, tempbuf[25] = 0;
	write(fil,&tempbuf[0],26);
	outp(0x3c7,0);
	for(i=0;i<768;i++)
		tempbuf[i] = (inp(0x3c9)<<2);
	for(i=0;i<768;i+=3)
	{
		j = tempbuf[i];
		tempbuf[i] = tempbuf[i+2];
		tempbuf[i+2] = j;
	}
	if (write(fil,&tempbuf[0],768) < 768)
	{
		close(fil);
		capturecount--;
		return(-1);
	}

	bufplc = 0L;
	for(j=199;j>=0;j--)
	{
		outp(97,inp(97)|3);
		readrow(frameplace+j*320,bufplc);
		bufplc += 320L;
		if (bufplc >= 3200L)
		{
			if (write(fil,&tempbuf[0],3200) < 3200)
			{
				close(fil);
				capturecount--;
				return(-1);
			}
			bufplc = 0L;
		}
		outp(97,inp(97)&252);
	}
	close(fil);
	return(0);
}

int getcard()         //returns: 0 = nothing, 1 = TSENG ET4000, 2 = VESA
{
	int value, newvalue, oldvalue, bas, tseng, vesa, card;

	outp(0x3bf,3);                       //Check TSENG
	if ((inp(0x3cc)&1) == 1)
		outp(0x3d8, 0xa0);
	else
		outp(0x3b8, 0xa0);
	inp(0x3da); outp(0x3c0,0x16); oldvalue = inp(0x3c1);
	inp(0x3da); outp(0x3c0,0x16); newvalue = (oldvalue^16);
	outp(0x3c0, newvalue);
	inp(0x3da); outp(0x3c0,0x16); value = inp(0x3c1);
	inp(0x3da); outp(0x3c0,0x16); outp(0x3c0,oldvalue);
	tseng = 0;
	if (value == newvalue)
	{
		if ((inp(0x3cc)&1) == 1)
			bas = 0x3d0;
		else
			bas = 0x3b0;
		outp(bas+4,0x33); oldvalue = inp(bas+5);
		outp(bas+5,oldvalue^15); newvalue = inp(bas+5);
		outp(bas+5,oldvalue);
		if (newvalue == (oldvalue^15))
			tseng = 4000;  //TSENG ET4000
		else
			tseng = 3000;  //TSENG ET3000
	}

	vesa = checkvesa();                   //Check VESA

	if (tseng == 4000)
		card = 1;
	else if (vesa == 1)
		card = 2;
	else
		card = 0;

	return(card);
}

loadpcx(char filename[20], long ptrlong)
{
	char *ptr;
	unsigned char dat;
	long fil, i, p, datcnt, leng;

	ptr = (char *)ptrlong;
	if ((fil = open(filename,O_BINARY|O_RDWR,S_IREAD)) == -1)
		return(-1);
	read(fil,&tempbuf[0],128);
	p = 64000;
	read(fil,&tempbuf[0],4096), datcnt = 0;
	while (p > 0)
	{
		dat = tempbuf[datcnt++];
		if (datcnt == 4096)
			read(fil,&tempbuf[0],4096), datcnt = 0;
		if ((dat&0xc0) == 0xc0)
		{
			leng = (dat&0x3f);
			dat = tempbuf[datcnt++];
			if (datcnt == 4096)
				read(fil,&tempbuf[0],4096), datcnt = 0;
			for(i=0;i<leng;i++)
			{
				*ptr = dat;
				ptr++;
			}
			p -= leng;
		}
		else
		{
			*ptr = dat;
			ptr++;
			p--;
		}
	}
	dat = tempbuf[datcnt++];
	if (datcnt == 4096)
		read(fil,&tempbuf[0],4096), datcnt = 0;
	if (dat == 0xc)
	{
		outp(0x3c8,0);
		for(i=0;i<768;i++)
		{
			dat = tempbuf[datcnt++];
			if (datcnt == 4096)
				read(fil,&tempbuf[0],4096), datcnt = 0;
			outp(0x3c9,dat>>2);
		}
	}
	close(fil);
	if ((videocard == 1) && (ptrlong == 0xa0000))
	{
		i = page+1; if (i == 3) i = 0;
		outp(0x3cd,(page<<4)+i), videomove();
		i++; if (i == 3) i = 0;
		outp(0x3cd,(page<<4)+i), videomove();
		outp(0x3cd,page);
	}
	return(0);
}

loadbmp(char filename[20], long ptrlong)
{
	char *ptr;
	long p;
	short fil, i;

	ptr = (char *)ptrlong;
	if ((fil = open(filename,O_BINARY|O_RDWR,S_IREAD)) == -1)
		return(-1);
	read(fil,&tempbuf[0],54);
	read(fil,&tempbuf[0],1024);
	outp(0x3c8,0);
	for(i=0;i<1024;i++)
		if ((i&3) != 0)
			outp(0x3c9,tempbuf[i^3]>>2);
	p = ((long)ptr)+64000L-320L;
	for(i=0;i<200;i++)
	{
		read(fil,&tempbuf[0],320);
		drawrow(p);
		p -= 320L;
	}
	close(fil);
	if ((videocard == 1) && (ptrlong == 0xa0000))
	{
		i = page+1; if (i == 3) i = 0;
		outp(0x3cd,(page<<4)+i), videomove();
		i++; if (i == 3) i = 0;
		outp(0x3cd,(page<<4)+i), videomove();
		outp(0x3cd,page);
	}
	return(0);
}

showmouse320200()
{
	long i;

	for(i=1;i<=4;i++)
	{
		if ((startumost[searchx+i] <= searchy) && (startdmost[searchx+i] > searchy))
			drawpixel(frameplace+(searchy)*320L+(searchx+i),31);
		if ((startumost[searchx-i] <= searchy) && (startdmost[searchx-i] > searchy))
			drawpixel(frameplace+(searchy)*320L+(searchx-i),31);
		if ((startumost[searchx] <= (searchy-i)) && (startdmost[searchx] > (searchy-i)))
			drawpixel(frameplace+(searchy-i)*320L+(searchx),31);
		if ((startumost[searchx] <= (searchy+i)) && (startdmost[searchx] > (searchy+i)))
			drawpixel(frameplace+(searchy+i)*320L+(searchx),31);
	}
}

editinput()
{
	long i, templong;

	readmouse();
	searchx += mousx;
	searchy += mousy;
	if (searchx < 4) searchx = 4;
	if (searchx > 315) searchx = 315;
	if (searchy < 4) searchy = 4;
	if (searchy > 195) searchy = 195;
	showmouse320200();
	if (searchstat >= 0)
	{
		if (keystatus[0x4a] > 0)  // -
		{
			if (searchstat == 0) wall[searchwall].shade++;
			if (searchstat == 1) sector[searchsector].shade1++;
			if (searchstat == 2) sector[searchsector].shade2++;
			asksave = 1;
			keystatus[0x4a] = 0;
		}
		if (keystatus[0x4e] > 0)  // +
		{
			if (searchstat == 0) wall[searchwall].shade--;
			if (searchstat == 1) sector[searchsector].shade1--;
			if (searchstat == 2) sector[searchsector].shade2--;
			asksave = 1;
			keystatus[0x4e] = 0;
		}
		if (keystatus[0xc9] > 0) // PGUP
		{
			if (searchstat == 0) sector[searchsector].z1 -= 4;
			if (searchstat == 1) sector[searchsector].z1 -= 4;
			if (searchstat == 2) sector[searchsector].z2 -= 4;
			asksave = 1;
			keystatus[0xc9] = 0;
		}
		if (keystatus[0xd1] > 0) // PGDN
		{
			if (searchstat == 0) sector[searchsector].z1 += 4;
			if (searchstat == 1) sector[searchsector].z1 += 4;
			if (searchstat == 2) sector[searchsector].z2 += 4;
			asksave = 1;
			keystatus[0xd1] = 0;
		}
		if (keystatus[0x0f] > 0)  //TAB
		{
			if (searchstat == 0)
			{
				temppicnum = wall[searchwall].picnum;
				tempshade = wall[searchwall].shade;
			}
			if (searchstat == 1)
			{
				temppicnum = sector[searchsector].picnum1;
				tempshade = sector[searchsector].shade1;
			}
			if (searchstat == 2)
			{
				temppicnum = sector[searchsector].picnum2;
				tempshade = sector[searchsector].shade2;
			}
			somethingintab = 1;
			keystatus[0x0f] = 0;
		}
		if ((keystatus[0x39] > 0) && (somethingintab > 0))  //SPACE
		{
			if (searchstat == 0)
			{
				wall[searchwall].picnum = temppicnum;
				wall[searchwall].shade = tempshade;
			}
			if (searchstat == 1)
			{
				sector[searchsector].picnum1 = temppicnum;
				sector[searchsector].shade1 = tempshade;
			}
			if (searchstat == 2)
			{
				sector[searchsector].picnum2 = temppicnum;
				sector[searchsector].shade2 = tempshade;
			}
			asksave = 1;
			keystatus[0x39] = 0;
		}
		if (keystatus[0x2f] > 0)  //V
		{
			if (searchstat == 0) templong = wall[searchwall].picnum;
			if (searchstat == 1) templong = sector[searchsector].picnum1;
			if (searchstat == 2) templong = sector[searchsector].picnum2;
			templong = gettile(templong);
			if (searchstat == 0) wall[searchwall].picnum = templong;
			if (searchstat == 1) sector[searchsector].picnum1 = templong;
			if (searchstat == 2) sector[searchsector].picnum2 = templong;
			asksave = 1;
			keystatus[0x2f] = 0;
		}
		if (keystatus[0x4b] > 0)  //4 (keypad)
		{
			if (searchstat == 0)
				if (wall[searchwall].xrepeat > 0)
					wall[searchwall].xrepeat--;
			if (searchstat > 0)
			{
				i = searchwall;
				do
				{
					if (wall[i].xrepeat > 0)
						wall[i].xrepeat--;
					i = wall[i].point2;
				}
				while (i != searchwall);
			}
			keystatus[0x4b] = 0;
		}
		if (keystatus[0x4d] > 0)  //6 (keypad)
		{
			if (searchstat == 0)
				if (wall[searchwall].xrepeat < 255)
					wall[searchwall].xrepeat++;
			if (searchstat > 0)
			{
				i = searchwall;
				do
				{
					if (wall[i].xrepeat < 255)
						wall[i].xrepeat++;
					i = wall[i].point2;
				}
				while (i != searchwall);
			}
			keystatus[0x4d] = 0;
		}
		if (keystatus[0x48] > 0)  //2 (keypad)
		{
			if (searchstat == 0)
				if (wall[searchwall].yrepeat > 0)
					wall[searchwall].yrepeat--;
			if (searchstat > 0)
			{
				i = searchwall;
				do
				{
					if (wall[i].yrepeat > 0)
						wall[i].yrepeat--;
					i = wall[i].point2;
				}
				while (i != searchwall);
			}
			keystatus[0x48] = 0;
		}
		if (keystatus[0x50] > 0)  //8 (keypad)
		{
			if (searchstat == 0)
				if (wall[searchwall].yrepeat < 255)
					wall[searchwall].yrepeat++;
			if (searchstat > 0)
			{
				i = searchwall;
				do
				{
					if (wall[i].yrepeat < 255)
						wall[i].yrepeat++;
					i = wall[i].point2;
				}
				while (i != searchwall);
			}
			keystatus[0x50] = 0;
		}
		if (keystatus[0x19] > 0)  // P (parallax)
		{
			sector[searchsector].cstat1 ^= 1;
			asksave = 1;
			keystatus[0x19] = 0;
		}
		if (keystatus[0x9c] > 0)  // Enter
		{
			overheadeditor();
			keystatus[0x9c] = 0;
		}
	}
	searchit = 1;
}

getsector()
{
	short i;

	if (inside(posx,posy,cursectnum) == 0)
	{
		for(i=0;i<numsectors;i++)
			if (inside(posx,posy,i) == 1)
			{
				cursectnum = i;
				break;
			}
	}
}

inside(long x, long y, short sectnum)
{
	long x1, y1, x2, y2, templong;
	short i, startwall, endwall, cnt;

	cnt = 0;
	startwall = sector[sectnum].wallptr;
	endwall = startwall + sector[sectnum].wallnum - 1;
	for(i=startwall;i<=endwall;i++)
	{
		x1 = wall[i].xa1; x2 = wall[wall[i].point2].xa1;
		if ((x1 >= x) || (x2 >= x))
		{
			y1 = wall[i].ya1; y2 = wall[wall[i].point2].ya1;
			if (y1 > y2)
			{
				templong = x1, x1 = x2, x2 = templong;
				templong = y1, y1 = y2, y2 = templong;
			}
			if ((y1 <= y) && (y2 > y))
				if (x1*(y-y2)+x2*(y1-y) <= x*(y1-y2))
					cnt ^= 1;
		}
	}
	return(cnt);
}

drawsvgarow(long picstart)
{
	long i, j, vidpos, vidpos2, dat, wallnum, xdime, ydime;
	char *picptr;

	if (videocard == 1)
		outp(0x3cd,(picstart>>4)&15);
	else
		vesasetpage((picstart>>4)&15);

	vidpos = 0xa0000;
	clearseg();
	for(wallnum=picstart;wallnum<picstart+16;wallnum++)
	{
		if (picsiz[wallnum] != 0)
		{
			picptr = (char *)(waloff[wallnum]);
			xdime = (1<<(picsiz[wallnum]&15));
			ydime = (1<<(picsiz[wallnum]>>4));

			if ((xdime <= 64) && (ydime <= 64))
			{
				vidpos2 = vidpos;
				for(i=0;i<xdime;i++)
				{
					for(j=0;j<ydime;j++)
					{
						dat = *picptr++;
						drawpixel(vidpos2, dat);
						vidpos2 += 1024L;
					}
					vidpos2 = (vidpos2&1023)+1+0xa0000;
				}
			}
			else
			{
				vidpos2 = vidpos;
				for(i=0;i<xdime;i+=2)
				{
					for(j=0;j<ydime;j+=2)
					{
						dat = *picptr;
						picptr += 2;
						drawpixel(vidpos2, dat);
						vidpos2 += 1024L;
					}
					vidpos2 = (vidpos2&1023)+1+0xa0000;
					picptr += ydime;
				}
			}
		}
		vidpos += 64;
	}
	return(0);
}

drawopenboxsvga(long picstart, char col)
{
	long i, vidpos, vidpos1, vidpos2, dat;

	dat = (long)col;
	dat += (dat<<8);
	if (videocard == 1)
		outp(0x3cd,(picstart>>4)&15);
	else
		vesasetpage((picstart>>4)&15);

	vidpos = ((picstart&15)<<6)+0xa0000;
	vidpos1 = vidpos;
	vidpos2 = vidpos+62;
	for(i=0;i<64;i++)
	{
		drawpixels(vidpos1,dat);
		drawpixels(vidpos2,dat);
		vidpos1 += 1024;
		vidpos2 += 1024;
	}
	vidpos1 = vidpos;
	vidpos2 = (63<<10)+vidpos;
	for(i=0;i<62;i+=2)
	{
		drawpixels(vidpos1,dat);
		drawpixels(vidpos1+1024L,dat);
		drawpixels(vidpos2,dat);
		drawpixels(vidpos2-1024L,dat);
		vidpos1+=2;
		vidpos2+=2;
	}
	return(0);
}

backupboxsvga(long picstart)
{
	long i, vidpos, vidpos1, vidpos2, tempbufcnt;

	if (videocard == 1)
		outp(0x3cd,((picstart>>4)&15)<<4);
	else
		vesasetpage(((picstart>>4)&15)<<4);

	vidpos = ((picstart&15)<<6)+0xa0000;
	vidpos1 = vidpos;
	vidpos2 = vidpos+62;
	tempbufcnt = 0;
	for(i=0;i<64;i++)
	{
		tempbuf[tempbufcnt++] = getpixel(vidpos1);
		tempbuf[tempbufcnt++] = getpixel(vidpos1+1);
		tempbuf[tempbufcnt++] = getpixel(vidpos2);
		tempbuf[tempbufcnt++] = getpixel(vidpos2+1);
		vidpos1 += 1024;
		vidpos2 += 1024;
	}
	vidpos1 = vidpos;
	vidpos2 = (63<<10)+vidpos;
	for(i=0;i<62;i+=2)
	{
		tempbuf[tempbufcnt++] = getpixel(vidpos1);
		tempbuf[tempbufcnt++] = getpixel(vidpos1+1);
		tempbuf[tempbufcnt++] = getpixel(vidpos1+1024);
		tempbuf[tempbufcnt++] = getpixel(vidpos1+1025);
		tempbuf[tempbufcnt++] = getpixel(vidpos2);
		tempbuf[tempbufcnt++] = getpixel(vidpos2+1);
		tempbuf[tempbufcnt++] = getpixel(vidpos2-1024);
		tempbuf[tempbufcnt++] = getpixel(vidpos2-1023);
		vidpos1+=2;
		vidpos2+=2;
	}
	return(0);
}

restoreboxsvga(long picstart)
{
	long i, vidpos, vidpos1, vidpos2, tempbufcnt;

	if (videocard == 1)
		outp(0x3cd,(picstart>>4)&15);
	else
		vesasetpage((picstart>>4)&15);

	vidpos = ((picstart&15)<<6)+0xa0000;
	vidpos1 = vidpos;
	vidpos2 = vidpos+62;
	tempbufcnt = 0;
	for(i=0;i<64;i++)
	{
		drawpixel(vidpos1,tempbuf[tempbufcnt++]);
		drawpixel(vidpos1+1,tempbuf[tempbufcnt++]);
		drawpixel(vidpos2,tempbuf[tempbufcnt++]);
		drawpixel(vidpos2+1,tempbuf[tempbufcnt++]);
		vidpos1 += 1024;
		vidpos2 += 1024;
	}
	vidpos1 = vidpos;
	vidpos2 = (63<<10)+vidpos;
	for(i=0;i<62;i+=2)
	{
		drawpixel(vidpos1,tempbuf[tempbufcnt++]);
		drawpixel(vidpos1+1,tempbuf[tempbufcnt++]);
		drawpixel(vidpos1+1024,tempbuf[tempbufcnt++]);
		drawpixel(vidpos1+1025,tempbuf[tempbufcnt++]);
		drawpixel(vidpos2,tempbuf[tempbufcnt++]);
		drawpixel(vidpos2+1,tempbuf[tempbufcnt++]);
		drawpixel(vidpos2-1024,tempbuf[tempbufcnt++]);
		drawpixel(vidpos2-1023,tempbuf[tempbufcnt++]);
		vidpos1+=2;
		vidpos2+=2;
	}
	return(0);
}

checkforscroll(long currentscreenplace, long realplace)
{
	long place;

	if (realplace < currentscreenplace)
	{
		currentscreenplace -= 16L;
		drawsvgarow(currentscreenplace);
	}
	if (realplace >= currentscreenplace+192)
	{
		drawsvgarow(currentscreenplace+192);
		currentscreenplace += 16L;
	}
	place = (((currentscreenplace>>4)&15)<<6);
	outp(0x3d4,0x33); outp(0x3d5,place>>8);
	outp(0x3d4,0xc); outp(0x3d5,place&255);
	return(currentscreenplace);
}

gettile(long templong)
{
	long i, j, otemplong;

	if (videocard == 1)
		setvmode(0x38);
	else
		vesasetmode(0x105);

	otemplong = templong;
	outp(0x3c8,0);
	for(i=0;i<768;i++)
		outp(0x3c9,palette[i]);
	j = (templong&0xfff0)-96;
	if (j < 0) j = 0;
	j = checkforscroll(j, templong);
	for(i=0;i<192;i+=16)
		drawsvgarow(i+j);
	backupboxsvga(templong);
	while ((keystatus[0x1c]|keystatus[1]) == 0)
	{
		restoreboxsvga(templong);
		if ((keystatus[0xcb] > 0) && (templong > 0))
			templong--, keystatus[0xcb] = 0;
		if ((keystatus[0xcd] > 0) && (templong < maxtiles-1))
			templong++, keystatus[0xcd] = 0;
		if ((keystatus[0xc8] > 0) && (templong >= 16))
			templong-=16, keystatus[0xc8] = 0;
		if ((keystatus[0xd0] > 0) && (templong < maxtiles-16))
			templong+=16, keystatus[0xd0] = 0;
		j = checkforscroll(j, templong);
		backupboxsvga(templong);

		readmouse();
		if ((mousx != 0) || (mousy != 0))
		{
			restoreboxsvga(templong);
			if ((mousx < 0) && ((templong&15) > 0)) templong--;
			if ((mousx > 0) && ((templong&15) < 15)) templong++;
			if ((mousy < 0) && (templong >= 16)) templong-=16;
			if ((mousy > 0) && (templong < maxtiles-16)) templong+=16;
			j = checkforscroll(j, templong);
			backupboxsvga(templong);
		}

		limitrate();
		drawopenboxsvga(templong,31);
		limitrate();
		drawopenboxsvga(templong,0);
	}
	if (keystatus[0x1c] == 0)
		templong = otemplong;
	else
		keystatus[0x1c] = 0;
	setvmode(0x13);
	outp(0x3c8,0);
	for(i=0;i<768;i++)
		outp(0x3c9,palette[i]);
	return(templong);
}

getangle(long xvect,long yvect)
{
	long d1, d2, d3, daang, incr, i;

	daang = 1024; incr = 512;
	for(i=0;i<10;i++)
	{
		if (yvect*sintable[(daang+512)&2047] > xvect*sintable[daang&2047])
			daang += incr;
		else
			daang -= incr;
		incr >>= 1;
	}
	d1 = yvect*sintable[(daang+2559)&1023] - xvect*sintable[(daang+2047)&1023];
	d2 = yvect*sintable[(daang+2560)&1023] - xvect*sintable[(daang+2048)&1023];
	d3 = yvect*sintable[(daang+2561)&1023] - xvect*sintable[(daang+2049)&1023];
	if (d1 < d2) daang--;
	if (d3 < d2) daang++;
	daang = ((daang+2048)&2047);
	return(daang);
}

drawline640400(long x0, long y0, long x1, long y1, char col)
{
	char maskl, maskr;
	long i, xdime, ydime, cnt, pinc, p;

	if ((x0 < 0) && (x1 < 0)) return(0);
	if ((x0 >= 640) && (x1 >= 640)) return(0);
	if ((y0 < 0) && (y1 < 0)) return(0);
	if ((y0 >= 400) && (y1 >= 400)) return(0);

	if (x0 == x1)
	{
		outp(0x3c4, 0x2);
		outp(0x3c5, 1<<(x0&3));
		if (y1 < y0)
			i = y0, y0 = y1, y1 = i;
		if (y0 < 0) y0 = 0;
		if (y1 >= 400) y1 = 399;
		ydime = y1-y0+1;
		p = ((y0*640+x0)>>2)+0xa0000;
		for(i=ydime;i>0;i--)
			drawpixel(p,col), p+=160;
		return(0);
	}
	if (y0 == y1)
	{
		if (x1 < x0)
			i = x0, x0 = x1, x1 = i;
		if (x0 < 0) x0 = 0;
		if (x1 >= 640) x1 = 639;

		maskl = 16-(1<<(x0&3));
		maskr = (1<<((x1&3)+1))-1;

		p = y0*160 + (x0>>2)+0xa0000;

		outp(0x3c4,2);
		if ((x0|3) == (x1|3))
		{
			outp(0x3c5,maskl&maskr);
			drawpixel(p,col);
		}
		else
		{
			outp(0x3c5,maskl);
			drawpixel(p,col), p++;
			outp(0x3c5,15);
			xdime = (((x1|3)-(x0|3))>>2)-1;
			for(i=xdime;i>0;i--)
				drawpixel(p,col), p++;
			outp(0x3c5,maskr);
			drawpixel(p,col);
		}
		return(0);
	}

	if ((x0 < 0) && (x0 != x1)) y0 -= ((y1-y0)*x0) / (x1-x0), x0 = 0;
	if ((x1 < 0) && (x0 != x1)) y1 -= ((y1-y0)*x1) / (x1-x0), x1 = 0;
	if ((y0 < 0) && (y0 != y1)) x0 -= ((x1-x0)*y0) / (y1-y0), y0 = 0;
	if ((y1 < 0) && (y0 != y1)) x1 -= ((x1-x0)*y1) / (y1-y0), y1 = 0;
	if ((x0 >= 640) && (x0 != x1)) y0 += ((y1-y0)*(639-x0)) / (x1-x0), x0 = 639;
	if ((x1 >= 640) && (x0 != x1)) y1 += ((y1-y0)*(639-x1)) / (x1-x0), x1 = 639;
	if ((y0 >= 400) && (y0 != y1)) x0 += ((x1-x0)*(399-y0)) / (y1-y0), y0 = 399;
	if ((y1 >= 400) && (y0 != y1)) x1 += ((x1-x0)*(399-y1)) / (y1-y0), y1 = 399;

	if ((x0 < 0) && (x1 < 0)) return(0);
	if ((x0 >= 640) && (x1 >= 640)) return(0);
	if ((y0 < 0) && (y1 < 0)) return(0);
	if ((y0 >= 400) && (y1 >= 400)) return(0);

	if (x1 > x0)
		xdime = x1-x0+1;
	else
		xdime = x0-x1+1;
	if (y1 > y0)
		ydime = y1-y0+1;
	else
		ydime = y0-y1+1;

	outp(0x3c4, 0x2);
	if (xdime >= ydime)
	{
		if (x1 < x0)
		{
			i = x0; x0 = x1; x1 = i;
			i = y0; y0 = y1; y1 = i;
		}
		p = y0 * 640 + x0;
		cnt = 0;
		if (y1 > y0) pinc = 640; else pinc = -640;
		for(i=xdime;i>0;i--)
		{
			drawpixel640400(p,col);
			cnt += ydime;
			if (cnt >= xdime)
				cnt -= xdime, p += pinc;
			p++;
		}
	}
	if (xdime < ydime)
	{
		if (y1 < y0)
		{
			i = x0; x0 = x1; x1 = i;
			i = y0; y0 = y1; y1 = i;
		}
		p = y0 * 640 + x0;
		cnt = 0;
		if (x1 > x0) pinc = 1; else pinc = -1;
		for(i=ydime;i>0;i--)
		{
			drawpixel640400(p,col);
			cnt += xdime;
			if (cnt >= ydime)
				cnt -= ydime, p += pinc;
			p += 640;
		}
	}
	return(0);
}

qsetmode320200()
{
	short i;

	if (videocard == 1)
	{
		outp(0x3c4, 0x4); outp(0x3c5, inp(0x3c5)|8);
		outp(0x3d4, 0x17); outp(0x3d5, inp(0x3d5)&(255-64));
		outp(0x3d4, 0x14); outp(0x3d5, inp(0x3d5)|64);
		outpw(0x3c4, 0x0f02), fillscreen(0);
		outp(0x3d4, 0x13); outp(0x3d5,40);
		outp(0x3d4, 0x14); outp(0x3d5, inp(0x3d5)&(255 - 32));
		outp(0x3d4, 0x17); outp(0x3d5, inp(0x3d5)&(255 - 8));
		outp(0x3d4, 0x9); outp(0x3d5, inp(0x3d5)|1);
		inp(0x3da); outp(0x3c0, 0x10 + 32); outp(0x3c0, inp(0x3c1)|64);
		outp(0x3ce, 0x5); outp(0x3cf, inp(0x3cf)&(255-32)|64);
	}
	else
	{
		setvmode(0x13);
		outp(0x3c8,0);
		for(i=0;i<768;i++)
			outp(0x3c9,palette[i]);
	}
	return(0);
}

qsetmode640400()
{
	short i;

	if (videocard == 1)
	{
		outp(0x3c4, 0x4); outp(0x3c5, inp(0x3c5)&(255-8));
		outp(0x3d4, 0x17); outp(0x3d5, inp(0x3d5)|64);
		outp(0x3d4, 0x14); outp(0x3d5, inp(0x3d5)&(255-64));
		outpw(0x3c4, 0x0f02), fillscreen(0);
		outp(0x3d4, 0x13); outp(0x3d5, 80);
		outp(0x3d4, 0x14); outp(0x3d5, inp(0x3d5)|32);
		outp(0x3d4, 0x17); outp(0x3d5, inp(0x3d5)|8);
		outp(0x3d4, 0x9); outp(0x3d5, inp(0x3d5)&(255-1));
		inp(0x3da); outp(0x3c0, 0x10 + 32); outp(0x3c0, inp(0x3c1)&(255-64));
		outp(0x3ce, 0x5); outp(0x3cf, inp(0x3cf)&(255-32)|64);
	}
	if (videocard == 2)
	{
		vesasetmode(0x101);
		outp(0x3c4, 0x4); outp(0x3c5, inp(0x3c5)&(255-8));
		outp(0x3d4, 0x17); outp(0x3d5, inp(0x3d5)|64);
		outp(0x3d4, 0x14); outp(0x3d5, inp(0x3d5)&(255-64));
		outp(0x3c8,0);
		for(i=0;i<768;i++)
			outp(0x3c9,palette[i]);
	}
	return(0);
}

drawboardlinrot(long posxe, long posye, short ange, long zoome, short danumwalls)
{
	long i, j, xp1, yp1, xp2, yp2, tempx, tempy, minx, miny, maxx, maxy;
	long tempx2, tempy2, templong;
	char col;

	if (danumwalls < 0)
		danumwalls = numwalls;
	if (anglelock == 0)
	{
		drawline640400(320,200,120,0,8);
		drawline640400(320,200,520,0,8);
	}

	if (anglelock == 0)
	{
		i = (320<<14)/zoom, j = (200<<14)/zoom;
		tempx = ((i*(long)sintable[(2048-ang)&2047]-j*(long)sintable[(2560-ang)&2047])>>14);
		tempy = ((i*(long)sintable[(2560-ang)&2047]+j*(long)sintable[(2048-ang)&2047])>>14);
		minx = tempx, miny = tempy, maxx = tempx, maxy = tempy;
		if (-tempx < minx) minx = -tempx;
		if (-tempy < miny) miny = -tempy;
		if (-tempx > maxx) maxx = -tempx;
		if (-tempy > maxy) maxy = -tempy;
		i = (320<<14)/zoom, j = (200<<14)/zoom;
		tempx = ((i*(long)sintable[(2048-ang)&2047]-j*(long)sintable[(2560-ang)&2047])>>14);
		tempy = ((i*(long)sintable[(2560-ang)&2047]+j*(long)sintable[(2048-ang)&2047])>>14);
		if (tempx < minx) minx = tempx;
		if (tempy < miny) miny = tempy;
		if (tempx > maxx) maxx = tempx;
		if (tempy > maxy) maxy = tempy;
		if (-tempx < minx) minx = -tempx;
		if (-tempy < miny) miny = -tempy;
		if (-tempx > maxx) maxx = -tempx;
		if (-tempy > maxy) maxy = -tempy;
		minx += posx, miny += posy, maxx += posx, maxy += posy;
	}
	if (grid == 1)
		for(i=0;i<=65536;i+=1024)
		{
			if (anglelock == 0)
			{
				if ((i >= minx) || (i <= maxx))
				{
					xp1 = (((long)(i-posxe)*(long)sintable[(2048+ange)&2047]-(long)(-posye)*(long)sintable[(2560+ange)&2047])>>14);
					yp1 = (((long)(i-posxe)*(long)sintable[(2560+ange)&2047]+(long)(-posye)*(long)sintable[(2048+ange)&2047])>>14);
					xp2 = (((long)(i-posxe)*(long)sintable[(2048+ange)&2047]-(long)(65536-posye)*(long)sintable[(2560+ange)&2047])>>14);
					yp2 = (((long)(i-posxe)*(long)sintable[(2560+ange)&2047]+(long)(65536-posye)*(long)sintable[(2048+ange)&2047])>>14);
					xp1 = ((xp1*zoome)>>14);
					yp1 = ((yp1*zoome)>>14);
					xp2 = ((xp2*zoome)>>14);
					yp2 = ((yp2*zoome)>>14);
					drawline640400(320-xp1,200-yp1,320-xp2,200-yp2,10);
				}
			}
			else
			{
				xp1 = posye;
				yp1 = i-posxe;
				xp2 = posye-65536;
				yp2 = i-posxe;
				xp1 = ((xp1*zoome)>>14);
				yp1 = ((yp1*zoome)>>14);
				xp2 = ((xp2*zoome)>>14);
				yp2 = ((yp2*zoome)>>14);
				if ((yp1 > -200) && (yp1 <= 200))
					drawline640400(320-xp1,200-yp1,320-xp2,200-yp2,10);
			}
			if (anglelock == 0)
			{
				if ((i >= miny) || (i <= maxy))
				{
					xp1 = (((long)(0-posxe)*(long)sintable[(2048+ange)&2047]-(long)(i-posye)*(long)sintable[(2560+ange)&2047])>>14);
					yp1 = (((long)(0-posxe)*(long)sintable[(2560+ange)&2047]+(long)(i-posye)*(long)sintable[(2048+ange)&2047])>>14);
					xp2 = (((long)(65536-posxe)*(long)sintable[(2048+ange)&2047]-(long)(i-posye)*(long)sintable[(2560+ange)&2047])>>14);
					yp2 = (((long)(65536-posxe)*(long)sintable[(2560+ange)&2047]+(long)(i-posye)*(long)sintable[(2048+ange)&2047])>>14);
					xp1 = ((xp1*zoome)>>14);
					yp1 = ((yp1*zoome)>>14);
					xp2 = ((xp2*zoome)>>14);
					yp2 = ((yp2*zoome)>>14);
					drawline640400(320-xp1,200-yp1,320-xp2,200-yp2,10);
				}
			}
			else
			{
				xp1 = posye-i;
				yp1 = -posxe;
				xp2 = posye-i;
				yp2 = 65536-posxe;
				xp1 = ((xp1*zoome)>>14);
				yp1 = ((yp1*zoome)>>14);
				xp2 = ((xp2*zoome)>>14);
				yp2 = ((yp2*zoome)>>14);
				if ((xp1 > -320) && (xp1 <= 320))
				drawline640400(320-xp1,200-yp1,320-xp2,200-yp2,10);
			}
		}
	for(i=0;i<danumwalls;i++)
	{
		if (wall[i].nextsector1 == -1)
			col = 31;
		else
			col = 159;  //191 is blue
		if (i == linehighlight)
			col += (numframes&1);

		j = 0;
		if (anglelock == 0)
		{
			tempx = wall[i].xa1;
			tempy = wall[i].ya1;
			tempx2 = wall[wall[i].point2].xa1;
			tempy2 = wall[wall[i].point2].ya1;
			if (((minx<=tempx) == (tempx<=maxx)) || ((minx<=tempx2) == (tempx2<=maxx)) || ((miny<=tempy) == (tempy<=maxy)) || ((miny<=tempy2) == (tempy2<=maxy)))
			{
				xp1 = (((long)(tempx-posxe)*(long)sintable[(2048+ange)&2047]-(long)(tempy-posye)*(long)sintable[(2560+ange)&2047])>>14);
				yp1 = (((long)(tempx-posxe)*(long)sintable[(2560+ange)&2047]+(long)(tempy-posye)*(long)sintable[(2048+ange)&2047])>>14);
				xp2 = (((long)(tempx2-posxe)*(long)sintable[(2048+ange)&2047]-(long)(tempy2-posye)*(long)sintable[(2560+ange)&2047])>>14);
				yp2 = (((long)(tempx2-posxe)*(long)sintable[(2560+ange)&2047]+(long)(tempy2-posye)*(long)sintable[(2048+ange)&2047])>>14);
				j = 1;
			}
		}
		else
		{
			xp1 = (long)(posye-wall[i].ya1);
			yp1 = (long)(wall[i].xa1-posxe);
			xp2 = (long)(posye-wall[wall[i].point2].ya1);
			yp2 = (long)(wall[wall[i].point2].xa1-posxe);
			j = 1;
		}
		if (j == 1)
		{
			xp1 = ((xp1*zoome)>>14);
			yp1 = ((yp1*zoome)>>14);
			xp2 = ((xp2*zoome)>>14);
			yp2 = ((yp2*zoome)>>14);
			drawline640400(320-xp1,200-yp1,320-xp2,200-yp2,col);
			if (zoome >= 256)
				if (((320-xp1) >= 2) && ((320-xp1) <= 637))
					if (((200-yp1) >= 2) && ((200-yp1) <= 397))
					{
						templong = (200-yp1)*640L+(320-xp1);
						drawpixel640400(templong-2-1280,216);
						drawpixel640400(templong-1-1280,218);
						drawpixel640400(templong+0-1280,220);
						drawpixel640400(templong+1-1280,218);
						drawpixel640400(templong+2-1280,216);

						drawpixel640400(templong-2+1280,216);
						drawpixel640400(templong-1+1280,218);
						drawpixel640400(templong+0+1280,220);
						drawpixel640400(templong+1+1280,218);
						drawpixel640400(templong+2+1280,216);

						drawpixel640400(templong-2-640,218);
						drawpixel640400(templong-2+0,220);
						drawpixel640400(templong-2+640,218);

						drawpixel640400(templong+2-640,218);
						drawpixel640400(templong+2+0,220);
						drawpixel640400(templong+2+640,218);
					}
		}
	}
	if (anglelock == 0)
	{
		drawline640400(320,193,320,207,63);
		drawline640400(320,193,327,200,63);
		drawline640400(320,193,313,200,63);
	}
	else
	{
		xp1 = (sintable[(ange+2560)&2047]>>11);
		yp1 = (sintable[(ange+2048)&2047]>>11);
		drawline640400(320+yp1,200-xp1,320-yp1,200+xp1,63);
		drawline640400(320+yp1,200-xp1,320-xp1,200-yp1,63);
		drawline640400(320+yp1,200-xp1,320+xp1,200+yp1,63);
	}
}

overheadeditor()
{
	long i, j, k, mousxplc, mousyplc, firstx, firsty, oposz;
	long templong, tempbufcnt, startwall, endwall, dax, day, x1, y1, x2, y2;
	short pag, newall, suckwall, sucksect, newnumwalls, tempshort;
	short pointhighlight;

	searchx <<= 1;
	searchy <<= 1;
	oposz = posz;

	qsetmode640400();
	outpw(0x3d4,0x000c);
	outpw(0x3d4,0x000d);
	pag = 0;
	tempbufcnt = 0;
	newnumwalls = -1;
	keystatus[0x9c] = 0;
	while ((keystatus[0x9c]|keystatus[0x1]) == 0)
	{

		readmouse();
		searchx += mousx;
		searchy += mousy;
		if (searchx < 8) searchx = 8;
		if (searchx > 631) searchx = 631;
		if (searchy < 8) searchy = 8;
		if (searchy > 391) searchy = 391;
		getpoint(searchx,searchy,&mousxplc,&mousyplc);
		linehighlight = getlinehighlight(mousxplc, mousyplc);

		if (newnumwalls >= 0)
		{
			dax = mousxplc;
			day = mousyplc;
			adjustmark(&dax,&day,newnumwalls);
			wall[newnumwalls].xa1 = dax;
			wall[newnumwalls].ya1 = day;
		}

		drawboardlinrot(posx,posy,ang,zoom,newnumwalls);
		drawline640400(searchx,searchy-8,searchx,searchy-1,31);
		drawline640400(searchx+1,searchy-8,searchx+1,searchy-1,31);
		drawline640400(searchx,searchy+2,searchx,searchy+9,31);
		drawline640400(searchx+1,searchy+2,searchx+1,searchy+9,31);
		drawline640400(searchx-8,searchy,searchx-1,searchy,31);
		drawline640400(searchx-8,searchy+1,searchx-1,searchy+1,31);
		drawline640400(searchx+2,searchy,searchx+9,searchy,31);
		drawline640400(searchx+2,searchy+1,searchx+9,searchy+1,31);

		if ((bstatus&1) > 0)                //drag points
		{
			if ((bstatus&1) > (obstatus&1))
				pointhighlight = getpointhighlight(mousxplc, mousyplc);
			if (pointhighlight >= 0)
			{
				dax = mousxplc;
				day = mousyplc;
				if (grid == 1)
				{
					if (labs(dax-((dax+512)&0xfffffc00))+labs(day-((day+512)&0xfffffc00)) < 192)
					{
						dax = ((dax+512)&0xfffffc00);
						day = ((day+512)&0xfffffc00);
					}
					if (labs(dax-((dax+128)&0xffffff00))+labs(day-((day+128)&0xffffff00)) < 96)
					{
						dax = ((dax+128)&0xffffff00);
						day = ((day+128)&0xffffff00);
					}
				}

				wall[pointhighlight].xa1 = dax;
				wall[pointhighlight].ya1 = day;

				tempshort = pointhighlight;    //search points CCW
				do
				{
					if (wall[tempshort].nextwall1 >= 0)
					{
						tempshort = wall[wall[tempshort].nextwall1].point2;
						wall[tempshort].xa1 = dax;
						wall[tempshort].ya1 = day;
					}
					else
					{
						tempshort = pointhighlight;    //search points CW if not searched all the way around
						do
						{
							if (wall[lastwall(tempshort)].nextwall1 >= 0)
							{
								tempshort = wall[lastwall(tempshort)].nextwall1;
								wall[tempshort].xa1 = dax;
								wall[tempshort].ya1 = day;
							}
							else
							{
								break;
							}
						}
						while (tempshort != pointhighlight);
						break;
					}
				}
				while (tempshort != pointhighlight);

				asksave = 1;
			}
		}
		if ((bstatus&2) > 0)
		{
			searchx = 320;
			searchy = 200;
			posx = mousxplc;
			posy = mousyplc;
		}

		posz = 256;
		getsector(), engineinput();
		if ((keystatus[0x1e] > 0) && (zoom < 2048)) zoom += (zoom>>4);
		if ((keystatus[0x2c] > 0) && (zoom > 48)) zoom -= (zoom>>4);
		if (keystatus[0x22] > 0)  // G (grid on/off)
			grid = 1-grid, keystatus[0x22] = 0;
		if (keystatus[0x32] > 0)  // M (anglelock mode)
			anglelock = 1-anglelock, keystatus[0x32] = 0;

		if (keystatus[0x39] > 0)  //Space
		{
			keystatus[0x39] = 0;
			adjustmark(&mousxplc,&mousyplc,newnumwalls);
			if ((newnumwalls < 0) && (numwalls < maxwalls-1))
			{
				firstx = mousxplc, firsty = mousyplc;  //firstpoint
				newnumwalls = numwalls;
				suckwall = -1;
				wall[newnumwalls].xa1 = mousxplc;
				wall[newnumwalls].ya1 = mousyplc;
				wall[newnumwalls].nextsector1 = -1;
				wall[newnumwalls].nextwall1 = -1;
				for(i=0;i<numwalls;i++)
					if ((wall[i].xa1 == mousxplc) && (wall[i].ya1 == mousyplc))
						suckwall = i;
				wall[newnumwalls].point2 = newnumwalls+1;
				newnumwalls++;
			}
			else
			{
				if ((firstx != mousxplc) || (firsty != mousyplc))  //nextpoint
				{
					j = 0;
					for(i=numwalls;i<newnumwalls;i++)
						if ((mousxplc == wall[i].xa1) && (mousyplc == wall[i].ya1))
							j = 1;
					if (j == 0)
					{
						wall[newnumwalls].xa1 = mousxplc;
						wall[newnumwalls].ya1 = mousyplc;
						wall[newnumwalls].nextsector1 = -1;
						wall[newnumwalls].nextwall1 = -1;
						for(i=0;i<numwalls;i++)
							if ((wall[i].xa1 == mousxplc) && (wall[i].ya1 == mousyplc))
								suckwall = i;
						wall[newnumwalls].point2 = newnumwalls+1;
						newnumwalls++;
					}
				}
				else if (newnumwalls >= numwalls+3)
				{
					wall[newnumwalls-1].point2 = numwalls;
					if (suckwall == -1)
					{
						k = -1;
						for(i=0;i<numsectors;i++)
							if (inside(firstx,firsty,i) == 1)
								k = i;
						if (k == -1)
						{
							if (clockdir(numwalls) == 1)
								flipwalls(numwalls,newnumwalls);
							sector[numsectors].wallptr = numwalls;
							sector[numsectors].wallnum = newnumwalls-numwalls;
							sector[numsectors].cstat1 = 0;
							sector[numsectors].cstat2 = 0;
							sector[numsectors].zosc1 = 0;
							sector[numsectors].zosc2 = 0;
							sector[numsectors].shade1 = 0;
							sector[numsectors].shade2 = 0;
							sector[numsectors].z1 = -32;
							sector[numsectors].z2 = 32;
							sector[numsectors].picnum1 = 0;
							sector[numsectors].picnum2 = 0;
							sector[numsectors].heinum1 = 0;
							sector[numsectors].heinum2 = 0;
							for(i=numwalls;i<newnumwalls;i++)
							{
								wall[i].cstat = 0;
								wall[i].shade = 0;
								wall[i].xrepeat = 8;
								wall[i].yrepeat = 8;
								wall[i].picnum = 0;
								wall[i].nextsector1 = -1;
								wall[i].nextwall1 = -1;
								wall[i].nextsector2 = -1;
								wall[i].nextwall2 = -1;
							}
							numsectors++;
						}
						else
						{
							if (clockdir(numwalls) == 0)
								flipwalls(numwalls,newnumwalls);

							j = newnumwalls-numwalls;

							sector[k].wallnum += j;
							for(i=k+1;i<numsectors;i++)
								sector[i].wallptr += j;
							suckwall = sector[k].wallptr;

							for(i=0;i<numwalls;i++)
							{
								if (wall[i].nextwall1 >= suckwall)
									wall[i].nextwall1 += j;
								if (wall[i].point2 >= suckwall)
									wall[i].point2 += j;
							}

							for(i=newnumwalls-1;i>=suckwall;i--)
								memcpy(&wall[i+j],&wall[i],sizeof(wall[0]));
							for(i=0;i<j;i++)
								memcpy(&wall[i+suckwall],&wall[i+newnumwalls],sizeof(wall[0]));

							for(i=suckwall;i<suckwall+j;i++)
							{
								wall[i].point2 += (suckwall-numwalls);

								wall[i].cstat = wall[suckwall+j].cstat;
								wall[i].shade = wall[suckwall+j].shade;

								dax = wall[i].xa1-wall[wall[i].point2].xa1;
								day = wall[i].ya1-wall[wall[i].point2].ya1;
								wall[i].xrepeat = (unsigned char)(ksqrt(dax*dax+day*day)>>(13-(picsiz[wall[suckwall+j].picnum]&15)));

								wall[i].yrepeat = wall[suckwall+j].yrepeat;
								wall[i].picnum = wall[suckwall+j].picnum;

								wall[i].nextsector1 = -1;
								wall[i].nextwall1 = -1;
								wall[i].nextsector2 = -1;
								wall[i].nextwall2 = -1;
							}
						}
					}
					else
					{
						if (clockdir(numwalls) == 1)
							flipwalls(numwalls,newnumwalls);

						sector[numsectors].wallptr = numwalls;
						sector[numsectors].wallnum = newnumwalls-numwalls;
						for(i=0;i<numsectors;i++)
						{
							startwall = sector[i].wallptr;
							endwall = startwall + sector[i].wallnum - 1;
							if ((suckwall >= startwall) && (suckwall <= endwall))
								sucksect = i;
						}
						sector[numsectors].cstat1 = sector[sucksect].cstat1;
						sector[numsectors].cstat2 = sector[sucksect].cstat2;
						sector[numsectors].zosc1 = sector[sucksect].zosc1;
						sector[numsectors].zosc2 = sector[sucksect].zosc2;
						sector[numsectors].shade1 = sector[sucksect].shade1;
						sector[numsectors].shade2 = sector[sucksect].shade2;
						sector[numsectors].z1 = sector[sucksect].z1;
						sector[numsectors].z2 = sector[sucksect].z2;
						sector[numsectors].picnum1 = sector[sucksect].picnum1;
						sector[numsectors].picnum2 = sector[sucksect].picnum2;
						sector[numsectors].heinum1 = sector[sucksect].heinum1;
						sector[numsectors].heinum2 = sector[sucksect].heinum2;
						for(i=numwalls;i<newnumwalls;i++)
						{
							wall[i].cstat = wall[suckwall].cstat;
							wall[i].shade = wall[suckwall].shade;

							dax = wall[i].xa1-wall[wall[i].point2].xa1;
							day = wall[i].ya1-wall[wall[i].point2].ya1;
							wall[i].xrepeat = (unsigned char)(ksqrt(dax*dax+day*day)>>(13-(picsiz[wall[suckwall].picnum]&15)));

							wall[i].yrepeat = wall[suckwall].yrepeat;
							wall[i].picnum = wall[suckwall].picnum;

							wall[i].nextsector1 = -1;
							wall[i].nextwall1 = -1;
							x1 = wall[i].xa1;
							y1 = wall[i].ya1;
							x2 = wall[wall[i].point2].xa1;
							y2 = wall[wall[i].point2].ya1;
							for(j=0;j<numsectors;j++)
							{
								startwall = sector[j].wallptr;
								endwall = startwall + sector[j].wallnum - 1;
								for(k=startwall;k<=endwall;k++)
								{
									if ((wall[k].xa1 == x2) && (wall[k].ya1 == y2))
										if ((wall[wall[k].point2].xa1 == x1) && (wall[wall[k].point2].ya1 == y1))
										{
											wall[i].nextsector1 = j;
											wall[i].nextwall1 = k;
											wall[k].nextsector1 = numsectors;
											wall[k].nextwall1 = i;
										}
								}
							}

							wall[i].nextsector2 = -1;
							wall[i].nextwall2 = -1;
						}
						numsectors++;
					}
					numwalls = newnumwalls;
					newnumwalls = -1;
					asksave = 1;
				}
				else
				{
					newnumwalls = -1;
				}
			}
		}

		if ((keystatus[0x0e] > 0) && (linehighlight >= 0)) //Backspace
		{
			if (newnumwalls > numwalls)
			{
				newnumwalls--;
				asksave = 1;
				keystatus[0x0e] = 0;
			}
			if (newnumwalls == numwalls)
			{
				newnumwalls = -1;
				asksave = 1;
				keystatus[0x0e] = 0;
			}
		}

		if (((keystatus[0xd3]&keystatus[0x1d]) > 0) && (linehighlight >= 0)) //Delete
		{
			sucksect = cursectnum;
			startwall = sector[sucksect].wallptr;
			endwall = startwall + sector[sucksect].wallnum - 1;
			j = sector[sucksect].wallnum;

			for(i=sucksect;i<numsectors-1;i++)
			{
				memcpy(&sector[i],&sector[i+1],sizeof(sector[0]));
				sector[i].wallptr -= j;
			}
			numsectors--;
			deletewalls(startwall,endwall);
			asksave = 1;
			keystatus[0xd3] = 0;
		}

		if ((keystatus[0xd3] > 0) && (linehighlight >= 0))  //MiniDelete
		{
			for(i=0;i<numsectors;i++)
			{
				startwall = sector[i].wallptr;
				endwall = startwall + sector[i].wallnum - 1;
				if ((linehighlight >= startwall) && (linehighlight <= endwall))
					sucksect = i;
			}
			sector[sucksect].wallnum--;
			for(i=sucksect+1;i<numsectors;i++)
				sector[i].wallptr--;

			j = wall[linehighlight].point2;
			wall[linehighlight].point2 = wall[j].point2;
			wall[linehighlight].xa1 = ((wall[linehighlight].xa1+wall[j].xa1)>>1);
			wall[linehighlight].ya1 = ((wall[linehighlight].ya1+wall[j].ya1)>>1);
			deletewalls(j,j);
			asksave = 1;
			keystatus[0xd3] = 0;
		}

		if (videocard == 1) outpw(0x3d4,0x33+(pag<<8));
		pag++;
		if (pag == 3) pag = 0;
		if (videocard == 1) outp(0x3cd,pag);
		numframes++;
		totalclock += ((long)clockspeed), clockspeed = 0;
		outpw(0x3c4,0x0f02), fillscreen(0);
	}
	qsetmode320200();
	if (videocard == 1) outpw(0x3d4,0x0033);

	posz = oposz;
	searchx >>= 1;
	searchy >>= 1;
}

getpoint(long searchxe, long searchye, long *x1, long *y1)
{
	long x2, y2;

	x2 = ((searchxe-320)<<14)/zoom;
	y2 = ((searchye-200)<<14)/zoom;
	if (anglelock == 0)
	{
		*x1 = ((x2*(long)sintable[(2048-ang)&2047]-y2*(long)sintable[(2560-ang)&2047])>>14);
		*y1 = ((x2*(long)sintable[(2560-ang)&2047]+y2*(long)sintable[(2048-ang)&2047])>>14);
	}
	else
	{
		*x1 = -y2;
		*y1 = x2;
	}
	*x1 += posx;
	*y1 += posy;
	if (*x1 < 0) *x1 = 0;
	if (*x1 >= 65536) *x1 = 65535;
	if (*y1 < 0) *y1 = 0;
	if (*y1 >= 65536) *y1 = 65535;
}

getlinehighlight(long xplc, long yplc)
{
	long i, dst, dist, closest;

	if (numwalls == 0)
		return(-1);
	dist = 0x7fffffff;
	closest = numwalls-1;
	for(i=0;i<numwalls;i++)
	{
		dst = labs(xplc - ((wall[i].xa1 + wall[wall[i].point2].xa1)>>1));
		dst += labs(yplc - ((wall[i].ya1 + wall[wall[i].point2].ya1)>>1));
		if (dst <= dist)
			dist = dst, closest = i;
	}
	return(closest);
}

getpointhighlight(long xplc, long yplc)
{
	long i, dst, dist, closest;

	if (numwalls == 0)
		return(-1);
	dist = 512;
	closest = -1;
	for(i=0;i<numwalls;i++)
	{
		dst = labs(xplc - wall[i].xa1) + labs(yplc - wall[i].ya1);
		if (dst <= dist)
			dist = dst, closest = i;
	}
	return(closest);
}

adjustmark(long *xplc, long *yplc, short danumwalls)
{
	long i, dst, dist, tempx, tempy;

	if (danumwalls < 0)
		danumwalls = numwalls;
	dist = 192;
	tempx = *xplc;
	tempy = *yplc;
	for(i=0;i<danumwalls;i++)
	{
		dst = labs((*xplc) - wall[i].xa1) + labs((*yplc) - wall[i].ya1);
		if (dst < dist)
		{
			dist = dst;
			tempx = wall[i].xa1;
			tempy = wall[i].ya1;
		}
	}
	if (grid == 1)
	{
		if (dist == 192L)
		{
			dst = labs((*xplc)-((*xplc+512)&0xfffffc00))+labs((*yplc)-((*yplc+512)&0xfffffc00));
			if (dst < dist)
			{
				dist = dst;
				tempx = ((*xplc+512)&0xfffffc00);
				tempy = ((*yplc+512)&0xfffffc00);
			}
		}
		if (dist == 192L)
		{
			dst = labs((*xplc)-((*xplc+128)&0xffffff00))+labs((*yplc)-((*yplc+128)&0xffffff00));
			if ((dst < dist) && (dst < 96))
			{
				dist = dst;
				tempx = ((*xplc+128)&0xffffff00);
				tempy = ((*yplc+128)&0xffffff00);
			}
		}
	}
	*xplc = tempx;
	*yplc = tempy;
	return(0);
}

ksqrt(long num)
{
	long root, temp;

	root = 128;
	do
	{
		temp = root;
		root = ((root+(num/root))>>1);
	}
	while (labs(temp-root) > 1);
	return(root);
}

clockdir(short wallstart)   //Returns: 0 is CW, 1 is CCW
{
	short i, themin;
	long minx, templong, x0, x1, x2, y0, y1, y2;

	minx = 0x7fffffff;
	themin = -1;
	i = wallstart-1;
	do
	{
		i++;
		if (wall[wall[i].point2].xa1 < minx)
		{
			minx = wall[wall[i].point2].xa1;
			themin = i;
		}
	}
	while ((wall[i].point2 != wallstart) && (i < maxwalls));

	x0 = wall[themin].xa1;
	y0 = wall[themin].ya1;
	x1 = wall[wall[themin].point2].xa1;
	y1 = wall[wall[themin].point2].ya1;
	x2 = wall[wall[wall[themin].point2].point2].xa1;
	y2 = wall[wall[wall[themin].point2].point2].ya1;

	if ((y1 >= y2) && (y1 <= y0)) return(0);
	if ((y1 >= y0) && (y1 <= y2)) return(1);

	templong = (x0-x1)*(y2-y1) - (x2-x1)*(y0-y1);
	if (templong < 0)
		return(0);
	else
		return(1);
}

flipwalls(short numwalls, short newnumwalls)
{
	long i, j, nume, templong;

	nume = newnumwalls-numwalls;

	for(i=numwalls;i<numwalls+(nume>>1);i++)
	{
		j = numwalls+newnumwalls-i-1;
		templong = wall[i].xa1; wall[i].xa1 = wall[j].xa1; wall[j].xa1 = templong;
		templong = wall[i].ya1; wall[i].ya1 = wall[j].ya1; wall[j].ya1 = templong;
	}
}

copytilepiece(walnume1,x1,y1,xsiz,ysiz,walnume2,x2,y2,shadeoffs)
long walnume1, x1, y1, xsiz, ysiz, walnume2, x2, y2;
char shadeoffs;
{
	long xsiz1, ysiz1, xsiz2, ysiz2, picplc1, picplc2, i, j, ptr1, ptr2;

	if ((picsiz[walnume1] != 0) && (picsiz[walnume2] != 0))
	{
		xsiz1 = (1<<(picsiz[walnume1]&15));
		ysiz1 = (1<<(picsiz[walnume1]>>4));
		xsiz2 = (1<<(picsiz[walnume2]&15));
		ysiz2 = (1<<(picsiz[walnume2]>>4));
		for(i=x1;i<x1+xsiz;i++)
			for(j=y1;j<y1+ysiz;j++)
				if ((i < xsiz1) && (j < ysiz1) && (i-x1+x2 < xsiz2) && (j-y1+y2 < ysiz2))
				{
					ptr1 = waloff[walnume1] + i * ysiz1 + j;
					ptr2 = waloff[walnume2] + (i-x1+x2) * ysiz2 + (j-y1+y2);
					memcopybyte(ptr2,ptr1,shadeoffs);
				}
	}
}

showpositioncapture(char filename[20])
{
	long i, fil, framenumber;

	if ((fil = open(filename,O_BINARY|O_RDWR,S_IREAD)) != -1)
	{
		read(fil,&framenumber,4);
		while ((framenumber > 0) && ((keystatus[0x1c]|keystatus[0x9c]|keystatus[1]|keystatus[0x39]) == 0))
		{
			read(fil,&posx,4);
			read(fil,&posy,4);
			read(fil,&posz,4);
			read(fil,&ang,2);

			draw3dscreen();
			nextpage();

			framenumber--;
		}
		close(fil);
	}
}

printext(long x, long y, char buffer[42])
{
	long p, bufplc;
	char *ptr, ch;
	unsigned char dat;
	short i, dax, day;

	bufplc = waloff[texttilenum];
	p = y*320+x+frameplace;
	i = 0;
	while (buffer[i] != 0)
	{
		ch = buffer[i];
		ptr = (char *)(bufplc+(((long)ch&15)<<9)+(((long)ch>>4)<<3));
		for(dax=0;dax<8;dax++)
		{
			for(day=0;day<8;day++)
			{
				dat = *ptr;
				if (dat != 255) drawpixel(p,(long)dat);
				p += 320;
				ptr++;
			}
			p = (p-(320*8)+1);
			ptr += (64-8);
		}
		i++;
	}
	return(0);
}

lastwall(short pointhighlight)
{
	long cnt, tempshort;

	tempshort = pointhighlight;
	cnt = maxwalls;
	while ((wall[tempshort].point2 != pointhighlight) && (cnt > 0))
	{
		tempshort = wall[tempshort].point2;
		cnt--;
	}
	if (cnt <= 0)
		tempshort = pointhighlight;
	return(tempshort);
}

deletewalls(long startwall, long endwall)
{
	long i, j;

	j = endwall-startwall+1;

	for(i=startwall;i<=endwall;i++)
		if (wall[i].nextwall1 != -1)
		{
			wall[wall[i].nextwall1].nextwall1 = -1;
			wall[wall[i].nextwall1].nextsector1 = -1;
		}
	for(i=startwall;i<numwalls-j;i++)
		memcpy(&wall[i],&wall[i+j],sizeof(wall[0]));
	numwalls -= j;
	for(i=0;i<numwalls;i++)
	{
		if (wall[i].nextwall1 >= startwall)
		{
			wall[i].nextsector1--;
			wall[i].nextwall1 -= j;
		}
		if (wall[i].point2 >= startwall)
			wall[i].point2 -= j;
	}
}

// snotbuf[0] = ((i/100)%10)+48;
// snotbuf[1] = ((i/10)%10)+48;
// snotbuf[2] = (i%10)+48;
// snotbuf[3] = 0;
// printext(0L,0L,snotbuf);
