#include <dos.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <conio.h>

static char *vox;
static char scanext[256], scanextcnt;
static short int sintable[2048];
static long tantable[1024], voxsiz, voxsqd, times80[256];
static long pageoffset;
static unsigned long scanplc[256];

volatile unsigned char readch, oldreadch, keystatus[256], extended;
void (__interrupt __far *oldhand)();
void (__interrupt __far *oldkeyhandler)();

#pragma aux setvmode =\
	"int 10h",\
	parm [eax]\

#pragma aux drawpixel =\
	"mov byte ptr es:[edi+0x000a0000], al",\
	parm [edi][al]\
	modify [edi al]\

#pragma aux readkey =\
	"in al, 0x60",\
	"mov readch, al",\
	"in al, 0x61",\
	"or al, 0x80",\
	"out 0x61, al",\
	"and al, 0x7f",\
	"out 0x61, al",\

#pragma aux interruptend =\
	"mov al, 20h",\
	"out 20h, al",\

void __interrupt __far keyhandler()
{
	oldreadch = readch;
	readkey();
	if ((readch|1) == 0xe1)
		extended = 128;
	else
	{
		if (oldreadch != readch)
			keystatus[(readch&127)+extended] = ((readch>>7)^1);
		extended = 0;
	}
	interruptend();
}

main(short int argc,char **argv)
{
	long int i, posx, posy, posz, ang;

	voxsiz = 16;
	if (argc == 2)
	{
		voxsiz = atoi(argv[1]);
		if (voxsiz < 8) voxsiz = 8;
		if (voxsiz > 128) voxsiz = 128;
	}
	voxsqd = voxsiz*voxsiz;
	vox = (char *)malloc(voxsiz*voxsiz*voxsiz);

	setvmode(0x13);
	outp(0x3d4,0x14); outp(0x3d5, inp(0x3d5)&(255-64));  //WORD MODE
	outp(0x3d4,0x17); outp(0x3d5, inp(0x3d5)|64);        //2-WORD MODE
	outp(0x3c4,0x4); outp(0x3c5, inp(0x3c5)&247);        //CHAIN 4 OFF
	outp(0x3d4,0x13); outp(0x3d5, 40);                   //Bytes per line

	loadtables();
	setupvoxbuffer();
	for(i=0;i<256;i++)
		times80[i] = i*80;

	oldkeyhandler = _dos_getvect(0x9);
	_disable(); _dos_setvect(0x9, keyhandler); _enable();

	posx = (voxsiz<<15), posy = (voxsiz<<15), posz = (voxsiz<<15), ang = 384;
	pageoffset = 0;
	while (keystatus[1] == 0)
	{
		drawvoxel(posx,posy,posz,ang);

		outp(0x3d4,0xc); outp(0x3d5,(pageoffset>>8));
		pageoffset += 16384;
		if (pageoffset > 32768) pageoffset = 0;

		if (keystatus[0x1e] > 0) posz-=8192;
		if (keystatus[0x2c] > 0) posz+=8192;
		if (keystatus[0xcb] > 0) ang -= 32;
		if (keystatus[0xcd] > 0) ang += 32;
		if (keystatus[0x2a] == 0)
		{
			if (keystatus[0xc8] > 0)
			{
				posx += sintable[(ang+2560)&2047];
				posy += sintable[ang&2047];
			}
			if (keystatus[0xd0] > 0)
			{
				posx -= sintable[(ang+2560)&2047];
				posy -= sintable[ang&2047];
			}
		}
		else
		{
			if (keystatus[0xc8] > 0)
			{
				posx += (sintable[(ang+2560)&2047]<<1);
				posy += (sintable[ang&2047]<<1);
			}
			if (keystatus[0xd0] > 0)
			{
				posx -= (sintable[(ang+2560)&2047]<<1);
				posy -= (sintable[ang&2047]<<1);
			}
		}
	}
	_dos_setvect(0x9, oldkeyhandler);
	free(vox);
	setvmode(0x3);
}

loadtables()
{
	short int fil, i;

	if ((fil = open("tables.dat",O_BINARY|O_RDWR,S_IREAD)) != -1)
	{
		read(fil,&sintable[0],4096);
		read(fil,&tantable[0],4096);
		close(fil);
	}
	return(0);
}

setupvoxbuffer()
{
	long int i, j, k;

	for(i=0;i<voxsiz;i++)
		for(j=0;j<voxsiz;j++)
			for(k=0;k<voxsiz;k++)
			{
				vox[i*voxsqd+j*voxsiz+k] = 0;
				if (k < (voxsiz>>1))
					if ((i-(voxsiz>>1))*(i-(voxsiz>>1))+(j-(voxsiz>>1))*(j-(voxsiz>>1))+(k-(voxsiz>>1))*(k-(voxsiz>>1)) >= ((voxsiz>>1)-1)*((voxsiz>>1)-1))
						vox[i*voxsqd+j*voxsiz+k] = rand()%72+32;
				if ((i == 0) || (j == 0) || (k == 0) || (i == voxsiz-1) || (j == voxsiz-1) || (k == voxsiz-1))
					vox[i*voxsqd+j*voxsiz+k] = ((i+j+k)&63)+32;
				if (((k&3) == 0) && (k > (voxsiz>>1)) && ((i == (voxsiz>>2)) || (i == voxsiz-(voxsiz>>2)) || (j == (voxsiz>>2)) || (j == voxsiz-(voxsiz>>2))))
					vox[i*voxsqd+j*voxsiz+k] = ((i+j+k)&63)+32;
			}
}

drawvoxel(posx, posy, posz, ang)
long int posx, posy, posz, ang;
{
	char dat;
	short int i, j, k;
	long int xc, incrx, incry, incrz, xyz, incrzx, incrzy, xyc, zc1, zc2;
	long int voxplc, xd, yd, zd, snx, sny, thexyc, p, scancnt;

	for(i=0;i<=25;i++)
		scanext[i] = (i<<3);
	for(i=0;i<320;i+=2)
	{
		p = pageoffset+(i>>2);
		outp(0x3c4,2); outp(0x3c5,3<<(i&3));

		xc = (i-160)*100;
		incrx = sintable[(ang+2560)&2047] - ((xc*sintable[ang&2047])>>14);
		incry = ((xc*sintable[(ang+2560)&2047])>>14) + sintable[ang&2047];
		if (incrx < 0)
		{
			xd = -voxsqd;
			incrx = -incrx;
			snx = (posx&65535);
		}
		else
		{
			xd = voxsqd;
			snx = (posx&65535^65535);
		}
		if (incry < 0)
		{
			yd = -voxsiz;
			incry = -incry;
			sny = (posy&65535);
		}
		else
		{
			yd = voxsiz;
			sny = (posy&65535^65535);
		}
		thexyc = ((incry*snx - incrx*sny)>>16);

		scanextcnt = 26;
		k = 0;
		while (k < scanextcnt)
		{
			j = scanext[k++];
			if (j < 100)
			{
				incrz = (100-j)*100;
				zd = -1;
				zc1 = (posz&65535) + (incrz*(snx^65535)) / (incrx+1);
				zc2 = (posz&65535) + (incrz*(sny^65535)) / (incry+1);
			}
			else
			{
				incrz = (j-100)*100;
				zd = 1;
				zc1 = (posz&65535^65535) + (incrz*(snx^65535)) / (incrx+1);
				zc2 = (posz&65535^65535) + (incrz*(sny^65535)) / (incry+1);
			}

			xyc = thexyc;
			incrzx = (incrz<<16) / (incrx+1);
			incrzy = (incrz<<16) / (incry+1);
        
			voxplc = (posx>>16)*voxsqd+(posy>>16)*voxsiz+(posz>>16);
			scancnt = 256L;

			do
			{
				if (xyc < 0)
				{
					zc1 -= incrzx;
					if (zc1 < 0)
					{
						zc1 += 65536;
						zc2 += 65536;
						voxplc += zd;
						if (vox[voxplc] != 0)
						{
							dat = vox[voxplc]+144;
							break;
						}
					}
					xyc += incry;
					voxplc += xd;
					if (vox[voxplc] != 0)
					{
						dat = vox[voxplc];
						break;
					}
				}
				else
				{
					zc2 -= incrzy;
					if (zc2 < 0)
					{
						zc1 += 65536;
						zc2 += 65536;
						voxplc += zd;
						if (vox[voxplc] != 0)
						{
							dat = vox[voxplc]+144;
							break;
						}
					}
					xyc -= incrx;
					voxplc += yd;
					if (vox[voxplc] != 0)
					{
						dat = vox[voxplc]+72;
						break;
					}
				}
			}
			while (scancnt-- > 0);
			scanplc[j] = voxplc+(((unsigned long)dat)<<24);
			switch(j&7)
			{
				case 0:
					drawpixel(p+times80[j],dat);
					if (j > 0)
					{
						if (scanplc[j] == scanplc[j-8])
						{
							drawpixel(p+times80[j]-560,dat);
							drawpixel(p+times80[j]-480,dat);
							drawpixel(p+times80[j]-400,dat);
							drawpixel(p+times80[j]-320,dat);
							drawpixel(p+times80[j]-240,dat);
							drawpixel(p+times80[j]-160,dat);
							drawpixel(p+times80[j]-80,dat);
						}
						else
							scanext[scanextcnt++] = j-4;
					}
					break;
				case 4:
					drawpixel(p+times80[j],dat);
					if (scanplc[j] == scanplc[j-4])
					{
						drawpixel(p+times80[j]-240,dat);
						drawpixel(p+times80[j]-160,dat);
						drawpixel(p+times80[j]-80,dat);
					}
					else
						scanext[scanextcnt++] = j-2;
					if (scanplc[j] == scanplc[j+4])
					{
						drawpixel(p+times80[j]+80,dat);
						drawpixel(p+times80[j]+160,dat);
						drawpixel(p+times80[j]+240,dat);
					}
					else
						scanext[scanextcnt++] = j+2;
					break;
				case 2: case 6:
					drawpixel(p+times80[j],dat);
					if (scanplc[j] == scanplc[j-2])
						drawpixel(p+times80[j]-80,dat);
					else
						scanext[scanextcnt++] = j-1;
					if (scanplc[j] == scanplc[j+2])
						drawpixel(p+times80[j]+80,dat);
					else
						scanext[scanextcnt++] = j+1;
					break;
				default:
					drawpixel(p+times80[j],dat);
					break;
			}
		}
	}
	return(0);
}
