// "Build Engine & Tools" Copyright (c) 1993-1997 Ken Silverman
// Ken Silverman's official web site: "http://www.advsys.net/ken"
// See the included license file "BUILDLIC.TXT" for license info.

#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <dos.h>
#include <stdlib.h>
#include <conio.h>
#include <stdio.h>

#define maxlines 2048
#define numwalls 64
#define maxscandist 32
#define numtimecounts 9
#define linxa1 0
#define linya1 2
#define linxa2 4
#define linya2 6
#define linpicnum 8
#define linpicnum1 10
#define linpicnum2 12
#define linshade 14
#define linshade1 15
#define linshade2 16
#define linz1 17
#define linz2 18
#define lingot 19
#define lincstat 20

static long timetakercount[numtimecounts], takernum;
static char timenames[numtimecounts][25] =
{
	"Main loop               ",
	"Addbox                  ",
	"Shell sort              ",
	"Infront sort            ",
	"Individual wall loop    ",
	"Vline calculations      ",
	"Vline drawing           ",
	"Hline calculations      ",
	"Hline drawing           ",
};
static int jmptable[10], numlines, capturecount;
static unsigned pic, walseg[numwalls+1], lin;
static unsigned pageoffset, lastpageoffset, centeroffset, scrsize;
static long posx, posy, posz, vel, svel, hvel, angvel, moustat, newposz;
static long xb1[64], yb1[64], xb2[64], yb2[64];
static int theline[64], ceilingstat, uxsave[242], dxsave[242];
static unsigned char swapped[64];
static int vidmode, xdim, ydim, halfwidth, eigthwidth, halfheight, clockspeed;
static int xtotaldim, ytotaldim, closestline, lintalplc, numhits;
static int sintable[2048];
static long tantable[1024], totalclock, numframes;
static unsigned int lintaladdr[4097], lintal, lintalcnt, ang, dside;
static unsigned char tempbuf[4096], palette[768];
static int umost[361], dmost[361], uplace[361], dplace[361], linum[361];
static unsigned char uzmost[361], dzmost[361];
static unsigned char readch, oldreadch, keystatus[256], extended;
void interrupt far ksmhandler(void);
void (interrupt far *oldhand)();
void interrupt far keyhandler(void);
void (interrupt far *oldkeyhandler)();


main(int argc,char **argv)
{
	unsigned l;
	long templong, templong1, snx, sny;
	int i, j, k, mousx, mousy;
	int px, py, x1, y1, x2, y2, xdir, ydir, scandist, x, y, ox, oy;
	int incrx, incry, cnt1, cnt2;

	ceilingstat = 1;
	vidmode = 0;
	xdim = 320;
	ydim = 200;
	xtotaldim = 320;
	ytotaldim = 200;
	scrsize = 18000;
	if (argc == 3)
	{
		xdim = (atoi(argv[1])&0xfff8);
		ydim = (atoi(argv[2])&0xfff8);
		if (xdim > 360) xdim = 360;
		if (xdim < 8) xdim = 8;
		if (ydim > 240) ydim = 240;
		if (ydim < 8) ydim = 8;
		if ((xdim > 320) || (ydim > 200))
		{
			vidmode = 1;
			xtotaldim = 360;
			ytotaldim = 240;
			scrsize = 21600;
		}
	}
	centeroffset = (((ytotaldim-ydim)>>1)*90) + ((xtotaldim-xdim)>>3);
	if ((pic = convallocate(numwalls*256)) == -1)
	{
		printf("Allocation failed!");
		exit(-1);
	}
	if ((lin = convallocate(maxlines<<1)) == -1)    //2048 abs. Max here (64K)
	{
		printf("Allocation failed!");
		exit(-1);
	}
	if ((lintal = convallocate(maxlines>>3)) == -1)      //Possibly Dangerous!
	{
		printf("Allocation failed!");
		exit(-1);
	}
	l = pic;
	for(i=0;i<numwalls;i++)
	{
		walseg[i] = l;
		l += 256;
	}
	for(i=0;i<242;i++)
	{
		uxsave[i] = -1;
		dxsave[i] = -1;
	}
	capturecount = 0;
	pageoffset = 0;
	lastpageoffset = 0;
	dside = ydim;
	halfwidth = (xdim>>1);
	eigthwidth = (xdim>>3);
	halfheight = (ydim>>1);
	loadtables();
	moustat = setupmouse();
	k = 0;
	if ((i = open("palette.dat",O_BINARY|O_RDWR,S_IREAD)) != -1)
	{
		read(i,&palette[0],768);
		close(i);
	}
	setgamevidmode();
	_asm \
	{
		cld
		mov dx, 0x3c4
		mov al, 2
		out dx, al
		inc dx
		mov al, 15
		out dx, al
		mov ax, 0xa000
		mov es, ax
		mov cx, 32768
		xor ax, ax
		xor di, di
		rep stosw
	}
	for(i=0;i<256;i++)
		keystatus[i] = 0;
	oldkeyhandler = _dos_getvect(0x9);
	_disable(); _dos_setvect(0x9, keyhandler); _enable();
	clockspeed = 0;
	totalclock = 0;
	takernum = 0;
	for(i=0;i<numtimecounts;i++)
		timetakercount[i] = 0;
	outp(0x43,54); outp(0x40,4972&255); outp(0x40,4972>>8);
	oldhand = _dos_getvect(0x8);
	_disable(); _dos_setvect(0x8, ksmhandler); _enable();
	loadpics();
	loadboard();
	if (argc == 5)
	{
		posx = (((long)atoi(argv[3]))<<5);
		posy = (((long)atoi(argv[4]))<<5);
	}
	posz = 40960L;
	mousx = 0;
	mousy = 0;
	numframes = 0;
	lintalplc = 0;
	outp(67,182);
	outp(66,255);
	outp(66,20);
	while (keystatus[1] == 0)
	{
		if (ceilingstat != 1)
			_asm \
			{
				mov dx, 0x3c4
				mov al, 0x2
				out dx, al
				inc dx
				mov al, 0xf
				out dx, al
				cld
				mov ax, 0xa000
				mov es, ax
				mov di, pageoffset
				mov cx, scrsize
				shr cx, 1
				shr cx, 1
				mov ax, 0xe3e3
				rep stosw
				mov cx, scrsize
				shr cx, 1
				shr cx, 1
				mov ax, 0x8484
				rep stosw
			}
		for(i=0;i<xdim;i++)
		{
			umost[i] = 0;
			dmost[i] = dside;
		}
		_asm \
		{
			mov es, lin       ;for(i=0;i<numlines;i++) got[i] = 0;
			mov cx, numlines
			xor bx, bx
			xor al, al
cleargots:
			mov byte ptr es:[bx+lingot], al
			add bx, 32
			loop cleargots
		}

		numhits = xdim;
		closestline = 0;
		px = (int)(posx>>10); py = (int)(posy>>10);
		incrx = sintable[(ang+256)&2047];
		incry = sintable[(ang+1792)&2047];
		xdir = 1; if (incrx < 0) xdir = -1, incrx = -incrx;
		ydir = 1; if (incry < 0) ydir = -1, incry = -incry;
		x1 = px, y1 = py, x2 = px, y2 = py;
		snx = (long)(posx&1023); if (xdir == 1) snx ^= 1023;
		sny = (long)(posy&1023); if (ydir == 1) sny ^= 1023;
		cnt1 = (int)((snx*incry - sny*incrx)>>10);
		snx = (long)(posx&1023); if (ydir == -1) snx ^= 1023;
		sny = (long)(posy&1023); if (xdir == 1) sny ^= 1023;
		cnt2 = (int)((snx*incrx - sny*incry)>>10);
		addbox(px,py), drawlines();
		scandist = 1;
		outp(97,inp(97)|2);
		while ((numhits > 0) && (scandist < maxscandist))
		{
			if (cnt1 < 0)
				cnt1 += incry, x1 += xdir;
			else
				cnt1 -= incrx, y1 += ydir;
			if (cnt2 < 0)
				cnt2 += incrx, x2 -= ydir;
			else
				cnt2 -= incry, y2 += xdir;
			x = x1, y = y1;
			addbox(x,y);
			while ((x != x2) || (y != y2))
			{
				ox = x, oy = y;
				if ((ox >= px) && (oy < py+scandist)) y++;
				if ((ox <= px) && (oy > py-scandist)) y--;
				if ((oy >= py) && (ox > px-scandist)) x--;
				if ((oy <= py) && (ox < px+scandist)) x++;
				addbox(x,y);
			}
			drawlines();
			scandist++;
		}
		outp(97,inp(97)&252);
		lastpageoffset = pageoffset;
		outp(0x3d4,0xc); outp(0x3d5,lastpageoffset>>8);
		outp(0x3d4,0xd); outp(0x3d5,lastpageoffset&255);
		if (vidmode == 0)
		{
			if (lastpageoffset < 18176) pageoffset = 18176;
			if (lastpageoffset == 18176) pageoffset = 36352;
			if (lastpageoffset > 18176) pageoffset = 0;
		}
		else
		{
			if (lastpageoffset < 21760) pageoffset = 21760;
			if (lastpageoffset == 21760) pageoffset = 43520;
			if (lastpageoffset > 21760) pageoffset = 0;
		}
		numframes++;
		vel = 0;
		svel = 0;
		hvel = 0;
		angvel = 0;
		if (moustat == 0)
			_asm \
			{
				mov ax, 11
				int 0x33
				sar cx, 1
				sar dx, 1
				mov mousx, cx
				mov mousy, dx
			}
		vel = -mousy;
		if (mousx != 0) angvel = mousx;
		if ((keystatus[0x4b]|keystatus[0x4d]|keystatus[0x48]|keystatus[0x50]) != 0)
		{
			if ((keystatus[0x4b] > 0) && (xdim > 8)) xdim -= 8;
			if ((keystatus[0x4d] > 0) && (xdim < 360)) xdim += 8;
			if ((keystatus[0x48] > 0) && (ydim > 8)) ydim -= 8;
			if ((keystatus[0x50] > 0) && (ydim < 240)) ydim += 8;
			if ((vidmode == 0) && ((xdim > 320) || (ydim > 200)))
			{
				vidmode = 1;
				xtotaldim = 360;
				ytotaldim = 240;
				scrsize = 21600;
				setgamevidmode();
			}
			else if ((vidmode == 1) && ((xdim <= 320) && (ydim <= 200)))
			{
				vidmode = 0;
				xtotaldim = 320;
				ytotaldim = 200;
				scrsize = 18000;
				setgamevidmode();
			}
			_asm \
			{
				cld
				mov dx, 0x3c4
				mov al, 2
				out dx, al
				inc dx
				mov al, 15
				out dx, al
				mov ax, 0xa000
				mov es, ax
				mov cx, 32768
				xor ax, ax
				xor di, di
				rep stosw
			}
			dside = ydim;
			halfwidth = (xdim>>1);
			eigthwidth = (xdim>>3);
			halfheight = (ydim>>1);
			centeroffset = (((ytotaldim-ydim)>>1)*90) + ((xtotaldim-xdim)>>3);
		}
		if (keystatus[0x4e] > 0) ceilingstat = 1;
		if (keystatus[0x4a] > 0) ceilingstat = 0;
		if (keystatus[0x9c] > 0) ceilingstat = 2;
		if (keystatus[0x53] > 0) ceilingstat = 3;
		if (keystatus[0x52] > 0) ceilingstat = 4;
		if (keystatus[0x9d] == 0)
		{
			if (keystatus[0xcb] > 0) angvel -= 24;
			if (keystatus[0xcd] > 0) angvel += 24;
		}
		else
		{
			if (keystatus[0xcb] > 0) svel += 12;
			if (keystatus[0xcd] > 0) svel -= 12;
		}
		if (keystatus[0x1e] > 0) hvel -= 12;
		if (keystatus[0x2c] > 0) hvel += 12;
		if (keystatus[0xc8] > 0) vel += 12;
		if (keystatus[0xd0] > 0) vel -= 12;
		if (keystatus[0x2a] > 0)
		{
			vel += vel;
			svel += svel;
		}
		if (keystatus[0x19] > 0)
		{
			keystatus[0x19] = 0;
			while ((keystatus[0x19] == 0) && (keystatus[1] == 0));
			keystatus[0x19] = 0;
		}
		if (keystatus[88] > 0)
		{
			screencapture();
			keystatus[88] = 0;
		}
		if (angvel != 0)
		{
			ang += ((angvel*clockspeed)>>3);
			ang = (ang+2048)&2047;
		}
		if (vel != 0)
		{
			posx += (int)((((long)vel)*((long)clockspeed)*sintable[(2560+ang)&2047])>>15);
			posy += (int)((((long)vel)*((long)clockspeed)*sintable[(2048+ang)&2047])>>15);
		}
		if (svel != 0)
		{
			posx += (int)((((long)svel)*((long)clockspeed)*sintable[(2048+ang)&2047])>>15);
			posy -= (int)((((long)svel)*((long)clockspeed)*sintable[(2560+ang)&2047])>>15);
		}
		if (hvel != 0)
		{
			posz += (long)((hvel*clockspeed)<<3);
			if (posz < 0) posz = 0;
			if (posz > 65535) posz = 65535;
		}
		posz = (posz+posz+posz+newposz)>>2;
		totalclock += ((long)clockspeed);
		clockspeed = 0;
	}

	outp(0x43,54); outp(0x40,255); outp(0x40,255);
	_disable(); _dos_setvect(0x8, oldhand); _enable();
	_disable(); _dos_setvect(0x9, oldkeyhandler); _enable();
	convdeallocate(pic);
	convdeallocate(lin);
	convdeallocate(lintal);
	_asm \
	{
		mov ax, 0x3
		int 0x10
	}
	templong = 0;
	for(i=0;i<numtimecounts;i++)
		templong += timetakercount[i];
	if (templong != 0)
		for(i=0;i<numtimecounts;i++)
		{
			templong1 = (10000L*timetakercount[i])/templong;
			printf("%d - %2d.%1d%1d%% - %s\n",i,(int)(templong1/100),(int)((templong1/10)%10),(int)(templong1%10),timenames[i]);
		}
	if (totalclock != 0)
	{
		templong1 = (numframes*24000L)/totalclock;
		printf("%d.%1d%1d frames per second\n",(int)(templong1/100),(int)((templong1/10)%10),(int)(templong1%10));
	}
	printf("Lintalcnt = %d",lintalcnt);
}

vline(x, y, y2, siz, walnume, picplc, ubord, dbord, shadeoffs)
int x, y, y2, siz, walnume, picplc, ubord, dbord, shadeoffs;
{
	unsigned int picseg, thesiz, start, minincr;
	unsigned long templong;

	if (ceilingstat >= 3)
		walnume = 55;
	picseg = walseg[walnume];
	thesiz = y2-y;
	if ((y > dbord) || (y+thesiz < ubord) || (thesiz <= 0) || (siz <= 0))
		return(0);
	takernum = 5;
	minincr = 0;
	if (y < ubord)
	{
		templong = (((unsigned long)(ubord-y))<<17) / (((unsigned long)siz));
		minincr = (unsigned)((templong<<8)&65535);
		picplc += (int)(templong>>8);
		thesiz -= (ubord-y);
		y = ubord;
	}
	if (y+thesiz > dbord)
		thesiz -= ((y+thesiz)-dbord);
	if ((thesiz < 1) || (thesiz > dside))
		return(-1);
	start = y*90+(x>>2)+pageoffset+centeroffset;
	templong = 33554432L / ((long)siz);
	takernum = 6;
	if (siz <= 512)
	{
		_asm \
		{
			mov ax, 0xa000
			mov es, ax
			mov di, start
			mov cx, thesiz
			mov si, picplc
			mov bx, si
			and bx, 0x003f
			and si, 0x0fc0
			mov ax, word ptr templong[0]
			mov cs:[machvline1+2], ax
			mov ax, word ptr templong[2]
			mov cs:[machvline2+2], ax
			mov dx, minincr
			mov ah, byte ptr shadeoffs[0]
			push ds
			mov ds, picseg
beginvline1:
			mov al, byte ptr ds:[bx+si]
			add al, ah
			mov byte ptr es:[di], al
			add di, 90
machvline1: add dx, 0x8888
machvline2: adc bx, 0x8888
			and bl, 63
			loop beginvline1
			pop ds
		}
	}
	else
	{
		_asm \
		{
			mov ax, 0xa000
			mov es, ax
			mov di, start
			mov si, picplc
			mov bx, si
			and bx, 0x003f
			and si, 0x0fc0
			mov ax, word ptr templong[0]
			mov word ptr cs:[machvline3+2], ax
			mov dx, minincr
			mov cx, thesiz
			mov ah, byte ptr shadeoffs[0]
			push ds
			mov ds, picseg
beginvline2:
			mov al, byte ptr ds:[bx+si]
			inc bl
			and bl, 63
			add al, ah
skiploadnext:
			mov byte ptr es:[di], al
			add di, 90
			sub cx, 1
			jz enditvline2
machvline3: add dx, 0x8888
			jnc skiploadnext
			jmp beginvline2
enditvline2:
			pop ds
		}
	}
}

convallocate(numparagraphs)
unsigned numparagraphs;
{
	unsigned segnum;

	_asm \
	{
		mov ah, 0x48
		mov bx, numparagraphs
		int 0x21
		mov segnum, ax
		jnc allocated
		mov segnum, -1
allocated:
	}
	return(segnum);
}

convdeallocate(segnum)
unsigned segnum;
{
	_asm \
	{
		mov ah, 0x49
		mov es, segnum
		int 0x21
		mov segnum, 0
		jnc deallocated
		mov segnum, -1
deallocated:
	}
	return(segnum);
}

void interrupt far keyhandler()
{
	oldreadch = readch;
	_asm \
	{
		in al, 0x60
		mov readch, al
		in al, 0x61
		or al, 0x80
		out 0x61, al
		and al, 0x7f
		out 0x61, al
	}
	if ((readch|1) == 0xe1)
		extended = 128;
	else
	{
		if (oldreadch != readch)
			keystatus[(readch&127)+extended] = ((readch>>7)^1);
		extended = 0;
	}
	_asm \
	{
		mov al, 0x20
		out 0x20, al
	}
}

loadtables()
{
	int fil, i;

	if ((fil = open("tables.dat",O_BINARY|O_RDWR,S_IREAD)) != -1)
	{
		read(fil,&sintable[0],4096);
		read(fil,&tantable[0],4096);
		close(fil);
	}
}

loadboard()
{
	long zx, templong, clipx, clipy, xz, yz, oldxz, oldyz;
	long xa1a, xa1b, xa1, ya1, xa2, ya2;
	int fil, i, j, gap, temp;
	char tempchar;

	if ((fil = open("boards.dat",O_BINARY|O_RDWR,S_IREAD)) != -1)
	{
		read(fil,&i,2), posx = (((long)i)<<5);
		read(fil,&i,2), posy = (((long)i)<<5);
		read(fil,&ang,2);
		read(fil,&numlines,2);
		read(fil,&tempbuf[0],numlines<<1);
		for(i=0;i<numlines;i++)          //xa1[i] = (((long)tempbuf[i])<<5);
			_asm \
			{
				mov es, lin
				mov di, i
				shl di, 1
				mov ax, word ptr tempbuf[di]
				mov cl, 4
				shl di, cl
				mov cl, 5
				rol ax, cl
				mov dx, ax
				and ax, 0xffe0
				and dx, 0x001f
				mov es:[di+linxa1+0], ax
			}
		read(fil,&tempbuf[0],numlines<<1);
		for(i=0;i<numlines;i++)          //ya1[i] = (((long)tempbuf[i])<<5);
			_asm \
			{
				mov es, lin
				mov di, i
				shl di, 1
				mov ax, word ptr tempbuf[di]
				mov cl, 4
				shl di, cl
				mov cl, 5
				rol ax, cl
				mov dx, ax
				and ax, 0xffe0
				and dx, 0x001f
				mov es:[di+linya1+0], ax
			}
		read(fil,&tempbuf[0],numlines<<1);
		for(i=0;i<numlines;i++)          //xa2[i] = (((long)tempbuf[i])<<5);
			_asm \
			{
				mov es, lin
				mov di, i
				shl di, 1
				mov ax, word ptr tempbuf[di]
				mov cl, 4
				shl di, cl
				mov cl, 5
				rol ax, cl
				mov dx, ax
				and ax, 0xffe0
				and dx, 0x001f
				mov es:[di+linxa2+0], ax
			}
		read(fil,&tempbuf[0],numlines<<1);
		for(i=0;i<numlines;i++)          //ya2[i] = (((long)tempbuf[i])<<5);
			_asm \
			{
				mov es, lin
				mov di, i
				shl di, 1
				mov ax, word ptr tempbuf[di]
				mov cl, 4
				shl di, cl
				mov cl, 5
				rol ax, cl
				mov dx, ax
				and ax, 0xffe0
				and dx, 0x001f
				mov es:[di+linya2+0], ax
			}
		read(fil,&tempbuf[0],numlines<<1);
		for(i=0;i<numlines;i++)          //picnum[i] = tempbuf[i];
			_asm \
			{
				mov es, lin
				mov di, i
				shl di, 1
				mov ax, word ptr tempbuf[di]
				mov cl, 4
				shl di, cl
				mov es:[di+linpicnum], ax
			}
		read(fil,&tempbuf[0],numlines<<1);
		for(i=0;i<numlines;i++)          //picnum1[i] = tempbuf[i];
			_asm \
			{
				mov es, lin
				mov di, i
				shl di, 1
				mov ax, word ptr tempbuf[di]
				mov cl, 4
				shl di, cl
				mov es:[di+linpicnum1], ax
			}
		read(fil,&tempbuf[0],numlines<<1);
		for(i=0;i<numlines;i++)          //picnum2[i] = tempbuf[i];
			_asm \
			{
				mov es, lin
				mov di, i
				shl di, 1
				mov ax, word ptr tempbuf[di]
				mov cl, 4
				shl di, cl
				mov es:[di+linpicnum2], ax
			}
		read(fil,&tempbuf[0],numlines);
		for(i=0;i<numlines;i++)          //shade[i] = tempbuf[i];
			_asm \
			{
				mov es, lin
				mov di, i
				mov al, tempbuf[di]
				mov cl, 5
				shl di, cl
				mov byte ptr es:[di+linshade], al
			}
		read(fil,&tempbuf[0],numlines);
		for(i=0;i<numlines;i++)          //shade1[i] = tempbuf[i];
			_asm \
			{
				mov es, lin
				mov di, i
				mov al, tempbuf[di]
				mov cl, 5
				shl di, cl
				mov byte ptr es:[di+linshade1], al
			}
		read(fil,&tempbuf[0],numlines);
		for(i=0;i<numlines;i++)          //shade2[i] = tempbuf[i];
			_asm \
			{
				mov es, lin
				mov di, i
				mov al, tempbuf[di]
				mov cl, 5
				shl di, cl
				mov byte ptr es:[di+linshade2], al
			}
		read(fil,&tempbuf[0],numlines);
		for(i=0;i<numlines;i++)          //z1[i] = tempbuf[i];
			_asm \
			{
				mov es, lin
				mov di, i
				mov al, tempbuf[di]
				mov cl, 5
				shl di, cl
				mov byte ptr es:[di+linz1], al
			}
		read(fil,&tempbuf[0],numlines);
		for(i=0;i<numlines;i++)          //z2[i] = tempbuf[i];
			_asm \
			{
				mov es, lin
				mov di, i
				mov al, tempbuf[di]
				mov cl, 5
				shl di, cl
				mov byte ptr es:[di+linz2], al
			}
		read(fil,&tempbuf[0],numlines);
		for(i=0;i<numlines;i++)          //cstat[i] = tempbuf[i];
			_asm \
			{
				mov es, lin
				mov di, i
				mov al, tempbuf[di]
				mov cl, 5
				shl di, cl
				mov byte ptr es:[di+lincstat], al
			}
		close(fil);
		gap = (numlines>>1);
		do
		{
			for(i=0;i<numlines-gap;i++)
			{
				_asm \
				{
					mov es, lin
					mov di, word ptr i           ;xa1a = xa1[i]
					mov cl, 5
					shl di, cl
					mov ax, word ptr es:[di+linxa1+0]
					mov word ptr xa1a[0], ax
					mov word ptr xa1a[2], 0
					mov di, word ptr i           ;xa1b = xa1[i+gap]
					add di, word ptr gap
					mov cl, 5
					shl di, cl
					mov ax, word ptr es:[di+linxa1+0]
					mov word ptr xa1b[0], ax
					mov word ptr xa1b[2], 0
				}
				temp = i;
				while ((xa1a > xa1b) && (temp >= 0))
				{
					_asm \
					{
						mov es, lin        ;Swap lin[temp] with lin[temp+gap]
						mov si, temp
						mov di, si
						add di, gap
						mov cl, 5
						shl si, cl
						shl di, cl
						mov cx, 8
swapemupforloadingboard:
						mov ax, word ptr es:[si+0]
						xchg ax, word ptr es:[di+0]
						mov word ptr es:[si+0], ax
						mov ax, word ptr es:[si+2]
						xchg ax, word ptr es:[di+2]
						mov word ptr es:[si+2], ax
						add si, 4
						add di, 4
						loop swapemupforloadingboard
					}
					temp -= gap;
					_asm \
					{
						mov es, lin
						mov di, word ptr i           ;xa1a = xa1[i]
						mov cl, 5
						shl di, cl
						mov ax, word ptr es:[di+linxa1+0]
						mov word ptr xa1a[0], ax
						mov word ptr xa1a[2], 0
						mov di, word ptr i           ;xa1b = xa1[i+gap]
						add di, word ptr gap
						mov cl, 5
						shl di, cl
						mov ax, word ptr es:[di+linxa1+0]
						mov word ptr xa1b[0], ax
						mov word ptr xa1b[2], 0
					}
				}
			}
			gap >>= 1;
		}
		while (gap > 0);
		for(i=0;i<=4096;i++)
			lintaladdr[i] = 0;
		for(i=0;i<64;i++)
			tempbuf[i] = 0;
		lintalcnt = 0;
		for(i=0;i<numlines;i++)
		{
			_asm \
			{
				mov es, lin
				mov di, word ptr i
				mov cl, 5
				shl di, cl
				mov ax, word ptr es:[di+linxa1]
				mov word ptr xa1[0], ax
				mov word ptr xa1[2], 0
				mov ax, word ptr es:[di+linya1]
				mov word ptr ya1[0], ax
				mov word ptr ya1[2], 0
				mov ax, word ptr es:[di+linxa2]
				mov word ptr xa2[0], ax
				mov word ptr xa2[2], 0
				mov ax, word ptr es:[di+linya2]
				mov word ptr ya2[0], ax
				mov word ptr ya2[2], 0
			}
			if (xa1 == xa2)
			{
				if (ya1 > ya2)
					templong = ya1, ya1 = ya2, ya2 = templong;
				addit((int)(xa1>>10),(int)(ya1>>10),i);
				if (((ya2-1)>>10) != (ya1>>10))
					addit((int)(xa1>>10),(int)(ya1>>10)+1,i);
			}
			else if (ya1 == ya2)
			{
				if (xa1 > xa2)
					templong = xa1, xa1 = xa2, xa2 = templong;
				addit((int)(xa1>>10),(int)(ya1>>10),i);
				if (((xa2-1)>>10) != (xa1>>10))
					addit((int)(xa1>>10)+1,(int)(ya1>>10),i);
			}
			else
			{
				clipx = -1L, clipy = -1L;
				if ((xa2 > xa1) && ((xa2&1023) == 0)) clipx = xa2;
				if ((ya2 > ya1) && ((ya2&1023) == 0)) clipy = ya2;
				if ((xa1 > xa2) && ((xa1&1023) == 0)) clipx = xa1;
				if ((ya1 > ya2) && ((ya1&1023) == 0)) clipy = ya1;
				xz = xa1, yz = ya1;
				if ((xz != clipx) && (yz != clipy))
					addit((int)(xz>>10),(int)(yz>>10),i);
				for(zx=1;zx<=8;zx++)
				{
					oldxz = xz, oldyz = yz;
					xz = (((xa2*zx) + (xa1*(8-zx)))>>3);
					yz = (((ya2*zx) + (ya1*(8-zx)))>>3);
					if (((xz|1023) != (oldxz|1023)) || ((yz|1023) != (oldyz|1023)))
						if ((xz != clipx) && (yz != clipy))
							addit((int)(xz>>10),(int)(yz>>10),i);
				}
			}
		}
		for(i=1;i<=4096;i++)
			lintaladdr[i] = lintaladdr[i-1] + lintaladdr[i];
	}
}

hline (xl, xr, yp, zd, walnume, shadeoffs)
unsigned xl, xr, yp;
int walnume, shadeoffs;
long zd;
{
	unsigned char planum;
	int i, siz, picplc, temp1, temp2;
	unsigned p, x, y, xc, yc, hd;

	if (ceilingstat >= 3)
		walnume = 54;
	if (xl > xr)
	{
		i = xl, xl = xr, xr = i;
	}
	siz = xr-xl;
	if ((siz < 0) || (yp == halfheight))
		return(-1);
	takernum = 7;
	_asm \
	{
		mov cx, pageoffset        ;p = yp*90+(xl>>2)+pageoffset;
		add cx, centeroffset
		mov bx, xl
		shr bx, 1
		shr bx, 1
		add cx, bx
		mov ax, yp
		mov bx, 90
		mul bx
		add cx, ax
		mov p, cx
							  ;hd = (zd<<11)/((long)((xtotaldim>>3)*abs(halfheight-yp)));
		mov ax, halfheight
		sub ax, yp
		jns skipnegatize
		neg ax
skipnegatize:
		mov bx, xtotaldim
		mov cl, 3
		shr bx, cl
		mul bx
		mov di, ax                  ;di is temp second half of equation
		mov ax, word ptr zd[0]
		mov dx, word ptr zd[2]
		mov dh, dl
		mov dl, ah
		mov ah, al
		xor al, al
		cmp di, dx
		ja dontreturn
	}
	return(-1);
	_asm \
	{
dontreturn:
		div di
		mov hd, ax
										 ;temp1 = ((hd*sintable[(2048+ang)&2047])>>13);
		mov bx, ang
		and bx, 1023
		shl bx, 1
		mov ax, word ptr sintable[bx]
		mul hd
		mov cl, 13
		shr ax, cl
		ror dx, cl
		and dx, 0xfff8
		or ax, dx
		mov temp1, ax
		test ang, 1024
		jz skipsinsign1
		neg temp1                                          ;negatize templong1
skipsinsign1:
																		 ;xc = (unsigned)temp1;
		mov ax, temp1
		mov xc, ax
										 ;temp2 = ((hd*sintable[(2560+ang)&2047])>>13);
		mov bx, ang
		add bx, 2560
		and bx, 1023
		shl bx, 1
		mov ax, word ptr sintable[bx]
		mul hd
		mov cl, 13
		shr ax, cl
		ror dx, cl
		and dx, 0xfff8
		or ax, dx
		mov temp2, ax
		mov bx, ang
		add bx, 512
		test bx, 1024
		jz skipsinsign2
		neg temp2                                           ;negatize templong2
skipsinsign2:
																		 ;yc = (unsigned)temp2;
		mov ax, temp2
		mov yc, ax
					  ;xc1 = ((long)temp1)*xl-((long)temp1+(long)temp2)*(xdim>>1);
		mov ax, temp1
		add ax, temp2
		sar ax, 1
		imul xdim
		mov bx, ax
		mov cx, dx
		mov ax, temp1
		imul xl
		sub ax, bx
		sbb dx, cx
													  ;x = (unsigned)((xc1-(posx<<8))>>2);
		mov bh, byte ptr posx[0]
		xor bl, bl
		mov cx, word ptr posx[1]
		sub ax, bx
		sbb dx, cx
		shr dx, 1
		rcr ax, 1
		shr dx, 1
		rcr ax, 1
		mov x, ax
					;yc1 = ((long)temp2)*xl+((long)temp1-(long)temp2)*(xdim>>1);
		mov ax, temp1
		sub ax, temp2
		sar ax, 1
		imul xdim
		mov bx, ax
		mov cx, dx
		mov ax, temp2
		imul xl
		add ax, bx
		adc dx, cx
													  ;y = (unsigned)((yc1+(posy<<8))>>2);
		mov bh, byte ptr posy[0]
		xor bl, bl
		mov cx, word ptr posy[1]
		add ax, bx
		adc dx, cx
		shr dx, 1
		rcr ax, 1
		shr dx, 1
		rcr ax, 1
		mov y, ax

		mov word ptr takernum[0], 8
		mov ax, 0xa000
		mov es, ax
		mov al, 0x11
		mov cl, byte ptr xl[0]
		and cl, 3
		rol al, cl
		mov planum, al
		cld
	}
	if (siz < 8)
	{
		_asm \
		{
			mov dx, 0x3c4
			mov al, 0x2
			out dx, al
			inc dx
			mov al, planum
			out dx, al

			mov cx, word ptr siz
			inc cx
			mov di, p
			mov si, x
			mov dx, y
			mov bx, walnume
			shl bx, 1
			add si, xc
			add dx, yc
			sar xc, 1
			sar yc, 1
			sar xc, 1
			sar yc, 1
			push ds
			mov ds, word ptr walseg[bx]
begindrawsmall:
			add si, xc
			add dx, yc
			mov bx, si
			mov bl, dh
			shr bh, 1
			shr bh, 1
			shr bx, 1
			shr bx, 1
			mov ah, ds:[bx]
			add ah, byte ptr shadeoffs[0]
			mov bx, dx
			mov es:[di], ah
			mov dx, 0x3c5
			rol byte ptr planum, 1
			mov al, byte ptr planum
			out dx, al
			mov dx, bx
			cmp al, 0x11
			ja skipdrawpre
			inc di
skipdrawpre:
			loop begindrawsmall
			pop ds
		}
		takernum = 4;
		return(0);
	}
	_asm \
	{
		mov dx, 0x3c4
		mov al, 0x2
		out dx, al
		inc dx
		mov al, planum
		out dx, al

		mov cx, siz
		sub cx, 4
		shr cx, 1
		mov ch, cl
		mov cl, 2
		mov di, p
		mov si, x
		mov dx, y
		mov ax, xc
		mov word ptr cs:[mach1a+2], ax
		mov word ptr cs:[mach3a+2], ax
		mov ax, yc
		mov word ptr cs:[mach2a+2], ax
		mov word ptr cs:[mach4a+2], ax
		mov ax, shadeoffs
		mov byte ptr cs:[mach5a+1], al
		mov byte ptr cs:[mach6a+2], al
		mov bx, walnume
		shl bx, 1
		push ds
		mov ds, word ptr walseg[bx]
begindrawa:
mach1a:add si, 0x1234
mach2a:add dx, 0x1234
		mov bx, si
		mov bl, dh
		shr bh, cl
		shr bx, cl
		mov al, ds:[bx]
mach3a:add si, 0x1234
mach4a:add dx, 0x1234
		mov bx, si
		mov bl, dh
		shr bh, cl
		shr bx, cl
		mov ah, ds:[bx]
mach5a:add al, 0x12
mach6a:add ah, 0x12
		stosw
		sub ch, 4
		jnc begindrawa
		cmp ch, -2
		jl enda
		add si, word ptr cs:[mach1a+2]
		add dx, word ptr cs:[mach2a+2]
		mov bx, si
		mov bl, dh
		shr bh, cl
		shr bx, cl
		mov al, ds:[bx]
		add al, byte ptr cs:[mach5a+1]
		stosb
enda:
		pop ds

		mov dx, 0x3c5
		rol byte ptr planum, 1
		rol byte ptr planum, 1
		mov al, byte ptr planum
		out dx, al
		cmp al, 0x22
		ja skipdrawb
		inc p
skipdrawb:

		mov cx, siz
		sub cx, 6
		shr cx, 1
		mov ch, cl
		mov cl, 2
		mov di, p
		mov si, x
		mov dx, y
		mov ax, xc
		mov word ptr cs:[mach1b+2], ax
		mov word ptr cs:[mach3b+2], ax
		sar ax, 1
		add si, ax
		mov ax, yc
		mov word ptr cs:[mach2b+2], ax
		mov word ptr cs:[mach4b+2], ax
		sar ax, 1
		add dx, ax
		mov ax, shadeoffs
		mov byte ptr cs:[mach5b+1], al
		mov byte ptr cs:[mach6b+2], al
		mov bx, walnume
		shl bx, 1
		push ds
		mov ds, word ptr walseg[bx]
begindrawb:
mach1b:add si, 0x1234
mach2b:add dx, 0x1234
		mov bx, si
		mov bl, dh
		shr bh, cl
		shr bx, cl
		mov al, ds:[bx]
mach3b:add si, 0x1234
mach4b:add dx, 0x1234
		mov bx, si
		mov bl, dh
		shr bh, cl
		shr bx, cl
		mov ah, ds:[bx]
mach5b:add al, 0x12
mach6b:add ah, 0x12
		stosw
		sub ch, 4
		jnc begindrawb
		cmp ch, -2
		jl endb
		add si, word ptr cs:[mach1b+2]
		add dx, word ptr cs:[mach2b+2]
		mov bx, si
		mov bl, dh
		shr bh, cl
		shr bx, cl
		mov al, ds:[bx]
		add al, byte ptr cs:[mach5b+1]
		stosb
endb:
		pop ds
	}
	_asm \
	{
		mov dx, 0x3c5                          ;Higher precision
		ror byte ptr planum, 1
		mov al, byte ptr planum
		out dx, al
		cmp al, 0x88
		jne skipdrawc
		dec p
skipdrawc:

		mov cx, siz
		sub cx, 5
		shr cx, 1
		mov ch, cl
		mov cl, 2
		mov di, p
		mov si, x
		mov dx, y
		mov ax, xc
		mov word ptr cs:[mach1c+2], ax
		mov word ptr cs:[mach3c+2], ax
		sar ax, 1
		sar ax, 1
		add si, ax
		mov ax, yc
		mov word ptr cs:[mach2c+2], ax
		mov word ptr cs:[mach4c+2], ax
		sar ax, 1
		sar ax, 1
		add dx, ax
		mov ax, shadeoffs
		mov byte ptr cs:[mach5c+1], al
		mov byte ptr cs:[mach6c+2], al
		mov bx, walnume
		shl bx, 1
		push ds
		mov ds, word ptr walseg[bx]
begindrawc:
mach1c:add si, 0x1234
mach2c:add dx, 0x1234
		mov bx, si
		mov bl, dh
		shr bh, cl
		shr bx, cl
		mov al, ds:[bx]
mach3c:add si, 0x1234
mach4c:add dx, 0x1234
		mov bx, si
		mov bl, dh
		shr bh, cl
		shr bx, cl
		mov ah, ds:[bx]
mach5c:add al, 0x12
mach6c:add ah, 0x12
		stosw
		sub ch, 4
		jnc begindrawc
		cmp ch, -2
		jl endc
		add si, word ptr cs:[mach1c+2]
		add dx, word ptr cs:[mach2c+2]
		mov bx, si
		mov bl, dh
		shr bh, cl
		shr bx, cl
		mov al, ds:[bx]
		add al, byte ptr cs:[mach5c+1]
		stosb
endc:
		pop ds

		mov dx, 0x3c5
		rol byte ptr planum, 1
		rol byte ptr planum, 1
		mov al, byte ptr planum
		out dx, al
		cmp al, 0x22
		ja skipdrawd
		inc p
skipdrawd:

		mov cx, siz
		sub cx, 7
		shr cx, 1
		mov ch, cl
		mov cl, 2
		mov di, p
		mov si, x
		mov dx, y
		mov ax, xc
		mov word ptr cs:[mach1d+2], ax
		mov word ptr cs:[mach3d+2], ax
		sar ax, 1
		add si, ax
		sar ax, 1
		add si, ax
		mov ax, yc
		mov word ptr cs:[mach2d+2], ax
		mov word ptr cs:[mach4d+2], ax
		sar ax, 1
		add dx, ax
		sar ax, 1
		add dx, ax
		mov ax, shadeoffs
		mov byte ptr cs:[mach5d+1], al
		mov byte ptr cs:[mach6d+2], al
		mov bx, walnume
		shl bx, 1
		push ds
		mov ds, word ptr walseg[bx]
begindrawd:
mach1d:add si, 0x1234
mach2d:add dx, 0x1234
		mov bx, si
		mov bl, dh
		shr bh, cl
		shr bx, cl
		mov al, ds:[bx]
mach3d:add si, 0x1234
mach4d:add dx, 0x1234
		mov bx, si
		mov bl, dh
		shr bh, cl
		shr bx, cl
		mov ah, ds:[bx]
mach5d:add al, 0x12
mach6d:add ah, 0x12
		stosw
		sub ch, 4
		jnc begindrawd
		cmp ch, -2
		jl endd
		add si, word ptr cs:[mach1d+2]
		add dx, word ptr cs:[mach2d+2]
		mov bx, si
		mov bl, dh
		shr bh, cl
		shr bx, cl
		mov al, ds:[bx]
		add al, byte ptr cs:[mach5d+1]
		stosb
endd:
		pop ds
	}
	takernum = 4;
}

loadpics()
{
	int fil, i;

	if ((fil=open("pics.dat",O_BINARY|O_RDONLY,S_IREAD))==-1)
		return(-1);
	read(fil,&tempbuf[0],1024);
	for(i=0;i<numwalls;i++)
	{
		read(fil,&tempbuf[0],4096);
		_asm \
		{
			cld
			mov cx, 2048
			mov bx, i
			shl bx, 1
			mov es, word ptr walseg[bx]
			xor di, di
			mov si, offset tempbuf
			mov ax, seg tempbuf
			push ds
			mov ds, ax
			rep movsw
			pop ds
		}
	}
	close(fil);
}

void interrupt far ksmhandler()
{
	clockspeed++;
	timetakercount[takernum]++;
	_asm \
	{
		mov al, 0x20
		out 0x20, al
	}
}

setmode360240()
{
	outp(0x3c4,0x4); outp(0x3c5,0x6);
	outp(0x3c4,0x0); outp(0x3c5,0x1);
	outp(0x3c2,0xe7);
	outp(0x3c4,0x0); outp(0x3c5,0x3);
	outp(0x3d4,0x11); outp(0x3d5,inp(0x3d5)&0x7f);      //Unlock indices 0-7
	outp(0x3d4,0x0); outp(0x3d5,0x6b);
	outp(0x3d4,0x1); outp(0x3d5,0x59);
	outp(0x3d4,0x2); outp(0x3d5,0x5a);
	outp(0x3d4,0x3); outp(0x3d5,0x8e);
	outp(0x3d4,0x4); outp(0x3d5,0x5e);
	outp(0x3d4,0x5); outp(0x3d5,0x8a);
	outp(0x3d4,0x6); outp(0x3d5,0xd);
	outp(0x3d4,0x7); outp(0x3d5,0x3e);
	outp(0x3d4,0x9); outp(0x3d5,0x41);
	outp(0x3d4,0x10); outp(0x3d5,0xea);
	outp(0x3d4,0x11); outp(0x3d5,0xac);
	outp(0x3d4,0x12); outp(0x3d5,0xdf);
	outp(0x3d4,0x13); outp(0x3d5,45);
	outp(0x3d4,0x14); outp(0x3d5,0x0);
	outp(0x3d4,0x15); outp(0x3d5,0xe7);
	outp(0x3d4,0x16); outp(0x3d5,0x6);
	outp(0x3d4,0x17); outp(0x3d5,0xe3);
}

addit(x,y, i)
int x, y, i;
{
	int p, hereplc, zz;

	p = (x<<6) + y + 1;
	hereplc = 0;
	for(zz=0;zz<x;zz++)
		hereplc += tempbuf[zz];
	for(zz=(x<<6)+1;zz<p;zz++)
		hereplc += lintaladdr[zz];
	for(zz=lintalcnt;zz>hereplc;zz--)
		_asm \
		{
			mov es, lintal
			mov bx, word ptr zz
			shl bx, 1
			mov ax, word ptr es:[bx-2]
			mov word ptr es:[bx], ax
		}
	lintaladdr[p]++;
	tempbuf[x]++;
	_asm \
	{
		mov es, lintal
		mov bx, word ptr hereplc
		shl bx, 1
		mov ax, word ptr i
		mov word ptr es:[bx], ax
	}
	lintalcnt++;
}

setupmouse()
{
	int snot;

	_asm \
	{
		mov snot, 0
		mov ax, 0x3533
		int 0x21
		mov dx, es
		or dx, bx
		jz nomouse
		cmp es:[bx], 0xcf
		je nomouse
		jmp skipnomouse
		mov ax, 0
		int 0x33
nomouse:
		mov snot, -1
skipnomouse:
	}
	return(snot);
}

int infront(l1, l2)
int l1, l2;
{
	int danger;
	long tempheight, temp1, temp2;

	if ((yb1[l1] < yb1[l2]) == (yb1[l1] < yb2[l2]))
		if ((yb2[l1] < yb1[l2]) == (yb2[l1] < yb2[l2]))
			if ((yb1[l1] < yb1[l2]) == (yb2[l1] < yb1[l2]))
			{
				if (yb1[l1]+yb2[l1] > yb1[l2]+yb2[l2])
					return(0);
				else
					return(1);
			}
	if ((xb1[l1] == xb2[l1]) && (xb2[l1] == xb2[l2]))
	{
		if (yb1[l1] > yb1[l2]) return(0);
		if (yb1[l1] < yb1[l2]) return(1);
		if (yb2[l1] > yb2[l2]) return(0);
		if (yb2[l1] < yb2[l2]) return(1);
		return(-1);
	}
	danger = 0;
	if ((xb1[l1] == xb1[l2]) && (yb1[l1] == yb1[l2])) danger |= 1;
	if ((xb2[l1] == xb2[l2]) && (yb2[l1] == yb2[l2])) danger |= 2;
	if ((xb1[l1] <= xb2[l2]) && (xb1[l1] >= xb1[l2]) && ((danger&1) == 0))
	{
		temp1 = (xb1[l2]-xb1[l1])*yb1[l2];
		temp2 = (xb2[l2]-xb1[l1])*yb2[l2];
		if (temp1 != temp2)
		{
			tempheight = yb1[l2] - ((yb2[l2]-yb1[l2])*temp1)/(temp2-temp1);
			if (tempheight > yb1[l1])
				return(1);
			else
				return(0);
		}
	}
	if ((xb1[l2] <= xb2[l1]) && (xb1[l2] >= xb1[l1]) && ((danger&1) == 0))
	{
		temp1 = (xb1[l1]-xb1[l2])*yb1[l2];
		temp2 = (xb2[l1]-xb1[l2])*yb2[l2];
		if (temp1 != temp2)
		{
			tempheight = yb1[l1] - ((yb2[l1]-yb1[l1])*temp1)/(temp2-temp1);
			if (tempheight > yb1[l2])
				return(0);
			else
				return(1);
		}
	}
	if ((xb2[l1] <= xb2[l2]) && (xb2[l1] >= xb1[l2]) && ((danger&2) == 0))
	{
		temp1 = (xb2[l2]-xb2[l1])*yb2[l2];
		temp2 = (xb1[l2]-xb2[l1])*yb1[l2];
		if (temp1 != temp2)
		{
			tempheight = yb2[l2] - ((yb1[l2]-yb2[l2])*temp1)/(temp2-temp1);
			if (tempheight > yb2[l1])
				return(1);
			else
				return(0);
		}
	}
	if ((xb2[l2] <= xb2[l1]) && (xb2[l2] >= xb1[l1]) && ((danger&2) == 0))
	{
		temp1 = (xb2[l1]-xb2[l2])*yb2[l1];
		temp2 = (xb1[l1]-xb2[l2])*yb1[l1];
		if (temp1 != temp2)
		{
			tempheight = yb2[l1] - ((yb1[l1]-yb2[l1])*temp1)/(temp2-temp1);
			if (tempheight > yb2[l2])
				return(0);
			else
				return(1);
		}
	}
	return(-1);
}

screencapture()
{
	char filename[15];
	int i, j, k, fil, bufplc;
	unsigned int plc;

	strcpy(filename,"capturxx.bmp");
	filename[6] = (capturecount/10)+48;
	filename[7] = (capturecount%10)+48;
	capturecount++;
	if ((fil=open(filename,O_BINARY|O_CREAT|O_WRONLY,S_IWRITE))==-1)
	{
		capturecount--;
		return(-1);
	}
	tempbuf[0] = 66, tempbuf[1] = 77, tempbuf[2] = 90, tempbuf[3] = 34;
	tempbuf[4] = 0, tempbuf[5] = 0, tempbuf[6] = 0, tempbuf[7] = 0;
	tempbuf[8] = 0, tempbuf[9] = 0, tempbuf[10] = 26, tempbuf[11] = 3;
	tempbuf[12] = 0, tempbuf[13] = 0, tempbuf[14] = 12, tempbuf[15] = 0;
	tempbuf[16] = 0, tempbuf[17] = 0, tempbuf[18] = 64, tempbuf[19] = 1;
	tempbuf[20] = 200, tempbuf[21] = 0, tempbuf[22] = 1, tempbuf[23] = 0;
	tempbuf[24] = 8, tempbuf[25] = 0;
	write(fil,&tempbuf[0],26);
	outp(0x3c7,0);
	for(i=0;i<768;i++)
		tempbuf[i] = (inp(0x3c9)<<2);
	for(i=0;i<768;i+=3)
	{
		j = tempbuf[i];
		tempbuf[i] = tempbuf[i+2];
		tempbuf[i+2] = j;
	}
	if (write(fil,&tempbuf[0],768) < 768)
	{
		close(fil);
		capturecount--;
		return(-1);
	}
	k = pageoffset;
	pageoffset = lastpageoffset;
	outp(0x3ce,4);
	_asm \
	{
		mov ax, 0xa000
		mov es, ax
	}
	bufplc = 0;
	for(j=199;j>=0;j--)
	{
		outp(97,inp(97)|3);
		plc = pageoffset+j*90;
		if (vidmode == 1)
			plc += 1800;
		_asm \
		{
			mov bx, offset tempbuf
			add bx, bufplc
			mov cx, bx
			add cx, 320
			mov di, plc
			mov dx, 0x3cf
			xor ah, ah
capturebegin:
			mov al, ah
			out dx, al
			mov al, byte ptr es:[di]
			mov byte ptr ds:[bx], al
			inc ah
			cmp ah, 4
			jne captureskip
			xor ah, ah
			inc di
captureskip:
			inc bx
			cmp bx, cx
			jl capturebegin
		}
		bufplc += 320;
		if (bufplc >= 3200)
		{
			if (write(fil,&tempbuf[0],3200) < 3200)
			{
				close(fil);
				pageoffset = k;
				capturecount--;
				return(-1);
			}
			bufplc = 0;
			outp(0x3ce,4);
			_asm \
			{
				mov ax, 0xa000
				mov es, ax
			}
		}
		outp(97,inp(97)&252);
	}
	close(fil);
	pageoffset = k;
}

setgamevidmode()
{
	int i;

	_asm \
	{
		mov ax, 0x13
		int 0x10
	}
	if (vidmode == 1)
		setmode360240();
	outp(0x3c8,0);
	for(i=0;i<768;i++)
		outp(0x3c9,palette[i]);
	outp(0x3d4,0x14); outp(0x3d5, inp(0x3d5)&(255-64));  //WORD MODE
	outp(0x3d4,0x17); outp(0x3d5, inp(0x3d5)|64);        //2-WORD MODE
	outp(0x3c4,0x4); outp(0x3c5, inp(0x3c5)&247);        //CHAIN 4 OFF
	outp(0x3d4,0x13); outp(0x3d5, 45);                   //Bytes per line
}

addbox(x,y)
int x,y;
{
	unsigned int p;
	long templong, templong1, xp1, yp1, xp2, yp2, xa1, ya1, xa2, ya2;
	unsigned char got, cstat;
	int i, j, k, good, linti;

	p = ((x&63)<<6)+(y&63);
	if (lintaladdr[p] == lintaladdr[p+1])
		return(0);
	takernum = 1;
	for(i=lintaladdr[p];i<lintaladdr[p+1];i++)
	{
		_asm \
		{
			mov es, lintal          ;linti = lintal[i]
			mov bx, i
			shl bx, 1
			mov di, word ptr es:[bx]
			mov linti, di

			mov es, lin             ;got = got[linti]
			mov cl, 5
			shl di, cl
			mov al, byte ptr es:[di+lingot]
			mov byte ptr got, al
		}
		if (got == 0)
		{
			_asm \
			{
				mov es, lin
				mov di, word ptr linti
				mov cl, 5
				shl di, cl
				mov ax, word ptr es:[di+linxa1]
				mov word ptr xa1[0], ax
				mov word ptr xa1[2], 0
				mov ax, word ptr es:[di+linya1]
				mov word ptr ya1[0], ax
				mov word ptr ya1[2], 0
				mov ax, word ptr es:[di+linxa2]
				mov word ptr xa2[0], ax
				mov word ptr xa2[2], 0
				mov ax, word ptr es:[di+linya2]
				mov word ptr ya2[0], ax
				mov word ptr ya2[2], 0
				mov al, byte ptr es:[di+lincstat]
				mov byte ptr cstat, al
				mov byte ptr es:[di+lingot], 1
			}
			xp1 = (((xa1-posx)*sintable[(2048+ang)&2047]-(ya1-posy)*sintable[(2560+ang)&2047])>>15);
			yp1 = (((xa1-posx)*sintable[(2560+ang)&2047]+(ya1-posy)*sintable[(2048+ang)&2047])>>15);
			xp2 = (((xa2-posx)*sintable[(2048+ang)&2047]-(ya2-posy)*sintable[(2560+ang)&2047])>>15);
			yp2 = (((xa2-posx)*sintable[(2560+ang)&2047]+(ya2-posy)*sintable[(2048+ang)&2047])>>15);
			swapped[lintalplc] = 0;
			if (xp1 > xp2)
			{
				templong = xp1; xp1 = xp2; xp2 = templong;
				templong = yp1; yp1 = yp2; yp2 = templong;
				swapped[lintalplc] = 1;
			}
			good = 7;
			if ((labs(xp1) <= yp1) && (yp1 > 0)) good--;
			if ((labs(xp2) <= yp2) && (yp2 > 0)) good-=2;
			if ((good == 7) && (yp1 > 0) && (yp2 > 0) && ((xp1 < 0) == (0 < xp2)))
				good = 3;
			yb1[lintalplc] = -1;
			yb2[lintalplc] = -1;
			if (good < 7)
			{
				if ((good&3) > 0)
					templong = xp2*yp1-xp1*yp2;
				if ((good&1) > 0)
				{
					if (xp1 < 0)
					{
						xb1[lintalplc] = (long)(xdim-1);
						templong1 = yp2-yp1+xp2-xp1;
						if (templong1 != 0)
							yb1[lintalplc] = templong/templong1;
					}
					else
					{
						xb1[lintalplc] = 0L;
						templong1 = yp1-yp2+xp2-xp1;
						if (templong1 != 0)
							yb1[lintalplc] = templong/templong1;
					}
				}
				else
				{
					if (yp1 != 0)
						xb1[lintalplc] = ((long)halfwidth)-((xp1*((long)halfwidth))/yp1);
					yb1[lintalplc] = yp1;
				}
				if ((good&2) > 0)
				{
					if (xp2 < 0)
					{
						xb2[lintalplc] = (long)(xdim-1);
						templong1 = yp2-yp1+xp2-xp1;
						if (templong1 != 0)
							yb2[lintalplc] = templong/templong1;
					}
					else
					{
						xb2[lintalplc] = 0L;
						templong1 = yp1-yp2+xp2-xp1;
						if (templong1 != 0)
							yb2[lintalplc] = templong/templong1;
					}
				}
				else
				{
					if (yp2 != 0)
						xb2[lintalplc] = ((long)halfwidth)-((xp2*((long)halfwidth))/yp2);
					yb2[lintalplc] = yp2;
				}
			}
			theline[lintalplc] = linti;
			if (xb1[lintalplc] > xb2[lintalplc])
			{
				templong = xb1[lintalplc]; xb1[lintalplc] = xb2[lintalplc]; xb2[lintalplc] = templong;
				templong = yb1[lintalplc]; yb1[lintalplc] = yb2[lintalplc]; yb2[lintalplc] = templong;
				swapped[lintalplc] ^= 1;
			}
			if ((yb1[lintalplc] > 0L) && (yb2[lintalplc] > 0L) && (xb1[lintalplc] != xb2[lintalplc]))
				if ((cstat != 3) || (swapped[lintalplc] == 0))
					lintalplc++;
		}
	}
	takernum = 0;
}

drawlines()
{
	unsigned char tempchar;
	unsigned int uplccnt, dplccnt;
	long templong, templong1, templong2, dztop, dzbot, thez;
	long x2, y2, x3, y3;
	long hc1, hc2, hc3, hi1, hi2, hi3, yn1, yn2;
	long xa1, ya1, xa2, ya2, yh1, yh2, yh3, yh4;
	int shade, shade1, shade2, picnum, picnum1, picnum2;
	unsigned char z1, z2, cstat;
	int uplc, dplc, theplc, theceilpic, theshade;
	int i, j, k, m, n, o, good, gap;
	int um, dm, up, dp;
	int temp, llinum, rlinum, closest;

	if (lintalplc == 0)
		return(0);
	takernum = 2;
	gap = (lintalplc>>1);
	while (gap > 0)
	{
		for(i=0;i<lintalplc-gap;i++)
		{
			temp = i;
			if (yb1[temp] < yb2[temp]) templong1 = yb1[temp]; else templong1 = yb2[temp];
			if (yb1[temp+gap] < yb2[temp+gap]) templong2 = yb1[temp+gap]; else templong2 = yb2[temp+gap];
			while (templong1 > templong2)
			{
				templong = xb1[temp]; xb1[temp] = xb1[temp+gap]; xb1[temp+gap] = templong;
				templong = yb1[temp]; yb1[temp] = yb1[temp+gap]; yb1[temp+gap] = templong;
				templong = xb2[temp]; xb2[temp] = xb2[temp+gap]; xb2[temp+gap] = templong;
				templong = yb2[temp]; yb2[temp] = yb2[temp+gap]; yb2[temp+gap] = templong;
				j = theline[temp]; theline[temp] = theline[temp+gap]; theline[temp+gap] = j;
				tempchar = swapped[temp]; swapped[temp] = swapped[temp+gap]; swapped[temp+gap] = tempchar;
				temp -= gap;
				if (temp < 0)
					break;
				if (yb1[temp] < yb2[temp]) templong1 = yb1[temp]; else templong1 = yb2[temp];
				if (yb1[temp+gap] < yb2[temp+gap]) templong2 = yb1[temp+gap]; else templong2 = yb2[temp+gap];
			}
		}
		gap >>= 1;
	}
	for(i=0;i<lintalplc;i++)
	{
		takernum = 3;
		closest = i;
		do
		{
			_asm \
			{
				mov es, lin      ;load up z1/z2[theline[closest]]
				mov bx, closest
				shl bx, 1
				mov di, word ptr theline[bx]
				mov cl, 5
				shl di, cl
				mov al, byte ptr es:[di+linz1]
				mov byte ptr z1, al
				mov al, byte ptr es:[di+linz2]
				mov byte ptr z2, al
			}
			good = 0;
			if ((z2 >= uzmost[xb1[closest]+1]) || (z1 <= dzmost[xb1[closest]+1]) || (z2 >= uzmost[xb2[closest]-1]) || (z1 <= dzmost[xb2[closest]-1]))
				good = 1;
			if (good == 0)
				closest++;
		}
		while ((good == 0) && (closest < lintalplc));
		if (closest >= lintalplc)
		{
			i = lintalplc;
			numhits = 0;
			break;
		}
		m = 1;
		n = 0;
		while ((m == 1) && (n < (lintalplc-i)))
		{
			m = 0;
			n++;
			for(j=i+1;j<lintalplc;j++)
			{
				_asm \
				{
					mov es, lin      ;load up z1/z2[theline[j]]
					mov bx, j
					shl bx, 1
					mov di, word ptr theline[bx]
					mov cl, 5
					shl di, cl
					mov al, byte ptr es:[di+linz1]
					mov byte ptr z1, al
					mov al, byte ptr es:[di+linz2]
					mov byte ptr z2, al
				}
				if ((z2 >= uzmost[xb1[j]+1]) || (z1 <= dzmost[xb1[j]+1]) || (z2 >= uzmost[xb2[j]-1]) || (z1 <= dzmost[xb2[j]-1]))
					if ((xb1[j] <= xb2[closest]) && (xb1[closest] <= xb2[j]))
					{
						m = infront(j,closest);
						if (m == 0)
						{
							_asm \
							{
								mov es, lin       ;load up cstat[theline[closest]]
								mov bx, closest
								shl bx, 1
								mov di, word ptr theline[bx]
								mov cl, 5
								shl di, cl
								mov al, byte ptr es:[di+lincstat]
								mov cstat, al     ;cstat[theline[closest]]
							}
							if (cstat == 3)
								if ((xb1[j] >= xb1[closest]) && (xb2[j] <= xb2[closest]))
								{
									lintalplc--;
									xb1[j] = xb1[lintalplc];
									yb1[j] = yb1[lintalplc];
									xb2[j] = xb2[lintalplc];
									yb2[j] = yb2[lintalplc];
									theline[j] = theline[lintalplc];
									swapped[j] = swapped[lintalplc];
									if (closest == lintalplc) closest = j;
								}
						}
						if (m == 1)
						{
							closest = j;
							break;
						}
					}
			}
			if (closest != i)
			{
				templong = xb1[i]; xb1[i] = xb1[closest]; xb1[closest] = templong;
				templong = yb1[i]; yb1[i] = yb1[closest]; yb1[closest] = templong;
				templong = xb2[i]; xb2[i] = xb2[closest]; xb2[closest] = templong;
				templong = yb2[i]; yb2[i] = yb2[closest]; yb2[closest] = templong;
				m = theline[i]; theline[i] = theline[closest]; theline[closest] = m;
				tempchar = swapped[i]; swapped[i] = swapped[closest]; swapped[closest] = tempchar;
			}
			good = 1;
			for(j=xb1[i];j<=xb2[i];j++)
				if (umost[j] < dmost[j])
					good = 0;
			if (good == 0)
			{
				takernum = 4;
				_asm \
				{
					mov es, lin
					mov bx, i
					shl bx, 1
					mov di, word ptr theline[bx]
					mov cl, 5
					shl di, cl
					mov ax, word ptr es:[di+linxa1]
					mov word ptr xa1[0], ax
					mov word ptr xa1[2], 0
					mov ax, word ptr es:[di+linya1]
					mov word ptr ya1[0], ax
					mov word ptr ya1[2], 0
					mov ax, word ptr es:[di+linxa2]
					mov word ptr xa2[0], ax
					mov word ptr xa2[2], 0
					mov ax, word ptr es:[di+linya2]
					mov word ptr ya2[0], ax
					mov word ptr ya2[2], 0
					mov ax, word ptr es:[di+linpicnum]
					mov word ptr picnum, ax
					mov ax, word ptr es:[di+linpicnum1]
					mov word ptr picnum1, ax
					mov ax, word ptr es:[di+linpicnum2]
					mov word ptr picnum2, ax
					mov al, byte ptr es:[di+linshade]
					mov byte ptr shade, al
					mov al, byte ptr es:[di+linshade1]
					mov byte ptr shade1, al
					mov al, byte ptr es:[di+linshade2]
					mov byte ptr shade2, al
					mov al, byte ptr es:[di+linz1]
					mov byte ptr z1, al
					mov al, byte ptr es:[di+linz2]
					mov byte ptr z2, al
					mov al, byte ptr es:[di+lincstat]
					mov byte ptr cstat, al
				}
				if ((closestline == 0) && ((cstat&2) > 0))
				{
					if (swapped[i] == 0)
						newposz = (((long)z2)<<8)-8192;
					else
						newposz = (((long)z1)<<8)-8192;
					closestline = 1;
				}
				dztop = posz-(((long)z1)<<8);
				dzbot = posz-(((long)z2)<<8);
				if ((yb1[i] != 0) && (yb2[i] != 0))
				{
					yh1 = ((long)halfheight)-((dztop*xdim<<2)/(xtotaldim*yb1[i]));
					yh2 = ((long)halfheight)-((dzbot*xdim<<2)/(xtotaldim*yb1[i]));
					yh3 = ((long)halfheight)-((dztop*xdim<<2)/(xtotaldim*yb2[i]));
					yh4 = ((long)halfheight)-((dzbot*xdim<<2)/(xtotaldim*yb2[i]));
					yn1 = 524288L / yb1[i];
					yn2 = 524288L / yb2[i];
				}
				uplc = ((int)yh1), uplccnt = 0;
				dplc = ((int)yh2), dplccnt = 0;

				if ((swapped[i] == 0) && (ceilingstat != 2))
				{
					hc3 = ((((long)((yn1*xdim)))/xtotaldim)<<16);

					x3 = (((long)((xdim>>1)-xb1[i]))<<19)/(yn1*(xdim>>3));
					y3 = (2097152L/yn1);
					x2 = posx-((x3*sintable[(2048-ang)&2047] - y3*sintable[(2560-ang)&2047])>>15);
					y2 = posy-((x3*sintable[(2560-ang)&2047] + y3*sintable[(2048-ang)&2047])>>15);
					if (labs(xa1-xa2) >= labs(ya1-ya2))
						llinum = (int)(((labs(xa1-x2))<<12)/labs(xa2-xa1));
					else
						llinum = (int)(((labs(ya1-y2))<<12)/labs(ya2-ya1));

					x3 = (((long)((xdim>>1)-xb2[i]))<<19)/(yn2*(xdim>>3));
					y3 = (2097152L/yn2);
					x2 = posx-((x3*sintable[(2048-ang)&2047] - y3*sintable[(2560-ang)&2047])>>15);
					y2 = posy-((x3*sintable[(2560-ang)&2047] + y3*sintable[(2048-ang)&2047])>>15);
					if (labs(xa1-xa2) >= labs(ya1-ya2))
						rlinum = (int)(((labs(xa1-x2))<<12)/labs(xa2-xa1));
					else
						rlinum = (int)(((labs(ya1-y2))<<12)/labs(ya2-ya1));
				}

				templong = xb2[i]-xb1[i]+1;
				hi1 = ((yh3-yh1)<<16)/templong;
				hi2 = ((yh4-yh2)<<16)/templong;
				hi3 = ((((yn2-yn1)*xdim)/xtotaldim)<<16)/templong;

				if (swapped[i] == 0)
				{
					um = uplc;
					dm = dplc-1;
				}
				else
				{
					um = dplc-1;
					dm = uplc;
				}
				up = um-1;
				dp = dm+1;
				if (dmost[xb1[i]]-1 <= up)
				{
					up = dmost[xb1[i]]-1;
					um = up+1;
				}
				if (umost[xb1[i]] >= dp)
				{
					dp = umost[xb1[i]];
					dm = dp-1;
				}
				if (um < -1) um = -1;
				if (dm < -1) dm = -1;
				if (up < -1) up = -1;
				if (dp < -1) dp = -1;
				if (um > ytotaldim) um = ytotaldim;
				if (dm > ytotaldim) dm = ytotaldim;
				if (up > ytotaldim) up = ytotaldim;
				if (dp > ytotaldim) dp = ytotaldim;

				for(j=xb1[i];j<=xb2[i];j++)
				{
					if ((ceilingstat > 0) && (ceilingstat < 4))
					{
						if ((cstat&1) > 0)
						{
							if (swapped[i] == 0)
								theplc = uplc-1, thez = dztop, theceilpic = picnum1, theshade = shade1;
							else
								theplc = dplc, thez = dzbot, theceilpic = picnum2, theshade = shade2;
							if (theplc < -1) theplc = -1;
							if (theplc > ytotaldim) theplc = ytotaldim;
							if (theplc > dmost[j]-1) theplc = dmost[j]-1;
							while ((umost[j] > um) && (um < ytotaldim))
							{
								if (uxsave[um+1] >= 0)
								{
									hline(uxsave[um+1],j-1,um,thez,theceilpic,theshade);
									uxsave[um+1] = -1;
								}
								um++;
							}
							while ((theplc < up) && (up >= 0))
							{
								if (uxsave[up+1] >= 0)
								{
									hline(uxsave[up+1],j-1,up,thez,theceilpic,theshade);
									uxsave[up+1] = -1;
								}
								up--;
							}
							while ((umost[j] < um) && (um > 0))
							{
								um--;
								if (up >= um)
									 uxsave[um+1] = j;
							}
							while ((theplc > up) && (up < ytotaldim-1))
							{
								up++;
								if (up >= um)
									uxsave[up+1] = j;
							}
						}
						if ((cstat&2) > 0)
						{
							if (swapped[i] == 0)
								theplc = dplc, thez = -dzbot, theceilpic = picnum2, theshade = shade2;
							else
								theplc = uplc-1, thez = -dztop, theceilpic = picnum1, theshade = shade1;
							if (theplc < -1) theplc = -1;
							if (theplc > ytotaldim) theplc = ytotaldim;
							if (theplc < umost[j]) theplc = umost[j];
							while ((dmost[j]-1 < dm) && (dm >= 0))
							{
								if (dxsave[dm+1] >= 0)
								{
									hline(dxsave[dm+1],j-1,dm,thez,theceilpic,theshade);
									dxsave[dm+1] = -1;
								}
								dm--;
							}
							while ((theplc > dp) && (dp < ytotaldim))
							{
								if (dxsave[dp+1] >= 0)
								{
									hline(dxsave[dp+1],j-1,dp,thez,theceilpic,theshade);
									dxsave[dp+1] = -1;
								}
								dp++;
							}
							while ((dmost[j]-1 > dm) && (dm < ytotaldim-1))
							{
								dm++;
								if (dp <= dm)
									dxsave[dm+1] = j;
							}
							while ((theplc < dp) && (dp > 0))
							{
								dp--;
								if (dp <= dm)
									dxsave[dp+1] = j;
							}
						}
					}
					uplace[j] = uplc;
					dplace[j] = dplc;

					_asm \
					{
						mov ax, word ptr hi1[0]
						mov dx, word ptr hi1[2]
						add uplccnt, ax
						adc uplc, dx
						mov ax, word ptr hi2[0]
						mov dx, word ptr hi2[2]
						add dplccnt, ax
						adc dplc, dx
					}
				}
				takernum = 5;
				for(j=xb1[i];j<=xb2[i];j++)
				{
					templong1 = (yh4-yh3)*(j-xb1[i]);
					templong2 = (yh2-yh1)*(xb2[i]-j);
					if (templong1 != -templong2)
						linum[j] = llinum+(templong1*(rlinum-llinum))/(templong1+templong2);
				}
				takernum = 4;
				for(j=xb1[i];j<=xb2[i];j++)
				{
					if (dmost[j] >= umost[j])
					{
						if ((uplace[j] <= dmost[j]) && (dplace[j] > umost[j]) && (ceilingstat != 2))
							if (swapped[i] == 0)
								if ((hc3&0xffff0000) != 0)
								{
									outp(0x3c4,2); outp(0x3c5,1<<(j&3));
									vline(j,uplace[j],dplace[j],(int)(hc3>>16),picnum,linum[j]&0xfc0,umost[j],dmost[j],shade);
									takernum = 4;
								}
						if (cstat == 6)
						{
							if (uplace[j] <= dmost[j])
								dmost[j] = uplace[j];
							if (swapped[i] == 0)
								dzmost[j] == z1;
							else
								dzmost[j] == z2;
						}
						else if (cstat == 9)
						{
							if (dplace[j] >= umost[j])
								umost[j] = dplace[j];
							if (swapped[i] == 0)
								uzmost[j] == z2;
							else
								uzmost[j] == z1;
						}
						if ((cstat == 3) || (umost[j] > dmost[j]))
						{
							umost[j] = 100;
							dmost[j] = 99;
							numhits--;
							if (numhits < 0)
							{
								j = xb2[i];
								i = lintalplc;
								break;
							}
						}
					}
					hc3 += hi3;
				}
				if ((ceilingstat > 0) && (ceilingstat < 4))
				{
					if ((cstat&1) > 0)
						for(m=um;m<=up;m++)
							if (uxsave[m+1] >= 0)
							{
								if (swapped[i] == 0)
									hline(uxsave[m+1],j-1,m,dztop,picnum1,shade1);
								else
									hline(uxsave[m+1],j-1,m,dzbot,picnum2,shade2);
								uxsave[m+1] = -1;
							}
					if ((cstat&2) > 0)
						for(m=dp;m<=dm;m++)
							if (dxsave[m+1] >= 0)
							{
								if (swapped[i] == 0)
									hline(dxsave[m+1],j-1,m,-dzbot,picnum2,shade2);
								else
									hline(dxsave[m+1],j-1,m,-dztop,picnum1,shade1);
								dxsave[m+1] = -1;
							}
				}
			}
		}
	}
	lintalplc = 0;
	takernum = 0;
}
