// "Build Engine & Tools" Copyright (c) 1993-1997 Ken Silverman
// Ken Silverman's official web site: "http://www.advsys.net/ken"
// See the included license file "BUILDLIC.TXT" for license info.

//Added numextras
//Added sector[].ceilingpal
//Added sector[].floorpal
//Added sector[].visibility
//Split sector[].tag to sector[].lotag and sector[].hitag
//Added sector[].extra
//Expanded wall[].cstat to a short
//Split wall[].tag to wall[].lotag and wall[].hitag
//Added wall[].extra
//Split sprite[].tag to sprite[].lotag and sprite[].hitag
//Got rid of sprite[].extra (the void * mess)
//Added sprite[].extra

#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>

#define MAXSECTORS 1024
#define MAXWALLS 4096
#define MAXSPRITES 4096
#define MAXEXTRAS 4096

#define MAXSTATUS 1024
#define MAXTILES 4096

static struct sectortype
{
	short unsigned int wallptr, wallnum;
	char ceilingstat, ceilingxpanning, ceilingypanning;
	signed char ceilingshade;
	long ceilingz;
	short ceilingpicnum, ceilingheinum;
	char floorstat, floorxpanning, floorypanning;
	signed char floorshade;
	long floorz;
	short floorpicnum, floorheinum;
	long tag;
};

static struct walltype
{
	long x, y;
	short point2;
	char cstat;
	signed char shade;
	unsigned char xrepeat, yrepeat, xpanning, ypanning;
	short picnum, overpicnum;
	short nextsector1, nextwall1;
	short nextsector2, nextwall2;
	long tag;
};

static struct spritetype
{
	long x, y, z;
	char cstat;
	signed char shade;
	unsigned char xrepeat, yrepeat;
	short picnum, ang, xvel, yvel, zvel, owner;
	short sectnum, statnum;
	long tag;
	char *extra;
};

static struct newsectortype
{
	short unsigned int wallptr, wallnum;
	short ceilingpicnum, floorpicnum;
	short ceilingheinum, floorheinum;
	long ceilingz, floorz;
	signed char ceilingshade, floorshade;
	char ceilingxpanning, floorxpanning;
	char ceilingypanning, floorypanning;
	char ceilingstat, floorstat;
	char ceilingpal, floorpal;
	char visibility;
	short lotag, hitag;
	short extra;
};

static struct newwalltype
{
	long x, y;
	short point2;
	short picnum, overpicnum;
	signed char shade;
	short cstat;
	unsigned char xrepeat, yrepeat, xpanning, ypanning;
	short nextsector1, nextwall1;
	short nextsector2, nextwall2;
	short lotag, hitag;
	short extra;
};

static struct newspritetype
{
	long x, y, z;
	char cstat;
	signed char shade;
	unsigned char xrepeat, yrepeat;
	short picnum, ang, xvel, yvel, zvel, owner;
	short sectnum, statnum;
	short lotag, hitag;
	short extra;
};

static struct newextratype
{
	short tag[4];
	short point2;
};

//**************************************************************************

static struct sectortype sector[MAXSECTORS];
static struct walltype wall[MAXWALLS];
static struct spritetype sprite[MAXSPRITES];

static struct newsectortype newsector[MAXSECTORS];
static struct newwalltype newwall[MAXWALLS];
static struct newspritetype newsprite[MAXSPRITES];
static struct newextratype newextra[MAXEXTRAS];

static long posx, posy, posz, mapversion;
static short int ang, cursectnum;
static short numsectors, numwalls, numsprites, numextras;

main(int argc, char **argv)
{
	char ch, *filename;
	long i, j;

	if (argc != 2)
	{
		printf("CONVMAP5 [filename]\n");
		printf("This CONVMAP5.EXE converts map version 4 to map version 5\n");
		exit(0);
	}
	strcpy(filename,argv[1]);
	if(strchr(filename,'.') == 0)
		strcat(filename,".map");

	printf("Converting from map version 4 to map version 5...\n");
	printf("Adding new fields to structures...\n");

	if ((i = loadoldboard(filename)) < 0)
	{
		if (i == -1) printf("%s not found.\n",filename);
		if (i == -2) printf("Cannot convert %s.\n",filename);
		exit(0);
	}

	for(i=0;i<numsectors;i++)
	{
		newsector[i].wallptr = sector[i].wallptr;
		newsector[i].wallnum = sector[i].wallnum;
		newsector[i].ceilingpicnum = sector[i].ceilingpicnum;
		newsector[i].floorpicnum = sector[i].floorpicnum;
		newsector[i].ceilingheinum = sector[i].ceilingheinum;
		newsector[i].floorheinum = sector[i].floorheinum;
		newsector[i].ceilingz = sector[i].ceilingz;
		newsector[i].floorz = sector[i].floorz;
		newsector[i].ceilingshade = sector[i].ceilingshade;
		newsector[i].floorshade = sector[i].floorshade;
		newsector[i].ceilingxpanning = sector[i].ceilingxpanning;
		newsector[i].floorxpanning = sector[i].floorxpanning;
		newsector[i].ceilingypanning = sector[i].ceilingypanning;
		newsector[i].floorypanning = sector[i].floorypanning;
		newsector[i].ceilingstat = sector[i].ceilingstat;
		newsector[i].floorstat = sector[i].floorstat;
		newsector[i].ceilingpal = 0;
		newsector[i].floorpal = 0;
		newsector[i].visibility = 0;
		newsector[i].lotag = (sector[i].tag&65535);
		newsector[i].hitag = (sector[i].tag>>16);
		newsector[i].extra = -1;
	}
	for(i=0;i<numwalls;i++)
	{
		newwall[i].x = wall[i].x;
		newwall[i].y = wall[i].y;
		newwall[i].point2 = wall[i].point2;
		newwall[i].picnum = wall[i].picnum;
		newwall[i].overpicnum = wall[i].overpicnum;
		newwall[i].shade = wall[i].shade;
		newwall[i].cstat = wall[i].cstat;
		newwall[i].xrepeat = wall[i].xrepeat;
		newwall[i].yrepeat = wall[i].yrepeat;
		newwall[i].xpanning = wall[i].xpanning;
		newwall[i].ypanning = wall[i].ypanning;
		newwall[i].nextsector1 = wall[i].nextsector1;
		newwall[i].nextwall1 = wall[i].nextwall1;
		newwall[i].nextsector2 = wall[i].nextsector2;
		newwall[i].nextwall2 = wall[i].nextwall2;
		newwall[i].lotag = (wall[i].tag&65535);
		newwall[i].hitag = (wall[i].tag>>16);
		newwall[i].extra = -1;
	}
	for(i=0;i<numsprites;i++)
	{
		newsprite[i].x = sprite[i].x;
		newsprite[i].y = sprite[i].y;
		newsprite[i].z = sprite[i].z;
		newsprite[i].cstat = sprite[i].cstat;
		newsprite[i].shade = sprite[i].shade;
		newsprite[i].xrepeat = sprite[i].xrepeat;
		newsprite[i].yrepeat = sprite[i].yrepeat;
		newsprite[i].picnum = sprite[i].picnum;
		newsprite[i].ang = sprite[i].ang;
		newsprite[i].xvel = sprite[i].xvel;
		newsprite[i].yvel = sprite[i].yvel;
		newsprite[i].zvel = sprite[i].zvel;
		newsprite[i].owner = sprite[i].owner;
		newsprite[i].sectnum = sprite[i].sectnum;
		newsprite[i].statnum = sprite[i].statnum;
		newsprite[i].lotag = (sprite[i].tag&65535);
		newsprite[i].hitag = (sprite[i].tag>>16);
		newsprite[i].extra = -1;
	}

	mapversion = 5L;
	numextras = 0;
	if (savenewboard(filename) == 0)
		printf("%s successfully converted.\n",filename);
	else
		printf("%s NOT converted for some stupid reason.\n",filename);
}

loadoldboard(char *filename)
{
	short int fil, i;

	if ((fil = open(filename,O_BINARY|O_RDWR,S_IREAD)) == -1)
		return(-1);

	read(fil,&mapversion,4);
	if (mapversion != 4L) return(-2);
	read(fil,&posx,4);
	read(fil,&posy,4);
	read(fil,&posz,4);
	read(fil,&ang,2);
	read(fil,&cursectnum,2);
	read(fil,&numsectors,2);
	read(fil,&numwalls,2);
	read(fil,&numsprites,2);
	for(i=0;i<numsectors;i++) read(fil,&sector[i],sizeof(struct sectortype));
	for(i=0;i<numwalls;i++) read(fil,&wall[i],sizeof(struct walltype));
	for(i=0;i<numsprites;i++) read(fil,&sprite[i],sizeof(struct spritetype));
	close(fil);

	return(0);
}

savenewboard(char *filename)
{
	short int fil, i;

	if ((fil = open(filename,O_BINARY|O_TRUNC|O_CREAT|O_WRONLY,S_IWRITE)) == -1)
		return(-1);

		 //Version control
	write(fil,&mapversion,4);

		 //Starting position
	write(fil,&posx,4);
	write(fil,&posy,4);
	write(fil,&posz,4);
	write(fil,&ang,2);
	write(fil,&cursectnum,2);

		//Sectors
	write(fil,&numsectors,2);
	for(i=0;i<numsectors;i++)
		write(fil,&newsector[i],sizeof(struct newsectortype));

		//Walls
	write(fil,&numwalls,2);
	for(i=0;i<numwalls;i++)
		write(fil,&newwall[i],sizeof(struct newwalltype));

		//Sprites
	write(fil,&numsprites,2);
	for(i=0;i<numsprites;i++)
		write(fil,&newsprite[i],sizeof(struct newspritetype));

		//Extras
	write(fil,&numextras,2);
	for(i=0;i<numextras;i++)
		write(fil,&newextra[i],sizeof(struct newextratype));

	close(fil);

	return(0);
}
