// "Build Engine & Tools" Copyright (c) 1993-1997 Ken Silverman
// Ken Silverman's official web site: "http://www.advsys.net/ken"
// See the included license file "BUILDLIC.TXT" for license info.

#include <string.h>
#include <stdlib.h>
#include "spw_int.h"
#include "build.h"
#include "names.h"

/***************************************************************************
	KEN'S TAG DEFINITIONS:      (Please define your own tags for your games)

 sector[?].tag = 0    Normal sector
 sector[?].tag = 1    If you are on a sector with this tag, then all sectors
								 with same hi tag as this are operated.  Once.
 sector[?].tag = 2    Same as sector[?].tag = 1 but this is retriggable.
 sector[?].tag = 3    A really stupid sector that really does nothing now.
 sector[?].tag = 4    A sector where you are put closer to the floor
								 (such as the slime in DOOM1.DAT)
 sector[?].tag = 5    A really stupid sector that really does nothing now.
 sector[?].tag = 6    A normal door - instead of pressing D, you tag the
								 sector with a 6.  The reason I make you edit doors
								 this way is so that can program the doors
								 yourself.
 sector[?].tag = 7    A door the goes down to open.
 sector[?].tag = 8    A door that opens horizontally in the middle.
 sector[?].tag = 9    A sliding door that opens vertically in the middle.
								 -Example of the advantages of not using BSP tree.
 sector[?].tag = 10   A warping sector with floor and walls that shade.
 sector[?].tag = 11   A sector with all walls that do X-panning.
 sector[?].tag = 12   A sector with walls using the dragging function.
 sector[?].tag = 13   A sector with some swinging doors in it.
 sector[?].tag = 14   A revolving door sector.
 sector[?].tag = 15   A subway track.
 sector[?].tag = 16   A true double-sliding door.
 sector[?].tag = 17   A true double-sliding door for subways only.

	wall[?].tag = 0    Normal wall
	wall[?].tag = 1    Y-panning wall
	wall[?].tag = 2    Switch - If you flip it, then all sectors with same hi
								 tag as this are operated.
	wall[?].tag = 3    Marked wall to detemine starting dir. (sector tag 12)
	wall[?].tag = 4    Mark on the shorter wall closest to the pivot point
								 of a swinging door. (sector tag 13)
	wall[?].tag = 5    Mark where a subway should stop. (sector tag 15)
	wall[?].tag = 6    Mark for true double-sliding doors (sector tag 16)
	wall[?].tag = 7    Water fountain

 sprite[?].tag = 0    Normal sprite
 sprite[?].tag = 1    If you press space bar on an AL, and the AL is tagged
								 with a 1, he will turn evil.
 sprite[?].tag = 2    When this sprite is operated, a bomb is shot at its
								 position.

KEN'S STATUS DEFINITIONS:  (Please define your own statuses for your games)
 status = 0            Inactive sprite
 status = 1            Active monster sprite
 status = 2            Monster that becomes active only when it sees you
 status = 3            Smoke on the wall for chainguns
 status = 4            Splashing sprites (When you shoot slime)
 status = 5            Explosion!
 status = 6            Travelling bullet
 status = 7            Bomb sprial-out explosion
 status = MAXSTATUS    Non-existent sprite (this will be true for your
														  code also)
**************************************************************************/

	//Shared player variables
static long posx[MAXPLAYERS], posy[MAXPLAYERS], posz[MAXPLAYERS];
static long horiz[MAXPLAYERS], zoom[MAXPLAYERS], hvel[MAXPLAYERS];
static short ang[MAXPLAYERS], cursectnum[MAXPLAYERS], ocursectnum[MAXPLAYERS];
static short playersprite[MAXPLAYERS], deaths[MAXPLAYERS];
static long lastchaingun[MAXPLAYERS], selectedgun[MAXPLAYERS+1];
static long health[MAXPLAYERS], score[MAXPLAYERS], saywatchit[MAXPLAYERS];
static short numbombs[MAXPLAYERS], screensize[MAXPLAYERS];
static char spaceflag[MAXPLAYERS], shootflag[MAXPLAYERS], dimflag[MAXPLAYERS];
static char dimensionmode[MAXPLAYERS];
static char revolvesyncstat[MAXPLAYERS];
static short revolvesyncang[MAXPLAYERS], revolvesyncrotang[MAXPLAYERS];
static long revolvesyncx[MAXPLAYERS], revolvesyncy[MAXPLAYERS];

	//Old player positions - for smoothing out frames in multiplayer mode
static long oposx[MAXPLAYERS], oposy[MAXPLAYERS], oposz[MAXPLAYERS];
static long ohoriz[MAXPLAYERS], ozoom[MAXPLAYERS];
static short oang[MAXPLAYERS];

	//Game recording variables
static long reccnt, recstat = 1;       //Can only record 1-player game
static signed char recsyncvel[16384][2];
static signed char recsyncsvel[16384][2];
static signed char recsyncangvel[16384][2];
static short recsyncbits[16384][2];
static unsigned char recsynctics[16384];

  //Multiplayer syncing variables
static signed char syncvel[MAXPLAYERS+1], osyncvel[MAXPLAYERS+1];
static signed char syncsvel[MAXPLAYERS+1], osyncsvel[MAXPLAYERS+1];
static signed char syncangvel[MAXPLAYERS+1], osyncangvel[MAXPLAYERS+1];
static short syncbits[MAXPLAYERS+1], osyncbits[MAXPLAYERS+1];
static unsigned char synctics, lastsynctics;

	//Miscellaneous variables
static char tempbuf[576], boardfilename[80];
static short screenpeek = 0, oldmousebstatus = 0;
static short neartagsector, neartagwall, neartagsprite;
static long lockclock, neartagdist, neartaghitdist;
extern char username[MAXPLAYERS][50];
extern long frameplace, pageoffset, ydim16;
extern void __interrupt __far ksmhandler(void);

	//Board animation variables
static short warpsectorlist[16], warpsectorcnt, warpsectorflag[MAXPLAYERS];
static short xpanningsectorlist[16], xpanningsectorcnt;
static short ypanningwalllist[64], ypanningwallcnt;
static short floorpanninglist[64], floorpanningcnt;
static short dragsectorlist[16], dragxdir[16], dragydir[16], dragsectorcnt;
static long dragx1[16], dragy1[16], dragx2[16], dragy2[16], dragfloorz[16];
static short swingcnt, swingwall[32][8], swingsector[32];
static short swingangopen[32], swingangclosed[32], swingangopendir[32];
static short swingang[32], swinganginc[32];
static long swingx[32][8], swingy[32][8];
static short revolvesector[4], revolveang[4], revolveclip[4], revolvecnt;
static long revolvex[4][16], revolvey[4][16];
static long revolvepivotx[4], revolvepivoty[4];
static short subwaytracksector[4][128], subwaynumsectors[4], subwaytrackcnt;
static long subwaystop[4][8], subwaystopcnt[4];
static long subwaytrackx1[4], subwaytracky1[4];
static long subwaytrackx2[4], subwaytracky2[4];
static long subwayx[4], subwaygoalstop[4], subwayvel[4], subwaypausetime[4];
static short waterfountainwall[MAXPLAYERS], waterfountaincnt[MAXPLAYERS];
static short slimesoundcnt[MAXPLAYERS];

	//Variables that let you type messages to other player
static char getmessage[162], getmessageleng;
static long getmessagetimeoff;
static char typemessage[162], typemessageleng = 0, typemode = 0;
static char scantoasc[128] =
{
	0,0,'1','2','3','4','5','6','7','8','9','0','-','=',0,0,
	'q','w','e','r','t','y','u','i','o','p','[',']',0,0,'a','s',
	'd','f','g','h','j','k','l',';',39,'`',0,92,'z','x','c','v',
	'b','n','m',',','.','/',0,'*',0,32,0,0,0,0,0,0,
	0,0,0,0,0,0,0,'7','8','9','-','4','5','6','+','1',
	'2','3','0','.',0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
};
static char scantoascwithshift[128] =
{
	0,0,'!','@','#','$','%','^','&','*','(',')','_','+',0,0,
	'Q','W','E','R','T','Y','U','I','O','P','{','}',0,0,'A','S',
	'D','F','G','H','J','K','L',':',34,'~',0,'|','Z','X','C','V',
	'B','N','M','<','>','?',0,'*',0,32,0,0,0,0,0,0,
	0,0,0,0,0,0,0,'7','8','9','-','4','5','6','+','1',
	'2','3','0','.',0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
};

	//Here are some nice in-line assembly pragmas that make life easier.
	//(at least for Ken)
#pragma aux setvmode =\
	"int 0x10",\
	parm [eax]\

main(short int argc,char **argv)
{
	long i, j, k, smoothratio, waitplayers;

	if (argc >= 2)
	{
		strcpy(&boardfilename,argv[1]);
		if(strchr(boardfilename,'.') == 0)
			strcat(boardfilename,".map");
	}
	else
		strcpy(&boardfilename,"nukeland.map");

	initengine();

	initsb();                                      //Init digitized sound
	if (option[2] == 1)                            //Init music
	{
		if (strcmp(boardfilename,"klab.map") == 0)
			loadmusic("klabsong.ksm");
		else
			loadmusic("eviltune.ksm");
		musicon(); kenchaintimer(ksmhandler,120);
	}

	loadpics("tiles000.art");                      //Load artwork

		//Here's an example of TRUE ornamented walls
		//The allocatepermanenttile should be called right after loadpics
		//Since it resets the tile cache for each call.
	if (allocatepermanenttile(4095,64,64) >= 0)    //If enough memory
	{
			//My face with an explosion written over it
		copytilepiece(KENPICTURE,0,0,64,64,4095,0,0);
		copytilepiece(EXPLOSION,0,0,64,64,4095,0,0);
	}

	prepareboard(boardfilename);                   //Load board

	if (option[4] > 0)
	{
		if ((option[4] >= 1) && (option[4] <= 4))
			waitplayers = 2;
		else
			waitplayers = option[4]-3;

			//Allocate palette lookup tables for other players
		for(i=1;i<waitplayers;i++)
		{
			for(j=0;j<256;j++)
				tempbuf[j] = j;

			switch(i)
			{
				case 1:
					for(j=0;j<32;j++)
						tempbuf[j+192] = j+128;  //remap green to red (player 2)
					break;
				case 2:
					for(j=0;j<32;j++)
						tempbuf[j+192] = j+32;   //remap green to blue (player 3)
					break;
				case 3:
					for(j=0;j<32;j++)
						tempbuf[j+192] = j+224;  //remap green to purple (player 4)
					break;
				case 4:
					for(j=0;j<32;j++)
						tempbuf[j+192] = j+64;   //remap green to orange (player 5)
					break;
			}
			allocatespritepalookup(i,tempbuf);
		}

		if (waitplayers == 2)
		{
			printext256(68L,84L,31,0,"Waiting for other player...",0);
		}
		else
		{
			sprintf(tempbuf,"Waiting for %ld players...",waitplayers);
			printext256(68L,84L,31,0,tempbuf,0);
		}

		sendlogon();

		k = -1;
		while (numplayers < waitplayers)
		{
			if ((option[4] >= 5) && (k != numplayers))
			{
				k = numplayers;

				i = connecthead; j = 100;
				while (i != -1)
				{
					printext256(96L,j,31,0,"                        ",0);
					printext256(96L,j,31,0,username[i],0);
					j += 8;

					i = connectpoint2[i];
				}
			}

			getpackets();
			if (keystatus[1] > 0)
			{
				if (numplayers > 1)            //Signing off
					sendlogoff();

				if (option[2] == 1)
				{
					kenchaintimer(0,0); musicoff();
				}
				uninitengine();
				uninitsb();
				setvmode(0x3);        //Set back to text mode
				exit(0);
			}

			//if (keystatus[0x1c] > 0)
			//{
			//   if (numplayers >= 2)
			//      if (myconnectnum == masterconnectnum)
			//      {
			//         //startgame
			//      }
			//   keystatus[0x1c] = 0;
			//}
		}
		screenpeek = myconnectindex;
	}

	reccnt = 0;

	initplayersprites();
	resettiming();

		//This is the whole game loop!
	while ((keystatus[1] == 0) && (sector[cursectnum[myconnectindex]].tag != 65535))
	{
		if (numplayers <= 1)
		{
			drawscreen(0,256L);
			getinput(0);
			movethings();
		}
		else
		{
			if ((myconnectnum != masterconnectnum) && (lastsynctics != 0))
				drawscreen(screenpeek,((totalclock-*lastpacket2clock)<<8)/lastsynctics);
			else
				drawscreen(screenpeek,256L);
			sync();
		}
	}

	for(i=0;i<65536;i++)
		i += 0;

	if (numplayers > 1)            //Signing off
		sendlogoff();

	if (option[2] == 1)
	{
		kenchaintimer(0,0); musicoff();
	}
	uninitengine();
	uninitsb();
	setvmode(0x3);        //Set back to text mode
	showengineinfo();     //Show speed statistics

	if (keystatus[1] == 2) printf("Skipped a packet\n");

	return(0);
}

operatesector(short dasector)
{     //Door code
	short startwall, endwall, wallfind[2];
	long i, j, k, s, nexti, good, cnt, datag;
	long dax, day, daz, dax2, day2, daz2, centx, centy;
	
	datag = (sector[dasector].tag&65535);

	startwall = sector[dasector].wallptr;
	endwall = startwall + sector[dasector].wallnum - 1;
	centx = 0L, centy = 0L;
	for(i=startwall;i<=endwall;i++)
	{
		centx += wall[i].x;
		centy += wall[i].y;
	}
	centx /= (endwall-startwall+1);
	centy /= (endwall-startwall+1);

		//Simple door that moves up  (tag 8 is a combination of tags 6 & 7)
	if ((datag == 6) || (datag == 8))    //If the sector in front is a door
	{
		i = getanimationgoal(&sector[dasector].ceilingz);
		if (i >= 0)      //If door already moving, reverse its direction
		{
			if (datag == 8)
				daz = ((sector[dasector].ceilingz+sector[dasector].floorz)>>1);
			else
				daz = sector[dasector].floorz;

			if (animategoal[i] == daz)
				animategoal[i] = sector[nextsectorneighborz(dasector,sector[dasector].floorz,-1,-1)].ceilingz;
			else
				animategoal[i] = daz;
		}
		else      //else insert the door's ceiling on the animation list
		{
			if (sector[dasector].ceilingz == sector[dasector].floorz)
				daz = sector[nextsectorneighborz(dasector,sector[dasector].floorz,-1,-1)].ceilingz;
			else
			{
				if (datag == 8)
					daz = ((sector[dasector].ceilingz+sector[dasector].floorz)>>1);
				else
					daz = sector[dasector].floorz;
			}
			if ((j = setanimation(&sector[dasector].ceilingz,daz,128)) >= 0)
			{
				wsay("updowndr.wav",4096L,sndvol(centx,centy));
			}
		}
	}
		//Simple door that moves down
	if ((datag == 7) || (datag == 8)) //If the sector in front's elevator
	{
		i = getanimationgoal(&sector[dasector].floorz);
		if (i >= 0)      //If elevator already moving, reverse its direction
		{
			if (datag == 8)
				daz = ((sector[dasector].ceilingz+sector[dasector].floorz)>>1);
			else
				daz = sector[dasector].ceilingz;

			if (animategoal[i] == daz)
				animategoal[i] = sector[nextsectorneighborz(dasector,sector[dasector].ceilingz,1,1)].floorz;
			else
				animategoal[i] = daz;

		}
		else      //else insert the elevator's ceiling on the animation list
		{
			if (sector[dasector].floorz == sector[dasector].ceilingz)
				daz = sector[nextsectorneighborz(dasector,sector[dasector].ceilingz,1,1)].floorz;
			else
			{
				if (datag == 8)
					daz = ((sector[dasector].ceilingz+sector[dasector].floorz)>>1);
				else
					daz = sector[dasector].ceilingz;
			}
			if ((j = setanimation(&sector[dasector].floorz,daz,128)) >= 0)
				wsay("updowndr.wav",4096L,sndvol(centx,centy));
		}
	}

	if (datag == 9)   //Smooshy-wall sideways double-door
	{
		//find any points with either same x or same y coordinate
		//  as center (centx, centy) - should be 2 points found.
		wallfind[0] = -1;
		wallfind[1] = -1;
		for(i=startwall;i<=endwall;i++)
			if ((wall[i].x == centx) || (wall[i].y == centy))
			{
				if (wallfind[0] == -1)
					wallfind[0] = i;
				else
					wallfind[1] = i;
			}

		for(j=0;j<2;j++)
		{
			if ((wall[wallfind[j]].x == centx) && (wall[wallfind[j]].y == centy))
			{
				//find what direction door should open by averaging the
				//  2 neighboring points of wallfind[0] & wallfind[1].
				i = wallfind[j]-1; if (i < startwall) i = endwall;
				dax2 = ((wall[i].x+wall[wall[wallfind[j]].point2].x)>>1)-wall[wallfind[j]].x;
				day2 = ((wall[i].y+wall[wall[wallfind[j]].point2].y)>>1)-wall[wallfind[j]].y;
				if (dax2 != 0)
				{
					dax2 = wall[wall[wall[wallfind[j]].point2].point2].x;
					dax2 -= wall[wall[wallfind[j]].point2].x;
					setanimation(&wall[wallfind[j]].x,wall[wallfind[j]].x+dax2,4);
					setanimation(&wall[i].x,wall[i].x+dax2,4);
					setanimation(&wall[wall[wallfind[j]].point2].x,wall[wall[wallfind[j]].point2].x+dax2,4);
				}
				else if (day2 != 0)
				{
					day2 = wall[wall[wall[wallfind[j]].point2].point2].y;
					day2 -= wall[wall[wallfind[j]].point2].y;
					setanimation(&wall[wallfind[j]].y,wall[wallfind[j]].y+day2,4);
					setanimation(&wall[i].y,wall[i].y+day2,4);
					setanimation(&wall[wall[wallfind[j]].point2].y,wall[wall[wallfind[j]].point2].y+day2,4);
				}
			}
			else
			{
				i = wallfind[j]-1; if (i < startwall) i = endwall;
				dax2 = ((wall[i].x+wall[wall[wallfind[j]].point2].x)>>1)-wall[wallfind[j]].x;
				day2 = ((wall[i].y+wall[wall[wallfind[j]].point2].y)>>1)-wall[wallfind[j]].y;
				if (dax2 != 0)
				{
					setanimation(&wall[wallfind[j]].x,centx,4);
					setanimation(&wall[i].x,centx+dax2,4);
					setanimation(&wall[wall[wallfind[j]].point2].x,centx+dax2,4);
				}
				else if (day2 != 0)
				{
					setanimation(&wall[wallfind[j]].y,centy,4);
					setanimation(&wall[i].y,centy+day2,4);
					setanimation(&wall[wall[wallfind[j]].point2].y,centy+day2,4);
				}
			}
		}
		wsay("updowndr.wav",4096L,sndvol(centx,centy));
	}

	if (datag == 13)  //Swinging door
	{
		for(i=0;i<swingcnt;i++)
		{
			if (swingsector[i] == dasector)
			{
				if (swinganginc[i] == 0)
				{
					if (swingang[i] == swingangclosed[i])
					{
						swinganginc[i] = swingangopendir[i];
						wsay("opendoor.wav",4096L,sndvol(centx,centy));
					}
					else
						swinganginc[i] = -swingangopendir[i];
				}
				else
				{
					swinganginc[i] = -swinganginc[i];
				}
			}
		}
	}

	if (datag == 16)  //True sideways double-sliding door
	{
		 //get 2 closest line segments to center (dax, day)
		wallfind[0] = -1;
		wallfind[1] = -1;
		for(i=startwall;i<=endwall;i++)
			if (wall[i].tag == 6)
			{
				if (wallfind[0] == -1)
					wallfind[0] = i;
				else
					wallfind[1] = i;
			}

		for(j=0;j<2;j++)
		{
			if ((((wall[wallfind[j]].x+wall[wall[wallfind[j]].point2].x)>>1) == centx) && (((wall[wallfind[j]].y+wall[wall[wallfind[j]].point2].y)>>1) == centy))
			{     //door was closed
					//find what direction door should open
				i = wallfind[j]-1; if (i < startwall) i = endwall;
				dax2 = wall[i].x-wall[wallfind[j]].x;
				day2 = wall[i].y-wall[wallfind[j]].y;
				if (dax2 != 0)
				{
					dax2 = wall[wall[wall[wall[wallfind[j]].point2].point2].point2].x;
					dax2 -= wall[wall[wall[wallfind[j]].point2].point2].x;
					setanimation(&wall[wallfind[j]].x,wall[wallfind[j]].x+dax2,4);
					setanimation(&wall[i].x,wall[i].x+dax2,4);
					setanimation(&wall[wall[wallfind[j]].point2].x,wall[wall[wallfind[j]].point2].x+dax2,4);
					setanimation(&wall[wall[wall[wallfind[j]].point2].point2].x,wall[wall[wall[wallfind[j]].point2].point2].x+dax2,4);
				}
				else if (day2 != 0)
				{
					day2 = wall[wall[wall[wall[wallfind[j]].point2].point2].point2].y;
					day2 -= wall[wall[wall[wallfind[j]].point2].point2].y;
					setanimation(&wall[wallfind[j]].y,wall[wallfind[j]].y+day2,4);
					setanimation(&wall[i].y,wall[i].y+day2,4);
					setanimation(&wall[wall[wallfind[j]].point2].y,wall[wall[wallfind[j]].point2].y+day2,4);
					setanimation(&wall[wall[wall[wallfind[j]].point2].point2].y,wall[wall[wall[wallfind[j]].point2].point2].y+day2,4);
				}
			}
			else
			{    //door was not closed
				i = wallfind[j]-1; if (i < startwall) i = endwall;
				dax2 = wall[i].x-wall[wallfind[j]].x;
				day2 = wall[i].y-wall[wallfind[j]].y;
				if (dax2 != 0)
				{
					setanimation(&wall[wallfind[j]].x,centx,4);
					setanimation(&wall[i].x,centx+dax2,4);
					setanimation(&wall[wall[wallfind[j]].point2].x,centx,4);
					setanimation(&wall[wall[wall[wallfind[j]].point2].point2].x,centx+dax2,4);
				}
				else if (day2 != 0)
				{
					setanimation(&wall[wallfind[j]].y,centy,4);
					setanimation(&wall[i].y,centy+day2,4);
					setanimation(&wall[wall[wallfind[j]].point2].y,centy,4);
					setanimation(&wall[wall[wall[wallfind[j]].point2].point2].y,centy+day2,4);
				}
			}
		}
		wsay("updowndr.wav",4096L,sndvol(centx,centy));
	}
}

operatesprite(short dasprite)
{
	long datag;

	datag = (sprite[dasprite].tag&65535);

	if (datag == 2)    //A sprite that shoots a bomb
	{
		shootgun(dasprite,
					sprite[dasprite].x,sprite[dasprite].y,sprite[dasprite].z,
					sprite[dasprite].ang,100L,sprite[dasprite].sectnum,2);
	}
}

changehealth(short snum, short deltahealth)
{
	long dax, day;
	short good, k, startwall, endwall, s;
	
	if (health[snum] > 0)
	{
		health[snum] += deltahealth;
		if (health[snum] > 999) health[snum] = 999;
	
		if (health[snum] <= 0)
		{
			health[snum] = -1;
			wsay("death.wav",4096L,sndvol(posx[snum],posy[snum]));
			sprite[playersprite[snum]].picnum = SKELETON;
		}

		if ((snum == screenpeek) && (dimensionmode[screenpeek] == 3) && (screensize[screenpeek] <= 320))
		{
			if (health[snum] > 0)
				sprintf(&tempbuf,"Health: %3d",health[snum]);
			else
				sprintf(&tempbuf,"YOU STINK!!");

			printext(160-(strlen(tempbuf)<<2),176,tempbuf,ALPHABET,80);
		}
	}
	return(health[snum] <= 0);      //You were just injured
}

changescore(short snum, short deltascore)
{
	score[snum] += deltascore;

	if ((snum == screenpeek) && (dimensionmode[screenpeek] == 3) && (screensize[screenpeek] <= 320))
	{
		sprintf(&tempbuf,"Score:%ld",score[snum]);
		printext(160-(strlen(tempbuf)<<2),192,tempbuf,ALPHABET,80);
	}
}

prepareboard(char *daboardfilename)
{
	short startwall, endwall, dasector;
	long i, j, k, s, dax, day, daz, dax2, day2;

	getmessageleng = 0;
	typemessageleng = 0;

	randomseed = 17L;

		//Clear (do)animation's list
	animatecnt = 0;
	typemode = 0;

	if (loadboard(daboardfilename,&posx[0],&posy[0],&posz[0],&ang[0],&cursectnum[0]) == -1)
	{
		printf("Board not found\n");
		exit(0);
	}
	precache();

	for(i=0;i<MAXPLAYERS;i++)
	{
		posx[i] = posx[0];
		posy[i] = posy[0];
		posz[i] = posz[0];
		ang[i] = ang[0];
		cursectnum[i] = cursectnum[0];
		ocursectnum[i] = cursectnum[0];
		horiz[i] = 100;
		lastchaingun[i] = 0;
		health[i] = 100;
		score[i] = 0L;
		dimensionmode[i] = 3;
		numbombs[i] = -1;
		zoom[i] = 768L;
		deaths[i] = 0L;
		playersprite[i] = -1;
		saywatchit[i] = -1;
		spaceflag[i] = 0;
		shootflag[i] = 0;
		dimflag[i] = 0;
		screensize[i] = 320;
	}

	setup3dscreen();

	for(i=0;i<MAXPLAYERS+1;i++)
	{
		osyncvel[i] = 0;
		osyncsvel[i] = 0;
		osyncangvel[i] = 0;
		osyncbits[i] = 0;
	}

		//Scan sector tags

	for(i=0;i<MAXPLAYERS;i++)
	{
		warpsectorflag[i] = 0;  //Make sure you don't warp more than once
		waterfountainwall[i] = -1;
		waterfountaincnt[i] = 0;
	}
	slimesoundcnt[i] = 0;
	warpsectorcnt = 0;      //Make a list of warping sectors
	xpanningsectorcnt = 0;  //Make a list of wall x-panning sectors
	floorpanningcnt = 0;    //Make a list of slime sectors
	dragsectorcnt = 0;      //Make a list of moving platforms
	swingcnt = 0;           //Make a list of swinging doors
	revolvecnt = 0;         //Make a list of revolving doors
	subwaytrackcnt = 0;     //Make a list of subways

	for(i=0;i<numsectors;i++)
	{
		switch(sector[i].tag&65535)
		{
			case 4:
				floorpanninglist[floorpanningcnt++] = i;
				break;
			case 10:
				warpsectorlist[warpsectorcnt++] = i;
				break;
			case 11:
				xpanningsectorlist[xpanningsectorcnt++] = i;
				break;
			case 12:
				dasector = i;
				dax = 0x7fffffff;
				day = 0x7fffffff;
				dax2 = 0x80000000;
				day2 = 0x80000000;
				startwall = sector[i].wallptr;
				endwall = startwall+sector[i].wallnum-1;
				for(j=startwall;j<=endwall;j++)
				{
					if (wall[j].x < dax) dax = wall[j].x;
					if (wall[j].y < day) day = wall[j].y;
					if (wall[j].x > dax2) dax2 = wall[j].x;
					if (wall[j].y > day2) day2 = wall[j].y;
					if (wall[j].tag == 3) k = j;
				}
				if (wall[k].x == dax) dragxdir[dragsectorcnt] = -16;
				if (wall[k].y == day) dragydir[dragsectorcnt] = -16;
				if (wall[k].x == dax2) dragxdir[dragsectorcnt] = 16;
				if (wall[k].y == day2) dragydir[dragsectorcnt] = 16;

				dasector = wall[startwall].nextsector1;
				dragx1[dragsectorcnt] = 0x7fffffff;
				dragy1[dragsectorcnt] = 0x7fffffff;
				dragx2[dragsectorcnt] = 0x80000000;
				dragy2[dragsectorcnt] = 0x80000000;
				startwall = sector[dasector].wallptr;
				endwall = startwall+sector[dasector].wallnum-1;
				for(j=startwall;j<=endwall;j++)
				{
					if (wall[j].x < dragx1[dragsectorcnt]) dragx1[dragsectorcnt] = wall[j].x;
					if (wall[j].y < dragy1[dragsectorcnt]) dragy1[dragsectorcnt] = wall[j].y;
					if (wall[j].x > dragx2[dragsectorcnt]) dragx2[dragsectorcnt] = wall[j].x;
					if (wall[j].y > dragy2[dragsectorcnt]) dragy2[dragsectorcnt] = wall[j].y;
				}

				dragx1[dragsectorcnt] += (wall[sector[i].wallptr].x-dax);
				dragy1[dragsectorcnt] += (wall[sector[i].wallptr].y-day);
				dragx2[dragsectorcnt] -= (dax2-wall[sector[i].wallptr].x);
				dragy2[dragsectorcnt] -= (day2-wall[sector[i].wallptr].y);

				dragfloorz[dragsectorcnt] = sector[i].floorz;

				dragsectorlist[dragsectorcnt++] = i;
				break;
			case 13:
				startwall = sector[i].wallptr;
				endwall = startwall+sector[i].wallnum-1;
				for(j=startwall;j<=endwall;j++)
				{
					if (wall[j].tag == 4)
					{
						k = wall[wall[wall[wall[j].point2].point2].point2].point2;
						if ((wall[j].x == wall[k].x) && (wall[j].y == wall[k].y))
						{     //Door opens counterclockwise
							swingwall[swingcnt][0] = j;
							swingwall[swingcnt][1] = wall[j].point2;
							swingwall[swingcnt][2] = wall[wall[j].point2].point2;
							swingwall[swingcnt][3] = wall[wall[wall[j].point2].point2].point2;
							swingangopen[swingcnt] = 1536;
							swingangclosed[swingcnt] = 0;
							swingangopendir[swingcnt] = -1;
						}
						else
						{     //Door opens clockwise
							swingwall[swingcnt][0] = wall[j].point2;
							swingwall[swingcnt][1] = j;
							swingwall[swingcnt][2] = lastwall(j);
							swingwall[swingcnt][3] = lastwall(swingwall[swingcnt][2]);
							swingangopen[swingcnt] = 512;
							swingangclosed[swingcnt] = 0;
							swingangopendir[swingcnt] = 1;
						}
						for(k=0;k<4;k++)
						{
							swingx[swingcnt][k] = wall[swingwall[swingcnt][k]].x;
							swingy[swingcnt][k] = wall[swingwall[swingcnt][k]].y;
						}

						swingsector[swingcnt] = i;
						swingang[swingcnt] = swingangclosed[swingcnt];
						swinganginc[swingcnt] = 0;
						swingcnt++;
					}
				}
				break;
			case 14:
				startwall = sector[i].wallptr;
				endwall = startwall+sector[i].wallnum-1;
				dax = 0L;
				day = 0L;
				for(j=startwall;j<=endwall;j++)
				{
					dax += wall[j].x;
					day += wall[j].y;
				}
				revolvepivotx[revolvecnt] = dax / (endwall-startwall+1);
				revolvepivoty[revolvecnt] = day / (endwall-startwall+1);

				k = 0;
				for(j=startwall;j<=endwall;j++)
				{
					revolvex[revolvecnt][k] = wall[j].x;
					revolvey[revolvecnt][k] = wall[j].y;
					k++;
				}
				revolvesector[revolvecnt] = i;
				revolveang[revolvecnt] = 0;

					//Make revolving door clip only if neighbor ceiling z doesn't match.
				revolveclip[revolvecnt] = 1;
				if (sector[i].ceilingz == sector[wall[startwall].nextsector1].ceilingz)
					revolveclip[revolvecnt] = 0;

				revolvecnt++;
				break;
			case 15:
				subwaytracksector[subwaytrackcnt][0] = i;

				subwaystopcnt[subwaytrackcnt] = 0;
				dax = 0x7fffffff;
				day = 0x7fffffff;
				dax2 = 0x80000000;
				day2 = 0x80000000;
				startwall = sector[i].wallptr;
				endwall = startwall+sector[i].wallnum-1;
				for(j=startwall;j<=endwall;j++)
				{
					if (wall[j].x < dax) dax = wall[j].x;
					if (wall[j].y < day) day = wall[j].y;
					if (wall[j].x > dax2) dax2 = wall[j].x;
					if (wall[j].y > day2) day2 = wall[j].y;
				}
				for(j=startwall;j<=endwall;j++)
				{
					if (wall[j].tag == 5)
					{
						if ((wall[j].x > dax) && (wall[j].y > day) && (wall[j].x < dax2) && (wall[j].y < day2))
						{
							subwayx[subwaytrackcnt] = wall[j].x;
						}
						else
						{
							subwaystop[subwaytrackcnt][subwaystopcnt[subwaytrackcnt]] = wall[j].x;
							subwaystopcnt[subwaytrackcnt]++;
						}
					}
				}

				for(j=1;j<subwaystopcnt[subwaytrackcnt];j++)
					for(k=0;k<j;k++)
						if (subwaystop[subwaytrackcnt][j] < subwaystop[subwaytrackcnt][k])
						{
							s = subwaystop[subwaytrackcnt][j];
							subwaystop[subwaytrackcnt][j] = subwaystop[subwaytrackcnt][k];
							subwaystop[subwaytrackcnt][k] = s;
						}

				subwaygoalstop[subwaytrackcnt] = 0;
				for(j=0;j<subwaystopcnt[subwaytrackcnt];j++)
					if (labs(subwaystop[subwaytrackcnt][j]-subwayx[subwaytrackcnt]) < labs(subwaystop[subwaytrackcnt][subwaygoalstop[subwaytrackcnt]]-subwayx[subwaytrackcnt]))
						subwaygoalstop[subwaytrackcnt] = j;

				subwaytrackx1[subwaytrackcnt] = dax;
				subwaytracky1[subwaytrackcnt] = day;
				subwaytrackx2[subwaytrackcnt] = dax2;
				subwaytracky2[subwaytrackcnt] = day2;

				subwaynumsectors[subwaytrackcnt] = 1;
				for(j=0;j<numsectors;j++)
					if (j != i)
					{
						startwall = sector[j].wallptr;
						if (wall[startwall].x > subwaytrackx1[subwaytrackcnt])
							if (wall[startwall].y > subwaytracky1[subwaytrackcnt])
								if (wall[startwall].x < subwaytrackx2[subwaytrackcnt])
									if (wall[startwall].y < subwaytracky2[subwaytrackcnt])
									{
										if (sector[j].tag == 16)
											sector[j].tag = 17;     //Make special subway door

										if (sector[j].floorz != sector[i].floorz)
										{
											sector[j].ceilingstat |= 64;
											sector[j].floorstat |= 64;
										}
										subwaytracksector[subwaytrackcnt][subwaynumsectors[subwaytrackcnt]] = j;
										subwaynumsectors[subwaytrackcnt]++;
									}
					}

				subwayvel[subwaytrackcnt] = 64;
				subwaypausetime[subwaytrackcnt] = 720;
				subwaytrackcnt++;
				break;
		}
	}


		//Scan wall tags

	ypanningwallcnt = 0;
	for(i=0;i<numwalls;i++)
	{
		if (wall[i].tag == 1)
			ypanningwalllist[ypanningwallcnt++] = i;
	}


		//Scan sprite picnum's

	for(i=0;i<MAXSPRITES;i++)
		if (sprite[i].statnum < MAXSTATUS)   //That is, if sprite exists
		{
			switch(sprite[i].picnum)
			{
				case BROWNMONSTER:           //All cases here put the sprite
				case DOOMGUY:                //on waiting for you (list 2)
					changespritestat(i,2);
					sprite[i].tag = 100;
					break;
			}
			switch(sprite[i].picnum)
			{
				case AL:                     //These sprites will be
				case BROWNMONSTER:           //sensitive to the hitscan function
				case EVILAL:
				case DOOMGUY:
					sprite[i].cstat |= 1;     //Set the hitscan sensitivity bit
					break;
			}
		}

		//Map starts out completed
	for(i=0;i<(MAXWALLS>>3);i++)
		show2dwall[i] = 0xff;
	for(i=0;i<(MAXSPRITES>>3);i++)
		show2dsprite[i] = 0xff;

		//Map starts out blank; you map out what you see in 3D mode
	//automapping = 1;

	lockclock = 0;

	dax = 160-(screensize[screenpeek]>>1);
	dax2 = dax+screensize[screenpeek]-1;
	day = 84-(((screensize[screenpeek]*168)/320)>>1);
	day2 = day + ((screensize[screenpeek]*168)/320)-1;
	setview(dax,day,dax2,day2);
}


checktouchsprite(short snum)
{
	long i, nexti;

	if ((cursectnum[snum] < 0) || (cursectnum[snum] >= numsectors))
		return;

	i = headspritesect[cursectnum[snum]];
	while (i != -1)
	{
		nexti = nextspritesect[i];
		if ((labs(posx[snum]-sprite[i].x)+labs(posy[snum]-sprite[i].y) < 512) && (labs((posz[snum]>>8)-((sprite[i].z>>8)-(tilesizy[sprite[i].picnum]>>1))) <= 40))
		{
			switch(sprite[i].picnum)
			{
				case COINSTACK:
					wsay("getstuff.wav",4096L,sndvol(sprite[i].x,sprite[i].y));
					changescore(snum,1000);
					changehealth(snum,25);
					deletesprite((short)i);
					break;
				case GIFTBOX:
					wsay("getstuff.wav",4096L,sndvol(sprite[i].x,sprite[i].y));
					changescore(snum,1000);
					changehealth(snum,10);
					deletesprite((short)i);
					break;
				case COIN:
					wsay("getstuff.wav",4096L,sndvol(sprite[i].x,sprite[i].y));
					changescore(snum,100);
					changehealth(snum,5);
					deletesprite((short)i);
					break;
				case DIAMONDS:
					wsay("getstuff.wav",4096L,sndvol(sprite[i].x,sprite[i].y));
					changescore(snum,500);
					changehealth(snum,15);
					deletesprite((short)i);
					break;
				case CANON:
					if (numbombs[snum] == -1)
					{
						wsay("getstuff.wav",3584L,sndvol(sprite[i].x,sprite[i].y));
						changescore(snum,1000);
						if (snum == myconnectindex)
							keystatus[4] = 1;
						numbombs[snum] = 32767;
					}
					break;
			}
		}

		i = nexti;
	}
}

shootgun(short snum, long x, long y, long z,
			short daang, long dahoriz, short dasectnum, char guntype)
{
	short hitsect, hitwall, hitsprite, daang2;
	long i, j, daz2, hitx, hity, hitz;

	switch(guntype)
	{
		case 0:    //Shoot silver sphere bullet
			wsay("shoot2.wav",4096L,sndvol(x,y));
			j = insertsprite(dasectnum,6);
			sprite[j].x = x;
			sprite[j].y = y;
			sprite[j].z = z+(8<<8);
			sprite[j].cstat = 0;        //Hitscan does not hit other bullets
			sprite[j].picnum = BULLET;
			sprite[j].shade = 0;
			sprite[j].xrepeat = 64;
			sprite[j].yrepeat = 64;
			sprite[j].ang = daang;
			sprite[j].xvel = (sintable[(daang+2560)&2047]>>5);
			sprite[j].yvel = (sintable[(daang+2048)&2047]>>5);
			sprite[j].zvel = ((100-dahoriz)*64);
			sprite[j].owner = snum+4096;
			sprite[j].tag = 0L;
			break;
		case 1:    //Shoot chain gun
			daang2 = ((daang + 2048 + (krand()&31)-16)&2047);
			daz2 = ((100-dahoriz)*2000) + ((krand()-32768)>>1);

			hitscan(x,y,z,dasectnum,                   //Start position
				sintable[(daang2+2560)&2047],           //X vector of 3D ang
				sintable[(daang2+2048)&2047],           //Y vector of 3D ang
				daz2,                                   //Z vector of 3D ang
				&hitsect,&hitwall,&hitsprite,&hitx,&hity,&hitz);

			if (wall[hitwall].picnum == KENPICTURE)
			{
				if (waloff[4095] >= 0)
					wall[hitwall].picnum = 4095;
				wsay("hello.wav",4096L,sndvol(wall[hitwall].x,wall[hitwall].y));   //Ken says, "Hello... how are you today!"
			}
			else if (((hitwall < 0) && (hitsprite < 0) && (hitz >= z) && (sector[hitsect].floorpicnum == SLIME)) || ((hitwall >= 0) && (wall[hitwall].picnum == SLIME)))
			{    //If you shoot slime, make a splash
				wsay("splash.wav",4096L+(rand()&511)-256,sndvol(hitx,hity));

				j = insertsprite(hitsect,4);
				sprite[j].x = hitx;
				sprite[j].y = hity;
				sprite[j].z = hitz;
				sprite[j].cstat = 2;
				sprite[j].picnum = SPLASH;
				sprite[j].shade = 0;
				sprite[j].xrepeat = 64;
				sprite[j].yrepeat = 64;
				sprite[j].ang = daang;
				sprite[j].xvel = 0;
				sprite[j].yvel = 0;
				sprite[j].zvel = 0;
				sprite[j].owner = snum+4096;
				sprite[j].tag = 63;       //Time left for splash to be alive
			}
			else
			{
				wsay("shoot.wav",4096L,sndvol(hitx,hity));

				if ((hitsprite >= 0) && (sprite[hitsprite].statnum < MAXSTATUS))
					switch(sprite[hitsprite].picnum)
					{
						case BROWNMONSTER:
							if (sprite[hitsprite].tag > 0)
								sprite[hitsprite].tag -= 10;
							if (sprite[hitsprite].tag > 0)
							{
								wsay("hurt.wav",4096L+(rand()&511)-256,sndvol(hitx,hity));
								if (sprite[hitsprite].tag <= 25)
									sprite[hitsprite].cstat |= 2;
							}
							else
							{
								wsay("blowup.wav",4096L,sndvol(hitx,hity));
								sprite[hitsprite].picnum = GIFTBOX;
								sprite[hitsprite].cstat &= ~3;    //Should not clip
								changespritestat(hitsprite,0);
								changescore(snum,200);

								j = insertsprite(hitsect,5);
								sprite[j].x = hitx;
								sprite[j].y = hity;
								sprite[j].z = hitz+(32<<8);
								sprite[j].cstat = 0;        //Hitscan does not hit explosion
								sprite[j].picnum = EXPLOSION;
								sprite[j].shade = -4;
								sprite[j].xrepeat = 64;
								sprite[j].yrepeat = 64;
								sprite[j].ang = daang;
								sprite[j].xvel = 0;
								sprite[j].yvel = 0;
								sprite[j].zvel = 0;
								sprite[j].owner = snum+4096;
								sprite[j].tag = 31;  //Time left for explosion to stay
							}
							break;
						case EVILAL:
							wsay("blowup.wav",4096L,sndvol(hitx,hity));
							sprite[hitsprite].picnum = EVILALGRAVE;
							sprite[hitsprite].xrepeat = (krand()&63)+16;
							sprite[hitsprite].yrepeat = sprite[hitsprite].xrepeat;
							changespritestat(hitsprite,0);

							j = insertsprite(hitsect,5);
							sprite[j].x = hitx;
							sprite[j].y = hity;
							sprite[j].z = hitz+(32<<8);
							sprite[j].cstat = 0;        //Hitscan does not hit explosion
							sprite[j].picnum = EXPLOSION;
							sprite[j].shade = -4;
							sprite[j].xrepeat = 64;
							sprite[j].yrepeat = 64;
							sprite[j].ang = daang;
							sprite[j].xvel = 0;
							sprite[j].yvel = 0;
							sprite[j].zvel = 0;
							sprite[j].owner = snum+4096;
							sprite[j].tag = 31;  //Time left for explosion to stay
							break;
						case DOOMGUY:
							j = connecthead;
							while (j != -1)
							{
								if (playersprite[j] == hitsprite)
								{
									wsay("ouch.wav",4096L,sndvol(hitx,hity));
									changehealth(j,-10);
									break;
								}
								j = connectpoint2[j];
							}
							break;
					}

				j = insertsprite(hitsect,3);
				sprite[j].x = hitx;
				sprite[j].y = hity;
				sprite[j].z = hitz+(8<<8);
				sprite[j].cstat = 2;        //Hitscan does not hit smoke on wall
				sprite[j].picnum = EXPLOSION;
				sprite[j].shade = -4;
				sprite[j].xrepeat = 16;    // 25% normal size
				sprite[j].yrepeat = 16;    // 25% normal size
				sprite[j].ang = daang;
				sprite[j].xvel = 0;
				sprite[j].yvel = 0;
				sprite[j].zvel = 0;
				sprite[j].owner = snum+4096;
				sprite[j].tag = 63;       //Time left for smoke to be alive
					//Sprite starts out with center exactly on wall.
					//This moves it back enough to see it at all angles.
				movesprite(j,-(sintable[(512+daang)&2047]>>7),-(sintable[daang]>>7),0L,128L,4<<8,4<<8,1,4L);
			}
			break;
		case 2:    //Shoot bomb
			wsay("shoot3.wav",4096L,sndvol(x,y));
			j = insertsprite(dasectnum,6);
			sprite[j].x = x;
			sprite[j].y = y;
			sprite[j].z = z+(8<<8);
			sprite[j].cstat = 0;        //Hitscan does not hit other bullets
			sprite[j].picnum = BOMB;
			sprite[j].shade = 0;
			sprite[j].xrepeat = 16;
			sprite[j].yrepeat = 16;
			sprite[j].ang = daang;
			sprite[j].xvel = (sintable[(daang+2560)&2047]>>6);
			sprite[j].yvel = (sintable[(daang+2048)&2047]>>6);
			sprite[j].zvel = ((80-dahoriz)*64);
			sprite[j].owner = snum+4096;
			sprite[j].tag = 0L;
			break;
	}
}

analyzesprites(short daang)
{
	long i, j, k, m, scrx, scry, lockx, locky, dist, mindist;
	short spritetarget;

		//Between drawrooms() and drawmasks()
		//is the perfect time to animate sprites / lock on a target

	for(i=0;i<spritesortcnt;i++)
	{
		j = thesprite[i];
		switch(sprite[j].picnum)
		{
			case DOOMGUY:
					//Get which of the 8 angles of the sprite to draw (0-7)
					//k ranges from 0-7
				k = ((((sprite[j].ang+3072+128-daang)&2047)>>8)&7);
					//This guy has only 5 pictures for 8 angles (3 are x-flipped)
				if (k <= 4)
				{
					spritepicnum[i] += (k<<2);
					sprite[j].cstat &= (255-4);   //clear x-flipping bit
				}
				else
				{
					spritepicnum[i] += ((8-k)<<2);
					sprite[j].cstat |= 4;         //set x-flipping bit
				}

				if (playersprite[screenpeek] == j)
					thesprite[i] = -1;

				m = 0;
				k = connecthead;
				while (k != -1)
				{
					if (playersprite[k] == j)
						spritepal[i] = m;
					m++;
					k = connectpoint2[k];
				}

				break;
		}
	}

	//if (keystatus[0x3a] > 0)  // (Caps Lock) - Hold to lock on target
	//{
	//   mindist = 0x7fffffff;
	//   for(i=0;i<spritesortcnt;i++)
	//      if ((spritepicnum[i] == AL) || (spritepicnum[i] == BROWNMONSTER) || (spritepicnum[i] == EVILAL) || (spritepicnum[i] == DOOMGUY))
	//      {
	//         getspritescreencoord((short)i,&scrx,&scry);
	//         dist = labs(scrx-160) + labs(scry-100);
	//         if (dist < mindist)
	//         {
	//            lockx = scrx;
	//            locky = scry;
	//            mindist = dist;
	//         }
	//      }
	//   if (mindist < 0x7fffffff)
	//   {
	//      ang[?] += ((lockx-160)>>3);
	//      horiz[0] -= ((locky-100)>>3);
	//   }
	//}
}

tagcode()
{
	short startwall, endwall, dasector, p, oldang;
	long i, nexti, j, k, s, dax, day, daz, dax2, day2, cnt, good;

	p = connecthead;
	while (p != -1)
	{
		if ((sector[cursectnum[p]].tag&65535) == 1)
		{
			for(i=0;i<numsectors;i++)
				if ((sector[i].tag|65535) == (sector[cursectnum[p]].tag|65535))
					if ((sector[i].tag&65535) != 1)
						operatesector(i);
			i = headspritestat[0];
			while (i != -1)
			{
				nexti = nextspritestat[i];
				if ((sprite[i].tag|65535) == (sector[cursectnum[p]].tag|65535))
					operatesprite(i);
				i = nexti;
			}

			sector[cursectnum[p]].tag = 0;
		}
		if (((sector[cursectnum[p]].tag&65535) == 2) && (cursectnum[p] != ocursectnum[p]))
		{
			for(i=0;i<numsectors;i++)
				if ((sector[i].tag|65535) == (sector[cursectnum[p]].tag|65535))
					if ((sector[i].tag&65535) != 1)
						operatesector(i);
			i = headspritestat[0];
			while (i != -1)
			{
				nexti = nextspritestat[i];
				if ((sprite[i].tag|65535) == (sector[cursectnum[p]].tag|65535))
					operatesprite(i);
				i = nexti;
			}
		}

		ocursectnum[p] = cursectnum[p];

		p = connectpoint2[p];
	}

	for(i=0;i<warpsectorcnt;i++)
	{
		dasector = warpsectorlist[i];
		j = ((lockclock&127)>>2);
		if (j >= 16) j = 31-j;
		{
			sector[dasector].ceilingshade = j;
			sector[dasector].floorshade = j;
			startwall = sector[dasector].wallptr;
			endwall = startwall+sector[dasector].wallnum-1;
			for(s=startwall;s<=endwall;s++)
				wall[s].shade = j;
		}
	}

	p = connecthead;
	while (p != -1)
	{
		if (sector[cursectnum[p]].tag == 10)  //warp sector
		{
			if (warpsectorflag[p] == 0)
			{
				wsay("warp.wav",4096L,sndvol(posx[p],posy[p]));

				for(i=0;i<warpsectorcnt;i++)
				{
					if (warpsectorlist[i] == cursectnum[p])
					{
						if (i == warpsectorcnt-1)
							cursectnum[p] = warpsectorlist[0];
						else
							cursectnum[p] = warpsectorlist[i+1];
						break;
					}
				}
					//Find center of sector
				startwall = sector[cursectnum[p]].wallptr;
				endwall = startwall+sector[cursectnum[p]].wallnum-1;
				dax = 0L, day = 0L;
				for(s=startwall;s<=endwall;s++)
				{
					dax += wall[s].x, day += wall[s].y;
					if (wall[s].nextsector1 >= 0)
						i = s;
				}
				posx[p] = dax / (endwall-startwall+1);
				posy[p] = day / (endwall-startwall+1);
				posz[p] = sector[cursectnum[p]].floorz-(32<<8);
				updatesector(posx[p],posy[p],&cursectnum[p]);
				dax = ((wall[i].x+wall[wall[i].point2].x)>>1);
				day = ((wall[i].y+wall[wall[i].point2].y)>>1);
				ang[p] = getangle(dax-posx[p],day-posy[p]);

					//Update sprite representation of player
				setsprite(playersprite[p],posx[p],posy[p],posz[p]+(32<<8));
				sprite[playersprite[p]].ang = ang[p];

				warpsectorflag[p] = 1;
			}
		}
		else
			warpsectorflag[p] = 0;

		p = connectpoint2[p];
	}

	for(i=0;i<xpanningsectorcnt;i++)   //animate wall x-panning sectors
	{
		dasector = xpanningsectorlist[i];

		startwall = sector[dasector].wallptr;
		endwall = startwall+sector[dasector].wallnum-1;
		for(s=startwall;s<=endwall;s++)
			wall[s].xpanning = ((lockclock>>2)&255);
	}

	for(i=0;i<ypanningwallcnt;i++)
		wall[ypanningwalllist[i]].ypanning = ~(lockclock&255);

	for(i=0;i<floorpanningcnt;i++)   //animate floor of slime sectors
	{
		sector[floorpanninglist[i]].floorxpanning = ((lockclock>>2)&255);
		sector[floorpanninglist[i]].floorypanning = ((lockclock>>2)&255);
	}

	for(i=0;i<dragsectorcnt;i++)
	{
		dasector = dragsectorlist[i];

		startwall = sector[dasector].wallptr;
		endwall = startwall+sector[dasector].wallnum-1;

		if (wall[startwall].x+dragxdir[i] < dragx1[i]) dragxdir[i] = 16;
		if (wall[startwall].y+dragydir[i] < dragy1[i]) dragydir[i] = 16;
		if (wall[startwall].x+dragxdir[i] > dragx2[i]) dragxdir[i] = -16;
		if (wall[startwall].y+dragydir[i] > dragy2[i]) dragydir[i] = -16;

		for(j=startwall;j<=endwall;j++)
			dragpoint(j,wall[j].x+dragxdir[i],wall[j].y+dragydir[i]);
		j = sector[dasector].floorz;
		sector[dasector].floorz = dragfloorz[i]+(sintable[(lockclock<<4)&2047]>>3);

		p = connecthead;
		while (p != -1)
		{
			if (cursectnum[p] == dasector)
			{
				posx[p] += dragxdir[i];
				posy[p] += dragydir[i];
				posz[p] += (sector[dasector].floorz-j);

					//Update sprite representation of player
				setsprite(playersprite[p],posx[p],posy[p],posz[p]+(32<<8));
				sprite[playersprite[p]].ang = ang[p];
			}

			p = connectpoint2[p];
		}
	}

	for(i=0;i<swingcnt;i++)
	{
		if (swinganginc[i] != 0)
		{
			oldang = swingang[i];
			for(j=0;j<(((long)synctics)<<2);j++)
			{
				swingang[i] = ((swingang[i]+2048+swinganginc[i])&2047);
				if (swingang[i] == swingangclosed[i])
				{
					wsay("closdoor.wav",4096L,sndvol(swingx[i][0],swingy[i][0]));
					swinganginc[i] = 0;
				}
				if (swingang[i] == swingangopen[i]) swinganginc[i] = 0;
			}
			for(k=1;k<=3;k++)
				rotatepoint(swingx[i][0],swingy[i][0],swingx[i][k],swingy[i][k],swingang[i],&wall[swingwall[i][k]].x,&wall[swingwall[i][k]].y);

			if (swinganginc[i] != 0)
			{
				p = connecthead;
				while (p != -1)
				{
					if (cursectnum[p] == swingsector[i])
					{
						good = 1;
						for(k=1;k<=3;k++)
						{
							dax = wall[swingwall[i][k]].x-posx[p]+64;
							day = wall[swingwall[i][k]].y-posy[p]+64;
							dax2 = wall[wall[swingwall[i][k]].point2].x-posx[p]+64;
							day2 = wall[wall[swingwall[i][k]].point2].y-posy[p]+64;
							if (clipinsidebox(dax,day,dax2,day2,128L) == 1)
							{
								good = 0;
								break;
							}
						}
						if (good == 0)
						{
							swingang[i] = oldang;
							for(k=1;k<=3;k++)
								rotatepoint(swingx[i][0],swingy[i][0],swingx[i][k],swingy[i][k],swingang[i],&wall[swingwall[i][k]].x,&wall[swingwall[i][k]].y);
							swinganginc[i] = -swinganginc[i];
							break;
						}
					}

					p = connectpoint2[p];
				}
			}

		}
	}

	for(i=0;i<revolvecnt;i++)
	{
		startwall = sector[revolvesector[i]].wallptr;
		endwall = startwall + sector[revolvesector[i]].wallnum - 1;

		revolveang[i] = ((revolveang[i]+2048-(((long)synctics)<<2))&2047);
		for(k=startwall;k<=endwall;k++)
		{
			rotatepoint(revolvepivotx[i],revolvepivoty[i],revolvex[i][k-startwall],revolvey[i][k-startwall],revolveang[i],&dax,&day);
			dragpoint(k,dax,day);
		}

		if (revolveclip[i] == 0)
		{
			p = connecthead;
			while (p != -1)
			{
				if (cursectnum[p] == revolvesector[i])
				{
					if (revolvesyncstat[p] != 0)
					{
						revolvesyncang[p] = ang[p];
						revolvesyncrotang[p] = 0;
						revolvesyncx[p] = posx[p];
						revolvesyncy[p] = posy[p];
					}
					else
					{
						revolvesyncang[p] = ((ang[p]+2048-revolvesyncrotang[p])&2047);
					}
					revolvesyncrotang[p] = ((revolvesyncrotang[p]+2048-(((long)synctics)<<2))&2047);

					rotatepoint(revolvepivotx[i],revolvepivoty[i],revolvesyncx[p],revolvesyncy[p],revolvesyncrotang[p],&posx[p],&posy[p]);
					ang[p] = ((revolvesyncang[p]+revolvesyncrotang[p])&2047);
				}

				p = connectpoint2[p];
			}
		}
		else
		{
			p = connecthead;
			while (p != -1)
			{
				if ((cursectnum[p] == revolvesector[i]) || (cursectnum[p] == wall[startwall].nextsector1))
				{
					cnt = 256;
					do
					{
						good = 1;
						for(k=startwall;k<=endwall;k++)
						{
							dax = wall[k].x-posx[p]+64;
							day = wall[k].y-posy[p]+64;
							dax2 = wall[wall[k].point2].x-posx[p]+64;
							day2 = wall[wall[k].point2].y-posy[p]+64;
							if (clipinsidebox(dax,day,dax2,day2,128L) == 1)
							{
								revolveang[i] = ((revolveang[i]+1)&2047);
								for(k=startwall;k<=endwall;k++)
								{
									rotatepoint(revolvepivotx[i],revolvepivoty[i],revolvex[i][k-startwall],revolvey[i][k-startwall],revolveang[i],&dax,&day);
									dragpoint(k,dax,day);
								}
								good = 0;
								cnt--;
								break;
							}
						}
					} while ((good == 0) && (cnt > 0));
				}

				p = connectpoint2[p];
			}
		}
	}

	for(i=0;i<subwaytrackcnt;i++)
	{
		dasector = subwaytracksector[i][0];
		startwall = sector[dasector].wallptr;
		endwall = startwall+sector[dasector].wallnum-1;
		for(k=startwall;k<=endwall;k++)
		{
			if (wall[k].x > subwaytrackx1[i])
				if (wall[k].y > subwaytracky1[i])
					if (wall[k].x < subwaytrackx2[i])
						if (wall[k].y < subwaytracky2[i])
							wall[k].x += (subwayvel[i]&0xfffffffc);
		}

		for(j=1;j<subwaynumsectors[i];j++)
		{
			dasector = subwaytracksector[i][j];

			startwall = sector[dasector].wallptr;
			endwall = startwall+sector[dasector].wallnum-1;
			for(k=startwall;k<=endwall;k++)
				wall[k].x += (subwayvel[i]&0xfffffffc);

			s = headspritesect[dasector];
			while (s != -1)
			{
				k = nextspritesect[s];
				sprite[s].x += (subwayvel[i]&0xfffffffc);
				s = k;
			}
		}

		p = connecthead;
		while (p != -1)
		{
			if (cursectnum[p] != subwaytracksector[i][0])
				if (sector[cursectnum[p]].floorz != sector[subwaytracksector[i][0]].floorz)
					if (posx[p] > subwaytrackx1[i])
						if (posy[p] > subwaytracky1[i])
							if (posx[p] < subwaytrackx2[i])
								if (posy[p] < subwaytracky2[i])
								{
									posx[p] += (subwayvel[i]&0xfffffffc);

										//Update sprite representation of player
									setsprite(playersprite[p],posx[p],posy[p],posz[p]+(32<<8));
									sprite[playersprite[p]].ang = ang[p];
								}

			p = connectpoint2[p];
		}

		subwayx[i] += (subwayvel[i]&0xfffffffc);

		k = subwaystop[i][subwaygoalstop[i]] - subwayx[i];
		if (k > 0)
		{
			if (k > 2048)
			{
				if (subwayvel[i] < 128) subwayvel[i]++;
			}
			else
				subwayvel[i] = (k>>4)+1;
		}
		else if (k < 0)
		{
			if (k < -2048)
			{
				if (subwayvel[i] > -128) subwayvel[i]--;
			}
			else
				subwayvel[i] = ((k>>4)-1);
		}

		if (((subwayvel[i]>>2) == 0) && (labs(k) < 2048))
		{
			if (subwaypausetime[i] == 720)
			{
				for(j=1;j<subwaynumsectors[i];j++)   //Open all subway doors
				{
					dasector = subwaytracksector[i][j];
					if (sector[dasector].tag == 17)
					{
						sector[dasector].tag = 16;
						operatesector(dasector);
						sector[dasector].tag = 17;
					}
				}
			}
			if ((subwaypausetime[i] >= 120) && (subwaypausetime[i]-((long)synctics) < 120))
			{
				for(j=1;j<subwaynumsectors[i];j++)   //Close all subway doors
				{
					dasector = subwaytracksector[i][j];
					if (sector[dasector].tag == 17)
					{
						sector[dasector].tag = 16;
						operatesector(dasector);
						sector[dasector].tag = 17;
					}
				}
			}

			subwaypausetime[i] -= ((long)synctics);
			if (subwaypausetime[i] < 0)
			{
				subwaypausetime[i] = 720;
				if (subwayvel[i] < 0)
				{
					subwaygoalstop[i]--;
					if (subwaygoalstop[i] < 0)
						subwaygoalstop[i] = 1;
				}
				if (subwayvel[i] > 0)
				{
					subwaygoalstop[i]++;
					if (subwaygoalstop[i] >= subwaystopcnt[i])
						subwaygoalstop[i] = subwaystopcnt[i]-2;
				}
			}
		}
	}
}

statuslistcode()
{
	short p, target, hitobject, daang;
	long i, nexti, j, nextj, k, l, dax, day, daz, dist, mindist;

	i = headspritestat[1];     //Go through active monster sprites
	while (i != -1)
	{
		nexti = nextspritestat[i];

			//Choose a target player
		mindist = 0x7fffffff; target = connecthead;
		p = connecthead;
		while (p != -1)
		{
			dist = labs(sprite[i].x-posx[p])+labs(sprite[i].y-posy[p]);
			if (dist < mindist) mindist = dist, target = p;

			p = connectpoint2[p];
		}

		switch(sprite[i].picnum)
		{
			case BROWNMONSTER:
				if ((krand()&63) == 0)    //brown monster decides to shoot bullet
				{
					if (cansee(posx[target],posy[target],posz[target],cursectnum[target],sprite[i].x,sprite[i].y,sprite[i].z-(tilesizy[sprite[i].picnum]<<7),sprite[i].sectnum) == 1)
					{
						wsay("zipguns.wav",5144L,sndvol(sprite[i].x,sprite[i].y));
						j = insertsprite(sprite[i].sectnum,6);
						sprite[j].x = sprite[i].x;
						sprite[j].y = sprite[i].y;
						sprite[j].z = sector[sprite[i].sectnum].floorz-(24<<8);
						sprite[j].cstat = 0;        //Hitscan does not hit other bullets
						sprite[j].picnum = BULLET;
						sprite[j].shade = 0;
						sprite[j].xrepeat = 64;
						sprite[j].yrepeat = 64;
						sprite[j].ang = (((getangle(posx[target]-sprite[j].x,posy[target]-sprite[j].y)+(krand()&15)-8)+2048)&2047);
						sprite[j].xvel = (sintable[(sprite[j].ang+2560)&2047]>>6);
						sprite[j].yvel = (sintable[(sprite[j].ang+2048)&2047]>>6);
						sprite[j].zvel = ((posz[target]+(8<<8)-sprite[j].z)<<8)/ksqrt((posx[target]-sprite[j].x)*(posx[target]-sprite[j].x)+(posy[target]-sprite[j].y)*(posy[target]-sprite[j].y));
						sprite[j].owner = i;
						sprite[j].tag = 0L;
					}
				}

				sprite[i].z = sector[sprite[i].sectnum].floorz;

				//Move brown monster
				dax = (sintable[(sprite[i].ang+512)&2047]>>8);
				day = (sintable[sprite[i].ang]>>8);
				if (movesprite(i,dax,day,0L,128L,4<<8,4<<8,2,((long)synctics)) != 0)
				{
					daang = (getangle(posx[target]-sprite[i].x,posy[target]-sprite[i].y)&2047);
					if (((sprite[i].ang+2048-daang)&2047) < 1024)
						sprite[i].ang = ((sprite[i].ang+2048-(synctics<<1))&2047);
					else
						sprite[i].ang = ((sprite[i].ang+(synctics<<1))&2047);
				}

				break;

			case EVILAL:
				if (sprite[i].yrepeat >= 38)
				{
					if (sprite[i].yrepeat < 64)
					{
						sprite[i].xrepeat++;
						sprite[i].yrepeat++;
					}
					else
					{
						if ((krand()&63) == 0)
						{                   //Al decides to reproduce
							j = insertsprite(sprite[i].sectnum,1);
							sprite[j].x = sprite[i].x;
							sprite[j].y = sprite[i].y;
							sprite[j].z = 0;
							sprite[j].cstat = 3;     //Hitscan should hit new monsters / Make him transluscent
							sprite[j].picnum = EVILAL;
							sprite[j].shade = 0;
							sprite[j].xrepeat = 38;
							sprite[j].yrepeat = 38;
							sprite[j].ang = (krand()&2047);
							sprite[j].xvel = 0;
							sprite[j].yvel = 0;
							sprite[j].zvel = 0;
							sprite[j].owner = i;
							sprite[j].tag = 0L;
						}
						if ((krand()&63) == 0)     //Al decides to shoot bullet
						{
							if (cansee(posx[target],posy[target],posz[target],cursectnum[target],sprite[i].x,sprite[i].y,sprite[i].z-(tilesizy[sprite[i].picnum]<<7),sprite[i].sectnum) == 1)
							{
								wsay("zipguns.wav",5144L,sndvol(sprite[i].x,sprite[i].y));
								j = insertsprite(sprite[i].sectnum,6);
								sprite[j].x = sprite[i].x;
								sprite[j].y = sprite[i].y;
								sprite[j].z = sector[sprite[i].sectnum].floorz-(24<<8);
								sprite[j].cstat = 0;            //Hitscan does not hit other bullets
								sprite[j].picnum = BULLET;
								sprite[j].shade = 0;
								sprite[j].xrepeat = 64;
								sprite[j].yrepeat = 64;
								sprite[j].ang = (((getangle(posx[target]-sprite[j].x,posy[target]-sprite[j].y)+(krand()&15)-8)+2048)&2047);
								sprite[j].xvel = (sintable[(sprite[j].ang+2560)&2047]>>6);
								sprite[j].yvel = (sintable[(sprite[j].ang+2048)&2047]>>6);
								sprite[j].zvel = ((posz[target]+(8<<8)-sprite[j].z)<<8)/ksqrt((posx[target]-sprite[j].x)*(posx[target]-sprite[j].x)+(posy[target]-sprite[j].y)*(posy[target]-sprite[j].y));
								sprite[j].owner = i;
								sprite[j].tag = 0L;
							}
						}

						sprite[i].z = sector[sprite[i].sectnum].floorz;

						//Move Al
						dax = (sintable[(sprite[i].ang+512)&2047]>>7);
						day = (sintable[sprite[i].ang]>>7);
						if (movesprite(i,dax,day,0L,128L,4<<8,4<<8,2,((long)synctics)) != 0)
						{
							if ((krand()&1) == 0)
								sprite[i].ang = getangle(posx[target]-sprite[i].x,posy[target]-sprite[i].y);
							else
								sprite[i].ang = (krand()&2047);
						}
					}
				}
				break;
		}

		i = nexti;
	}

	i = headspritestat[6];     //Go through travelling bullet sprites
	while (i != -1)
	{
		nexti = nextspritestat[i];

			 //If the sprite is a bullet then...
		if ((sprite[i].picnum == BULLET) || (sprite[i].picnum == BOMB))
		{
			if (sprite[i].picnum == BOMB)
			{
				sprite[i].z += sprite[i].zvel;
				sprite[i].zvel += (synctics<<5);
				if (sprite[i].z < sector[sprite[i].sectnum].ceilingz+(4<<8))
				{
					sprite[i].z = sector[sprite[i].sectnum].ceilingz+(4<<8);
					sprite[i].zvel = -(sprite[i].zvel>>1);
				}
				if (sprite[i].z > sector[sprite[i].sectnum].floorz-(4<<8))
				{
					sprite[i].z = sector[sprite[i].sectnum].floorz-(4<<8);
					sprite[i].zvel = -(sprite[i].zvel>>1);
				}

				dax = sprite[i].xvel;
				day = sprite[i].yvel;
				daz = 0;
			}
			else
			{
				dax = sprite[i].xvel;
				day = sprite[i].yvel;
				daz = sprite[i].zvel;
			}

			if ((hitobject = movesprite(i,dax,day,daz,64L,4<<8,4<<8,1,((long)synctics))) != 0)
			{
				if (sprite[i].picnum == BOMB)
				{
					j = insertsprite(sprite[i].sectnum,5);
					sprite[j].x = sprite[i].x;
					sprite[j].y = sprite[i].y;
					sprite[j].z = sprite[i].z+(32<<8);
					sprite[j].cstat = 0;        //Hitscan does not hit explosion
					sprite[j].picnum = EXPLOSION;
					sprite[j].shade = -4;
					sprite[j].xrepeat = 64;
					sprite[j].yrepeat = 64;
					sprite[j].ang = sprite[i].ang;
					sprite[j].xvel = 0;
					sprite[j].yvel = 0;
					sprite[j].zvel = 0;
					sprite[j].owner = sprite[i].owner;
					sprite[j].tag = 31;  //Time left for explosion to stay

					for(k=0;k<16;k++)
					{
						j = insertsprite(sprite[i].sectnum,7);
						sprite[j].x = sprite[i].x;
						sprite[j].y = sprite[i].y;
						sprite[j].z = sprite[i].z+(8<<8);
						sprite[j].cstat = 2;        //Hitscan does not hit smoke on wall
						sprite[j].picnum = EXPLOSION;
						sprite[j].shade = -4;
						sprite[j].xrepeat = 24;    // 25% normal size
						sprite[j].yrepeat = 24;    // 25% normal size
						sprite[j].ang = sprite[i].ang;
						sprite[j].xvel = (krand()&511)-256;
						sprite[j].yvel = (krand()&511)-256;
						sprite[j].zvel = (krand()&16384)-8192;
						sprite[j].owner = sprite[i].owner;
						sprite[j].tag = 96;       //Time left for smoke to be alive
					}

					j = connecthead;
					while (j != -1)
					{
						dist = (posx[j]-sprite[i].x)*(posx[j]-sprite[i].x);
						dist += (posy[j]-sprite[i].y)*(posy[j]-sprite[i].y);
						dist += ((posz[j]-sprite[i].z)>>4)*((posz[j]-sprite[i].z)>>4);
						if (dist < 4194304)
							if (cansee(posx[j],posy[j],posz[j],cursectnum[j],sprite[i].x,sprite[i].y,sprite[i].z-(tilesizy[sprite[i].picnum]<<7),sprite[i].sectnum) == 1)
							{
								k = ((32768/((dist>>16)+4))>>5);
								if (j == myconnectindex)
								{
									daang = getangle(posx[j]-sprite[i].x,posy[j]-sprite[i].y);
									dax = ((k*sintable[(daang+2560)&2047])>>14);
									day = ((k*sintable[(daang+2048)&2047])>>14);
									vel += ((dax*sintable[(ang[j]+2560)&2047]+day*sintable[(ang[j]+2048)&2047])>>14);
									svel += ((day*sintable[(ang[j]+2560)&2047]-dax*sintable[(ang[j]+2048)&2047])>>14);
								}
								if (changehealth(j,-k) == 1)  //If you're dead
									if (sprite[i].owner == j+4096)
										saywatchit[j] = lockclock+120;
							}

						j = connectpoint2[j];
					}

					for(k=1;k<=2;k++)         //Check for hurting monsters
					{
						j = headspritestat[k];
						while (j != -1)
						{
							nextj = nextspritestat[j];

							dist = (sprite[j].x-sprite[i].x)*(sprite[j].x-sprite[i].x);
							dist += (sprite[j].y-sprite[i].y)*(sprite[j].y-sprite[i].y);
							dist += ((sprite[j].z-sprite[i].z)>>4)*((sprite[j].z-sprite[i].z)>>4);
							if (dist < 4194304)
								if (cansee(sprite[j].x,sprite[j].y,sprite[j].z-(tilesizy[sprite[j].picnum]<<7),sprite[j].sectnum,sprite[i].x,sprite[i].y,sprite[i].z-(tilesizy[sprite[i].picnum]<<7),sprite[i].sectnum) == 1)
								{
									if (sprite[j].picnum == BROWNMONSTER)
									{
										sprite[j].picnum = GIFTBOX;
										sprite[j].cstat &= ~3;    //Should not clip
										changespritestat(j,0);
									}
									if (sprite[j].picnum == EVILAL)
									{
										sprite[j].picnum = EVILALGRAVE;
										sprite[j].xrepeat = (krand()&63)+16;
										sprite[j].yrepeat = sprite[j].xrepeat;
										changespritestat(j,0);
									}
								}

							j = nextj;
						}
					}

					wsay("blowup.wav",3840L,sndvol(sprite[i].x,sprite[i].y));
					deletesprite((short)i);
					goto bulletisdeletedskip;
				}

				if ((hitobject&0xc000) == 16384)  //Hits a ceiling / floor
				{
					wsay("bullseye.wav",4096L,sndvol(sprite[i].x,sprite[i].y));
					deletesprite((short)i);
					goto bulletisdeletedskip;
				}
				else if ((hitobject&0xc000) == 32768)  //Bullet hit a wall
				{
					if (wall[hitobject&4095].picnum == KENPICTURE)
					{
						if (waloff[4095] >= 0)
							wall[hitobject&4095].picnum = 4095;
						wsay("hello.wav",4096L,sndvol(sprite[i].x,sprite[i].y));   //Ken says, "Hello... how are you today!"
					}
					else
						wsay("bullseye.wav",4096L,sndvol(sprite[i].x,sprite[i].y));

					deletesprite((short)i);
					goto bulletisdeletedskip;
				}
				else if ((hitobject&0xc000) == 49152)  //Bullet hit a sprite
				{
						//Check if bullet hit any players...
					if (sprite[hitobject&4095].picnum == DOOMGUY)
					{
							//Find which player it was!
						j = connecthead;
						while (j != -1)
						{
							if (sprite[i].owner != j+4096)
								if (playersprite[j] == (hitobject&4095))
								{
									wsay("ouch.wav",4096L,sndvol(sprite[i].x,sprite[i].y));
									changehealth(j,-25);
									deletesprite((short)i);
									goto bulletisdeletedskip;
								}

							j = connectpoint2[j];
						}
					}
						//Check if bullet hit any monsters...
					j = (hitobject&4095);     //j is the spritenum that the bullet (spritenum i) hit
					if (sprite[i].owner != j)
					{
						switch(sprite[j].picnum)
						{
							case BROWNMONSTER:
								if (sprite[j].tag > 0)
									sprite[j].tag -= 25;
								if (sprite[j].tag > 0)
								{
									if (sprite[j].tag <= 25)
										sprite[j].cstat |= 2;
									wsay("hurt.wav",4096L+(rand()&511)-256,sndvol(sprite[i].x,sprite[i].y));

									j = connecthead;
									while (j != -1)
									{
										if (sprite[i].owner == j+4096)
											changescore(j,200);
										j = connectpoint2[j];
									}
								}
								else
								{
									wsay("blowup.wav",4096L,sndvol(sprite[i].x,sprite[i].y));
									sprite[j].picnum = GIFTBOX;
									sprite[j].cstat &= ~3;    //Should not clip

									k = insertsprite(sprite[j].sectnum,5);
									sprite[k].x = sprite[j].x;
									sprite[k].y = sprite[j].y;
									sprite[k].z = sprite[j].z;
									sprite[k].cstat = 0;        //Hitscan does not hit explosion
									sprite[k].picnum = EXPLOSION;
									sprite[k].shade = -4;
									sprite[k].xrepeat = 64;
									sprite[k].yrepeat = 64;
									sprite[k].ang = sprite[j].ang;
									sprite[k].xvel = 0;
									sprite[k].yvel = 0;
									sprite[k].zvel = 0;
									sprite[k].owner = j;
									sprite[k].tag = 31;  //Time left for explosion to stay

									changespritestat(j,0);

									j = connecthead;
									while (j != -1)
									{
										if (sprite[i].owner == j+4096)
											changescore(j,200);
										j = connectpoint2[j];
									}
								}
								deletesprite((short)i);
								goto bulletisdeletedskip;
							case EVILAL:
								wsay("blowup.wav",5144L,sndvol(sprite[i].x,sprite[i].y));
								sprite[j].picnum = EVILALGRAVE;
								sprite[j].xrepeat = (krand()&63)+16;
								sprite[j].yrepeat = sprite[j].xrepeat;
								changespritestat(j,0);

								j = connecthead;
								while (j != -1)
								{
									if (sprite[i].owner == j+4096)
										changescore(j,200);
									j = connectpoint2[j];
								}
								deletesprite((short)i);
								goto bulletisdeletedskip;
							case AL:
								wsay("blowup.wav",5144L,sndvol(sprite[i].x,sprite[i].y));
								sprite[j].xrepeat += 2;
								sprite[j].yrepeat += 2;
								if (sprite[j].yrepeat >= 38)
								{
									sprite[j].picnum = EVILAL;
									sprite[j].cstat |= 2;      //Make him transluscent
								}
								changespritestat(j,1);
								deletesprite((short)i);
								goto bulletisdeletedskip;
							default:
								wsay("bullseye.wav",4096L,sndvol(sprite[i].x,sprite[i].y));
								deletesprite((short)i);
								goto bulletisdeletedskip;
						}
					}
				}
			}
		}

bulletisdeletedskip:
		i = nexti;
	}

	j = connecthead;
	while (j != -1)
	{
		if ((saywatchit[j] >= lockclock-synctics) && (saywatchit[j] < lockclock))
			wsay("watchit.wav",4096L,sndvol(posx[j],posy[j]));

		j = connectpoint2[j];
	}

	i = headspritestat[2];     //Go through monster waiting for you list
	while (i != -1)
	{
		nexti = nextspritestat[i];

		p = connecthead;
		while (p != -1)
		{
				//Use dot product to see if monster's angle is towards you
			if (sintable[(sprite[i].ang+2560)&2047]*(posx[p]-sprite[i].x) + sintable[(sprite[i].ang+2048)&2047]*(posy[p]-sprite[i].y) >= 0)
			{
				if (cansee(posx[p],posy[p],posz[p],cursectnum[p],sprite[i].x,sprite[i].y,sprite[i].z-(tilesizy[sprite[i].picnum]<<7),sprite[i].sectnum) == 1)
				{
					changespritestat(i,1);
					wsay("iseeyou.wav",4096L,sndvol(sprite[i].x,sprite[i].y));
				}
			}
			p = connectpoint2[p];
		}

		i = nexti;
	}

	i = headspritestat[3];     //Go through smoke sprites
	while (i >= 0)
	{
		nexti = nextspritestat[i];

		sprite[i].z -= (((long)synctics)<<6);
		sprite[i].tag -= ((long)synctics);
		if (sprite[i].tag < 0)
			deletesprite(i);

		i = nexti;
	}

	i = headspritestat[4];     //Go through splash sprites
	while (i >= 0)
	{
		nexti = nextspritestat[i];

		sprite[i].tag -= ((long)synctics);
		sprite[i].picnum = SPLASH + ((63-sprite[i].tag)>>4);
		if (sprite[i].tag < 0)
			deletesprite(i);

		i = nexti;
	}

	i = headspritestat[5];     //Go through explosion sprites
	while (i >= 0)
	{
		nexti = nextspritestat[i];

		sprite[i].tag -= ((long)synctics);
		if (sprite[i].tag < 0)
			deletesprite(i);

		i = nexti;
	}

	i = headspritestat[7];     //Go through bomb spriral-explosion sprites
	while (i >= 0)
	{
		nexti = nextspritestat[i];

		sprite[i].x += ((sprite[i].xvel*synctics)>>4);
		sprite[i].y += ((sprite[i].yvel*synctics)>>4);
		sprite[i].z += ((sprite[i].zvel*synctics)>>4);

		sprite[i].zvel += (synctics<<7);
		if (sprite[i].z < sector[sprite[i].sectnum].ceilingz+(4<<8))
		{
			sprite[i].z = sector[sprite[i].sectnum].ceilingz+(4<<8);
			sprite[i].zvel = -(sprite[i].zvel>>1);
		}
		if (sprite[i].z > sector[sprite[i].sectnum].floorz-(4<<8))
		{
			sprite[i].z = sector[sprite[i].sectnum].floorz-(4<<8);
			sprite[i].zvel = -(sprite[i].zvel>>1);
		}

		sprite[i].xrepeat = (sprite[i].tag>>2);
		sprite[i].yrepeat = (sprite[i].tag>>2);

		sprite[i].tag -= ((long)synctics);
		if (sprite[i].tag < 0)
			deletesprite(i);

		i = nexti;
	}
}

processinput(short snum)
{
	char *ptr;
	short startwall, endwall;
	long oldposx, oldposy, i, nexti;
	long j, k, doubvel, xvect, yvect, goalz, hiz, loz;
	long dax, day, dax2, day2, odax, oday, odax2, oday2;

		//SHARED KEYS:
		//Movement code
	if ((syncvel[snum]|syncsvel[snum]) != 0)
	{
		doubvel = (((long)synctics)<<((syncbits[snum]&256)>0));

		xvect = 0, yvect = 0;
		if (syncvel[snum] != 0)
		{
			xvect += ((((long)syncvel[snum])*doubvel*(long)sintable[(ang[snum]+2560)&2047])>>3);
			yvect += ((((long)syncvel[snum])*doubvel*(long)sintable[(ang[snum]+2048)&2047])>>3);
		}
		if (syncsvel[snum] != 0)
		{
			xvect += ((((long)syncsvel[snum])*doubvel*(long)sintable[(ang[snum]+2048)&2047])>>3);
			yvect += ((((long)syncsvel[snum])*doubvel*(long)sintable[(ang[snum]+1536)&2047])>>3);
		}
		clipmove(&posx[snum],&posy[snum],&posz[snum],&cursectnum[snum],xvect,yvect,128L,4<<8,4<<8,0);

		revolvesyncstat[snum] = 1;
	}
	else
	{
		revolvesyncstat[snum] = 0;
	}

		//Don't let you fall if you're just slightly over a cliff
			// This function returns the highest and lowest z's
			//    for an entire box, NOT just a point.  -Useful for clipping
	getzrange(posx[snum],posy[snum],posz[snum],cursectnum[snum],&hiz,&loz,128L,0);

	if (syncangvel[snum] != 0)          //ang += angvel * constant
	{                         //ENGINE calculates angvel for you
		doubvel = ((long)synctics);
		if ((syncbits[snum]&256) > 0)  //Lt. shift makes turn velocity 50% faster
			doubvel += (((long)synctics)>>1);
		ang[snum] += ((((long)syncangvel[snum])*doubvel)>>4);
		ang[snum] = (ang[snum]+2048)&2047;
	}

	if (health[snum] < 0)
	{
		health[snum] -= (long)synctics;
		if (health[snum] <= -120)
		{
			hvel[snum] = 0;
			if (snum == myconnectindex)
				vel = 0, svel = 0, angvel = 0, keystatus[3] = 1;

			deaths[snum]++;
			health[snum] = 100;
			numbombs[snum] = -1;

			findrandomspot(&posx[snum],&posy[snum],&cursectnum[snum]);
			posz[snum] = sector[cursectnum[snum]].floorz-(32<<8);
			ang[snum] = (krand()&2047);

			setsprite(playersprite[snum],posx[snum],posy[snum],posz[snum]+(32<<8));
			sprite[playersprite[snum]].picnum = DOOMGUY;
			sprite[playersprite[snum]].ang = ang[snum];
			sprite[playersprite[snum]].xrepeat = 64;
			sprite[playersprite[snum]].yrepeat = 64;

			if ((snum == screenpeek) && (dimensionmode[screenpeek] == 3) && (screensize[screenpeek] <= 320))
			{
				sprintf(&tempbuf,"Deaths: %d",deaths[snum]);
				printext(160-(strlen(tempbuf)<<2),184,tempbuf,ALPHABET,80);
				sprintf(&tempbuf,"Health: %3d",health[snum]);
				printext(160-(strlen(tempbuf)<<2),176,tempbuf,ALPHABET,80);
			}
		}
		else
		{
			sprite[playersprite[snum]].xrepeat = ((128+health[snum])>>1);
			sprite[playersprite[snum]].yrepeat = ((128+health[snum])>>1);

			hvel[snum] += (synctics<<2);
			posz[snum] += hvel[snum];
			if (posz[snum] > loz-(4<<8))
			{
				posz[snum] = loz-(4<<8);
				hvel[snum] = 0;
			}
		}
		return;
	}

	if (((syncbits[snum]&8) > 0) && (horiz[snum] > 100-(YDIM>>1))) horiz[snum] -= 4;     //-
	if (((syncbits[snum]&4) > 0) && (horiz[snum] < 100+(YDIM>>1))) horiz[snum] += 4;   //+

	goalz = loz-(32<<8);         //32 pixels above floor
	if (sector[cursectnum[snum]].tag == 4)   //slime sector
	{
		goalz = loz-(8<<8);
		if (posz[snum] >= goalz-(2<<8))
		{
			clipmove(&posx[snum],&posy[snum],&posz[snum],&cursectnum[snum],-(((long)synctics)<<14),-(((long)synctics)<<14),128L,4<<8,4<<8,0);

			if (slimesoundcnt[snum] >= 0)
			{
				slimesoundcnt[snum] -= synctics;
				while (slimesoundcnt[snum] < 0)
				{
					slimesoundcnt[snum] += 120;
					wsay("slime.wav",4096L,sndvol(posx[snum],posy[snum]));
				}
			}
		}
	}
	if (goalz < hiz+(16<<8))   //ceiling&floor too close
		goalz = ((loz+hiz)>>1);
	//goalz += mousz;
	if ((syncbits[snum]&1) > 0)                         //A (stand high)
	{
		goalz -= (16<<8);
		if ((syncbits[snum]&256) > 0)                    //Either shift key
			goalz -= (24<<8);
	}
	if ((syncbits[snum]&2) > 0)                         //Z (stand low)
	{
		goalz += (12<<8);
		if ((syncbits[snum]&256) > 0)                    //Either shift key
			goalz += (12<<8);
	}
	if ((sector[cursectnum[snum]].floorstat&2) > 0) //Groudraw
	{
		if (waloff[sector[cursectnum[snum]].floorheinum] == -1) loadtile(sector[cursectnum[snum]].floorheinum);
		ptr = (char *)(waloff[sector[cursectnum[snum]].floorheinum]+(((posx[snum]>>4)&63)<<6)+((posy[snum]>>4)&63));
		goalz -= ((*ptr)<<8);
	}

	if (goalz != posz[snum])
	{
		if (posz[snum] < goalz) hvel[snum] += (synctics<<3);
		if (posz[snum] > goalz) hvel[snum] = (((goalz-posz[snum])*synctics)>>5);

		posz[snum] += hvel[snum];
		if (posz[snum] > loz-(4<<8)) posz[snum] = loz-(4<<8), hvel[snum] = 0;
		if (posz[snum] < hiz+(4<<8)) posz[snum] = hiz+(4<<8), hvel[snum] = 0;
	}

	if (dimensionmode[snum] == 2)
	{
		if (((syncbits[snum]&32) > 0) && (zoom[snum] > 48)) zoom[snum] -= (zoom[snum]>>4);
		if (((syncbits[snum]&16) > 0) && (zoom[snum] < 4096)) zoom[snum] += (zoom[snum]>>4);
	}
	else
	{
		if (((syncbits[snum]&32) > 0) && (screensize[snum] > 64))
		{
			screensize[snum] -= 8;
			if (screenpeek == snum)
			{
				dax = 160-(screensize[snum]>>1);
				dax2 = dax+screensize[snum]-1;
				day = 84-(((screensize[snum]*168)/320)>>1);
				day2 = day + ((screensize[snum]*168)/320)-1;
				setview(dax,day,dax2,day2);

				odax = 160-((screensize[snum]+8)>>1);
				odax2 = dax+(screensize[snum]+8)-1;
				oday = 84-((((screensize[snum]+8)*168)/320)>>1);
				oday2 = day + (((screensize[snum]+8)*168)/320)-1;

				permanentwritesprite(0,0,BACKGROUND,0,odax,oday,dax-1,oday2);
				permanentwritesprite(0,0,BACKGROUND,0,dax2+1,oday,odax2,oday2);
				permanentwritesprite(0,0,BACKGROUND,0,dax,oday,dax2,day-1);
				permanentwritesprite(0,0,BACKGROUND,0,dax,day2+1,dax2,oday2);

				if (screensize[screenpeek] == 320)
				{
					permanentwritesprite(0,200-32,STATUSBAR,0,0,0,319,199);  //Draw status bar only once

					sprintf(&tempbuf,"Deaths: %d",deaths[screenpeek]);
					printext(160-(strlen(tempbuf)<<2),184,tempbuf,ALPHABET,80);

					sprintf(&tempbuf,"Health: %3d",health[screenpeek]);
					printext(160-(strlen(tempbuf)<<2),176,tempbuf,ALPHABET,80);

					sprintf(&tempbuf,"Score:%ld",score[screenpeek]);
					printext(160-(strlen(tempbuf)<<2),192,tempbuf,ALPHABET,80);
				}
			}
		}
		if (((syncbits[snum]&16) > 0) && (screensize[snum] <= 320))
		{
			screensize[snum] += 8;
			if (screenpeek == snum)
			{
				if (screensize[snum] > 320)
				{
					dax = 0; day = 0;
					dax2 = 319; day2 = 199;
				}
				else
				{
					dax = 160-(screensize[snum]>>1);
					dax2 = dax+screensize[snum]-1;
					day = 84-(((screensize[snum]*168)/320)>>1);
					day2 = day + ((screensize[snum]*168)/320)-1;
				}
				setview(dax,day,dax2,day2);
			}
		}
	}

		//Update sprite representation of player
		//   -should be after movement, but before shooting code
	setsprite(playersprite[snum],posx[snum],posy[snum],posz[snum]+(32<<8));
	sprite[playersprite[snum]].ang = ang[snum];

	if ((cursectnum[snum] < 0) || (cursectnum[snum] >= numsectors))
	{       //How did you get in the wrong sector?
		wsay("ouch.wav",4096L,sndvol(posx[snum],posy[snum]));
		changehealth(snum,-synctics);
	}
	else if (hiz+(8<<8) > loz)
	{       //Ceiling and floor are smooshing you!
		wsay("ouch.wav",4096L,sndvol(posx[snum],posy[snum]));
		changehealth(snum,-synctics);
	}

	if (waterfountainwall[snum] >= 0)
		if ((wall[neartagwall].tag != 7) || ((syncbits[snum]&1024) == 0))
		{
			if (wall[waterfountainwall[snum]].overpicnum == USEWATERFOUNTAIN)
				wall[waterfountainwall[snum]].overpicnum = WATERFOUNTAIN;
			else if (wall[waterfountainwall[snum]].picnum == USEWATERFOUNTAIN)
				wall[waterfountainwall[snum]].picnum = WATERFOUNTAIN;

			waterfountainwall[snum] = -1;
		}

	if ((syncbits[snum]&1024) > 0)  //Space bar
	{
		neartag(posx[snum],posy[snum],posz[snum],cursectnum[snum],ang[snum],&neartagsector,&neartagwall,&neartagsprite,&neartaghitdist,1024L);  //1024 is range (1024 = 1 largest grid unit)

		if (wall[neartagwall].tag == 7)  //Water fountain
		{
			if (wall[neartagwall].overpicnum == WATERFOUNTAIN)
			{
				wsay("water.wav",4096L,sndvol(posx[snum],posy[snum]));
				wall[neartagwall].overpicnum = USEWATERFOUNTAIN;
				waterfountainwall[snum] = neartagwall;
			}
			else if (wall[neartagwall].picnum == WATERFOUNTAIN)
			{
				wsay("water.wav",4096L,sndvol(posx[snum],posy[snum]));
				wall[neartagwall].picnum = USEWATERFOUNTAIN;
				waterfountainwall[snum] = neartagwall;
			}

			if (waterfountainwall[snum] >= 0)
			{
				waterfountaincnt[snum] -= synctics;
				while (waterfountaincnt[snum] < 0)
				{
					waterfountaincnt[snum] += 120;
					wsay("water.wav",4096L,sndvol(posx[snum],posy[snum]));
					changehealth(snum,2);
				}
			}
		}

		if (spaceflag[snum] == 0)
		{
			if (neartagsector >= 0)
			{
				if ((sector[neartagsector].tag&0xffff0000) == 0)
					operatesector(neartagsector);
				spaceflag[snum] = 1;
			}

			if (neartagwall >= 0)
			{
				if ((wall[neartagwall].tag&65535) == 2)  //Switch
				{
					for(i=0;i<numsectors;i++)
						if ((sector[i].tag|65535) == (wall[neartagwall].tag|65535))
							if ((sector[i].tag&65535) != 1)
								operatesector(i);
					i = headspritestat[0];
					while (i != -1)
					{
						nexti = nextspritestat[i];
						if ((sprite[i].tag|65535) == (wall[neartagwall].tag|65535))
							operatesprite(i);
						i = nexti;
					}

					j = wall[neartagwall].overpicnum;
					if (j == SWITCH1ON)                     //1-time switch
					{
						wall[neartagwall].overpicnum = GIFTBOX;
						wall[neartagwall].tag = 0;
					}
					if (j == GIFTBOX)                       //1-time switch
					{
						wall[neartagwall].overpicnum = SWITCH1ON;
						wall[neartagwall].tag = 0;
					}
					if (j == SWITCH2ON) wall[neartagwall].overpicnum = SWITCH2OFF;
					if (j == SWITCH2OFF) wall[neartagwall].overpicnum = SWITCH2ON;

					spaceflag[snum] = 1;
				}
			}

			if (neartagsprite >= 0)
			{
				if (sprite[neartagsprite].tag == 1)
				{     //if you're shoving innocent little AL around, he gets mad!
					if (sprite[neartagsprite].picnum == AL)
					{
						sprite[neartagsprite].picnum = EVILAL;
						sprite[neartagsprite].cstat |= 2;   //Make him transluscent
						sprite[neartagsprite].xrepeat = 38;
						sprite[neartagsprite].yrepeat = 38;
					}
					changespritestat(neartagsprite,1);
					spaceflag[snum] = 1;
				}
			}
		}
	}
	else
	{
		spaceflag[snum] = 0;
	}

	if ((syncbits[snum]&2048) > 0)      //Shoot a bullet
	{
		switch((syncbits[snum]>>13)&7)
		{
			case 0:
				if (shootflag[snum] == 0)
				{
					shootgun(snum,posx[snum],posy[snum],posz[snum],ang[snum],horiz[snum],cursectnum[snum],0);
					shootflag[snum] = 1;
				}
				break;
			case 1:
				if (lockclock > lastchaingun[snum]+8)
				{
					lastchaingun[snum] = lockclock;
					shootgun(snum,posx[snum],posy[snum],posz[snum],ang[snum],horiz[snum],cursectnum[snum],1);
				}
				break;
			case 2:
				if (shootflag[snum] == 0)
				{
					if (numbombs[snum] > 0)
					{
						shootgun(snum,posx[snum],posy[snum],posz[snum],ang[snum],horiz[snum],cursectnum[snum],2);
						numbombs[snum]--;
					}
					shootflag[snum] = 1;
				}
				break;
		}
	}
	else
	{
		shootflag[snum] = 0;
	}

	if ((syncbits[snum]&4096) > 0)  //Keypad enter
	{
		if (dimflag[snum] == 0)
		{
			if (dimensionmode[snum] == 3)
			{
				dimensionmode[snum] = 2;
				if (screenpeek == snum) qsetmode640480();
			}
			else
			{
				dimensionmode[snum] = 3;
				if (screenpeek == snum) setup3dscreen();
			}
			dimflag[snum] = 1;
		}
	}
	else
	{
		dimflag[snum] = 0;
	}
}

drawscreen(short snum, long smoothratio)
{
	long i, j, charsperline, templong;
	long daposx[MAXPLAYERS], daposy[MAXPLAYERS], daposz[MAXPLAYERS];
	long dahoriz[MAXPLAYERS], dazoom[MAXPLAYERS];
	short daang[MAXPLAYERS], dasectnum;

		//Smoothratio is for frame interpolation
		//interpolates positions between last and 2nd to last frames
	if (smoothratio < 0) smoothratio = 0;
	if (smoothratio > 256) smoothratio = 256;

	i = connecthead;
	while (i != -1)
	{
		daposx[i] = oposx[i] + (((posx[i]-oposx[i])*smoothratio)>>8);
		daposy[i] = oposy[i] + (((posy[i]-oposy[i])*smoothratio)>>8);
		daposz[i] = oposz[i] + (((posz[i]-oposz[i])*smoothratio)>>8);
		dahoriz[i] = ohoriz[i] + (((horiz[i]-ohoriz[i])*smoothratio)>>8);
		daang[i] = oang[i] + (((((ang[i]+1024-oang[i])&2047)-1024)*smoothratio)>>8);
		dazoom[i] = ozoom[i] + (((zoom[i]-ozoom[i])*smoothratio)>>8);
		i = connectpoint2[i];
	}

	//i = connecthead;   //Swap em'
	//while (i != -1)
	//{
	//   templong = sprite[playersprite[i]].x; sprite[playersprite[i]].x = daposx[i]; daposx[i] = templong;
	//   templong = sprite[playersprite[i]].y; sprite[playersprite[i]].y = daposy[i]; daposy[i] = templong;
	//   templong = sprite[playersprite[i]].z; sprite[playersprite[i]].z = daposz[i]; daposz[i] = templong+(32<<8);
	//   templong = sprite[playersprite[i]].ang; sprite[playersprite[i]].ang = daang[i]; daang[i] = templong;
	//   i = connectpoint2[i];
	//}

	if (dimensionmode[snum] == 3)
	{
		dasectnum = cursectnum[snum];
		drawrooms(daposx[snum],daposy[snum],daposz[snum],daang[snum],dahoriz[snum],&dasectnum);
		analyzesprites(daang[snum]);

		drawmasks();

		if (lockclock < 384)
		{
			if (lockclock < 128)
				rotatesprite(160L<<16,100L<<16,lockclock<<9,lockclock<<4,DEMOSIGN);
			else if (lockclock < 256)
				rotatesprite(160L<<16,100L<<16,65536,0,DEMOSIGN);
			else
				rotatesprite(160L<<16,100L<<16,(384-lockclock)<<9,lockclock<<4,DEMOSIGN);
		}

		if (numbombs[screenpeek] > 0)
			overwritesprite(160L,YDIM-32,GUNONBOTTOM,sector[cursectnum[screenpeek]].floorshade,1|2);

		if (health[screenpeek] <= 0)
			rotatesprite(160L<<16,100L<<16,(-health[screenpeek])<<11,(-health[screenpeek])<<5,NO);
	}

	else if (dimensionmode[snum] == 2)
	{
		draw2dscreen(daposx[snum],daposy[snum],daang[snum],dazoom[snum],0);  // 0 is grid type, range:0-5
	}

	if (typemode == 1)
	{
		if (dimensionmode[snum] == 2) charsperline = 80;
		if (dimensionmode[snum] == 3) charsperline = 40;

		for(i=0;i<=typemessageleng;i+=charsperline)
		{
			for(j=0;j<charsperline;j++)
				tempbuf[j] = typemessage[i+j];
			if (typemessageleng < i+charsperline)
			{
				tempbuf[(typemessageleng-i)] = '_';
				tempbuf[(typemessageleng-i)+1] = 0;
			}
			else
				tempbuf[charsperline] = 0;
			if (dimensionmode[snum] == 3)
				printext256(0L,(i/charsperline)<<3,183,-1,tempbuf,0);
			else
				printext16(0L,((i/charsperline)<<3)+(pageoffset/640),10,-1,tempbuf,0);
		}
	}
	if (getmessageleng > 0)
	{
		if (dimensionmode[snum] == 2) charsperline = 80;
		if (dimensionmode[snum] == 3) charsperline = 40;

		for(i=0;i<=getmessageleng;i+=charsperline)
		{
			for(j=0;j<charsperline;j++)
				tempbuf[j] = getmessage[i+j];
			if (getmessageleng < i+charsperline)
				tempbuf[(getmessageleng-i)] = 0;
			else
				tempbuf[charsperline] = 0;

			if (dimensionmode[snum] == 3)
				printext256(0L,((i/charsperline)<<3)+(YDIM-32-8)-(((getmessageleng-1)/charsperline)<<3),151,-1,tempbuf,0);
			else
			{
					//350 lines high for qsetmode350
					//336 lines high for qsetmode480 (+144 lines for status bar)
				printext16(0L,((i/charsperline)<<3)+(480-8)-(((getmessageleng-1)/charsperline)<<3)+(pageoffset/640),4,-1,tempbuf,0);
			}
		}
		if (totalclock > getmessagetimeoff)
			getmessageleng = 0;
	}
	if ((numplayers >= 2) && (screenpeek != myconnectindex))
	{
		if (option[4] >= 5)
			strcpy(tempbuf,username[screenpeek]);
		else
			strcpy(tempbuf,"Other");

		if (dimensionmode[snum] == 3)
			printext256(160-(strlen(tempbuf)<<2),0,24,-1,tempbuf,0);
		else
			printext16(320-(strlen(tempbuf)<<2),(pageoffset/640),7,-1,tempbuf,0);
	}

	nextpage();
}

movethings()
{
	short i, j;

	//printf("%x",((long)synctics));
	//printf("%x",randomseed);
	//i = connecthead;
	//while (i != -1)
	//{
	//   printf("%x ",syncvel[i]);
	//   printf("%x ",syncsvel[i]);
	//   printf("%x ",syncangvel[i]);
	//   printf("%x",syncbits[i]);
	//   i = connectpoint2[i];
	//}
	//printf("\n");

	lastsynctics = synctics;
	i = connecthead;
	while (i != -1)
	{
		oposx[i] = posx[i]; oposy[i] = posy[i]; oposz[i] = posz[i];
		ohoriz[i] = horiz[i]; oang[i] = ang[i]; ozoom[i] = zoom[i];
		i = connectpoint2[i];
	}

	if ((numplayers <= 2) && (recstat == 1))
	{
		i = connecthead; j = 0;
		while (i != -1)
		{
			recsyncvel[reccnt][j] = syncvel[i];
			recsyncsvel[reccnt][j] = syncsvel[i];
			recsyncangvel[reccnt][j] = syncangvel[i];
			recsyncbits[reccnt][j] = syncbits[i];

			j++;
			i = connectpoint2[i];
		}
		recsynctics[reccnt] = synctics;
		reccnt++; if (reccnt > 16383) reccnt = 16383;
	}

	lockclock += ((long)synctics);

	i = connecthead;
	while (i != -1)
	{
		processinput(i);       //Move player
		checktouchsprite(i);   //Pick up coins
		i = connectpoint2[i];
	}

	doanimations(((long)synctics));
	tagcode();            //Door code, moving sector code, other stuff
	statuslistcode();     //Monster / bullet code / explosions

	synctics = 0;
}

getinput(short snum)
{
	static struct Spw_IntPacket packet, oldpacket;
	long i, j, k;
	short mousx, mousy, bstatus;

	if (typemode == 0)           //if normal game keys active
	{
		if ((keystatus[0x2a]&keystatus[0x36]&keystatus[0x13]) > 0)   //Sh.Sh.R (replay)
		{
			keystatus[0x13] = 0;
			playback();
		}

		if (keystatus[keys[13]] > 0)
		{
			keystatus[keys[13]] = 0;

			screenpeek = connectpoint2[screenpeek];
			if (screenpeek == -1) screenpeek = connecthead;

			if (dimensionmode[screenpeek] == 2) qsetmode640480();
			if (dimensionmode[screenpeek] == 3) setup3dscreen();
		}

		if (keystatus[0x2] > 0) selectedgun[snum] = 0, keystatus[0x2] = 0;
		if (keystatus[0x3] > 0) selectedgun[snum] = 1, keystatus[0x3] = 0;
		if (keystatus[0x4] > 0) selectedgun[snum] = 2, keystatus[0x4] = 0;
		if (keystatus[0x5] > 0) selectedgun[snum] = 3, keystatus[0x5] = 0;
		if (keystatus[0x6] > 0) selectedgun[snum] = 4, keystatus[0x6] = 0;
		if (keystatus[0x7] > 0) selectedgun[snum] = 5, keystatus[0x7] = 0;
		if (keystatus[0x8] > 0) selectedgun[snum] = 6, keystatus[0x8] = 0;
		if (keystatus[0x9] > 0) selectedgun[snum] = 7, keystatus[0x9] = 0;
	}

	syncvel[snum] = vel;
	syncsvel[snum] = svel;
	syncangvel[snum] = angvel;

	switch(moustat)
	{
		case 0:
			bstatus = 0;
			break;
		case 1:
			getmousevalues(&mousx,&mousy,&bstatus);
			i = syncangvel[snum];
			i += (mousx<<3);
			if (i < -128) i = -128;
			if (i > 127) i = 127;
			syncangvel[snum] = i;

			i = syncvel[snum];
			i -= (mousy<<3);
			if (i < -128) i = -128;
			if (i > 127) i = 127;
			syncvel[snum] = i;
			break;
		case 2:
			SPW_IntCommand(SPW_INTCMD_GFORCE, &packet);
			if (packet.period == 0)
				memcpy(&packet,&oldpacket,sizeof(packet));
			else
				memcpy(&oldpacket,&packet,sizeof(packet));

			//mousz = 0;
			bstatus = packet.button;

			if (labs((long)packet.rx) > 4096)
			{
				//horiz[snum] = (((horiz[snum]*7)+(100+(packet.rx>>8)))>>3);

				i = 0;
				j = labs((long)packet.ry);
				if (labs((long)packet.tz) > labs((long)packet.ry))
				{
					i = 1;
					j = labs((long)packet.tz);
				}
				if (labs((long)packet.tx) > j) i = 2;

				switch(i)
				{
					case 0: syncangvel[snum] = -(packet.ry>>6); break;
					case 1: syncvel[snum] = -(packet.tz>>4); break;
					case 2: syncsvel[snum] = -(packet.tx>>4); break;
				}

				//if (labs((long)packet.ty) > 512) mousz = -packet.ty;
			}
			else
			{
				//horiz[snum] = ((horiz[snum]+(horiz[snum]<<1)+100)>>2);
				syncangvel[snum] = -(packet.ry>>6);
				syncvel[snum] = -(packet.tz>>4);
				syncsvel[snum] = -(packet.tx>>4);
				//if (labs((long)packet.ty) > 512) mousz = -packet.ty;
			}

			//if (syncvel[snum] < -128) syncvel[snum] = -128;
			//if (syncvel[snum] > 127) syncvel[snum] = 127;
			//if (syncsvel[snum] < -128) syncsvel[snum] = -128;
			//if (syncsvel[snum] > 127) syncsvel[snum] = 127;
			//if (syncangvel[snum] < -128) syncangvel[snum] = -128;
			//if (syncangvel[snum] > 127) syncangvel[snum] = 127;
			break;
	}

	syncbits[snum] = (selectedgun[snum]<<13);
	if (typemode == 0)           //if normal game keys active
	{
		syncbits[snum] += keystatus[keys[8]];
		syncbits[snum] += (keystatus[keys[9]]<<1);
		syncbits[snum] += (keystatus[keys[10]]<<2);
		syncbits[snum] += (keystatus[keys[11]]<<3);
		syncbits[snum] += (keystatus[keys[14]]<<4);
		syncbits[snum] += (keystatus[keys[15]]<<5);
		syncbits[snum] += (keystatus[keys[4]]<<8);
		syncbits[snum] += (((keystatus[keys[7]]==1)|((bstatus&6)>(oldmousebstatus&6)))<<10);
		syncbits[snum] += (((keystatus[keys[6]]==1)|((bstatus&1)>(oldmousebstatus&1)))<<11);
		syncbits[snum] += ((keystatus[keys[12]]==1)<<12);

		oldmousebstatus = bstatus;
		if (((syncbits[snum]&2048) > 0) && (selectedgun[snum] == 1))
			oldmousebstatus &= ~1;     //Allow continous fire with mouse for chain gun
	}

	if (snum < MAXPLAYERS)
	{
		i = (totalclock-lockclock);    //Make sure synctics as char won't
		if (i > 255) i = 255;          //go over 255 (i is a long)
		synctics = i;
	}

		//PRIVATE KEYS:

	//if (numplayers == 2)
	//{
	//       //Record & send a sound to remote computer
	//   if (keystatus[0x1c] != 0)
	//   {
	//      if (lockclock > 240)
	//      {
	//         record();
	//         keystatus[0x1c] = 0;
	//      }
	//      else
	//      {
	//         keystatus[0x1c] = 0;
	//      }
	//   }
	//}

	if (keystatus[0xb7] > 0)  //Printscreen
	{
		printscreeninterrupt();
		keystatus[0xb7] = 0;
	}
	if (keystatus[0x58] > 0)   //F12
	{
		screencapture();
		keystatus[0x58] = 0;
	}
	if (typemode == 0)           //if normal game keys active
	{
		if (keystatus[0x19] > 0)  //P
		{
			keystatus[0x19] = 0;
			parallaxtype++;
			if (parallaxtype > 2) parallaxtype = 0;
		}
		if (keystatus[0x2f] > 0)  //V
		{
			keystatus[0x2f] = 0;
			visibility--;
			if (visibility == 7)
				visibility = 15;
		}

		if ((keystatus[keys[16]]) > 0)   //Typing mode
		{
			keystatus[keys[16]] = 0;
			typemode = 1;
		}
	}
	else
	{
		for(i=0;i<128;i++)
			if (keystatus[i] > 0)
			{
				if ((keystatus[0x2a]|keystatus[0x36]) > 0)
				{
					if (scantoascwithshift[i] != 0)
					{
						if (typemessageleng < 159)
							typemessage[typemessageleng++] = scantoascwithshift[i];
						keystatus[i] = 0;
					}
				}
				else
				{
					if (scantoasc[i] != 0)
					{
						if (typemessageleng < 159)
							typemessage[typemessageleng++] = scantoasc[i];
						keystatus[i] = 0;
					}
				}
			}

		if (keystatus[0xe] > 0)   //Backspace
		{
				//Here's a trick of making key repeat after a 1/2 second
			if ((keystatus[0xe] == 1) || (keystatus[0xe] == 60))
			{
				if (typemessageleng > 0)
					typemessageleng--;
				else
					typemessageleng = 0, typemode = 0;
			}
			if (keystatus[0xe] < 60)
			{
				keystatus[0xe] += synctics;
				if (keystatus[0xe] > 60) keystatus[0xe] = 60;
			}
		}

		if ((keystatus[0x1c]|keystatus[0x9c]) > 0)  //Either ENTER
		{
			if (typemessageleng > 0)
			{
				i = 0;
				tempbuf[i++] = 4;          //Sending text is message type 4
				for(j=0;j<typemessageleng;j++)
					tempbuf[i++] = typemessage[j];

				//sendpacket(-1,tempbuf,i);    2 methods of sending packets...
				if (numplayers == 2)
				{
					if (myconnectnum == connectnum[connecthead])
						sendpacket(connectnum[connectpoint2[connecthead]],tempbuf,i);
					else
						sendpacket(connectnum[connecthead],tempbuf,i);
				}
				else if (numplayers > 2)
					sendpacket(-1,tempbuf,i);
			}

			typemessageleng = 0;
			keystatus[0x1c] = 0;
			keystatus[0x9c] = 0;
			typemode = 0;
		}

		if ((keystatus[0xf]) > 0) //Tab
		{
			keystatus[0xf] = 0;
			typemode = 0;
		}
	}
}

sync()
{
	long i, j, k;

	if (myconnectnum == masterconnectnum) //I am the MASTER!
	{
		getinput(myconnectindex);
		getpackets();

		if (synctics > (option[5]<3))      //1 for <= 9600 baud, 0 for faster
		{
			tempbuf[0] = 2;
			tempbuf[1] = synctics;

			j = ((numplayers+1)>>1)+2;
			for(k=2;k<j;k++)
				tempbuf[k] = 0;

			i = connecthead; k = (2<<3);
			while (i != -1)
			{
				if (syncvel[i] != osyncvel[i])
				{
					tempbuf[j++] = syncvel[i];
					osyncvel[i] = syncvel[i];
					tempbuf[k>>3] |= (1<<(k&7));
				}
				if (syncsvel[i] != osyncsvel[i])
				{
					tempbuf[j++] = syncsvel[i];
					osyncsvel[i] = syncsvel[i];
					tempbuf[k>>3] |= (2<<(k&7));
				}
				if (syncangvel[i] != osyncangvel[i])
				{
					tempbuf[j++] = syncangvel[i];
					osyncangvel[i] = syncangvel[i];
					tempbuf[k>>3] |= (4<<(k&7));
				}
				if (syncbits[i] != osyncbits[i])
				{
					tempbuf[j++] = (syncbits[i]&255);
					tempbuf[j++] = ((syncbits[i]>>8)&255);
					osyncbits[i] = syncbits[i];
					tempbuf[k>>3] |= (8<<(k&7));
				}
				k += 4;

				i = connectpoint2[i];
			}

			if (numplayers == 2)
			{
				if (myconnectnum == connectnum[connecthead])
					sendpacket(connectnum[connectpoint2[connecthead]],tempbuf,j);
				else
					sendpacket(connectnum[connecthead],tempbuf,j);
			}
			else if (numplayers > 2)
				sendpacket(-1,tempbuf,j);

			movethings();  //Move EVERYTHING (you too!)
		}
	}
	else                        //I am a SLAVE
	{
		getinput(MAXPLAYERS);

		tempbuf[0] = 3;
		tempbuf[1] = 0;
		j = 2;
		if (syncvel[MAXPLAYERS] != osyncvel[MAXPLAYERS])
		{
			tempbuf[j++] = syncvel[MAXPLAYERS];
			osyncvel[MAXPLAYERS] = syncvel[MAXPLAYERS];
			tempbuf[1] |= 1;
		}
		if (syncsvel[MAXPLAYERS] != osyncsvel[MAXPLAYERS])
		{
			tempbuf[j++] = syncsvel[MAXPLAYERS];
			osyncsvel[MAXPLAYERS] = syncsvel[MAXPLAYERS];
			tempbuf[1] |= 2;
		}
		if (syncangvel[MAXPLAYERS] != osyncangvel[MAXPLAYERS])
		{
			tempbuf[j++] = syncangvel[MAXPLAYERS];
			osyncangvel[MAXPLAYERS] = syncangvel[MAXPLAYERS];
			tempbuf[1] |= 4;
		}
		if ((syncbits[MAXPLAYERS]&255) != (osyncbits[MAXPLAYERS]&255))
		{
			tempbuf[j++] = (syncbits[MAXPLAYERS]&255);
			tempbuf[1] |= 8;
		}
		if ((syncbits[MAXPLAYERS]|255) != (osyncbits[MAXPLAYERS]|255))
		{
			tempbuf[j++] = ((syncbits[MAXPLAYERS]>>8)&255);
			tempbuf[1] |= 16;
		}
		osyncbits[MAXPLAYERS] = syncbits[MAXPLAYERS];

		sendpacket(masterconnectnum,tempbuf,j);
		getpackets();
	}
}

getpackets()
{
	short otherconnectnum, otherconnectindex, tempbufleng;
	long i, j, k;

	while ((tempbufleng = getpacket(&otherconnectnum,tempbuf)) > 0)
	{
		//for(i=0;i<tempbufleng;i++)
		//   printf("%d ",tempbuf[i]);
		//printf("\n");

		switch(tempbuf[0])
		{
			case 2:  //[2] (receive master sync buffer)
				synctics = tempbuf[1];

				j = ((numplayers+1)>>1)+2; k = (2<<3);
				i = connecthead;
				while (i != -1)
				{
					if ((tempbuf[k>>3]&(1<<(k&7))) > 0) syncvel[i] = tempbuf[j++];
					if ((tempbuf[k>>3]&(2<<(k&7))) > 0) syncsvel[i] = tempbuf[j++];
					if ((tempbuf[k>>3]&(4<<(k&7))) > 0) syncangvel[i] = tempbuf[j++];
					if ((tempbuf[k>>3]&(8<<(k&7))) > 0)
					{
						syncbits[i] = ((short)tempbuf[j])+(((short)tempbuf[j+1])<<8);
						j += 2;
					}

					k += 4;

					i = connectpoint2[i];
				}

				movethings();        //Move all players and sprites

				break;
			case 3:  //[3] (receive slave sync buffer)
				otherconnectindex = connectindex[otherconnectnum];
				j = 2;
				if ((tempbuf[1]&1) > 0)
					syncvel[otherconnectindex] = tempbuf[j++];
				if ((tempbuf[1]&2) > 0)
					syncsvel[otherconnectindex] = tempbuf[j++];
				if ((tempbuf[1]&4) > 0)
					syncangvel[otherconnectindex] = tempbuf[j++];
				if ((tempbuf[1]&8) > 0)
				{
					syncbits[otherconnectindex] &= 0xff00;
					syncbits[otherconnectindex] |= ((short)tempbuf[j++]);
				}
				if ((tempbuf[1]&16) > 0)
				{
					syncbits[otherconnectindex] &= 0x00ff;
					syncbits[otherconnectindex] |= (((short)tempbuf[j++])<<8);
				}

				break;
			case 4:
				getmessageleng = 0;
				for(j=1;j<tempbufleng;j++)
					getmessage[getmessageleng++] = tempbuf[j];
				getmessagetimeoff = totalclock+360+(getmessageleng<<4);
				break;
			case 255:  //[255] (logout)
				keystatus[1] = 1;
				break;
		}
	}
}

sndvol(long xplc, long yplc)
{
	long sqrdist;

	sqrdist = (posx[screenpeek]-xplc)*(posx[screenpeek]-xplc);
	sqrdist += (posy[screenpeek]-yplc)*(posy[screenpeek]-yplc);
	if (sqrdist < 2097152)
		return(255);
	else
		return(1023 / ((sqrdist>>21)+4));
}

initplayersprites()
{
	short i;

	i = connecthead;
	while (i != -1)
	{
		playersprite[i] = insertsprite(cursectnum[i],0);

		sprite[playersprite[i]].x = posx[i];
		sprite[playersprite[i]].y = posy[i];
		sprite[playersprite[i]].z = posz[i]+(32<<8);
		sprite[playersprite[i]].cstat = 1;            //Hitscan should hit opponent
		sprite[playersprite[i]].picnum = DOOMGUY;
		sprite[playersprite[i]].shade = 0;
		sprite[playersprite[i]].xrepeat = 64;
		sprite[playersprite[i]].yrepeat = 64;
		sprite[playersprite[i]].ang = ang[i];
		sprite[playersprite[i]].xvel = 0;
		sprite[playersprite[i]].yvel = 0;
		sprite[playersprite[i]].zvel = 0;
		sprite[playersprite[i]].owner = i+4096;
		sprite[playersprite[i]].tag = 0L;

		i = connectpoint2[i];
	}
}

playback()
{
	long i, j, k;

	recstat = 0;

	while (keystatus[1] == 0)
	{
		prepareboard(boardfilename);

		initplayersprites();
		resettiming();

		i = 0;
		while (i < reccnt)
		{
			if (keystatus[1] > 0) break;

			if (keystatus[keys[13]] > 0)
			{
				keystatus[keys[13]] = 0;

				screenpeek = connectpoint2[screenpeek];
				if (screenpeek == -1) screenpeek = connecthead;

				if (dimensionmode[screenpeek] == 2) qsetmode640480();
				if (dimensionmode[screenpeek] == 3) setup3dscreen();
			}
			if (keystatus[keys[12]] > 0)
			{
				keystatus[keys[12]] = 0;

				if (dimensionmode[screenpeek] == 3)
				{
					dimensionmode[screenpeek] = 2;
					qsetmode640480();
				}
				else
				{
					dimensionmode[screenpeek] = 3;
					setup3dscreen();
				}
			}

			while ((totalclock > lockclock) && (i < reccnt))
			{
				j = connecthead; k = 0;
				while (j != -1)
				{
					syncvel[j] = recsyncvel[i][k];
					syncsvel[j] = recsyncsvel[i][k];
					syncangvel[j] = recsyncangvel[i][k];
					syncbits[j] = recsyncbits[i][k];
					k++;
					j = connectpoint2[j];
				}
				synctics = recsynctics[i];

				movethings();
				i++;
			}

			while (totalclock <= lockclock)
			{
				if (lastsynctics != 0)
					drawscreen(screenpeek,256L+((totalclock-lockclock)<<8)/lastsynctics);
				else
					drawscreen(screenpeek,256L);
			}
		}
	}

	if (option[2] == 1)
	{
		kenchaintimer(0,0); musicoff();
	}
	uninitengine();
	uninitsb();
	setvmode(0x3);        //Set back to text mode
	exit(0);
}

setup3dscreen()
{
	long i, dax, day, dax2, day2;

	setgamemode();

	if (screensize[screenpeek] > 320)
	{
		dax = 0; day = 0;
		dax2 = 319; day2 = 199;
	}
	else
	{
		dax = 160-(screensize[screenpeek]>>1);
		dax2 = dax+screensize[screenpeek]-1;
		day = 84-(((screensize[screenpeek]*168)/320)>>1);
		day2 = day + ((screensize[screenpeek]*168)/320)-1;
	}
	setview(dax,day,dax2,day2);

	if (screensize[screenpeek] < 320)
		permanentwritesprite(0,0,BACKGROUND,0,0,0,319,199);      //Draw background

	if (screensize[screenpeek] <= 320)
	{
		permanentwritesprite(0,200-32,STATUSBAR,0,0,0,319,199);  //Draw status bar only once

		sprintf(&tempbuf,"Deaths: %d",deaths[screenpeek]);
		printext(160-(strlen(tempbuf)<<2),184,tempbuf,ALPHABET,80);

		sprintf(&tempbuf,"Health: %3d",health[screenpeek]);
		printext(160-(strlen(tempbuf)<<2),176,tempbuf,ALPHABET,80);

		sprintf(&tempbuf,"Score:%ld",score[screenpeek]);
		printext(160-(strlen(tempbuf)<<2),192,tempbuf,ALPHABET,80);
	}
}

findrandomspot(long *x, long *y, short *sectnum)
{
	short startwall, endwall, s;
	long dax, day, minx, maxx, miny, maxy, cnt, k;

	cnt = 256;
	while (cnt > 0)
	{
		do
		{
			k = (krand()%numsectors);
		} while ((sector[k].ceilingz >= sector[k].floorz) || (sector[k].tag != 0) || ((sector[k].floorstat&2) != 0));

		startwall = sector[k].wallptr;
		endwall = startwall+sector[k].wallnum-1;
		if (endwall > startwall)
		{
			dax = 0L;
			day = 0L;
			minx = 0x7fffffff; maxx = 0x80000000;
			miny = 0x7fffffff; maxy = 0x80000000;

			for(s=startwall;s<=endwall;s++)
			{
				dax += wall[s].x;
				day += wall[s].y;
				if (wall[s].x < minx) minx = wall[s].x;
				if (wall[s].x > maxx) maxx = wall[s].x;
				if (wall[s].y < miny) miny = wall[s].y;
				if (wall[s].y > maxy) maxy = wall[s].y;
			}
			if ((maxx-minx > 256) && (maxy-miny > 256))
			{
				dax /= (endwall-startwall+1);
				day /= (endwall-startwall+1);
				if (inside(dax,day,k) == 1)
				{
					*x = dax;
					*y = day;
					*sectnum = k;
					return;
				}
			}
		}
		cnt--;
	}
}

/**************************************************************************

	This is my exact movesprite code in ENGINE.C:

movesprite(short spritenum, long xchange, long ychange, long zchange, long walldist, long ceildist, long flordist, char cliptype, long numtics)
{
	long daz, templong;
	short retval, dasectnum;

	dasectnum = sprite[spritenum].sectnum;

	daz = sprite[spritenum].z;
	daz -= ((tilesizy[sprite[spritenum].picnum]*sprite[spritenum].yrepeat)<<1);
	retval = clipmove(&sprite[spritenum].x,&sprite[spritenum].y,&daz,&dasectnum,((xchange*numtics)<<11),((ychange*numtics)<<11),walldist,ceildist,flordist,cliptype);
	if ((dasectnum != sprite[spritenum].sectnum) && (dasectnum >= 0))
		changespritesect(spritenum,dasectnum);

	daz = sprite[spritenum].z + ((zchange*numtics)>>3);
	if ((daz > sector[dasectnum].ceilingz) && (daz <= sector[dasectnum].floorz))
		sprite[spritenum].z = daz;
	else
	{
		if (retval == 0)
			return(16384+dasectnum);
	}

	return(retval);
}

	This is my exact precache code in ENGINE.C:

precache()
{
	long i;
	short j;

	for(i=0;i<numsectors;i++)
	{
		j = sector[i].ceilingpicnum;
		if (waloff[j] == -1) loadtile(j);
		j = sector[i].floorpicnum;
		if (waloff[j] == -1) loadtile(j);
	}
	for(i=0;i<numwalls;i++)
	{
		j = wall[i].picnum;
		if (waloff[j] == -1) loadtile(j);
	}
	for(i=0;i<MAXSPRITES;i++)
		if (sprite[i].statnum < MAXSTATUS)
		{
			j = sprite[i].picnum;
			if (waloff[j] == -1) loadtile(j);
		}
}
***********************************************************************/
